/****
 *
 * Exemplo: Disperso com Endereamento Aberto
 *
 * Seo: 8.4.5
 *
 * Descrio:
 *    -> A tabela de busca  implementada usando disperso com
 *       endereamento aberto em que cada elemento contm uma
 *       chave e a posio do respectivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Tipo de busca: linear
 *
 ****/

/*********************** Includes *************************/

#include <math.h>      /* Funes matemticas   */
#include <stdlib.h>    /* Miscelnea de funes */
#include <stdio.h>     /* Funes de Entrada/Sada    */
#include <string.h>    /* Funes memXXX() e strXXX() */

/******************* Definies de Tipos ******************/

   /* Tipo usado para indicar o status */
   /* de um elemento da tabela         */
typedef enum {VAZIO, OCUPADO} tStatus;

   /* Tipo dos elementos da tabela de busca */
typedef struct no {
           unsigned chave;
           tStatus  status;
        } tElemento, *tTabela;

   /* Tipo de um ponteiro para uma funo de */
   /* disperso com endereamento aberto     */
typedef unsigned (*tDispersao) (unsigned, unsigned, unsigned);

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/******************* Aluses de Funes *******************/

extern unsigned FDispersaoLinear( unsigned chave, unsigned i,
                                  unsigned tamanhoTab );
extern unsigned FDispersaoQuadratica( unsigned chave, unsigned i,
                                      unsigned tamanhoTab );
extern unsigned FDispersaoDupla( unsigned chave, unsigned i,
                                 unsigned tamanhoTab );

/********************* Funes de disperso ********************/

/****
 *
 * FDispersaoLinear(): Calcula o cdigo disperso de uma chave, que
 *                      dado pela frmula:
 *                             h(c, i, m) = (c % m + i) % m,
 *                     em que c  o nmero formado pelos dgitos
 *                     da chave, m  o tamanho da tabela e i, o
 *                     ndice de sondagem, varia entre 0 e m - 1
 *                     (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      tamanhoTab (entrada) - nmero de posies da
 *                             tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoLinear( unsigned chave, unsigned i,
                           unsigned tamanhoTab )
{
   unsigned c = chave; /* Usa funo de disperso modular */

   return (c % tamanhoTab + i) % tamanhoTab;
}

/****
 *
 * FDispersaoQuadratica(): Calcula o cdigo disperso de uma chave
 *                         dado pela frmula:
 *                             h(c, i, m) = (c % m + i*i) % m,
 *                         em que c  o nmero formado pelos dgitos
 *                         da chave, m  o tamanho da tabela e i, o
 *                         ndice de sondagem, varia entre 0 e m - 1
 *                         (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      tamanhoTab (entrada) - nmero de posies da tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoQuadratica( unsigned chave, unsigned i,
                               unsigned tamanhoTab )
{
   unsigned c = chave; /* Usa funo de disperso modular */

   return (c % tamanhoTab + i*i) % tamanhoTab;
}

unsigned FSecundaria(unsigned c, unsigned m)
{
   return (c + 3)%m;
}

/****
 *
 * FDispersaoDupla(): Calcula o cdigo disperso de uma chave
 *                    usando a frmula:
 *                    h(c, i, m) = (c % m + i*(1 + c % (m-1))) % m,
 *                    em que c  o nmero formado pelos dgitos da
 *                    chave, m  o tamanho da tabela e i, o ndice
 *                    de sondagem,  um valor que varia entre 0 e
 *                    m-1 (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      tamanhoTab (entrada) - nmero de posies da
 *                             tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoDupla(unsigned chave, unsigned i, unsigned m)
{
   unsigned c = chave; /* Usa funo de disperso modular */

   return (c % m + i*FSecundaria(chave, m)) % m;
}

/****
 *
 * InsereDEA(): Faz insero numa tabela de disperso com
 *              endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *      chaveEIndice (entrada) - a chave de busca e seu
 *                               respectivo ndice
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio
 *
 ****/
int InsereDEA( tTabela tabela, unsigned tamanhoTab,
               unsigned chave, tDispersao fDispersao )
{
   unsigned i, pos;
   int      posInsercao = -1;

      /* Faz sucessivas sondagens at encontrar a chave */
      /* ou um local onde se pode fazer a insero.     */
   for (i = 0; i < tamanhoTab; ++i) {
      pos = fDispersao(chave, i, tamanhoTab);
      ASSEGURA( pos < tamanhoTab,
                "Valor invalido recebido por InsereDEA()" );

      if (tabela[pos].status == VAZIO && posInsercao < 0) {
         posInsercao = pos; /* A chave no foi encontrada */
         break; /* A chave ser inserida fora do lao */
      }

      if ( tabela[pos].status == OCUPADO &&
           tabela[pos].chave == chave )
         return 0; /* Chave encontrada */
   }

      /* Insere a chave na posio determinada no lao for */
   tabela[posInsercao].chave = chave;
   tabela[posInsercao].status = OCUPADO;

   return 1;
}

/****
 *
 * ConstroiTabela(): L arquivo de dados (binrio) contendo
 *                   os registros e constri uma tabela de
 *                   disperso
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo de
 *                         registros
 *      tamTabela (sada) - tamanho da tabela criada
 *
 * Retorno: Ponteiro para a tabela criada, se a funo
 *          for bem sucedida; NULL, em caso contrrio
 *
 ****/
tTabela ConstroiTabela( const unsigned ar[], unsigned nEl,
                        tDispersao fDispersao, unsigned tamTabela )
{
   unsigned i;
   tTabela  tabela;

   tabela = calloc(tamTabela, sizeof(tElemento));

   for (i = 0; i < nEl; ++i) {
      InsereDEA( tabela, tamTabela, ar[i], fDispersao );
   }

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como
 *                    parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(tTabela tabela, unsigned nElementos)
{
   unsigned i;

//   printf("\nNumero de elementos: %u", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      if (tabela[i].status == VAZIO)
         printf("V, ");
      else
         printf("%u, ", tabela[i].chave);

   printf("%u}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * BuscaDEA(): Executa uma busca simples numa tabela de disperso
 *             com endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *      chave (entrada) - a chave de busca
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: O nmero de sondagens
 *
 ****/
int BuscaDEA( tTabela tabela, unsigned tamanhoTab,
              unsigned chave, tDispersao fDispersao )
{
   unsigned i, pos;

      /* Faz sucessivas sondagens at encontrar a chave, */
      /* uma posio vazia ou a ltima sondagem          */
   for (i = 0; i < tamanhoTab; ++i) {
      pos = fDispersao(chave, i, tamanhoTab);
      ASSEGURA(pos < tamanhoTab,
               "Dispersao invalida recebido por BuscaDEA()");

      if (tabela[pos].status == VAZIO)
         break; /* A chave no foi encontrada */

      if (tabela[pos].status == OCUPADO &&
          tabela[pos].chave == chave )
         break; /* Chave encontrada */
   }

   return i + 1;  /* A chave no foi encontrada */
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa funes de busca, insero, remoo e
 *         atualizao usando disperso com encadeamento
 *
 * Parmetros: nenhum
 *
 * Retorno: 0
 *
 ****/
int main(void)
{
   unsigned  ar[] = { 34, 52, 76, 92, 118, 135, 140, 151,
                      166, 244, 371, 386, 444, 555, 660, 777 };
   unsigned tamAr = sizeof(ar)/sizeof(ar[0]);
   tTabela  tabela = NULL;
   unsigned tamTabela = 20;
   int      sondagens;
   unsigned busca[] = {34, 444, 555, 777, 244, 76, 16};
   unsigned i, tamBusca = sizeof(busca)/sizeof(busca[0]);

   printf("\n\t*** Dispersao Linear ***\n");

      /* Cria a tabela de disperso */
   tabela = ConstroiTabela( ar, tamAr, FDispersaoLinear,
                            tamTabela );

      /* Certifica-se que a criao da tabela foi OK */
   ASSEGURA(tabela, "\nNao foi possivel criar a tabela\n");

   ApresentaTabela(tabela, tamTabela);

   for (i = 0; i < tamBusca; ++i) {
      sondagens = BuscaDEA(tabela, tamTabela, busca[i],
                           FDispersaoLinear);
      printf("\n\t>>> Sondagens de %u: %d", busca[i], sondagens);
   }

   free(tabela);

   printf("\n\n\t*** Dispersao Dupla ***\n");

      /* Cria a tabela de disperso */
   tabela = ConstroiTabela( ar, tamAr, FDispersaoDupla,
                            tamTabela );

      /* Certifica-se que a criao da tabela foi OK */
   ASSEGURA(tabela, "\nNao foi possivel criar a tabela\n");

   ApresentaTabela(tabela, tamTabela);

   for (i = 0; i < tamBusca; ++i) {
      sondagens = BuscaDEA(tabela, tamTabela, busca[i],
                           FDispersaoDupla);
      printf("\n\t>>> Sondagens de %u: %d", busca[i], sondagens);
   }

   free(tabela);

   printf("\n\n\t*** FIM ***\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

{140, 660, 777, V, 244, 444, 166, 386, V, V, V, 151, 52, 92, 34, 135, 76, 371, 118, 555}
        >>> Sondagens de 34: 1
        >>> Sondagens de 444: 2
        >>> Sondagens de 555: 5
        >>> Sondagens de 777: 6
        >>> Sondagens de 244: 1
        >>> Sondagens de 76: 1
        >>> Sondagens de 16: 8

        *** Dispersao Dupla ***

{140, V, V, 660, 244, 371, 166, 92, V, 555, V, 151, 52, 386, 34, 135, 76, 777, 118, 444}
        >>> Sondagens de 34: 1
        >>> Sondagens de 444: 6
        >>> Sondagens de 555: 4
        >>> Sondagens de 777: 1
        >>> Sondagens de 244: 1
        >>> Sondagens de 76: 1
        >>> Sondagens de 16: 7

**********************************************************/
