/****
 *
 * Exemplo: Disperso com Endereamento Aberto
 *
 * Seo: 8.4.5
 *
 * Descrio:
 *    -> A tabela de busca  implementada usando disperso com
 *       endereamento aberto em que cada elemento contm uma
 *       chave e a posio do respectivo registro no arquivo
 *
 *    -> Os registros so mantidos em arquivo
 *
 *    -> Tipo de busca: linear
 *
 ****/

/*********************** Includes *************************/

#include <math.h>      /* Funes matemticas   */
#include <stdlib.h>    /* Miscelnea de funes */
#include <stdio.h>     /* Funes de Entrada/Sada    */
#include <string.h>    /* Funes memXXX() e strXXX() */

/******************* Definies de Tipos ******************/

   /* Tipo usado para indicar o status */
   /* de um elemento da tabela         */
typedef enum {VAZIO, OCUPADO} tStatus;

   /* Tipo dos elementos da tabela de busca */
typedef struct no {
           unsigned chave;
           tStatus  status;
        } tElemento, *tTabela;

   /* Tipo de um ponteiro para uma funo de */
   /* disperso com endereamento aberto     */
typedef unsigned (*tDispersao) (unsigned, unsigned, unsigned);

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/******************* Aluses de Funes *******************/

extern unsigned FDispersaoLinear( unsigned chave, unsigned i,
                                  unsigned m );
extern unsigned FDispersaoQuadratica( unsigned chave, unsigned i,
                                      unsigned m );
extern unsigned FDispersaoDupla( unsigned chave, unsigned i,
                                 unsigned m );

/********************* Funes de disperso ********************/

/****
 *
 * FDispersaoLinear(): Calcula o cdigo disperso de uma chave, que
 *                      dado pela frmula:
 *                             h(c, i, m) = (c % m + i) % m,
 *                     em que c  o nmero formado pelos dgitos
 *                     da chave, m  o tamanho da tabela e i, o
 *                     ndice de sondagem, varia entre 0 e m - 1
 *                     (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      m (entrada) - nmero de posies da tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoLinear(unsigned chave, unsigned i, unsigned m)
{
   unsigned c = (2*chave + 5)%m;

   return (c % m + i) % m;
}

/****
 *
 * FDispersaoQuadratica(): Calcula o cdigo disperso de uma chave
 *                         dado pela frmula:
 *                             h(c, i, m) = (c % m + i*i) % m,
 *                         em que c  o nmero formado pelos dgitos
 *                         da chave, m  o tamanho da tabela e i, o
 *                         ndice de sondagem, varia entre 0 e m - 1
 *                         (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      m (entrada) - nmero de posies da tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoQuadratica(unsigned chave, unsigned i, unsigned m)
{
   unsigned c = (2*chave + 5)%m;

   return (c % m + i*i) % m;
}

unsigned FSecundaria(unsigned c)
{
   return 7 - c%7;
}

/****
 *
 * FDispersaoDupla(): Calcula o cdigo disperso de uma chave
 *                    usando a frmula:
 *                    h(c, i, m) = (c % m + i*(1 + c % (m-1))) % m,
 *                    em que c  o nmero formado pelos dgitos da
 *                    chave, m  o tamanho da tabela e i, o ndice
 *                    de sondagem,  um valor que varia entre 0 e
 *                    m-1 (inclusive).
 *
 * Parmetros:
 *      chave (entrada) - a chave que ter seu cdigo disperso
 *                        calculado
 *      i (entrada) - ndice da sondagem
 *      m (entrada) - nmero de posies da
 *                             tabela de disperso
 *
 * Retorno: O cdigo disperso da chave recebida como parmetro
 *
 ****/
unsigned FDispersaoDupla(unsigned chave, unsigned i, unsigned m)
{
   unsigned c = chave; /* Usa funo de disperso modular */

   return (c % m + i*FSecundaria(chave)) % m;
}

/****
 *
 * InsereDEA(): Faz insero numa tabela de disperso com
 *              endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      m (entrada) - tamanho da tabela de disperso
 *      chaveEIndice (entrada) - a chave de busca e seu
 *                               respectivo ndice
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio
 *
 ****/
int InsereDEA( tTabela tabela, unsigned m,
               unsigned chave, tDispersao fDispersao )
{
   unsigned i, pos;
   int      posInsercao = -1;

      /* Faz sucessivas sondagens at encontrar a chave */
      /* ou um local onde se pode fazer a insero.     */
   for (i = 0; i < m; ++i) {
      pos = fDispersao(chave, i, m);
      ASSEGURA(pos < m, "Valor invalido recebido por InsereDEA()");

      if (tabela[pos].status == VAZIO && posInsercao < 0) {
         posInsercao = pos; /* A chave no foi encontrada */
         break; /* A chave ser inserida fora do lao */
      }

      if (tabela[pos].status == OCUPADO && tabela[pos].chave == chave) {
         printf("\n\t\a>>> Chave %u nao foi inserida\n", chave);
         return 0; /* Chave encontrada */
      }
   }

   if (posInsercao < 0) {
      printf("\n\t\a>>> Chave %u nao foi inserida\n", chave);
      return 0;
   }

      /* Insere a chave na posio determinada no lao for */
   tabela[posInsercao].chave = chave;
   tabela[posInsercao].status = OCUPADO;

   return 1;
}

/****
 *
 * ConstroiTabela(): L arquivo de dados (binrio) contendo
 *                   os registros e constri uma tabela de
 *                   disperso
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo de
 *                         registros
 *      m (sada) - tamanho da tabela criada
 *
 * Retorno: Ponteiro para a tabela criada, se a funo
 *          for bem sucedida; NULL, em caso contrrio
 *
 ****/
tTabela ConstroiTabela( const unsigned ar[], unsigned nEl,
                        tDispersao fDispersao, unsigned m )
{
   unsigned i;
   tTabela  tabela;

   tabela = calloc(m, sizeof(tElemento));

   for (i = 0; i < nEl; ++i)
      InsereDEA( tabela, m, ar[i], fDispersao );

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero exibidos
 *      m (entrada) - nmero mnimo de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(tTabela tabela, unsigned m)
{
   unsigned i;

   printf("\n{");

   for (i = 0; i < m - 1; ++i)
      if (tabela[i].status == VAZIO)
         printf("V, ");
      else
         printf("%u, ", tabela[i].chave);

   printf("%u}", tabela[m - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa funes de busca, insero, remoo e
 *         atualizao usando disperso com encadeamento
 *
 * Parmetros: nenhum
 *
 * Retorno: 0
 *
 ****/
int main(void)
{
   unsigned  ar[] = { 13, 41, 12, 79, 21, 89, 11, 37, 33, 19, 16, 44, 3 };
   unsigned tamAr = sizeof(ar)/sizeof(ar[0]);
   tTabela  tabela = NULL;
   unsigned m = 13;

   printf("\n\t*** Q89: Dispersao Linear ***\n");

      /* Cria a tabela de disperso */
   tabela = ConstroiTabela(ar, tamAr, FDispersaoLinear, m);

      /* Certifica-se que a criao da tabela foi OK */
   ASSEGURA(tabela, "\nNao foi possivel criar a tabela\n");

   ApresentaTabela(tabela, m);

   free(tabela);

   printf("\n\n\t*** Q90: Dispersao Quadratica ***\n");

      /* Cria a tabela de disperso */
   tabela = ConstroiTabela(ar, tamAr, FDispersaoQuadratica, m);

      /* Certifica-se que a criao da tabela foi OK */
   ASSEGURA(tabela, "\nNao foi possivel criar a tabela\n");

   ApresentaTabela(tabela, m);

   free(tabela);

   printf("\n\n\t*** Q91: Dispersao Dupla ***\n");

      /* Cria a tabela de disperso */
   tabela = ConstroiTabela(ar, tamAr, FDispersaoDupla, m);

      /* Certifica-se que a criao da tabela foi OK */
   ASSEGURA(tabela, "\nNao foi possivel criar a tabela\n");

   ApresentaTabela(tabela, m);

   free(tabela);

   printf("\n\n\t*** FIM ***\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

        *** Q84: Dispersao Linear ***

{3, 89, 11, 12, 37, 13, 33, 79, 21, 41, 19, 16, 44}

        *** Q85: Dispersao Quadratica ***

        >>> Chave 3 nao foi inserida

{V, 89, 11, 12, 19, 13, 33, 79, 21, 41, 37, 16, 44}

        *** Q86: Dispersao Dupla ***

{13, 79, 41, 37, 11, 16, 19, 33, 21, 3, 44, 89, 12}

**********************************************************/
