/****
 *
 * Exemplo: Lista indexada no ordenada
 *
 * CriaArquivoBin.c: Programa que cria o arquivo binrio "CEPs.bin"
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* exit() e strtod()        */
#include <time.h>     /* Medio de tempo         */

/***************** Constantes Simblicas ******************/

#define MAX_NOME  40 /* Nmero mximo de caracteres num nome */
#define TAM_UF     2 /* Nmero de dgitos num UF */
#define TAM_CEP    8 /* Nmero de dgitos num CEP */
#define MAX_COMP  60 /* Nmero mximo de caracteres */
                     /* num complemento             */
#define MAX_TIPO_LOG 10 /* Nmero mximo de caracteres */
                        /* num tipo de logradouro      */
#define TAM_DNE   16 /* Nmero de dgitos numa chave DNE */

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

   /* Nome dos arquivos de dados */
#define NOME_ARQUIVO_BIN "CEPs.bin"

   /* Altere o nome para incluir o diretrio  */
   /* em que se encontra o arquivo "CEPs.txt" */
#define NOME_ARQUIVO_TXT "CEPs.txt"

/****************** Definies de Macros ******************/

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(cond, msg) if (!(cond)) {\
                               fprintf(stderr, "\n%s\n", msg);\
                               exit(1); \
                                }

/****************** Definies de Tipos *******************/

typedef struct {
           int  numero;
           char UF[TAM_UF];
           int  localidadeNumero;
           char nomeAbr[MAX_NOME];
           char nome[MAX_NOME];
           int  bairroInicio;
           int  bairroFim;
           char CEP[TAM_CEP];
           char complemento[MAX_COMP];
           char tipoLogradouro[MAX_TIPO_LOG];
           char statusTipoLogradouro;
           char nomeSemAcento[MAX_NOME];
           char chaveDNE[TAM_DNE];
         } tRegistro;

/***************** Definies de Funes ******************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
static char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
*
* ObtemTokens(): Separa um string em partes (tokens)
*
* Parmetros:
*     str (entrada/sada) - o string que ser separado
*                           ou NULL para continuar uma
*                           separao anterior
*     sep (entrada) - string contendo os separadores de tokens
*
* Retorno: Endereo do prximo token do string, se este
*          existir, ou NULL, em caso contrrio
*
* Observao: Esta funo NO  funcionalmente equivalente 
*             funo strtok() da biblioteca padro de C
*
****/
static char *ObtemTokens(char *str, char const *sep)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, /* Apontar para o string no qual a */
                   /* busca pelo token ser efetuada   */
               *inicio; /* Guardar o incio */
                        /* do token corrente */

   if (str != NULL)
      proximoToken = str;

   if (proximoToken == NULL)
      return NULL;

   if ((s = strpbrk(proximoToken, sep)) != NULL) {
      *s  = '\0';
      inicio = proximoToken;
      proximoToken = ++s;
   } else if (*proximoToken) {
      inicio = proximoToken;
      proximoToken = NULL;
   }

   return inicio;
}

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa
 *                    estrutura do tipo tRegistro.
 *
 * Parmetros: regCEP (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser
 *                                     convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por ObtemTokens() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por ObtemTokens().
 *
 *     3. A funo ObtemTokens()  usada em lugar de strtok()
 *        devido  presena de campos vazios no arquivo de
 *        texto, o que faz com que dois separadores apaream
 *        juntos. Nesse caso, a funo strtok() consideraria
 *        os dois separadores como sendo apenas um.
 *
 ****/
static tRegistro *LinhaEmRegistro(tRegistro *regCEP, char *linha)
{
   char *str; /* Apontar para cada token da linha */
   char *final; /* Final de uma converso numrica */

      /* Obtm o nmero e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(linha, "\t;\n");
   regCEP->numero = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm a UF e acrescenta-o ao  */
      /* respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->UF, str, TAM_UF);

      /* Obtm o nmero de localidade e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->localidadeNumero = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->localidadeNumero < 0,
            "Erro de conversao numerica");

      /* Obtm o nome abreviado e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nomeAbr, str, MAX_NOME);

      /* Obtm o nome e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nome, str, MAX_NOME);

      /* Obtm o bairro incio e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->bairroInicio = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroInicio < 0,
            "Erro de conversao numerica");

      /* Obtm o bairro final e acrescenta-o ao */
      /* respectivo campo da estrutura    */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->bairroFim = (int) strtol(str, &final, 10);
   ASSEGURA(!*final || regCEP->bairroFim < 0,
            "Erro de conversao numerica");

      /* Obtm o CEP e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->CEP, str, TAM_CEP);

      /* Obtm o complemento e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->complemento, str, MAX_COMP);

      /* Obtm o tipo de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->tipoLogradouro, str, MAX_TIPO_LOG);

      /* Obtm o tipo de logradouro e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   regCEP->statusTipoLogradouro = *str;

      /* Obtm o nome sem acento e acrescenta-o ao */
      /* respectivo campo da estrutura  */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->nomeSemAcento, str, MAX_NOME);

      /* Obtm a chave DNE e acrescenta-o */
      /* ao respectivo campo da estrutura */
   str = ObtemTokens(NULL, "\t;\n");
   strncpy(regCEP->chaveDNE, str, TAM_DNE);

   return regCEP;
}

/****
 *
 * CriaArquivoBin(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura
 *                   do tipo tRegistro e armazena a estrutura
 *                   num arquivo binrio.
 *
 * Parmetros:
 *     streamTexto (entrada) - stream associado ao arquivo que
 *                             ser lido
 *     streamBin (entrada) - stream associado ao arquivo que ser
 *                           escrito
 *
 * Retorno: Nmero de estruturas escritas no arquivo
 *          binrio ou um valor negativo, se ocorrer erro
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de
 *        texto que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoBin(const char *arqTexto, const char *arqBin)
{
   char     *linha; /* Apontar para cada linha lida */
   tRegistro umCEP; /* Dados de uma linha      */
                    /* convertida em estrutura */
   int       nRegistros = 0; /* Nmero de estruturas        */
                             /* escritas no arquivo binrio */
   FILE     *streamTexto, /* Associado ao arquivo de texto */
            *streamBin;   /* Associado ao arquivo binrio  */

      /* Arquivo binrio no foi aberto. Tenta abrir */
      /* arquivo de texto para leitura               */
   streamTexto = fopen(arqTexto, "r");

      /* Se o arquivo de texto no pode ser */
      /* aberto, nada mais pode ser feito   */
   if (!streamTexto) {
      return -1; /* Operao mal sucedida */
   }

      /* Tenta criar arquivo binrio. Se um arquivo */
      /* com esse nome existir ele ser destrudo   */
   streamBin = fopen(arqBin, "wb");

      /* Se o arquivo no pode ser aberto, */
      /* nada mais pode ser feito.         */
   if (!streamBin) {
         /* Antes de retornar, fecha arquivo aberto */
      fclose(streamTexto);

      return -1; /* Operao mal sucedida */
   }

      /* L cada linha do arquivo, cria um registro do   */
      /* tipo tRegistro e armazena-o no arquivo binrio. */
      /* O lao encerra quando h tentativa de leitura   */
      /* alm do final do arquivo de entrada, ou ocorre  */
      /* erro de leitura ou escrita.                     */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, streamTexto);

         /* Verifica se o final do arquivo foi atingido */
      if (feof(streamTexto)) {
         break;
      }

         /* Verifica se ocorreu erro de leitura */
      if (ferror(streamTexto)) {
         return -1; /* Operao mal sucedida */
      }

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umCEP, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umCEP, sizeof(umCEP), 1, streamBin);

      free(linha); /* A linha no  mais necessria */

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamBin)) { /* Ocorreu */
         return -1; /* Operao mal sucedida */
      }

      ++nRegistros; /* Mais um registro lido */
   }

   return nRegistros; /* No ocorreu nenhum erro */
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 ****/
void MedidaDeTempo(void)
{
   static int numeroDeChamadas = 0;
   static time_t t0;
   time_t tf;

   ++numeroDeChamadas;

   if (numeroDeChamadas%2) { /* Calcula instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta intervalos de    */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * main(): Executa a funo CriaArquivoBin()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int nRegs; /* Nmero de registros no arquivo binrio */

   MedidaDeTempo();
   printf("\nCriando arquivo...");

   nRegs = CriaArquivoBin(NOME_ARQUIVO_TXT, NOME_ARQUIVO_BIN);

   printf("\n...Operacao encerrada");
   MedidaDeTempo();

   if (nRegs >= 0) {
      printf( "\n\t>>> Criacao do arquivo binario foi "
             "bem sucedida\n" );
   } else {
      printf( "\n\t>>> Ocorreu erro na criacao do arquivo"
              " binario\n" );
   }

   return 0;
}

/****************** Resultado do Programa *****************

Criando arquivo...
...Operacao encerrada

        *** Tempo gasto na operacao: 18.00 segundos ***

        >>> Criacao do arquivo binario foi bem sucedida

**********************************************************/

