/****
 *
 * Ttulo: Tabela de busca sequencial com transposio usando lista
 *         simplesmente encadeada e chaves externas.
 *
 * Seo: 3.4.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Descrio: Funes para a criao e manejo de uma tabela debusca encadeada,
 *            cuja busca  sequencial com movimentao para o incio
 ****/

/************************ Includes ************************/

#include <stdio.h>    /* Entrada e sada         */
#include <stdlib.h>   /* exit()                  */
#include <string.h>   /* Funes strXXX()        */
#include "ASSEGURA.h" /* Macro ASSEGURA          */
#include "Tabela.h"   /* Interface deste arquivo */

/****************** Definies de Funes *****************/

/****
 *
 * IniciaListaSE(): Inicia uma lista simplesmente encadeada
 *
 * Parmetros: *lista (sada) - a lista que ser iniciada
 *
 * Retorno: Nada
 *
 * Observao: Esta funo independe do tipo de contedo
 *             efetivo armazenado em cada n da lista
 *
 ****/
void IniciaListaSE(tListaSE *lista)
{
   *lista = NULL;
}

/****
 *
 * ComprimentoListaSE(): Calcula o nmero de ns de uma lista
 *                       simplesmente encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para o primeiro n da lista
 *                       cujo comprimento ser calculado
 *
 * Retorno: O nmero de ns da referida lista
 *
 * Observao: Esta funo independe do tipo de contedo
 *             efetivo armazenado em cada n da lista
 *
 ****/
int ComprimentoListaSE(tListaSE lista)
{
   int tamanho = 0; /* Armazena o nmero de ns da lista */

      /* Acessa cada n da lista e conta */
      /* quantos ns so acessados       */
   while (lista) {
      ++tamanho; /* Mais um n foi encontrado */
      lista = lista->proximo; /* Passa para o prximo n */
   }

   return tamanho;
}

/****
 *
 * InsereLSE(): Insere um novo n no incio de uma lista simplesmente encadeada
 *
 * Parmetros:
 *    *lista (entrada/sada) - ponteiro para a lista na qual
 *                             ser feita a insero
 *     conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Nada
 *
 * Observao: Como a lista no obedece nenhuma ordem, pode-se
 *             inserir um n em qualquer posio da lista.
 *             Escolheu-se insero no incio da lista porque
 *              mais simples e eficiente.
 *
 ****/
void InsereLSE(tListaSE *lista, const tCEP_Ind *conteudo)
{
   tNoListaSE *ptrNovoNo; /* Apontar para o novo n alocado */

      /* Tenta alocar um novo n */
   ASSEGURA( ptrNovoNo = malloc(sizeof(tNoListaSE)),
             "Nao foi possivel alocar no'" );

      /* Armazena no novo n os dados recebidos como parmetro */
   ptrNovoNo->conteudo = *conteudo;

      /* O novo n apontar para o incio corrente da lista */
   ptrNovoNo->proximo = *lista;

      /* O incio da lista passa a apontar para o novo n */
   *lista = ptrNovoNo;
}

/****
 *
 * RemoveLSE(): Remove de uma lista simplesmente encadeada um n
 *             que contm um determinado valor especificado como
 *             parmetro
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro que representa
 *                             a lista na qual ser feita a remoo
 *     conteudo (entrada) - contedo do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveLSE(tListaSE *lista, tCEP chave)
{
   tListaSE p = *lista, /* p aponta para o n corrente     */
            q = NULL;   /* q aponta para o n anterior a p */

      /*                                                     */
      /* O lao while procura um n que cujo contedo case   */
      /* com o parmetro 'conteudo' e tem duas condies de  */
      /* parada: (1) todos os ns so acessados sem que o    */
      /* referido n tenha sido encontrado (neste  caso, p   */
      /* assume NULL); (2) o n que contm o valor procurado */
      /*  encontrado.                                       */
      /*                                                     */
   while (p && strcmp(p->conteudo.chave, chave)) {
      q = p; /* q passa a apontar para o n corrente */
      p = p->proximo; /* p passa a apontar para o prximo n */
   }

      /* Se p for NULL, a lista foi totalmente examinada   */
      /* sem que o n a ser removido tenha sido encontrado */
   if (!p) {
      return 1; /* Tentativa de remoo de n inexistente */
   }

      /****************************************************/
      /* Neste ponto, p aponta para o n a ser removido e */
      /* q aponta para o n imediatamente anterior a ele  */
      /****************************************************/

      /* Verifica se p aponta para o primeiro */
      /* n da lista porque remoo no incio */
      /* deve ser tratada separadamente       */
   if (p == *lista) { /* Remoo ser no incio da lista */
      (*lista) = p->proximo; /* Altera incio da lista */
   } else {/* N a ser removido NO  o primeiro da lista */
      q->proximo = p->proximo; /* Desvia o n anterior */
                               /* do n a ser removido */
   }

      /* Neste ponto, o n desejado no faz mais parte da */
      /* lista, mas ele ainda ocupa espao em memria.    */
      /* Portanto,  necessrio liberar este espao.      */
   free(p); /* Libera o n removido */

   return 0;
}

/****
 *
 * EstaVaziaLSE(): Verifica se uma lista simplesmente encadeada est vazia
 *
 * Parmetros:
 *     lista (entrada) - a lista que ser verificada
 *
 * Retorno: 1, quando a lista est vazia, ou 0, em caso contrrio
 *
 ****/
int EstaVaziaLSE(tListaSE lista)
{
   return lista == NULL;
}

/****
 *
 * BuscaComTransposicaoLSE(): Retorna o endereo do contedo efetivo
 *                         do n que possui uma determinada chave
 *                         numa lista simplesmente encadeada. Se
 *                         esse n for encontrado, troca-o de
 *                         posio com seu antecessor
 *
 * Parmetros:
 *      *lista (entrada) - a tabela de busca
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo dos dados que contm o referido contedo,
 *          se ele for encontrado. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaComTransposicaoLSE(tListaSE *lista, tCEP chave)
{
   tListaSE p,        /* p apontar para o n corrente     */
            q = NULL, /* q apontar para o antecessor de p */
            r = NULL; /* r apontar para o antecessor de q */

      /* Enquanto p no assume NULL ou a chave de um n no */
      /* casa com o parmetro 'chave', a busca prossegue    */
   for ( p = *lista; p && strcmp(p->conteudo.chave, chave);
         p = p->proximo) {
      r = q;
      q = p;
   }

      /* Se 'p' assume NULL  porque a chave especificada */
      /* como parmetro no foi encontrado                */
   if(!p)
      return NULL; /* Chave no foi encontrada */

      /*                                                  */
      /* Neste ponto, q aponta para o antecessor imediato */
      /* de p e r aponta para o antecessor imediato de q  */
      /*                                                  */

   if (!q) /* O n encontrado j  o primeiro da lista */
      return &p->conteudo; /* No  necessria a transposio */

   q->proximo = p->proximo;
   p->proximo = q;

   if (!r)
         /* N encontrado passa a ser o primeiro da lista */
      *lista = p;
   else
      r->proximo = p;

   return &p->conteudo;
}

/****
 *
 * BuscaComMovimentoLSE(): Retorna o endereo do contedo efetivo do
 *                      n que possui uma determinada chave numa
 *                      lista simplesmente encadeada e o movimenta
 *                      para o incio da lista
 *
 * Parmetros:
 *      *lista (entrada) - lista que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo dos dados que contm o referido contedo,
 *          se ele for encontrado. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaComMovimentoLSE(tListaSE *lista, tCEP chave)
{
   tListaSE p = *lista, /* p apontar para o n corrente     */
            q = NULL;   /* q apontar para o n anterior a p */

      /* Enquanto p no assume NULL ou a chave de um n no */
      /* casa com o parmetro 'chave', a busca prossegue    */
   while (p && strcmp(p->conteudo.chave, chave)) {
      q = p;
      p = p->proximo;
   }

      /* Se 'p' assume NULL  porque a chave especificada */
      /* como parmetro no foi encontrado                */
   if(!p)
      return NULL; /* Chave no foi encontrada */

      /* A chave foi encontrada. Se o n que a contm for o primeiro */
      /* da lista, a movimentao para o incio no  necessria.    */
      /* Caso contrrio, ela ser efetuada a seguir.                 */
   if (p != *lista) {
      q->proximo = p->proximo;
      p->proximo = *lista;
      *lista = p;
   }

      /* Neste ponto, p aponta para o n encontrado. Ento, */
      /* retorna-se o endereo de seu campo 'conteudo'.     */
   return &p->conteudo;
}

/****
 *
 * DestroiListaSE(): Libera o espao ocupado pelos ns de
 *                   uma lista encadeada, tornando-a vazia
 *
 * Parmetros:
 *     lista (entrada/sada) - endereo do ponteiro para a
 *                             lista simplesmente encadeada
 *                             que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiListaSE(tListaSE *lista)
{
   tListaSE p; /* Aponta para o prximo n a ser liberado */

   if (!*lista) { /* Verifica se a lista est vazia */
      return; /* Lista vazia no precisa ser destruda */
   }

   p = *lista; /* Faz p apontar para o incio da lista */

      /* Visita cada n da lista liberando-o */
   do {
         /* Passa para o prximo n antes que o n corrente */
         /* seja destrudo. No importa que o ponteiro para */
         /*o incio da lista seja alterado. Afinal, este    */
         /* ponteiro ser mesmo anulado.                    */
      *lista = (*lista)->proximo;

      free(p); /* Libera o espao do n corrente */

         /* Faz p apontar para o prximo n, cujo */
         /* endereo est armazenado em *lista    */
      p = *lista;
   } while (p);

      /* Neste ponto, a lista ficou vazia, pois o  */
      /* ltimo valor assumido por *lista foi NULL */
}

/****
 *
 * ProximoListaSE(): Encontra o prximo n de uma lista simplesmente encadeada
 *
 * Parmetros:
 *     lista (entrada) - ponteiro para a lista simplesmente encadeada
 *
 * Retorno: Endereo do contedo do prximo n
 *
 ****/
tCEP_Ind *ProximoListaSE(tListaSE lista)
{
   static tNoListaSE *proximoNo = NULL;

   if (lista) {
      if (!proximoNo) {
         proximoNo = lista;
      } else {
         proximoNo = proximoNo->proximo;
      }
      return &proximoNo->conteudo;
   } else {
      return NULL;
   }
}
