/****
 *
 * Arquivo: ArvoreBB2.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2016
 * ltima modificao: 22/07/2016
 *
 * Descrio: Implementao do mdulo ArvoreBB2
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>    /* Alocao dinmica           */
#include <string.h>    /* Funes strXXX() e memXXX() */
#include "ArvoreBB2.h" /* Interface deste mdulo      */
#include "ASSEGURA.h"  /* Macro ASSEGURA              */

/****************** Complemento de Tipo *******************/

struct rotArvoreBB2 {
   tNoBB2 *raiz; /* Raiz da rvore */
   int     nNos; /* Nmero de ns da rvore */
};

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

///****
// *
// * RotacaoSimples(): Efetua uma rotao simples sobre
// *                   um n de uma rvore binria de busca
// *                   na direo especificada
// *
// * Parmetros:
// *      *x (entrada/sada) - o n em torno do qual ser
// *                           efetuada a rotao
// *       dir (entrada) - sentido da rotao (0 = esquerda;
// *                       1 = direita)
// *
// * Retorno: A nova raiz aps a rotao
// *
// * Observao: Assume-se que todos os ns so vlidos
// *
// ****/
//static tNoBB2 *RotacaoSimples(tNoBB2 *x, tDirecao dir)
//{
//   tNoBB2 *y = x->filho[!dir]; /* Faz y apontar para o    */
//                              /* filho oposto da rotao */
//
//   x->filho[!dir] = y->filho[dir];
//   y->filho[dir] = x;
//
//   return y;
//}
//
///****
// *
// * RotacaoDupla(): Efetua uma rotao dupla sobre
// *                 um n de uma rvore binria de busca
// *                 na direo especificada
// *
// * Parmetros:
// *      *x (entrada/sada) - o pai do n em torno do qual
// *                           ser efetuada a primeira rotao
// *       dir (entrada) - sentido da rotao (0 = esquerda;
// *                       1 = direita)
// *
// * Retorno: A nova raiz aps a rotao
// *
// * Observao: Assume-se que todos os ns so vlidos
// *
// ****/
//static tNoBB2 *RotacaoDupla(tNoBB2 *x, tDirecao dir)
//{
//      /* Efetua uma rotao em torno do filho oposto de x */
//      /* na direo contrria especificada por 'dir'      */
//   x->filho[!dir] = RotacaoSimples(x->filho[!dir], !dir);
//
//      /* Agora gira o filho */
//   return RotacaoSimples(x, dir);
//}

/****
 *
 * NoNovoArvoreBB2(): Cria e inicia um n de uma rvore
 *                    binria de busca
 *
 * Parmetros:
 *       conteudo (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 * Observao: Esta funo no insere o novo n na rvore
 *
 ****/
static tNoBB2 *NoNovoArvoreBB2(const tConteudo *conteudo)
{
   tNoBB2 *pNovo = malloc(sizeof *pNovo);

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(pNovo, "Erro: Nao foi possivel alocar no'");

   pNovo->conteudo = *conteudo;

      /* O n ainda no possui filhos */
   pNovo->filho[ESQUERDA] = pNovo->filho[DIREITA] = NULL;

   return pNovo;
}

/****
 *
 * InsereABBRec(): Insere um novo n numa rvore binria de busca
 *
 * Parmetros:
 *    *raiz (entrada/sada) - raiz da rvore binria de busca
 *                            na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Endereo da raiz, se a insero for bem sucedida
 *          NULL, em caso contrrio
 *
 ****/
static tNoBB2 *InsereABBRec( tNoBB2 *raiz,
                             const tConteudo *conteudo )
{
   tDirecao dir; /* Direo (esquerda/direita) de descida */
   int      compara; /* Resultado de comparao de chaves */

      /* Se a rvore estiver vazia, cria uma raiz e retorna-a */
   if (!raiz) {
      return NoNovoArvoreBB2(conteudo);
   }

      /* Compara a chave a ser inserida com */
      /* a chave que se encontra na raiz    */
   compara = strcmp(raiz->conteudo.chave, conteudo->chave);

      /* Se a chave j existir, ela no ser inserida */
      /* novamente, pois ela  primria               */
   if (!compara) {
      return NULL; /* Chave j existe */
   }

      /*                                 */
      /*  preciso descer na rvore para */
      /* encontrar o local da insero   */
      /*                                 */

      /* 'dir' indicar se a descida ser pela esquerda */
      /* (dir < 0) ou pela direita (dir > 1)            */
   dir = (tDirecao) (compara < 0);

      /* Chamada recursiva para insero  */
      /* direita ou esquerda de uma folha  */
   raiz->filho[dir] = InsereABBRec(raiz->filho[dir], conteudo);

   return raiz;
}

/****
 *
 * RemoveABBRec(): Recursivamente, remove de uma rvore
 *                 binria de busca um n que contm a
 *                 chave especificada como parmetro
 *
 * Parmetros:
 *     raiz (entrada/sada) - ponteiro para a raiz da rvore
 *                            na qual ser feita a remoo
 *     chave (entrada/sada) - chave do n a ser removido
 *     removido (entrada/sada) - indica quando ocorre remoo
 *
 * Retorno: Endereo da raiz da rvore na qual ocorreu a
 *          remoo
 *
 * OBSERVAO IMPORTANTE: O parmetro chave pode ser alterado!
 *
 ****/
static tNoBB2 *RemoveABBRec( tNoBB2 *raiz, tChave chave,
                             int *removido )
{
   tNoBB2  *sucessor, /* Apontar para o sucessor de um n */
           *novaRaiz; /* Um ponteiro auxiliar */
   tDirecao dir;  /* Direo (esquerda/direita) de descida */
   int      compara; /* Resultado de comparao de chaves */

   if (!raiz) { /* rvore vazia */
      return raiz; /* No h mais o que fazer */
   } else {
         /* Compara a chave do n corrente   */
         /* com a chave do n a ser removido */
      compara = strcmp(raiz->conteudo.chave, chave);

         /* Verifica se a chave foi encontrada */
      if (!compara) { /* Chave encontrada */
         if (!raiz->filho[ESQUERDA] || !raiz->filho[DIREITA]) {
            /*                              */
            /* Caso trivial: O n no tem   */
            /* filho ou tem apenas um filho */
            /*                              */

               /* Faz 'novaRaiz' apontar para o filho direito  */
               /* da raiz se o filho esquerdo for NULL. Se o   */
               /* filho esquerdo da raiz no for NULL,         */
               /* 'novaRaiz' apontar para ele. Se os dois     */
               /* filhos forem NULL, novaRaiz tambm ser NULL */
            novaRaiz = raiz->filho[!raiz->filho[ESQUERDA]];

            free(raiz); /* Libera a antiga raiz */

            return novaRaiz; /* Retorna a nova raiz */
         } else { /* O n tem dois filhos */
            /*                           */
            /* Encontra o sucessor do n */
            /*                           */

               /* Primeiro, desce-se at o filho direito */
            sucessor = raiz->filho[DIREITA];

               /* Agora, desce-se sempre  esquerda */
               /* at encontrar uma folha           */
            while (sucessor->filho[ESQUERDA]) {
               sucessor = sucessor->filho[ESQUERDA];
            }

            /* Sucessor encontrado */

               /* Copia o contedo do sucessor */
               /* para o n que seria removido */
            raiz->conteudo = sucessor->conteudo;

               /* Copia a chave do sucessor para */
               /* a chave do n a ser removido   */
            strcpy(chave, sucessor->conteudo.chave);
         }

         *removido = 1; /* Houve remoo */
      }

      /*                                  */
      /* A chave ainda no foi encontrada */
      /* e a busca continua               */
      /*                                  */

         /* Obtm a direo da descida  */
         /* (0 = esquerda; 1 = direita) */
      dir = compara < 0;

         /* Efetua a remoo na subrvore indicada por 'dir' */
      raiz->filho[dir] = RemoveABBRec( raiz->filho[dir],
                                       chave, removido );
   }

   return raiz;
}

/****
 *
 * CaminhamentoInfixoRec(): Efetua um caminhamento em ordem
 *                          infixa numa rvore binria
 *
 * Parmetros:
 *       raiz (entrada) - ponteiro para a raiz da
 *                        rvore sobre a qual ser
 *                        efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
static void CaminhamentoInfixoRec( tNoBB2 *raiz, tOperacao op,
                                   FILE *stream )
{
   if (raiz) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixoRec( raiz->filho[ESQUERDA], op, stream );

      op(raiz->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixoRec( raiz->filho[DIREITA], op, stream );
   }
}

/****
 *
 * ProfundidadeABBRec(): Calcula a profundidade de uma
 *                       rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A profundidade da referida rvore
 *
 ****/
static int ProfundidadeABBRec(tNoBB2 *raiz)
{
   int profEsq, profDir;

   if (!raiz) {
      return 0; /* rvore vazia tem profundidade zero */
   } else {
        /* Calcula a profundidade da subrvore esquerda */
      profEsq = 1 + ProfundidadeABBRec(raiz->filho[ESQUERDA]);

        /* Calcula a profundidade da subrvore direita */
      profDir = 1 + ProfundidadeABBRec(raiz->filho[DIREITA]);

        /* A profundidade da rvore  a maior */
      return profEsq > profDir ? profEsq : profDir;
   }
}

/****
 *
 * EhArvoreBB2(): Verifica se uma rvore binria 
 *                realmente rvore binria de busca
 *
 * Parmetros:
 *      raiz (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: 1, se a rvore for uma rvore binria de busca;
 *          0, em caso contrrio
 *
 ****/
static int EhArvoreBB2(tNoBB2 *raiz)
{
   int    comparaEsq,
          comparaDir;
   tNoBB2 *noEsquerda,
          *noDireita;

   if (!raiz) {
      return 1;
   } else {
      noEsquerda = raiz->filho[ESQUERDA];
      noDireita = raiz->filho[DIREITA];

         /* Compara a chave do filho esquerdo */
         /* com a chave do seu pai            */
      if (noEsquerda) {
         comparaEsq = strcmp( noEsquerda->conteudo.chave,
                              raiz->conteudo.chave );
      }

         /* Compara a chave do filho direito */
         /* com a chave do seu pai           */
      if (noDireita) {
         comparaDir = strcmp( noDireita->conteudo.chave,
                              raiz->conteudo.chave );
      }

         /* ??? */
      if ( (noEsquerda && comparaEsq >= 0) ||
           (noDireita && comparaDir <= 0) ) {
         return EhArvoreBB2(noEsquerda) && EhArvoreBB2(noDireita);
      }
   }

   return 0;
}

/********************* Funes Globais ********************/

/****
 *
 * CriaArvoreBB2(): Cria uma rvore binria de busca vazia
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da rvore criada
 *
 ****/
tArvoreBB2 CriaArvoreBB2(void)
{
   tArvoreBB2 arvore;

      /* Aloca espao para a estrutura que armazena a rvore */
   arvore = malloc(sizeof(struct rotArvoreBB2));

      /* Garante que a alocao realmente ocorreu */
   ASSEGURA(arvore, "Nao foi possivel alocar arvore");

   arvore->raiz = NULL;
   arvore->nNos = 0;

   return arvore;
}

/****
 *
 * EstaVaziaArvoreBB2(): Verifica se uma rvore binria de busca
 *                       est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreBB2(tArvoreBB2 arvore)
{
   return arvore->nNos == 0;
}

/****
 *
 * DestroiArvoreBB2(): Destri uma rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada/sada) - ponteiro para a raiz da rvore
 *                              que ser destruda
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBB2(tArvoreBB2 arvore)
{
   tNoBB2 *noCorrente = arvore->raiz;
   tNoBB2 *q;

      /* Gira os filhos esquerdos de modo que  */
      /* a operao possa ser tratada como uma */
      /* destruio de lista encadeada         */
   while (noCorrente) {
      if (!noCorrente->filho[ESQUERDA]) {
            /* Como no h filho esquerdo, remove-se */
            /* o n corrente e passa-se adiante      */
         q = noCorrente->filho[DIREITA];
         free(noCorrente);
      } else {
            /* Gira o filho esquerdo e checa novamente */
         q = noCorrente->filho[ESQUERDA];
         noCorrente->filho[ESQUERDA] = q->filho[DIREITA];
         q->filho[DIREITA] = noCorrente;
      }

      noCorrente = q;
   }

   free(arvore);
}

/****
 *
 * BuscaArvoreBB2(): Retorna o endereo do contedo do n
 *                   que possui uma determinada chave numa
 *                   rvore binria de busca
 *
 * Parmetros:
 *      *arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tConteudo *BuscaArvoreBB2(tArvoreBB2 arvore, tChave chave)
{
   tNoBB2 *noCorrente = arvore->raiz;
   int    compara;

   while (noCorrente) {
         /* Compara a chave do n corrente */
         /* com a chave de busca           */
      compara = strcmp(noCorrente->conteudo.chave, chave);

         /* Verifica se a chave foi encontrada */
      if (!compara)
         break; /* Chave encontrada */

         /* Desce na direo indicada pelo */
         /* valor da varivel 'compara'    */
      noCorrente = noCorrente->filho[compara < 0];
   }

   return noCorrente ? &noCorrente->conteudo : NULL;
}

/****
 *
 * InsereArvoreBB2(): Insere um novo n numa rvore binria
 *                    de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore binria de busca na qual ser
 *                        feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: 1, se a insero for bem sucedida
 *          0, em caso contrrio
 *
 ****/
int InsereArvoreBB2( tArvoreBB2 arvore,
                     const tConteudo *conteudo )
{
   tNoBB2 *p;

      /* De fato, quem faz a insero  a funo InsereABBRec()*/
   p = InsereABBRec(arvore->raiz, conteudo);

      /* Altera a raiz da rvore apenas se ocorreu insero */
   if (p) {
         /* Altera a raiz da rvore para refletir a insero */
      arvore->raiz = p;

      ++arvore->nNos; /* Mais um n foi inserido */
   }

   return p ? 1 : 0;
}

/****
 *
 * RemoveArvoreBB2(): Remove de uma rvore binria de busca
 *                    um n que contm a chave especificada
 *                    como parmetro
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro que representa a rvore
 *                        na qual ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: 1, se a remoo for bem sucedida
 *          0, em caso contrrio (i.e., se a chave no
 *          for encontrada)
 *
 ****/
int RemoveArvoreBB2(tArvoreBB2 arvore, tChave chave)
{
   int houveRemocao = 0; /* Indicar se houve remoo */

      /* Chama a funo que realmente efetua a remoo */
   arvore->raiz = RemoveABBRec( arvore->raiz, chave,
                                &houveRemocao );

   return houveRemocao;
}

/****
 *
 * AlteraConteudoArvoreBB2(): Altera o contedo efetivo de um
 *                            n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreBB2( tArvoreBB2 arvore, tChave chave,
                            tConteudo *conteudo )
{
   tConteudo *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreBB2(arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /*                                            */
      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
      /*                                            */

      /* Verifica se a chave antiga se encontra na rvore */
   if (!BuscaArvoreBB2(arvore, chave)) {
      return 1; /* Chave no foi encontrada */
   }

      /* Verifica se a chave nova j se encontra na rvore */
   if (BuscaArvoreBB2(arvore, conteudo->chave)) {
      return 1; /* Chave nova j existe na rvore */
   }

      /* Remove o n contendo a chave antiga */
   ASSEGURA( RemoveArvoreBB2(arvore, chave),
             "Erro na alteracao de um no" );

   ASSEGURA( InsereArvoreBB2(arvore, conteudo),
             "Erro na alteracao de um no");

   return 0;
}

/****
 *
 * NumeroNosArvoreBB2(): Obtm o nmero de ns de uma rvore
 *                       binria de busca
 *
 * Parmetros:
 *     *arvore (entrada) - a rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreBB2(tArvoreBB2 arvore)
{
   return arvore->nNos;
}

/****
 *
 * CaminhamentoInfixo(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixo( tArvoreBB2 arvore, tOperacao op,
                         FILE *stream )
{
   CaminhamentoInfixoRec(arvore->raiz, op, stream);
}

/****
 *
 * ProfundidadeArvoreBB2(): Calcula a profundidade de uma
 *                          rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A profundidade da referida rvore
 *
 ****/
int ProfundidadeArvoreBB2(tArvoreBB2 arvore)
{
   return ProfundidadeABBRec(arvore->raiz);
}

/****
 *
 * MenorChaveArvoreBB2(): Obtm o contedo do n que possui
 *                        a menor chave armazenada numa rvore
 *                        binria de busca
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MenorChaveArvoreBB2(tArvoreBB2 arvore)
{
   tNoBB2 *p = arvore->raiz,
          *q; /* Apontar para o n que */
              /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->filho[ESQUERDA];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChaveArvoreBB2(): Obtm o contedo do n que possui
 *                        a maior chave armazenada numa rvore
 *                        binria de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MaiorChaveArvoreBB2(tArvoreBB2 arvore)
{
   tNoBB2 *p = arvore->raiz,
         *q; /* Apontar para o n que */
             /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->filho[DIREITA];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}

/****
 *
 * AvaliaArvoreBB2(): Funo acionadora de EhArvoreBB2()
 *
 * Parmetros:
 *      arvore (entrada) - ponteiro que representa a rvore
 *
 * Retorno: 1, se a rvore for uma rvore binria de busca;
            0, em caso contrrio
 *
 ****/
void AvaliaArvoreBB2(tArvoreBB2 arvore)
{
   if (EhArvoreBB2(arvore->raiz)) {
      puts("\nEssa nao e' uma arvore binaria de busca <<<<");
   } else {
      puts("\nEssa e' uma legitima arvore binaria de busca <<<<");
   }
}
