/****
 *
 * Arquivo: ArvoreAVL.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo ArvoreAVL
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada    */
#include <stdlib.h>   /* Funo exit()               */
#include <string.h>   /* Funes strXXX() e memXXX() */
#include "ArvoreAVL.h" /* Interface deste mdulo     */
#include "ASSEGURA.h"  /* Macro ASSEGURA             */

/****************** Definies de Macros ******************/

   /* Calcula o maior de dois nmeros */
#define MAIOR(a, b) (((a) > (b)) ? (a) : (b))

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

/****
 *
 * NovoNoAVL(): Constri um n de uma rvore AVL
 *
 * Parmetros:
 *       conteudo (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tNoAVL *NovoNoAVL(const tCEP_Ind *conteudo)
{
   tNoAVL *no;

      /* Tenta alocar o novo n */
   no = malloc(sizeof(tNoAVL));

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = *conteudo;
   no->altura = 1;  /* N novo  acrescentado como folha */
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * BalanceamentoAVL(): Calcula o balanceamento de um n
 *
 * Parmetros:
 *       pNo (entrada) - endereo do n
 *
 * Retorno: O balanceamento do n
 *
 ****/
static int BalanceamentoAVL(tArvoreAVL pNo)
{
   return pNo ? AlturaAVL(pNo->esquerda) - AlturaAVL(pNo->direita) : 0;
}

/****
 *
 * MenorNoAVL(): Encontra o menor n da vore cuja raiz  recebida como parmetro
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da vore
 *
 * Retorno: Endereo do menor n da rvore
 *
 ****/
static tArvoreAVL MenorNoAVL(tArvoreAVL arvore)
{
   ASSEGURA(arvore, "Arvore vazia nao tem menor no");

      /* Encontra o n mais  esquerda da rvore */
   while (arvore->esquerda)
      arvore = arvore->esquerda;

   return arvore;
}

/****
 *
 * RotacaoEsquerdaAVL(): executa uma rotao esquerda sobre uma rvore AVL
 *
 * Parmetros:
 *      x (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e as variveis locais so
 *             rotuladas de acordo com a figura no texto
 *
 ****/
static tArvoreAVL RotacaoEsquerdaAVL(tArvoreAVL x)
{
   tArvoreAVL y = x->direita,
              A2 = y->esquerda;

      /* Executa a rotao */
   y->esquerda = x;
   x->direita = A2;

      /* Atualiza as alturas */
   x->altura = MAIOR(AlturaAVL(x->esquerda), AlturaAVL(x->direita)) + 1;
   y->altura = MAIOR(AlturaAVL(y->esquerda), AlturaAVL(y->direita)) + 1;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * RotacaoDireitaAVL(): executa uma rotao direita sobre uma rvore AVL
 *
 * Parmetros:
 *      y (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e as variveis locais so
 *             rotuladas de acordo com a figura no texto
 *
 ****/
static tArvoreAVL RotacaoDireitaAVL(tArvoreAVL y)
{
   tArvoreAVL x = y->esquerda,
              A2 = x->direita;

      /* Executa a rotao */
   x->direita = y;
   y->esquerda = A2;

      /* Atualiza as alturas */
   y->altura = MAIOR(AlturaAVL(y->esquerda),
                     AlturaAVL(y->direita)) + 1;
   x->altura = MAIOR(AlturaAVL(x->esquerda),
                     AlturaAVL(x->direita)) + 1;

      /* Retorna a nova raiz */
   return x;
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaArvoreAVL(): Inicia uma rvore AVL
 *
 * Parmetros: *arvore (sada) - a arvore que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaArvoreAVL(tArvoreAVL *arvore)
{
   *arvore = NULL;
}

/****
 *
 * NumeroNosArvoreAVL(): Calcula o nmero de ns de uma rvore AVL
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreAVL(tArvoreAVL arvore)
{
      /* Se a rvore for vazia seu nmero de ns  0 */
   if (!arvore)
      return 0;

      /* Leva em considerao a raiz da rvore e os */
      /* ns das subrvores esquerda e direita      */
   return 1 + NumeroNosArvoreAVL(arvore->esquerda) +
              NumeroNosArvoreAVL(arvore->direita);
}

/****
 *
 * AlturaAVL(): Calcula a altura de uma rvore AVL
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A altura da rvore recebida como parmetro
 *
 ****/
int AlturaAVL(tArvoreAVL arvore)
{
   return arvore ? arvore->altura : 0;
}

/****
 *
 * BuscaArvoreAVL(): Retorna o endereo do contedo do n que
 *                   possui uma determinada chave numa rvore
 *                   binria de busca
 *
 * Parmetros:
 *       arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaArvoreAVL(tArvoreAVL arvore, tCEP chave)
{
   int compara; /* Resultado da comparao de duas chaves */

      /* Enquanto 'arvore' no assume NULL ou a chave de */
      /* um n no casa com 'chave', a busca prossegue   */
   while (arvore) {
         /* Compara a chave de busca com */
         /* a chave do n corrente       */
      compara = strcmp(chave, arvore->conteudo.chave);

         /* Verifica se a chave foi encontrada */
      if (!compara) {
         return &arvore->conteudo; /* Chave encontrada */
      }

         /* Se a chave de busca for menor do a chave do n   */
         /* corrente, desce-se pela subrvore esquerda. Caso */
         /* contrrio, desce-se pela subrvore direita.      */
      arvore = compara < 0 ? arvore->esquerda : arvore->direita;
   }

      /* A chave no foi encontrada */
   return NULL;
}

/****
 *
 * InsereAVL(): Insere um novo n numa rvore AVL
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *                        na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Endereo da raiz da rvore, se a insero ocorrer;
 *          NULL, se no houver insero porque a chave j se
 *          encontra na rvore
 *
 *
 ****/
tArvoreAVL InsereAVL(tArvoreAVL arvore, const tCEP_Ind *conteudo)
{
   int bal, /* BalanceamentoAVL de um n */
       compara; /* Comparao de duas chaves */

      /* Se a rvore estiver vazia, o novo n ser a raiz da rvore */
   if (!arvore)
      return NovoNoAVL(conteudo);

      /* Compara a chave do n a ser inserido com a chave do n corrente */
   compara = strcmp(conteudo->chave, arvore->conteudo.chave);

   if (compara < 0)
      arvore->esquerda = InsereAVL(arvore->esquerda, conteudo);
   else if (compara > 0)
      arvore->direita = InsereAVL(arvore->direita, conteudo);
   else
      return NULL; /* Chave j existe */

      /* Atualiza a altura desta arvore */
   arvore->altura = MAIOR(AlturaAVL(arvore->esquerda),
                          AlturaAVL(arvore->direita)) + 1;

      /* Obtm o balanceamento da rvore */
   bal = BalanceamentoAVL(arvore);

      /* Examina os 4 casos possveis de desbalanceamento */

   if (bal > 1) {
      /* Desbalanceamento esquerda-esquerda ou esquerda-direita */

      compara = strcmp(conteudo->chave, arvore->esquerda->conteudo.chave);

      if (compara < 0) /* Caso esquerda-esquerda */
         return RotacaoDireitaAVL(arvore);

      if (compara > 0) { /* Caso esquerda-direita */
         arvore->esquerda = RotacaoEsquerdaAVL(arvore->esquerda);
         return RotacaoDireitaAVL(arvore);
      }
   }

   if (bal < -1) {
      /* Desbalanceamento direita-direita ou direita-esquerda */

      compara = strcmp(conteudo->chave, arvore->direita->conteudo.chave);

      if (compara > 0) /* Caso direita-direita */
         return RotacaoEsquerdaAVL(arvore);

      if (compara < 0) { /* Caso direita-esquerda */
         arvore->direita = RotacaoDireitaAVL(arvore->direita);
         return RotacaoEsquerdaAVL(arvore);
      }
   }

   return arvore; /* No ocorreu desbalanceamento */
}

/****
 *
 * RemoveAVL(): Remove de uma rvore AVL um n que contm
 *              a chave especificada como parmetro
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro que representa a rvore na
 *                        qual ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: O primeiro parmetro, se a remoo for bem sucedida,
 *          NULL, se a rvore estiver vazia ou se ela se tornar
 *          tornar vazia aps a remoo
 *
 ****/
tArvoreAVL RemoveAVL(tArvoreAVL arvore, tCEP chave)
{
   int        bal,     /* BalanceamentoAVL de um n    */
              compara; /* Comparao de duas chaves */
   tArvoreAVL p;       /* Ponteiro auxiliar */

   if (!arvore)  /* Se a rvore estiver vazia, no h remoo */
      return NULL;

      /* Compara a chave do n a ser removido com a chave do n corrente */
   compara = strcmp(chave, arvore->conteudo.chave);

   if ( compara < 0 )
         /* O n a ser removido se encontra na subrvore esquerda */
      arvore->esquerda = RemoveAVL(arvore->esquerda, chave);
   else if( compara > 0 )
         /* O n a ser removido se encontra na subrvore direita */
      arvore->direita = RemoveAVL(arvore->direita, chave);
   else {
      /* Chave encontrada. Este  o n a ser removido */

         /* Se o n possui apenas um filho ou no possui nenhum */
         /* filho, p apontar para o n a ser removido          */
      if( (!arvore->esquerda) || (!arvore->direita) ) {
         p = arvore->esquerda ? arvore->esquerda : arvore->direita;

         if(!p) { /* N no possui filhos */
            p = arvore;
            arvore = NULL;
         } else /* N possui um nico filho */
               /* Copia o contedo do filho nico para o pai */
            *arvore = *p;

         free(p); /* Libera o devido n */
      } else { /* N possui dois filhos */
            /* Obtm o sucessor em ordem infixa do n a ser removido */
         p = MenorNoAVL(arvore->direita);

            /* Copia o contedo do sucessor para o n que seria removido */
         arvore->conteudo = p->conteudo;

            /* Remove o referido sucessor */
         arvore->direita = RemoveAVL(arvore->direita, p->conteudo.chave);
      }
   }

      /* Se a rvore s tinha um n e ele foi removido, o servio est completo */
   if (!arvore)
     return arvore;

      /* Atualiza a altura do n corrente */
   arvore->altura = MAIOR(AlturaAVL(arvore->esquerda),
                          AlturaAVL(arvore->direita)) + 1;

      /* Obtm o balanceamento desta rvore para  */
      /* verificar se ela se tornou desbalanceada */
   bal = BalanceamentoAVL(arvore);

      /* Examina os 4 casos possveis de desbalanceamento */

      /* Caso esquerda-esquerda */
   if (bal > 1 && BalanceamentoAVL(arvore->esquerda) >= 0)
      return RotacaoDireitaAVL(arvore);

      /* Caso esquerda-direita */
   if (bal > 1 && BalanceamentoAVL(arvore->esquerda) < 0) {
      arvore->esquerda =  RotacaoEsquerdaAVL(arvore->esquerda);
      return RotacaoDireitaAVL(arvore);
   }

      /* Caso direita-direita */
   if (bal < -1 && BalanceamentoAVL(arvore->direita) <= 0)
      return RotacaoEsquerdaAVL(arvore);

      /* Caso direita-esquerda */
   if (bal < -1 && BalanceamentoAVL(arvore->direita) > 0) {
      arvore->direita = RotacaoDireitaAVL(arvore->direita);
      return RotacaoEsquerdaAVL(arvore);
   }

   return arvore; /* No ocorreu desbalanceamento */
}

/****
 *
 * EstaVaziaArvoreAVL(): Verifica se uma rvore binria
 *                       est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreAVL(tArvoreAVL arvore)
{
   return arvore == NULL;
}

/****
 *
 * AlteraConteudoArvoreAVL(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreAVL( tArvoreAVL *arvore, tCEP chave,
                             tCEP_Ind *conteudo )
{
   tCEP_Ind *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreAVL(*arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /*                                            */
      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
      /*                                            */

      /* Verifica se a chave antiga se encontra na rvore */
   if (!BuscaArvoreAVL(*arvore, chave)) {
      return 1; /* Chave no foi encontrada */
   }

      /* Verifica se a chave nova j se encontra na rvore */
   if (BuscaArvoreAVL(*arvore, conteudo->chave)) {
      return 1; /* Chave nova j existe na rvore */
   }

      /* Remove o n contendo a chave antiga */
   *arvore = RemoveAVL(*arvore, chave);

   *arvore = InsereAVL( *arvore, conteudo );

   ASSEGURA(*arvore, "Erro na alteracao de um no");

   return 0;
}

/****
 *
 * DestroiArvoreAVL(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz
 *                               da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreAVL(tArvoreAVL arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreAVL(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreAVL(arvore->direita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * CaminhamentoInfixoAVL(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixoAVL( tArvoreAVL arvore, tOperacao op,
                         FILE *stream )
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixoAVL(arvore->esquerda, op, stream);

      op(arvore->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixoAVL(arvore->direita, op, stream);
   }
}

/****
 *
 * MenorChaveAVL(): Obtm o contedo do n que possui a menor
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MenorChaveAVL(tArvoreAVL p)
{
   tArvoreAVL q; /* Apontar para o n que */
                /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->esquerda;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChaveAVL(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MaiorChaveAVL(tArvoreAVL p)
{
   tArvoreAVL q; /* Apontar para o n que */
                /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->direita;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}
