/****
 *
 * Arquivo: ArvoreAVL.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo ArvoreAVL2
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>      /* Funes de Entrada/Sada    */
#include <stdlib.h>     /* Funo exit()               */
#include <string.h>     /* Funes strXXX() e memXXX() */
#include "ArvoreAVL2.h" /* Interface deste mdulo      */
#include "ASSEGURA.h"   /* Macro ASSEGURA              */

/****************** Definies de Macros ******************/

   /* Calcula o maior de dois nmeros */
#define MAIOR(a, b) (((a) > (b)) ? (a) : (b))

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

/****
 *
 * NovoNo(): Constri um n de uma rvore AVL
 *
 * Parmetros:
 *       conteudo (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tArvoreAVL2 NovoNo(const tConteudo *conteudo)
{
   tArvoreAVL2 pNo;

      /* Tenta alocar o novo n */
   pNo = malloc(sizeof(tNoAVL2));

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(pNo, "Erro: Nao foi possivel alocar no'");

   pNo->conteudo = *conteudo; /* Preenche o contedo do n */
   pNo->altura = 1;  /* N novo  acrescentado como folha */

      /* Este n ainda no tem filho */
   pNo->filho[0] = pNo->filho[1] = NULL;

   return pNo;
}

/****
 *
 * AlturaAVL(): Calcula a altura de uma rvore AVL
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A altura da rvore recebida como parmetro
 *
 ****/
static int AlturaAVL(tArvoreAVL2 arvore)
{
   return arvore ? arvore->altura : 0;
}

/****
 *
 * Balanceamento(): Calcula o balanceamento de um n
 *
 * Parmetros:
 *       pNo (entrada) - endereo do n
 *
 * Retorno: O balanceamento do n
 *
 ****/
static int Balanceamento(tArvoreAVL2 pNo)
{
   return pNo ? AlturaAVL(pNo->filho[0]) -
                AlturaAVL(pNo->filho[1])
              : 0;
}

/****
 *
 * MenorNo(): Encontra o menor n da vore cuja raiz 
 *            recebida como parmetro
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da vore
 *
 * Retorno: Endereo do menor n da rvore
 *
 ****/
static tArvoreAVL2 MenorNo(tArvoreAVL2 arvore)
{
   ASSEGURA(arvore, "Arvore vazia nao tem menor no");

      /* Encontra o n mais  esquerda da rvore */
   while (arvore->filho[0])
      arvore = arvore->filho[0];

   return arvore;
}

/****
 *
 * Rotacao(): executa uma rotao esquerda ou direta sobre um
 *            n de uma rvore AVL
 *
 * Parmetros:
 *      x (entrada) - raiz da rvore que sofrer rotao
 *      dir (entrada) - sentido da rotao (0 = esquerda;
 *                      1 = direita)
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e as variveis locais so
 *             rotuladas de acordo com a figura no texto
 *
 ****/
static tArvoreAVL2 Rotacao(tArvoreAVL2 x, int dir)
{
   tArvoreAVL2 y = x->filho[!dir],
               A2 = y->filho[dir];

   ASSEGURA( dir == 0 || dir == 1,
             "Sentido de rotacao incorreto" );

      /* Executa a rotao */
   y->filho[dir] = x;
   x->filho[!dir] = A2;

      /* Atualiza as alturas */
   x->altura = MAIOR(AlturaAVL(x->filho[dir]),
                     AlturaAVL(x->filho[!dir])) + 1;
   y->altura = MAIOR(AlturaAVL(y->filho[dir]),
                     AlturaAVL(y->filho[!dir])) + 1;

      /* Retorna a nova raiz */
   return y;
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaArvoreAVL(): Inicia uma rvore AVL
 *
 * Parmetros: *arvore (sada) - a arvore que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaArvoreAVL(tArvoreAVL2 *arvore)
{
   *arvore = NULL;
}

/****
 *
 * NumeroNosArvoreAVL(): Calcula o nmero de ns de uma rvore
 *                      AVL
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreAVL(tArvoreAVL2 arvore)
{
      /* Se a rvore for vazia seu nmero de ns  0 */
   if (!arvore) {
      return 0;
   }

      /* Leva em considerao a raiz da rvore e os */
      /* ns das subrvores esquerda e direita      */
   return 1 + NumeroNosArvoreAVL(arvore->filho[0]) +
              NumeroNosArvoreAVL(arvore->filho[1]);
}

/****
 *
 * ProfundidadeArvoreAVL(): Calcula a profundidade de uma
 *                         rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A profundidade da referida rvore
 *
 ****/
int ProfundidadeArvoreAVL(tArvoreAVL2 arvore)
{
      /* Profundidade e altura so a mesma coisa */
   return AlturaAVL(arvore);
}

/****
 *
 * BuscaArvoreAVL(): Retorna o endereo do contedo do n que
 *                   possui uma determinada chave numa rvore
 *                   binria de busca
 *
 * Parmetros:
 *       arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tConteudo *BuscaArvoreAVL(tArvoreAVL2 arvore, tChave chave)
{
   int compara; /* Resultado da comparao de duas chaves */

      /* Enquanto 'arvore' no assume NULL ou a chave de */
      /* um n no casa com 'chave', a busca prossegue   */
   while (arvore) {
         /* Compara a chave de busca com */
         /* a chave do n corrente       */
      compara = strcmp(chave, arvore->conteudo.chave);

         /* Verifica se a chave foi encontrada */
      if (!compara) {
         return &arvore->conteudo; /* Chave encontrada */
      }

         /* Se a chave de busca for menor do a chave do n   */
         /* corrente, desce-se pela subrvore esquerda. Caso */
         /* contrrio, desce-se pela subrvore direita.      */
      arvore = compara < 0 ? arvore->filho[0] : arvore->filho[1];
   }

      /* A chave no foi encontrada */
   return NULL;
}

/****
 *
 * InsereAVL(): Insere um novo n numa rvore AVL
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *                        na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: Endereo da raiz da rvore, se a insero ocorrer;
 *          NULL, se no houver insero porque a chave j se
 *          encontra na rvore
 *
 *
 ****/
tArvoreAVL2 InsereAVL( tArvoreAVL2 arvore,
                       const tConteudo *conteudo )
{
   int bal, /* Balanceamento de um n */
       compara; /* Comparao de duas chaves */

      /*********************/
      /* Efetua a insero */
      /*********************/

      /* Se a rvore estiver vazia, o novo */
      /* n ser a raiz da rvore          */
   if (!arvore)
      return NovoNo(conteudo);

      /* Compara a chave do n a ser inserido */
      /* com a chave do n corrente           */
   compara = strcmp(conteudo->chave, arvore->conteudo.chave);

   if (compara < 0)
      arvore->filho[0]  = InsereAVL(arvore->filho[0], conteudo);
   else if (compara > 0)
      arvore->filho[1] = InsereAVL(arvore->filho[1], conteudo);
   else
      return NULL; /* Chave j existe */

      /* Atualiza a altura desta arvore (???) */
   arvore->altura = MAIOR(AlturaAVL(arvore->filho[0]),
                          AlturaAVL(arvore->filho[1])) + 1;

      /* Obtm o balanceamento desta rvore para  */
      /* verificar se ela se tornou desbalanceada */
   bal = Balanceamento(arvore);

      /* Examina os 4 casos possveis de desbalanceamento */

   if (bal > 1) {
      compara = strcmp( conteudo->chave,
                        arvore->filho[0]->conteudo.chave );

         /* Caso esquerda-esquerda */
      if (compara < 0)
         return Rotacao(arvore, 1);

         /* Caso esquerda-direita */
      if (compara > 0) {
         arvore->filho[0] = Rotacao(arvore->filho[0], 0);
         return Rotacao(arvore, 1);
      }
   }

   if (bal < -1) {
      compara = strcmp( conteudo->chave,
                        arvore->filho[1]->conteudo.chave );

         /* Caso direita-direita */
      if (compara > 0)
         return Rotacao(arvore, 0);

         /* Caso direita-esquerda */
      if (compara < 0) {
         arvore->filho[1] = Rotacao(arvore->filho[1], 1);
         return Rotacao(arvore, 0);
      }
   }

      /* Retorna o parmetro inalterado */
   return arvore;
}

/****
 *
 * RemoveAVL(): Remove de uma rvore AVL um n que contm
 *              a chave especificada como parmetro
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro que representa a rvore na
 *                        qual ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: O primeiro parmetro, se a remoo for bem sucedida,
 *          NULL, se a rvore estiver vazia ou se ela se tornar
 *          tornar vazia aps a remoo
 *
 ****/
tArvoreAVL2 RemoveAVL(tArvoreAVL2 arvore, tChave chave)
{
   int         bal, /* Balanceamento de um n */
               compara; /* Comparao de duas chaves */
   tArvoreAVL2 p; /* Ponteiro auxiliar */

      /* Se a rvore estiver vazia, no h remoo */
   if (!arvore)
      return NULL;

      /* Compara a chave do n a ser removido */
      /* com a chave do n corrente           */
   compara = strcmp(chave, arvore->conteudo.chave);

   if ( compara < 0 )
         /* O n a ser removido se encontra */
         /* na subrvore esquerda           */
      arvore->filho[0] = RemoveAVL(arvore->filho[0], chave);
   else if( compara > 0 )
         /* O n a ser removido se encontra */
         /* na subrvore direita            */
      arvore->filho[1] = RemoveAVL(arvore->filho[1], chave);
   else {
      /* Chave encontrada. Este  o n a ser removido */

         /* O n possui apenas um filho ou no possui nenhum */
         /* filho. p apontar para o n a ser removido.      */
      if( (!arvore->filho[0]) || (!arvore->filho[1]) ) {
         p = arvore->filho[0] ? arvore->filho[0]
                              : arvore->filho[1];

         if(!p) { /* N no possui filhos */
            p = arvore;
            arvore = NULL;
         } else /* N possui um nico filho */
               /* Copia o contedo do filho nico para o pai */
            *arvore = *p;

         free(p); /* Libera o devido n */
      } else { /* N possui dois filhos */
            /* Obtm o sucessor em ordem infixa */
            /* do n a ser removido             */
         p = MenorNo(arvore->filho[1]);

            /* Copia o contedo do sucessor */
            /* para o n que seria removido */
         arvore->conteudo = p->conteudo;

            /* Remove o referido sucessor */
         arvore->filho[1] = RemoveAVL( arvore->filho[1],
                                       p->conteudo.chave );
      }
   }

      /* Se a rvore s tinha um n e ele foi */
      /* removido, o servio est completo    */
   if (!arvore)
     return arvore;

      /* Atualiza a altura do n corrente */
   arvore->altura = MAIOR(AlturaAVL(arvore->filho[0]),
                          AlturaAVL(arvore->filho[1])) + 1;

      /* Obtm o balanceamento desta rvore para  */
      /* verificar se ela se tornou desbalanceada */
   bal = Balanceamento(arvore);

      /* Examina os 4 casos possveis de desbalanceamento */

      /* Caso esquerda-esquerda */
   if (bal > 1 && Balanceamento(arvore->filho[0]) >= 0)
      return Rotacao(arvore, 1);

      /* Caso esquerda-direita */
   if (bal > 1 && Balanceamento(arvore->filho[0]) < 0) {
      arvore->filho[0] =  Rotacao(arvore->filho[0], 0);
      return Rotacao(arvore, 1);
   }

      /* Caso direita-direita */
   if (bal < -1 && Balanceamento(arvore->filho[1]) <= 0)
      return Rotacao(arvore, 0);

      /* Caso direita-esquerda */
   if (bal < -1 && Balanceamento(arvore->filho[1]) > 0) {
      arvore->filho[1] = Rotacao(arvore->filho[1], 1);
      return Rotacao(arvore, 0);
   }

      /* No ocorreu desbalanceamento */
   return arvore;
}

/****
 *
 * EstaVaziaArvoreAVL(): Verifica se uma rvore binria
 *                       est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreAVL(tArvoreAVL2 arvore)
{
   return arvore == NULL;
}

/****
 *
 * AlteraConteudoArvoreAVL(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreAVL( tArvoreAVL2 *arvore, tChave chave,
                             tConteudo *conteudo )
{
   tConteudo *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreAVL(*arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /*                                            */
      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
      /*                                            */

      /* Verifica se a chave antiga se encontra na rvore */
   if (!BuscaArvoreAVL(*arvore, chave)) {
      return 1; /* Chave no foi encontrada */
   }

      /* Verifica se a chave nova j se encontra na rvore */
   if (BuscaArvoreAVL(*arvore, conteudo->chave)) {
      return 1; /* Chave nova j existe na rvore */
   }

      /* Remove o n contendo a chave antiga */
   *arvore = RemoveAVL(*arvore, chave);

   *arvore = InsereAVL( *arvore, conteudo );

   ASSEGURA(*arvore, "Erro na alteracao de um no");

   return 0;
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz
 *                               da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBin(tArvoreAVL2 arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBin(arvore->filho[0]);

         /* Destri subrvore direita */
      DestroiArvoreBin(arvore->filho[1]);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * CaminhamentoInfixo(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixo( tArvoreAVL2 arvore, tOperacao op,
                         FILE *stream )
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixo(arvore->filho[0], op, stream);

      op(arvore->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixo(arvore->filho[1], op, stream);
   }
}

/****
 *
 * MenorChave(): Obtm o contedo do n que possui a menor
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MenorChave(tArvoreAVL2 p)
{
   tArvoreAVL2 q; /* Apontar para o n que */
                  /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->filho[0];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChave(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MaiorChave(tArvoreAVL2 p)
{
   tArvoreAVL2 q; /* Apontar para o n que */
                /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->filho[1];
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}
