/****
 *
 * Arquivo: ArvoreFunil.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2016
 * ltima modificao: 22/07/2016
 *
 * Descrio: Implementao do mdulo ArvoreFunil
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada    */
#include <stdlib.h>   /* Funo exit()               */
#include <string.h>   /* Funes strXXX() e memXXX() */
#include "ArvoreFunil.h" /* Interface deste mdulo   */
#include "ASSEGURA.h"    /* Macro ASSEGURA           */

/***************** Definies de Funes ******************/

/********************* Funes Locais *********************/

/****
 *
 * ConstroiNo(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *       item (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tArvoreFunil ConstroiNo(tConteudo item)
{
   tArvoreFunil no;

      /* Tenta alocar o novo n */
   no = malloc(sizeof(tNo));

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = item;
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * RotacaoDireita(): executa uma rotao direita sobre uma
 *                   rvore afunilada
 *
 * Parmetros:
 *      y (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e a varivel local so rotulados
 *             de acordo com a figura no texto, mas,
 *             diferentemente do que o ocorre no texto, x e y
 *             so ponteiros para ns e no ns.
 *
 ****/
static tArvoreFunil RotacaoDireita(tArvoreFunil x)
{
   tArvoreFunil y = x->esquerda; /* Guarda o endereo da nova */
                                 /* raiz aps a rotao       */

      /*                                             */
      /* A ordem das instrues a seguir  essencial */
      /*                                             */

      /* O filho direito de y torna-se filho esquerdo de x */
   x->esquerda = y->direita;

      /* x torna-se filho direito de y */
   y->direita = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * RotacaoEsquerda(): executa uma rotao esquerda sobre uma
 *                    rvore afunilada
 *
 * Parmetros:
 *      x (entrada) - raiz da rvore que sofrer rotao
 *
 * Retorno: Raiz da rvore aps a rotao
 *
 * Observao: O parmetro e a varivel local so rotulados
 *             de acordo com a figura no texto, mas,
 *             diferentemente do que o ocorre no texto, x e y
 *             so ponteiros para ns e no ns.
 *
 ****/
static tArvoreFunil RotacaoEsquerda(tArvoreFunil x)
{
   tArvoreFunil y = x->direita; /* Guarda o endereo da nova */
                                /* raiz aps a rotao       */

      /*                                             */
      /* A ordem das instrues a seguir  essencial */
      /*                                             */

      /* O filho esquerdo de y torna-se filho direito de x */
   x->direita = y->esquerda;

      /* x torna-se filho esquerdo de y */
   y->esquerda = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * MenorNo(): Encontra o menor n da vore cuja raiz 
 *            recebida como parmetro
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da vore
 *
 * Retorno: Endereo do menor n da rvore
 *
 ****/
static tArvoreFunil MenorNo(tArvoreFunil arvore)
{
   ASSEGURA(arvore, "Arvore vazia nao tem menor no");

      /* Encontra o n mais  esquerda da rvore */
   while (arvore->esquerda)
      arvore = arvore->esquerda;

   return arvore;
}

/****
 *
 * Afunila(): Afunila um n de uma rvore afunilada que
 *            contm uma chave especificada
 *
 * Parmetros:
 *      arvore (entrada/sada) - raiz da rvore que ser
 *                               afunilada
 *      chave (entrada) - chave do n que ser afunilado
 *
 * Retorno: Raiz da rvore aps o afunilamento
 *
 * Observao: Se a chave no for encontrada, o ltimo n
 *             acessado se tornar a raiz da rvore
 *
 ****/
static tArvoreFunil Afunila(tArvoreFunil arvore, tChave chave)
{
   int compara; /* Resultado de comparao de duas chaves */

      /* Se a rvore estiver vazia no h o que afunilar */
   if (!arvore)
      return arvore;

      /* Compara a chave do n a ser afunilado */
      /* com a chave da raiz                   */
   compara = strcmp(chave, arvore->conteudo.chave);

      /* Se a chave j estiver na raiz no h o que afunilar */
   if (!compara)
      return arvore;

      /* Verifica se a chave se encontra na subrvore esquerda */
   if (compara < 0) {
         /* Se a chave no se encontra na subrvore */
         /* esquerda, o servio est completo       */
      if (!arvore->esquerda)
         return arvore;

         /* Compara a chave do n a ser afunilado com */
         /* a chave da raiz da subrvore esquerda     */
      compara = strcmp(chave, arvore->esquerda->conteudo.chave);

      if (compara < 0) { /* Caso 2A: Efetua-se zig-zig */
            /* Primeiro, recursivamente, afunila-se o n */
            /* que contm a chave na subrvore esquerda  */
            /* da subrvore esquerda                     */
         arvore->esquerda->esquerda =
            Afunila(arvore->esquerda->esquerda, chave);

            /* Efetua a primeira rotao zig. A segunda */
            /* rotao ocorrer aps esta instruo if. */
         arvore = RotacaoDireita(arvore);
      } else if (compara > 0) { /* Caso 3A: Efetua-se zag-zig */
            /* Primeiro, recursivamente, afunila-se o n */
            /* que contm a chave na subrvore direita   */
            /* da subrvore esquerda                     */
         arvore->esquerda->direita =
            Afunila(arvore->esquerda->direita, chave);

            /* Efetua a rotao zag. A segunda rotao */
            /* ocorrer aps esta instruo if.        */
         if (arvore->esquerda->direita != NULL)
            arvore->esquerda=RotacaoEsquerda(arvore->esquerda);
      }

         /* Efetua uma rotao direita (zig) sobre a raiz */
         /* da rvore, a no ser que rvore tenha ficado  */
         /* sem filho esquerdo aps o afunilamento        */
      return !arvore->esquerda ? arvore: RotacaoDireita(arvore);
   } else { /* A chave se encontra na subrvore direita */
         /* Se a chave no se encontra na subrvore */
         /* direita, o servio est completo        */
      if (!arvore->direita)
         return arvore;

         /* Compara a chave do n a ser afunilado com */
         /* a chave da raiz da subrvore direita      */
      compara = strcmp(chave, arvore->direita->conteudo.chave);

      if (compara < 0) { /* Caso 3B: Efetua-se zig-zag */
            /* Primeiro, recursivamente, afunila-se o n */
            /* que contm a chave na subrvore esquerda */
            /* da subrvore direita                     */
         arvore->direita->esquerda =
            Afunila(arvore->direita->esquerda, chave);

            /* Efetua a rotao zig. A segunda rotao */
            /* ocorrer aps esta instruo if.        */
         if (arvore->direita->esquerda != NULL)
            arvore->direita = RotacaoDireita(arvore->direita);
      } else if (compara > 0) { /* Caso 2B: Zag-Zag */
            /* Primeiro, recursivamente, afunila-se o n */
            /* que contm a chave na subrvore direita */
            /* da subrvore direita                    */
         arvore->direita->direita =
            Afunila(arvore->direita->direita, chave);

            /* Efetua a primeira rotao zag. A segunda */
            /* rotao ocorrer aps esta instruo if. */
         arvore = RotacaoEsquerda(arvore);
      }

         /* Efetua uma rotao esquerda (zag) sobre a raiz */
         /* da rvore, a no ser que rvore tenha ficado   */
         /* sem filho direito aps o afunilamento          */
      return !arvore->direita ? arvore: RotacaoEsquerda(arvore);
   }
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaArvoreFunil(): Inicia uma rvore binria afunilada
 *
 * Parmetros: *arvore (sada) - a arvore que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaArvoreFunil(tArvoreFunil *arvore)
{
   *arvore = NULL;
}

/****
 *
 * NumeroNosArvoreFunil(): Calcula o nmero de ns de uma rvore
 *                      binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreFunil(tArvoreFunil arvore)
{
      /* Se a rvore for vazia seu nmero de ns  0 */
   if (!arvore) {
      return 0;
   }

      /* Leva em considerao a raiz da rvore e os */
      /* ns das subrvores esquerda e direita      */
   return 1 + NumeroNosArvoreFunil(arvore->esquerda) +
              NumeroNosArvoreFunil(arvore->direita);
}

/****
 *
 * ProfundidadeArvoreFunil(): Calcula a profundidade de uma
 *                         rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A profundidade da referida rvore
 *
 ****/
int ProfundidadeArvoreFunil(tArvoreFunil arvore)
{
   int profEsq, profDir;

   if (!arvore) {
      return 0; /* rvore vazia tem profundidade zero */
   } else {
        /* Calcula a profundidade da subrvore esquerda */
      profEsq = 1 + ProfundidadeArvoreFunil(arvore->esquerda);

        /* Calcula a profundidade da subrvore direita */
      profDir = 1 + ProfundidadeArvoreFunil(arvore->direita);

        /* A profundidade da rvore  a maior */
      return profEsq > profDir ? profEsq : profDir;
   }
}

/****
 *
 * BuscaArvoreFunil(): Retorna o endereo do contedo do n que
 *                     possui uma determinada chave numa rvore
 *                     afunilada
 *
 * Parmetros:
 *       arvore (entrada) - rvore que ser pesquisada
 *       chave (entrada) - chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tConteudo *BuscaArvoreFunil(tArvoreFunil *arvore, tChave chave)
{
      /* Verifica se a rvore est vazia */
   if (!*arvore) {
      return NULL; /* A rvore est vazia */
   }

      /* Afunila o n que contm a chave de busca */
   *arvore = Afunila(*arvore, chave);

      /* Se a chave foi encontrada, ela agora est na raiz */
   if (!strcmp(chave, (*arvore)->conteudo.chave)) {
      return &(*arvore)->conteudo;
   } else {
      return NULL;
   }
}

/****
 *
 * InsereNoArvoreFunil(): Insere um novo n numa rvore
 *                        afunilada
 *
 * Parmetros:
 *    *arvore (entrada/sada) - ponteiro para a raiz da rvore
 *                              na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: 1, se no houver insero porque a chave j existe
 *          0, se a insero ocorrer
 *
 ****/
int InsereNoArvoreFunil( tArvoreFunil *arvore,
                         tConteudo *conteudo )
{
   tArvoreFunil pNovoNo; /* Ponteiro para n  */
                         /* que ser inserido */
   int          compara; /* Resultado da comparao */
                         /* de duas chaves          */

      /* Se a rvore estiver vazia torna */
      /* o novo n sua raiz e retorna    */
   if (!*arvore) {
      *arvore = ConstroiNo(*conteudo);

      return 0;
   }

      /* Afunila o n que contm a chave do n a ser inserido */
   *arvore = Afunila(*arvore, conteudo->chave);

      /* Compara a chave do n a ser inserido com a chave que */
      /* se encontra na raiz da rvore aps o afunilamento    */
   compara = strcmp(conteudo->chave, (*arvore)->conteudo.chave);

      /* Verifica se a chave j existe */
   if (!compara) {
      return 1; /* A chave j existe */
   } else {
         /* Constri o novo n */
      pNovoNo = ConstroiNo(*conteudo);

         /* Faz com que o novo n se torne a raiz da rvore */
      if (compara < 0) {
         /*                               */
         /* A chave do novo n  menor do */
         /* que a chave da raiz atual     */
         /*                               */

            /* Faz o filho direito do novo */
            /* n apontar para a raiz      */
         pNovoNo->direita = *arvore;

            /* Faz o filho esquerdo do novo n apontar */
            /* para o filho esquerdo da raiz           */
         pNovoNo->esquerda = (*arvore)->esquerda;

            /* Torna nulo o filho esquerdo da raiz */
         (*arvore)->esquerda = NULL;
      } else {
         /*                               */
         /* A chave do novo n  maior do */
         /* que a chave da raiz atual.    */
         /*                               */

            /* Faz o filho esquerdo do novo */
            /* n apontar para a raiz       */
         pNovoNo->esquerda = *arvore;

            /* Faz o filho direito do novo n apontar */
            /* para o filho direito da raiz           */
         pNovoNo->direita = (*arvore)->direita;
         (*arvore)->direita = NULL;
      }

         /* O novo n passa a ser a nova raiz */
      *arvore = pNovoNo;
   }

   return 0; /* Insero foi OK */
}

/****
 *
 * RemoveNoArvoreFunil(): Remove de uma rvore afunilada um n
 *                        que contm a chave especificada como
 *                        parmetro
 *
 * Parmetros:
 *     arvore (entrada/sada) - endereo do ponteiro que
 *                              representa a rvore na qual
 *                              ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveNoArvoreFunil(tArvoreFunil *arvore, tChave chave)
{
   tArvoreFunil pE, /* Apontar para o filho esquerdo da raiz */
                pD, /* Apontar para o filho direito da raiz  */
                pS, /* Apontar para o sucessor da raiz */
                p;  /* Um ponteiro auxiliar */

      /* Verifica se a rvore est vazia */
   if (!*arvore) {
      return 1; /* No h o que remover */
   }

      /* Afunila o n que ser removido */
   *arvore = Afunila(*arvore, chave);

      /* Se a chave foi encontrada, ela agora est na raiz */
   if (strcmp(chave, (*arvore)->conteudo.chave)) {
      return 1; /* Chave no foi encontrada */
   }

   /*                                               */
   /* A chave foi encontrada e se encontra na raiz. */
   /* Portanto, a raiz da rvore ser removida.     */
   /*                                               */

      /* Verifica quantos filhos a raiz possui */
   if (!(*arvore)->esquerda && !(*arvore)->direita) {
         /* A raiz no tem filhos logo a rvore */
         /* ficar vazia aps a remoo         */
      free(*arvore);
      *arvore = NULL;
   } else if ((*arvore)->esquerda && (*arvore)->direita) {
      /*                        */
      /* A raiz tem dois filhos */
      /*                        */

         /* Faz pE apontar para o filho esquerdo da raiz */
      pE = (*arvore)->esquerda;

         /* Faz pD apontar para o filho direito da raiz */
      pD = (*arvore)->direita;

         /* A raiz j pode ser liberada */
      free(*arvore);

         /* Obtm o sucessor imediato da antiga raiz */
      pS = MenorNo(pD);

         /* Afunila o sucessor imediato da antiga raiz */
      pD = Afunila(pD, pS->conteudo.chave);

         /* Apesar de o sucessor ter sido movido para a raiz, */
         /* ele ainda deve ocupar o mesmo endereo em memria */
      ASSEGURA(pS == pD, "Sucessor nao esta' na raiz");

         /* O n apontado por pD no pode ter filho esquerdo */
      ASSEGURA(!pD->esquerda, "Sucessor tem filho esquerdo");

         /* Faz com que o filho esquerdo do sucessor */
         /* seja a rvore apontada por pE            */
      pD->esquerda = pE;

         /* Faz o ponteiro que representa a */
         /* rvore apontar para a nova raiz */
      *arvore = pD;
   } else {
      /*                                           */
      /* A raiz s tem um filho. Esse filho ser a */
      /* nova raiz e a raiz antiga ser removida.  */
      /*                                           */

         /* Guarda o endereo da raiz para remov-la depois */
      p = *arvore;

         /* Faz a raiz apontar para seu nico filho */
      *arvore = (*arvore)->esquerda ? (*arvore)->esquerda
                                    : (*arvore)->direita;

      free(p); /* Remove a raiz antiga */
   }

   return 0; /* A remoo foi bem sucedida */
}

/****
 *
 * EstaVaziaArvoreFunil(): Verifica se uma rvore binria
 *                      est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreFunil(tArvoreFunil arvore)
{
   return arvore == NULL;
}

/****
 *
 * AlteraConteudoArvoreFunil(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreFunil( tArvoreFunil *arvore, tChave chave,
                               tConteudo *conteudo )
{
   tConteudo *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreFunil(arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
   if (!RemoveNoArvoreFunil(arvore, chave)) {
      if (!InsereNoArvoreFunil( arvore, conteudo )) {
         return 0; /* Alterao OK */
      } else {
         return 1; /* Insero no foi possvel */
      }
   } else {
      return 1; /* Remoo no foi possvel */
   }
}

/****
 *
 * DestroiArvoreBin(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz
 *                               da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBin(tArvoreFunil arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBin(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBin(arvore->direita);

      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * CaminhamentoInfixo(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixo( tArvoreFunil arvore, tOperacao op,
                         FILE *stream )
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixo(arvore->esquerda, op, stream);

      op(arvore->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixo(arvore->direita, op, stream);
   }
}

/****
 *
 * MenorChave(): Obtm o contedo do n que possui a menor
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MenorChave(tArvoreFunil p)
{
   tArvoreFunil q; /* Apontar para o n que */
                /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->esquerda;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChave(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tConteudo *MaiorChave(tArvoreFunil p)
{
   tArvoreFunil q; /* Apontar para o n que */
                /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->direita;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}
