/****
 *
 * Arquivo: ArvoreBB.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 22/07/2012
 * ltima modificao: 22/07/2012
 *
 * Descrio: Implementao do mdulo Arvore
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada  */
#include <stdlib.h>   /* Funo exit()             */
#include <string.h>  /* Funes strXXX() e memXXX() */
#include "ArvoreBB.h" /* Interface deste mdulo */
#include "ASSEGURA.h" /* Macro ASSEGURA            */

/***************** Definies de Funes ******************/

/****
 *
 * ConstroiNoArvoreBB(): Constri um n de uma rvore binria
 *
 * Parmetros:
 *     item (entrada) - contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
static tArvoreBB ConstroiNoArvoreBB(tCEP_Ind item)
{
   tArvoreBB no;

   no = malloc(sizeof(tNoArvoreBB)); /* Tenta alocar o novo n */

      /* Se no houve alocao, aborta o programa */
   ASSEGURA(no, "Erro: Nao foi possivel alocar no'");

      /* Preenche o contedo do n */
   no->conteudo = item;
   no->esquerda = NULL; /* Este n ainda no tem filho */
   no->direita = NULL;

   return no;
}

/****
 *
 * RotacaoDireitaArvoreBB(): Efetua uma rotao direita sobre um
 *                   n de uma rvore binria
 *
 * Parmetros:
 *     x (entrada) - raiz da subrvore que sofrer a rotao
 *
 * Retorno: Endereo da nova raiz aps a rotao
 *
 * OBSERVAO: Esta funo no  usada por rvore binria
 *             de busca ordinria
 *
 ****/
static tArvoreBB RotacaoDireitaArvoreBB(tArvoreBB x)
{
   tArvoreBB y = x->esquerda; /* Guarda o endereo da nova raiz aps a rotao*/

      /* Garante que  possvel efetuar a rotao */
   ASSEGURA(y, "Rotacao direita impossivel: filho esquerdo e' nulo");

      /***********************************************/
      /* A ordem das instrues a seguir  essencial */
      /***********************************************/

      /* O filho direito de y torna-se filho esquerdo de x */
   x->esquerda = y->direita;

      /* x torna-se filho direito de y */
   y->direita = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * RotacaoEsquerdaArvoreBB(): Efetua uma rotao esquerda sobre um
 *                    n de uma rvore binria
 *
 * Parmetros:
 *     x (entrada) - raiz da subrvore que sofrer a rotao
 *
 * Retorno: Endereo da nova raiz aps a rotao
 *
 * OBSERVAO: Esta funo no  usada por rvore binria de busca ordinria
 *
 ****/
static tArvoreBB RotacaoEsquerdaArvoreBB(tArvoreBB x)
{
   tArvoreBB y = x->direita; /* Guarda o endereo da nova raiz aps a rotao */

      /* Garante que  possvel efetuar a rotao */
   ASSEGURA(y, "Rotacao esquerda impossivel: filho direito e' nulo");

      /***********************************************/
      /* A ordem das instrues a seguir  essencial */
      /***********************************************/

      /* O filho esquerdo de y torna-se filho direito de x */
   x->direita = y->esquerda;

      /* x torna-se filho esquerdo de y */
   y->esquerda = x;

   return y; /* Retorna a nova raiz */
}

/****
 *
 * IniciaArvoreBB(): Inicia uma rvore binria de busca
 *
 * Parmetros: *arvore (sada) - a rvore que ser iniciada
 *
 * Retorno: Nada
 *
 ****/
void IniciaArvoreBB(tArvoreBB *arvore)
{
   *arvore = NULL;
}

/****
 *
 * NumeroNosArvoreBB(): Calcula o nmero de ns de uma
 *                      rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: O nmero de ns da referida rvore
 *
 ****/
int NumeroNosArvoreBB(tArvoreBB arvore)
{
      /* Se a rvore for vazia seu nmero de ns  0 */
   if (!arvore) {
      return 0;
   }

      /* Leva em considerao a raiz da rvore e  */
      /* os ns das subrvores esquerda e direita */
   return 1 + NumeroNosArvoreBB(arvore->esquerda) +
              NumeroNosArvoreBB(arvore->direita);
}

/****
 *
 * AlturaArvoreBB(): Calcula a altura de uma rvore binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *
 * Retorno: A altura da referida rvore
 *
 ****/
int AlturaArvoreBB(tArvoreBB arvore)
{
   int profEsq, profDir;

   if (!arvore) {
      return 0; /* rvore vazia tem altura zero */
   } else {
        /* Calcula a altura da subrvore esquerda */
      profEsq = 1 + AlturaArvoreBB(arvore->esquerda);

        /* Calcula a altura da subrvore direita */
      profDir = 1 + AlturaArvoreBB(arvore->direita);

        /* A altura da rvore  a maior delas */
      return profEsq > profDir ? profEsq : profDir;
   }
}

/****
 *
 * BuscaArvoreBB(): Efetua uma rvore numa binria de busca
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore na
 *                        qual ser efetuada a busca
 *     chave (entrada) - a chave de busca
 *
 * Retorno: Endereo do contedo que contm a referida chave,
 *          se ela for encontrada. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *BuscaArvoreBB(tArvoreBB arvore, tCEP chave)
{
   int compara; /* Resultado da comparao de duas chaves */

      /* Enquanto 'arvore' no assume NULL ou a chave de */
      /* um n no casa com 'chave', a busca prossegue   */
   while (arvore) {
         /* Compara a chave de busca com */
         /* a chave do n corrente       */
      compara = strcmp(chave, arvore->conteudo.chave);

         /* Verifica se a chave foi encontrada */
      if (!compara)
         return &arvore->conteudo; /* Chave encontrada */

         /* Se a chave de busca for menor do que a chave do n */
         /* corrente, desce-se pela subrvore esquerda. Caso   */
         /* contrrio, desce-se pela subrvore direita.        */
      arvore = compara < 0 ? arvore->esquerda : arvore->direita;
   }

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * InsereArvoreBB(): Insere uma nova chave numa rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada/sada) - ponteiro para a raiz da rvore
 *                              na qual ser feita a insero
 *    *conteudo (entrada) - contedo do n que ser inserido
 *
 * Retorno: 1, se no houver insero porque a chave j existe
 *          0, se a insero ocorrer
 *
 ****/
int InsereArvoreBB( tArvoreBB *arvore, const tCEP_Ind *conteudo )
{
   tArvoreBB p, /* Ponteiro usado para descer na rvore */
             q, /* Ponteiro que apontar para o pai de p */
             pNovoNo; /* Ponteiro para n que ser inserido */
   int       compara; /* Resultado da comparao de duas chaves */

   p = *arvore; /* Comea a busca na raiz da rvore */
   q = NULL; /* A raiz no tem pai */

      /* Desce na rvore at encontrar um ponteiro nulo */
   while (p) {
         /* Compara a chave do novo n com a chave do n corrente */
      compara = strcmp(conteudo->chave, p->conteudo.chave);

         /* A chave  primria. Portanto, se ela */
         /* for encontrada, no haver insero. */
      if (!compara) /* A chave foi encontrada */
         return 1; /* No h mais nada a fazer */

         /* Faz q apontar para o n para o qual */
         /* p aponta antes que p mude de valor  */
      q = p;

         /* Se a chave de busca for menor do que a chave do n corrente, */
         /* desce-se  esquerda. Caso contrrio, desce-se  direita.     */
      if (compara < 0)
         p = p->esquerda;
      else
         p = p->direita;
   }

      /*                                              */
      /* Aqui, sabe-se que a chave no foi encontrada */
      /*                                              */

   pNovoNo = ConstroiNoArvoreBB(*conteudo);  /* Constri um novo n */

      /* Se a rvore no estiver vazia, q est apontando */
      /* para o n que ser o pai do novo n             */
   if (!q) /* A rvore estava vazia */
      *arvore = pNovoNo; /* Novo n ser a raiz */

      /* Se a chave do novo n for menor do que a chave */
      /* de seu pai, ele ser o filho da esquerda. Caso */
      /* contrrio, ele ser o filho da direita.        */
   else if (strcmp(conteudo->chave, q->conteudo.chave) < 0)
         /* q aponta para o n que ter um filho  esquerda */
      q->esquerda = pNovoNo;
   else
         /* q aponta para o n que ter um filho  direita */
      q->direita = pNovoNo;

   return 0; /* Insero foi OK */
}

/****
 *
 * RemoveArvoreBB(): Remove de uma rvore binria de busca
 *                   a chave especificada como parmetro
 *
 * Parmetros:
 *     arvore (entrada/sada) - endereo do ponteiro que
 *                              representa a rvore na qual
 *                              ser feita a remoo
 *     chave (entrada) - chave do n a ser removido
 *
 * Retorno: 0, se a remoo for bem sucedida, ou 1, se o n a
 *          ser removido no for encontrado.
 *
 ****/
int RemoveArvoreBB(tArvoreBB *arvore, tCEP chave)
{
   tArvoreBB p, /* Apontar para o n que ser removido */
             q, /* Apontar para o pai de p */
             subs, /* Apontar para o n que substituir p */
             filho, /* O filho esquerdo de 'subs' */
             pai; /* O pai de 'subs' */
   int       compara; /* Resultado da comparao de duas chaves */

   p = *arvore; /* Comea a busca na raiz da rvore */
   q = NULL; /* A raiz no tem pai */

      /* Procura o n a ser removido. O lao encerra quando */
      /* esse n ou um ponteiro nulo for encontrado.        */
   while (p) {
         /* Compara a chave que ser removida com a chave do n corrente */
      compara = strcmp(chave, p->conteudo.chave);

         /* Verifica se a chave foi encontrada */
      if (!compara)
         break; /* Chave encontrada */

         /* Faz q apontar para o n para o qual p */
         /* aponta antes de alterar o valor de p  */
      q = p;

         /* Se a chave procurada for menor do que a chave do n corrente, */
         /* desce-se pela subrvore esquerda. Caso contrrio, desce-se    */
         /* pela subrvore direita.                                       */
      p = compara < 0 ? p->esquerda : p->direita;
   }

      /* Se a chave no foi encontrada, a tarefa est encerrada */
   if (!p)
      return 1; /* No ocorreu remoo */

   /*************************************************************************/
   /* Neste ponto, p aponta para o n que ser removido e q aponta para seu */
   /* pai. O prximo passo  encontrar o n que substituir o n apontado   */
   /* por p. O ponteiro 'subs' apontar para esse n substituto.            */
   /*************************************************************************/

      /* Faz 'subs' apontar para o n que substituir p */
   if (!p->esquerda)
         /* O n a ser eliminado no possui nenhum filho ou s tem  */
         /* filho  direita. Esse filho ocupar o lugar do seu pai. */
       subs = p->direita;
   else if (!p->direita)
         /* O n a ser eliminado s tem filho  esquerda. */
         /* Esse filho ocupar o lugar do seu pai.        */
       subs = p->esquerda;
   else {
         /* p tem dois filhos. Neste caso, faz-se 'subs' */
         /* apontar para o sucessor em ordem infixa de p */
         /* e 'pai' apontar para o pai de 'subs'.        */
      pai = p;
      subs = p->direita;

         /* 'filho'  sempre o filho  esquerda de 'subs' */
      filho = subs->esquerda;

         /* Faz 'subs' apontar para o sucessor em ordem infixa de p e 'pai' */
         /* apontar para o pai de  'subs'. Usa-se o ponteiro 'filho' para   */
         /* descer na rvore sempre  esquerda at que 'filho' seja NULL.   */
      while (filho) {
         pai = subs;
         subs = filho;
         filho = subs->esquerda;
      }

         /* Neste ponto, sabe-se que 'subs'  o sucessor em ordem infixa */
         /* de p. Se 'pai' e 'p' estiverem apontando para o mesmo n,    */
         /* 'subs' no tem filho  esquerda.                             */
      if (pai != p) {
            /* Faz com que o filho direito de 'subs' */
            /* torne-se o filho esquerdo de 'pai'    */
         pai->esquerda = subs->direita;

            /* Substitui o filho direito de 'subs' pelo filho direito */
            /* de p, que aponta para o n a ser removido              */
         subs->direita = p->direita;
      }

         /* Substitui o filho esquerdo de 'subs' */
         /* pelo filho esquerdo de p             */
      subs->esquerda = p->esquerda;
   } /* else */

   /*                                     */
   /* Faz o n apontado por 'subs' ocupar */
   /* o lugar do n apontado por p        */
   /*                                     */

      /* Verifica se o n a ser removido  a raiz */
      /* pois a remoo da raiz  especial        */
   if (!q) /* O n removido era a raiz */
      *arvore = subs;
   else if (p == q->esquerda)
         /* O filho esquerdo de q ser removido */
      q->esquerda = subs;
   else
         /* O filho direito de q ser removido */
      q->direita = subs;

   free(p); /* Libera o n removido */

   return 0; /* A remoo foi bem sucedida */
}

/****
 *
 * EstaVaziaArvoreBB(): Verifica se uma rvore binria
 *                      est vazia
 *
 * Parmetros:
 *     arvore (entrada) - a rvore que ser verificada
 *
 * Retorno: 1, quando a arvore est vazia, ou
 *          0, em caso contrrio
 *
 ****/
int EstaVaziaArvoreBB(tArvoreBB arvore)
{
   return arvore == NULL;
}

/****
 *
 * AlteraConteudoArvoreBB(): Altera o contedo efetivo de um
 *                           n de uma rvore binria de busca
 *
 * Parmetros:
 *    *arvore (entrada) - rvore que contm o n que ser
 *                        alterado
 *     chave (entrada) - a chave do n
 *    *conteudo (entrada) - o novo contedo
 *
 * Retorno: 0, se a atualizao for bem sucedida, ou 1, se a
 *          atualizao no for possvel, porque o n no foi
 *          encontrado
 *
 ****/
int AlteraConteudoArvoreBB( tArvoreBB *arvore, tCEP chave,
                            const tCEP_Ind *conteudo )
{
   tCEP_Ind *ptrConteudo;

      /* Se a chave no foi alterada, encontra */
      /* o n e altera seu contedo            */
   if (!strcmp(chave, conteudo->chave)) {
      ptrConteudo = BuscaArvoreBB(*arvore, chave);

         /* Verifica se o n foi encontrado */
      if (ptrConteudo) {
            /* N foi encontrado e ser alterado */
         *ptrConteudo = *conteudo;

         return 0; /* Alterao OK */
      } else {
         return 1; /* N no foi encontrado */
      }
   }

      /* A chave foi alterada. Para no destruir a  */
      /* ordenao da rvore, remove-se o n antigo */
      /* e insere-se um novo n.                    */
   if (!RemoveArvoreBB(arvore, chave)) {
      if (!InsereArvoreBB( arvore, conteudo )) {
         return 0; /* Alterao OK */
      } else {
         return 1; /* Insero no foi possvel */
      }
   } else {
      return 1; /* Remoo no foi possvel */
   }
}

/****
 *
 * DestroiArvoreBB(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz
 *                               da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBB(tArvoreBB arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBB(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBB(arvore->direita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * CaminhamentoInfixoBB(): Efetua um caminhamento em ordem
 *                       infixa numa rvore binria
 *
 * Parmetros:
 *       arvore (entrada) - ponteiro para a raiz da
 *                          rvore sobre a qual ser
 *                          efetuado o caminhamento
 *       op (entrada) - ponteiro para a funo que representa
 *                      a operao de sada a ser efetuada
 *                      sobre a informao contida em cada n
 *       stream (entrada) - stream no qual o resultado da
 *                          operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void CaminhamentoInfixoBB( tArvoreBB arvore, tOperacao op,
                         FILE *stream )
{
   if (arvore) {
         /* Caminha na subrvore esquerda */
      CaminhamentoInfixoBB(arvore->esquerda, op, stream);

      op(arvore->conteudo, stream); /* Visita a raiz */

         /* Caminha na subrvore direita */
      CaminhamentoInfixoBB(arvore->direita, op, stream);
   }
}

/****
 *
 * MenorChaveArvoreBB(): Obtm o contedo do n que possui a menor
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a menor chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MenorChaveArvoreBB(tArvoreBB p)
{
   tArvoreBB q; /* Apontar para o n que */
                /* armazena a menor chave */

      /* Se a rvore estiver vazia no h menor chave */
   if (!p) {
      return NULL;
   }
      /* A menor chave encontra-se no n mais */
      /*  esquerda da rvore de busca        */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  esquerda */
      p = p->esquerda;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a menor chave      */
   return &q->conteudo;
}

/****
 *
 * MaiorChaveArvoreBB(): Obtm o contedo do n que possui a maior
 *               chave armazenada na rvore de busca
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore de busca
 *
 * Retorno: Endereo do contedo que contm a maior chave, se
 *          a rvore no estiver vazia. Caso contrrio, NULL.
 *
 ****/
tCEP_Ind *MaiorChaveArvoreBB(tArvoreBB p)
{
   tArvoreBB q; /* Apontar para o n que */
                /* armazena a maior chave */

      /* Se a rvore estiver vazia no h maior chave */
   if (!p) {
      return NULL;
   }
      /* A maior chave encontra-se no n   */
      /* mais  direita da rvore de busca */
   while (p) {
         /* Faz q apontar para o n para o qual p aponta */
      q = p;

         /* Faz p apontar para o n  direita */
      p = p->direita;
   }
      /* Neste ponto, q aponta para o n */
      /* que armazena a maior chave      */
   return &q->conteudo;
}
