/****
 *
 * Ttulo: Checando rvores Binrias de Busca
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 21/03/2017
 * ltima modificao: 21/03/2017
 *
 * Descrio: Verifica se uma rvore binria  uma rvore binria
 *            de busca
 *
 * Entrada: No h
 *
 * Sada: Uma mensagem indicando se a rvore binria criada
 *        pelo programa  uma rvore binria DE BUSCA
 *
 ****/

/*********************** Includes *************************/

#include<stdio.h>
#include<stdlib.h>

/************************* Macros *************************/

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

/****************** Definies de Tipos *******************/

typedef int tConteudo; /* Tipo do contedo efetivo de cada n */

   /* Tipo de n e tipo de ponteiro para n de uma rvore binria  */
typedef struct rotNoABB {
           struct rotNoABB *esquerda;
           tConteudo        conteudo;
           struct rotNoABB *direita;
        } tNoArvoreBB, *tArvoreBB;

typedef int (*tCompara) (const void *, const void *);

/***************** Definies de Funes ******************/

/****
 *
 * EhArvoreBinDeBusca(): Verifica se uma rvore binria  uma rvore binria
 *                    DE BUSCA
 *
 * Parmetros:
 *      p (entrada) - ponteiro para a raiz da rvore
 *      compara (entrada) - ponteiro para uma funo que compara contedos dos
 *                          ns de acordo com o mesmo critrio de comparao
 *                          de qsort()
 *
 * Retorno: 1, se a rvore for uma rvore binria de busca;
 *          0, em caso contrrio
 *
 ****/
int EhArvoreBinDeBusca(tNoArvoreBB *p, tCompara compara)
{
   static tNoArvoreBB *antecessor = NULL; /* Armazenar o antecessor de cada */
                                          /* n visitado em ordem infixa     */

      /* Efetua um caminhamento em ordem infixa guardando */
      /* o valor do n antecessor imediato do n corrente */
   if (p) {
         /* Verifica se a subrvore esquerda  de busca */
      if (!EhArvoreBinDeBusca(p->esquerda, compara))
         return 0;

         /* Se o contedo do n corrente for menor do que o contedo */
         /* do n antecessor, no se  trata de rvore de busca       */
      if ( antecessor && compara(&p->conteudo, &antecessor->conteudo) < 0 )
         return 0;

      antecessor = p; /* O antecessor passa a ser o n corrente */

         /* Verifica se a subrvore direita  de busca */
      return EhArvoreBinDeBusca(p->direita, compara);
   }

   return 1; /* Se ainda no houve retorno, a rvore foi aprovada */
}

/****
 *
 * DestroiArvoreBB(): Destri uma rvore binria
 *
 * Parmetros:
 *      arvore (entrada/sada) - ponteiro para a raiz da rvore binria
 *
 * Retorno: Nada
 *
 ****/
void DestroiArvoreBB(tArvoreBB arvore)
{
   if (arvore) {
         /* Destri subrvore esquerda */
      DestroiArvoreBB(arvore->esquerda);

         /* Destri subrvore direita */
      DestroiArvoreBB(arvore->direita);

         /* Remove a raiz */
      free(arvore); /* Libera a raiz */
   }
}

/****
 *
 * ComparaInts(): Compara dois elementos do tipo int
 *
 * Parmetros:
 *      *e1, *e2 (entrada) - os elementos
 *
 * Retorno:
 *     = 0, se *e1 == *e2
 *     < 0, se *e1 < *e2
 *     > 0, se *e1 > *e2
 *
 ****/
int ComparaInts(const void *e1, const void *e2)
{
   ASSEGURA(e1 && e2, "Elemento nulo recebido por ComparaInts()");

   return *(int *)e1 - *(int *)e2;
}

/****
 *
 * NoNovo(): Cria um n de uma rvore binria com o contedo
 *           especificado e subrvores vazias
 *
 * Parmetros:
 *      conteudo (entrada) - o contedo efetivo do n
 *
 * Retorno: Endereo do n criado
 *
 ****/
tNoArvoreBB *NoNovo(tConteudo conteudo)
{
   tNoArvoreBB *pNo = malloc(sizeof(tNoArvoreBB));

      /* Verifica se o n foi alocado */
   ASSEGURA(pNo, "Erro: Um no' da arvore nao foi alocado")

      /* Recheia o n */
   pNo->conteudo = conteudo;
   pNo->esquerda = pNo->direita = NULL;

   return pNo;
}

/****
 *
 * main(): Cria duas rvores binrias e verifica se elas so
 *         rvores binrias de busca
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tNoArvoreBB *raiz = NoNovo(12); /* Raiz da rvore da Figura (a) */

      /* Insere os demais ns na rvore da Figura (a) */
   raiz->esquerda = NoNovo(10);
   raiz->direita = NoNovo(13);
   raiz->esquerda->esquerda = NoNovo(7);
   raiz->esquerda->direita = NoNovo(11);
   raiz->esquerda->esquerda->esquerda = NoNovo(5);

      /* Testa se a rvore binria da Figura (a)  de busca */
   if (EhArvoreBinDeBusca(raiz, ComparaInts))
      printf("\n\t>>> A arvore da Figura (a) e' de busca");
   else
      printf("\n\t>>> A arvore da Figura (a) NAO e' de busca");

   DestroiArvoreBB(raiz); /* Destri a rvore da Figura (a) */

      /* Cria a rvore da Figura (b) */
   raiz = NoNovo(12);
   raiz->esquerda = NoNovo(10);
   raiz->direita = NoNovo(13);
   raiz->esquerda->esquerda = NoNovo(7);
   raiz->esquerda->direita = NoNovo(15);

      /* Testa se a rvore binria da Figura (b)  de busca */
   if (EhArvoreBinDeBusca(raiz, ComparaInts))
      printf("\n\t>>> A arvore da Figura (b) e' de busca");
   else
      printf("\n\t>>> A arvore da Figura (b) NAO e' de busca");

   return 0;
}

/****************** Resultado do Programa *****************

        >>> A arvore da Figura (a) e' de busca
        >>> A arvore da Figura (b) NAO e' de busca

**********************************************************/
