/****
 *
 * Ttulo do Programa: rvore B em Arquivo
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 25/04/2004
 * ltima modificao: 21/07/2008
 *
 * Descrio Geral: Este programa demonstra vrias
 *                  operaes utilizando rvores B
 *                  armazenadas em arquivo
 *
 * Entrada: Opes escolhidas atravs de menu
 *
 * Sada: Resultados das operaes escolhidas pelo usurio
 *        (v. exemplo de execuo do programa no final
 *        deste arquivo)
 *
 * Arquivos:
 *     ASSERT.h - macros de tratamento de exceo usadas pelo
 *                programa
 *     ArvoreB.h e - interface e implementao de operaes
 *     ArvoreB.c     operaes sobre rvores B armazenadas
 *                   em arquivo
 *     Registros.h e - interface e implementao de operaes
 *     Registros.c     sobre registros armazenados em arquivos
 *     main.c (este arquivo) - contm main() e outras funes
 *                             que testam as operaes sobre
 *                             rvores B armazenadas em arquivo
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de Entrada/Sada        */
#include <string.h>    /* Funes strXXX() e memXXX()     */
#include <time.h>      /* Medio de tempo                */
#include "ArvoreB.h"   /* Funes que lidam com rvores B */
#include "Registros.h" /* Funes de processamento        */
                       /* de registros                    */
#include "Leitura.h"   /* Funes de leitura via teclado  */
#include "ASSEGURA.h"  /* Macro ASSEGURA e outras macros  */
                       /* de tratamento de exceo        */

/***************** Constantes Simblicas ******************/

      /* Nome do arquivo de texto contendo os registros */
#define NOME_ARQUIVO_REGS "../CensoMEC2012.txt"

/* Nome do arquivo de registros */
#define NOME_ARQ_BIN "CensoMEC2012.bin"

   /* Nome do arquivo que contm a rvore */
//#define NOME_ARQ_ARV "I:\\Dados\\CensoMEC\\ArvoreB.dat"
#define NOME_ARQ_ARV "ArvoreB.dat"

   /* Nome do arquivo que contm a raiz rvore */
//#define NOME_ARQ_RAIZ    "I:\\Dados\\CensoMEC\\RaizB.dat"
#define NOME_ARQ_RAIZ "RaizB.dat"

   /* Nome do arquivo que contm as chaves ordenadas */
//#define NOME_ARQ_CHAVES  "I:\\Dados\\CensoMEC\\ChavesOrdenadas.txt"
#define NOME_ARQ_CHAVES  "ChavesOrdenadasB.txt"

   /* Nome inicial do arquivo de texto contendo os ns da rvore */
#define NOME_ARQ_NOS "NosArvoreB"

   /* Nome do arquivo de texto contendo registros em formato legvel */
#define NOME_ARQ_DEBUG   "RegistrosDebug.txt"

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/******************** Funes de Locais *******************/

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\n>>> Escolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t[%d] %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
static void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * TamanhoNoSemPreenchimento(): Calcula o tamanho de um n sem
 *                              o preenchimento de estrutura
 *                              efetuado pelo compilador
 *
 * Parmetros: Nenhum
 *
 * Retorno: O tamanho de um n sem preenchimento
 *
 * Observao: O propsito desta funo  apenas depurao e
 *             melhor entendimento do programa
 *
 ****/
static int TamanhoNoSemPreenchimento(void)
{
   int tn, /* Tamanho de um n sem preenchimento */
       ti = sizeof(int), /* Largura do tipo inteiro */
       tc = 2*sizeof(int); /* Tamanho de um par chave/ndice */

   tn = ti + (G - 1)*tc + G*ti;

   return tn;
}

/****
 *
 * ApresentaEstatisticas(): Apresenta informaes sobre o banco
 *                          de dados gerenciado pelo programa
 *
 * Parmetros:
 *       arvore - stream associado ao arquivo que armazena
 *                a rvore
 *       raiz - posio no arquivo da raiz da rvore
 *       registros - stream que representa os registros
 *
 * Retorno: Nada
 *
 * Observao: O propsito desta funo  apenas depurao e
 *             melhor entendimento do programa
 *
 ****/
void ApresentaEstatisticas( FILE *arvore, int raiz,
                            FILE *registros )
{
   tRegistro    umRegistro; /* Armazena um registro */
   int          nRegistros, /* Nmero de registros */
                nChaves, /* Nmero de chaves na rvore */
                prof; /* Profundidade da rvore de busca */
   tChaveIndice chaveIndice; /* Um par chave/ndice */

   printf( "\n\n>>>     Estatisticas do Banco de Dados <<<\n" );

   printf("\n\t>>> Grau da arvore: %d", (int) G);
   printf( "\n\t>>> Tamanho de um no' COM preenchimento: %d",
           (int) sizeof(tNoMultiMS) );
   printf( "\n\t>>> Tamanho de um no' SEM preenchimento: %d",
           TamanhoNoSemPreenchimento() );
   printf("\n\t>>> Tamanho de bloco: %d\n", TB);

   printf( "\n\t>>> Numero de nos da arvore: %d",
           NumeroDeNosB(arvore) );

   nRegistros = NumeroDeRegistros(registros, sizeof(umRegistro));

   printf( "\n\t>>> Numero de registros no arquivo: %d",
           nRegistros );

   nChaves = NumeroDeChavesB(arvore);
   printf("\n\t>>> Numero de chaves da arvore: %d", nChaves);

      /* Calcula a altura da rvore */
   prof = AlturaB(arvore, raiz);

   printf("\n\t>>> Profundidade da arvore: %d", prof);

   ASSEGURA( nChaves == nRegistros,
             "\n\nThis program is fucking wrong!!!" );

      /* Encontra a posio do registro */
      /* que contm a menor chave       */
   chaveIndice = MenorChaveB(arvore, raiz);

      /* Exibe o registro que contm a menor chave */
   if (chaveIndice.indice < 0)
      printf("\n\t>>> A menor chave nao foi encontrada");
   else if (LeRegistroEmArquivo(&umRegistro, registros,
                                chaveIndice.indice)){
      printf("\n\n\t>>> Registro com a menor chave <<<\n\n");
      ExibeRegistro(&umRegistro);
   } else
      printf( "\n\t>>> Ocorreu um erro ao tentar "
              "recuperar o registro com a menor chave" );

      /* Encontra a posio do registro */
      /* que contm a maior chave       */
   chaveIndice = MaiorChaveB(arvore, raiz);

      /* Exibe o registro que contm a maior chave */
   if (chaveIndice.indice < 0)
      printf("\n\t>>> A maior chave nao foi encontrada");
   else if (LeRegistroEmArquivo(&umRegistro, registros,
                                chaveIndice.indice)){
      printf("\n\n\t>>> Registro com a maior chave <<<\n\n");
      ExibeRegistro(&umRegistro);
   } else
      printf( "\n\t>>> Ocorreu um erro ao tentar "
              "recuperar o registro com a maior chave" );
}

/****
 *
 * ConstroiArvoreB(): L arquivo binrio de registros e
 *                   constri uma rvore B em arquivo
 *
 * Parmetros:
 *      streamReg (entrada) - stream associado ao arquivo que
 *                            contm os registros
 *      streamArvore (entrada) - stream associado ao arquivo
 *                               que conter a rvore
 *
 * Retorno: Posio da raiz da rvore no arquivo que a contm
 *
 ****/
int ConstroiArvoreB( FILE *streamReg, FILE *streamArvore )
{
   tChaveIndice chaveEIndice; /* Armazena um par chave/ndice */
   tRegistro    reg; /* Armazena um registro lido no arquivo */
   int          pos = 0, /* Posico de um registro no arquivo */
                raiz; /* Posico da raiz da rvore */
                      /* no arquivo que a armazena */

      /* Verifica se o arquivo de registros  NULL */
   ASSEGURA( streamReg,
             ERRO_STREAM_NULL(streamReg, ConstroiArvoreB) );

      /* Verifica se o arquivo da rvore  NULL */
   ASSEGURA( streamReg,
             ERRO_STREAM_NULL(streamArvore, ConstroiArvoreB) );

      /* Tenta mover o apontador de posio    */
      /* para o incio do arquivo de registros */
   MoveApontador(streamReg, 0, SEEK_SET);

      /* Inicialmente a rvore est vazia */
   raiz = POSICAO_NULA;

      /* L cada registro e armazena a chave e o respectivo  */
      /* ndice do registro no arquivo que armazena a rvore */
   while (1) {
         /* L um registro */
      fread(&reg, sizeof(reg), 1, streamReg);

         /* Se ocorreu erro ou o final de arquivo */
         /* foi atingido, encerra o lao          */
      if (ferror(streamReg) || feof(streamReg))
         break;

         /* Cria o contedo do par chave/ndice */
         /* que ser inserido na rvore         */
      chaveEIndice.chave = reg.codAlunoCurso;
      chaveEIndice.indice = pos++;

         /* Tenta armazenar o par chave/ndice na rvore. */
         /* Se isso for impossvel, a chave  repetida.   */
      ASSEGURA( InsereB( &raiz, &chaveEIndice, streamArvore ),
                "Insercao nao foi possivel" );
   }

      /* Certfica-se que no ocorreu erro de leitura */
   ASSEGURA(!ferror(streamReg), ERRO_FREAD(ConstroiArvoreB));

   return raiz;
}

/****
 *
 * IniciaArvoreB(): Verifica se a rvore de busca j existe.
 *                  Se ela j existir e o usurio desejar
 *                  reconstrui-la ou se ela no existir,
 *                  ela ser construda.
 *
 * Parmetros:
 *     nomeArqRegistros (entrada) - nome do arquivo de registros
 *     nomeArqArvore (entrada) - nome do arquivo que armazena a
 *                               rvore
 *     nomeArqRaiz (entrada) - nome do arquivo que armazena a
 *                             raiz da rvore
 *     *streamRegistros (entrada/sada) - stream associado ao
 *                                        arquivo de registros
 *     *streamArvore (entrada/sada) - stream associado ao
 *                                     arquivo que armazena
 *                                     a rvore
 *
 * Retorno: Posio da raiz da rvore no arquivo que a contm
 *
 ****/
int IniciaArvoreB( const char *nomeArqRegistros,
                   const char *nomeArqArvore,
                   const char *nomeArqRaiz,
                   FILE **streamRegistros,
                   FILE **streamArvore )
{
   int op, /* Opo do usurio */
       raiz; /* A raiz da rvore */

      /* Tenta abrir o arquivo de registros */
   *streamRegistros = AbreArquivo(nomeArqRegistros, "r+b");

      /* Verifica se o arquivo que contm a rvore e aquele   */
      /* que contm a posio da raiz dessa rvore j existem */
   if ( ArquivoExiste(nomeArqArvore) &&
        ArquivoExiste(nomeArqRaiz) ) {
      /*                                */
      /* rvore j existe. Se o usurio */
      /* desejar, ela ser reconstruda */
      /*                                */

         /* Oferece a opo ao usurio */
      printf( "\n\n\t>>> Arvore ja existe. "
              "Deseja substitui-la (s/n)? " );
      op = LeOpcao("sSnN");

         /* Verifica se o usurio deseja reconstrui a rvore */
      if ( op == 'n' || op == 'N') { /* Ele no deseja isso */
            /* Tenta abrir o arquivo que contm a rvore */
         *streamArvore = AbreArquivo(nomeArqArvore, "r+b");

         raiz = LeRaizB(nomeArqRaiz);

         return raiz; /* Servio completo */
      }
   }

   /*                                        */
   /* Neste ponto, ou a rvore ainda no     */
   /* existe ou usurio deseja reconstrui-la */
   /*                                        */

      /* Cria o arquivo e checa se abertura foi OK */
   *streamArvore = AbreArquivo(nomeArqArvore, "w+b");

      /* Constri a rvore e retorna a nova raiz */
   return ConstroiArvoreB(*streamRegistros, *streamArvore);
}

/****
 *
 * EscreveArrayChavesIndices(): Escreve num stream um array de
 *                              pares chave/ndice
 *
 * Parmetros:
 *      ci[] (entrada) - o array
 *      n (entrada) - nmero de elementos do array
 *      stream (entrada) - o stream no qual ocorrer a escrita
 *
 * Retorno: Nada
 *
 ****/
void EscreveArrayChavesIndices(const tChaveIndice ci[],
                               int n, FILE *stream)
{
   int i;

   fprintf(stream, "\nChaves e posicoes = { ");

   for (i = 0; i < n; ++i) {
      fprintf(stream, "(%ld, %d) ", ci[i].chave, ci[i].indice);
   }

   fprintf(stream, "}");
}

/****
 *
 * EscreveArrayFilhos(): Escreve num stream um array de ints que so
 *                       ndices que representam filhos de um n
 *
 * Parmetros:
 *      filhos[] (entrada) - o array
 *      n (entrada) - nmero de elementos do array
 *      stream (entrada) - o stream no qual ocorrer a escrita
 *
 * Retorno: Nada
 *
 ****/
void EscreveArrayFilhos(const int filhos[], int n, FILE *stream)
{
   int i;

   fprintf(stream, "\nFilhos = {%d", filhos[0]);

   for (i = 1; i < n; ++i) {
      fprintf(stream, ", %d", filhos[i]);
   }

   fprintf(stream, "}");
}

/****
 *
 * ExibeNo(): Escreve um n de uma rvore B num stream
 *
 * Parmetros:
 *      *no (entrada) - o n
 *      stream (entrada) - o stream no qual ocorrer a escrita
 *
 * Retorno: Um valor que indica o tipo de n escrito (interno
 *          ou folha)
 *
 ****/
void ExibeNo(const tNoMultiMS *no, FILE *stream)
{
   fprintf(stream, "Numero de chaves: %d", no->nFilhos - 1);
   fprintf(stream, "\nNumero de filhos: %d", no->nFilhos);
   EscreveArrayChavesIndices(no->chaves, no->nFilhos - 1, stream);
   EscreveArrayFilhos(no->filhos, no->nFilhos, stream);
}

/****
 *
 * EhFolha(): Verifica se um n de uma rvore B  uma folha
 *
 * Parmetros:
 *      *no (entrada) - o n
 *
 * Retorno: Um valor inteiro que indica se n  folha (1) ou no (0)
 *
 ****/
int EhFolha(const tNoMultiMS *no)
{
   int i,
       sim = 1;

      /* Se algum filho no for nulo, o n no  folha */
   for (i = 0; i < no->nFilhos; ++i) {
      if (no->filhos[i] != POSICAO_NULA)
         return !sim; /* No  folha */
   }

   return sim; /*  folha */
}

/****
 *
 * CriaArquivoDeNosB(): Apresenta os ns de uma rvore B num
 *                      arquivo de texto
 *
 * Parmetros:
 *      streamArv (entrada) - stream associado ao arquivo que
 *                            contm a rvore
 *      nomeIni (entrada) - nome inicial do arquivo que conter
 *                          as informaes sobre os ns da rvore
 *      posRaiz (entrada) - posio da raiz no arquivo que contm
 *                          a rvore
 *
 * Retorno: Nada
 *
 ****/
void CriaArquivoDeNosB( FILE *streamArv, const char *nomeIni,
                        int posRaiz )
{
   static int ordem = 0;
   int        i = 0,
              nNosInternos = 0,
              nFolhas = 0,
              nNosRemovidos = 0,
              nChaves = 0;
   tNoMultiMS  no;
   FILE      *streamNos;
   char      *nomeArqNos;

      /* O limite de arquivos criados por seo  99 */
   if (++ordem > 99) {
      printf("\n\t\a>>> Ja' criei arquivos demais <<<\n");
      return;
   }

      /* Aloca espao para o nome do arquivo de ns: */
      /* 1 (terminal) + 2 (ordem) + 4 (".txt") = 7   */
   nomeArqNos = malloc(strlen(nomeIni) + 7);
   ASSEGURA( nomeArqNos,
             "Nao foi possivel criar nome de arquivo de nos" );

      /* Cria o nome completo do arquivo */
   sprintf(nomeArqNos, "%s%.2d%s", nomeIni, ordem, ".txt");

      /* Tenta criar o aquivo */
   streamNos = fopen(nomeArqNos, "w");
   ASSEGURA(streamNos, "Nao foi possivel criar arquivo de nos");

      /* A leitura deve comear no incio */
      /* do arquivo que contm a rvore   */
   rewind(streamArv);

   while (1) {
      fread(&no, sizeof(no), 1, streamArv);

      if (feof(streamArv) || ferror(streamArv))
         break;

      fprintf(streamNos, "%d.\n\n", i++);

      ExibeNo(&no, streamNos);

      if (no.nFilhos <= 0)
         ++nNosRemovidos;
      else {
         nChaves += no.nFilhos - 1;

         if (EhFolha(&no))
            ++nFolhas;
         else
            ++nNosInternos;
      }

      fprintf(streamNos, "\n\n----------\n\n");
   }

   fprintf(streamNos, "\t>>> Grau da arvore (G): %d", (int) G);
   fprintf(streamNos, "\n\t>>> Posicao da raiz: %d", posRaiz);
   fprintf( streamNos, "\n\t>>> Profundidade da arvore: %d\n\n",
            AlturaB(streamArv, posRaiz) );

   fprintf( streamNos, "\t>>> Numero de nos internos: %d",
            nNosInternos );
   fprintf( streamNos, "\n\t>>> Numero de folhas: %d",
            nFolhas );
   fprintf( streamNos, "\n\t>>> Numero de nos removidos: %d",
            nNosRemovidos );
   fprintf( streamNos, "\n\t>>> Total de nos (incluindo removidos):"
            " %d", nFolhas + nNosInternos + nNosRemovidos );
   fprintf( streamNos, "\n\t>>> Numero total de chaves: %d\n\n",
            nChaves );

   fclose(streamNos);

   printf( "\n\t>>> Resultado escrito no arquivo: \n\t\t\"%s\"\n",
           nomeArqNos);

   free(nomeArqNos);
}

/****
 *
 * CriaArquivoRegistros(): Verifica se o arquivo de registros j
 *                         existe. Se ela j existir e o usurio
 *                         desejar reconstrui-lo ou se ele no
 *                         existir, ele ser construdo.
 *
 * Parmetros:
 *     nomeArqTexto (entrada) - nome do arquivo texto de registros
 *     nomeArqBin (entrada) - nome do arquivo binrio de registros
 *
 * Retorno: Stream associado ao arquivo binrio de registros
 *
 ****/
FILE *CriaArquivoRegistros( const char *nomeArqTexto,
                            const char *nomeArqBin )
{
   int   op, /* Opo do usurio */
         nRegistros, /* Nmero de registros no arquivo */
         maxRegistros; /* Nmero mximo de registros no arquivo */

      /* Verifica se o arquivo binrio de registros j existe */
   if (ArquivoExiste(nomeArqBin)) {
      /*                                */
      /* Arquivo j existe. Se o usurio */
      /* desejar, ele ser reconstrudo */
      /*                                */

         /* Oferece a opo ao usurio */
      op = LeSimNao( "Arquivo de registros ja' existe. "
                      "Deseja substitui-lo" );

         /* Verifica se o usurio deseja reconstruir o arquivo */
      if (op == 'n') /* Ele no deseja isso   */
         return AbreArquivo(nomeArqBin, "r+b");
   }

   /*                                        */
   /* Neste ponto, ou o arquivo ainda no    */
   /* existe ou usurio deseja reconstrui-lo */
   /*                                        */

   printf("\n\t>>> Digite o numero maximo de registros (0 = TODOS): ");
   maxRegistros = LeNatural();

   printf("\nCriando arquivo...\n");
   fflush(stdout);

      /* Cria o arquivo binrio */
   nRegistros = CriaArquivoBin( nomeArqTexto, nomeArqBin,
                                maxRegistros );

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, nomeArqBin );

      /* Abre e retorna stream assoaciado ao arquivo binrio */
   return AbreArquivo(nomeArqBin, "r+b");
}

/****
 *
 * EscreveRegistrosDebug(): Escreve o contedo de um arquivo
 *                          binrio de registros num arquivo de
 *                          texto de modo legvel
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 ****/
void EscreveRegistrosDebug(FILE *streamReg, const char *nomeArqTxt)
{
   int     i = 0, /* ndice do registro no arquivo */
           ocorreuErro = 0;
   tRegistroMEC  reg;
   FILE   *streamTxt;

   streamTxt = AbreArquivo(nomeArqTxt, "w");

   MoveApontador(streamReg, 0L, SEEK_SET);

   while (1) {
      fread(&reg, sizeof(reg), 1, streamReg);

      if (feof(streamReg) || ferror(streamReg))
         break;

      fprintf(streamTxt, "%d.\n\n", i++);

      if (ferror(streamTxt))
         break;

      EscreveRegistro(&reg, streamTxt);

      fprintf(streamTxt, "\n\n----------\n\n");

      if (ferror(streamTxt))
         break;
   }

   ocorreuErro = ferror(streamReg) || ferror(streamTxt);

   FechaArquivo(streamTxt, nomeArqTxt);

   if (ocorreuErro) {
      fclose(streamReg);
      printf("\n\t\a>>> Ocorreu erro de leitura ou escrita\n");
      exit(1);
   }
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa funes de busca multidirecional com a
 *         rvore B armazenada em arquivo
 *
 * Parmetros: nenhum
 *
 * Retorno: 0
 *
 ****/
int main(void)
{
   tRegistro   umRegistro;
   const char *opcoes[] = { "Consulta dados de um aluno",        /* 1 */
                            "Insere um aluno",                   /* 2 */
                            "Remove um aluno",                   /* 3 */
                            "Escreve chaves em ordem",           /* 4 */
                            "Escreve nos da arvore (depuracao)", /* 5 */
                            "Escreve registros (depuracao)",     /* 6 */
                            "Encerra o programa"                 /* 7 */
                          };
   int          op,  /* Opo do usurio  */
                pos, /* Posio de um registro em arquivo */
                posRaiz, /* Posio da raiz da rvore em arquivo */
                nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]),
                nRegs = 0, /* Nmero de registros no arquivo */
                sizeofLong = (int) sizeof(long);
   tChave       umaChave; /* Uma chave de busca */
   tChaveIndice chaveEIndice;
   FILE        *streamReg, /* Stream associado ao  */
                           /* arquivo de registros */
               *streamArv, /* Stream associado ao       */
                           /* arquivo contendo a rvore */
               *streamChaves; /* Stream associado a um     */
                              /* arquivo que armazena as   */
                              /* chaves em ordem crescente */

   printf("\n\t>>> sizeof(long) = %d", sizeofLong);

   if (sizeofLong >= 8) {
      printf("... OK\n");
   } else {
      printf("\n\t>>> Este programa pode nao funcionar adequadamente "
             "\n\t>>> quando sizeof(long) e' menor do que 8");
      op = LeSimNao("Deseja prosseguir");

      if (op == 'n') {
         printf("\n\t>>> Adquira um compilador decente e recompile "
                "\n\t>>> o programa. Bye.");
         return 1;
      }
   }

      /* Abre ou cria o arquivo de registros */
   streamReg = CriaArquivoRegistros(NOME_ARQUIVO_REGS, NOME_ARQ_BIN);

   MedidaDeTempo();
   printf("\nCriando arvore...");

      /* Inicia a rvore B e obtm a posio da */
      /* raiz da rvore no arquivo que a contm */
   posRaiz = IniciaArvoreB( NOME_ARQ_BIN, NOME_ARQ_ARV,
                            NOME_ARQ_RAIZ, &streamReg,
                            &streamArv );

   printf("\n\n\t>>> Arvore criada com sucesso");

      /* Calcula o nmero de registros */
   nRegs = NumeroDeRegistros(streamReg, sizeof(umRegistro));

   MedidaDeTempo();

   MedidaDeTempo();
   printf("\nChecando a consistencia do programa...");

   ApresentaEstatisticas(streamArv, posRaiz, streamReg);

   printf("\n\n...consistencia OK");
   MedidaDeTempo();

   while (1) { /* Lao principal do programa */
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("1234567");

         /* Verifica se o usurio deseja sair do programa */
      if (op == '7') {
            /* Grava em arquivo a posio da raiz da rvore */
         EscreveRaizB(NOME_ARQ_RAIZ, posRaiz);

         break; /* Encerra o lao */
      }

      if (op <= '3') { /* Estas opes precisam de uma chave */
         printf("\n\t>>> Digite o codigo aluno/curso: ");
         umaChave = LeNatural();
      }

      switch (op) {
         case '1': /* Procura o registro que */
                   /* possui uma dada chave  */

               /* Tenta encontrar a posio do registro */
               /* que contm a chave de busca           */
            pos = BuscaB(streamArv, posRaiz, umaChave);

               /* Verifica se a chave foi encontrada */
            if (pos < 0)
               printf("\n\t>>> A chave nao foi encontrada");
            else if (LeRegistroEmArquivo(&umRegistro, streamReg, pos)) {
               printf("\n\n\t>>> Registro com a chave %ld <<<\n",
                      umaChave);
               ExibeRegistro(&umRegistro);
               printf("\n\n"); /* Embelezamento */
            } else
               printf("\n\t>>> Erro tentando ler registro");

            break;

         case '2': /* Insere um novo registro, se a */
                   /* chave dada ainda no existe   */

               /* Tenta inserir o par chave/ndice na rvore. */
               /* Se insero for bem-sucedida, insere o      */
               /* respectivo registro no arquivo de registros */
            chaveEIndice.chave = umaChave;
            chaveEIndice.indice = nRegs;

            if (!InsereB(&posRaiz, &chaveEIndice, streamArv))
               printf("\n\t>>> A chave ja' existe na arvore");
            else {
                  /* A chave j foi armazenada na rvore. */
                  /* Resta inserir o respectivo registro  */
                  /* no arquivo.                          */
               InsereRegistroEmArquivo( chaveEIndice.chave,
                                        streamReg );

               nRegs++; /* O nmero de registros aumentou */

                  /* Garante que os buffers associados */
                  /* aos arquivos so descarregados    */
               fflush(streamArv);
               fflush(streamReg);

               printf("\n\t>>> Insercao bem-sucedida");
            }

            break;

         case '3': /* Remove uma chave da rvore */
            if (RemoveChaveB(streamArv, &posRaiz, umaChave)) {
               if (!streamArv) { /* rvore ficou vazia */
                  posRaiz = POSICAO_NULA;
                  streamArv = AbreArquivo(NOME_ARQ_ARV, "w+b");
               }

               printf( "\n\t>>> Remocao bem sucedida.");
            } else {
               printf( "\n\t>>> A remocao malsucedida.");
            }

            break;

         case '4':   /* Escreve em arquivo de texto  */
                     /* as chaves em ordem crescente */
           streamChaves = AbreArquivo(NOME_ARQ_CHAVES, "w");

               /* Faz um caminhamento em ordem infixa na */
               /* rvore e escreve as chaves encontradas */
               /* nesse processo num arquivo de texto    */
            CaminhamentoInfixoB(streamArv, posRaiz, streamChaves);
            printf( "\nAs chaves foram gravados em ordem "
                    "crescente no arquivo %s\n", NOME_ARQ_CHAVES );

               /* Fecha o arquivo de chaves ordenadas */
            FechaArquivo(streamChaves, NOME_ARQ_CHAVES);

            break;

         case '5': /* Apresenta os ns da rvore B+ */
                   /* num arquivo de texto          */
            CriaArquivoDeNosB(streamArv, NOME_ARQ_NOS, posRaiz);

            break;

         case '6': /* Apresenta os registros num arquivo */
                   /* de texto para depurao             */
            EscreveRegistrosDebug(streamReg, NOME_ARQ_DEBUG);
               printf( "\n\t>>> Os registros foram escritos no "
                       "arquivo: \"%s\"\n", NOME_ARQ_DEBUG );

            break;
      } /* switch */
   } /* while */

      /* Fecha os arquivos */
   FechaArquivo(streamReg,  NOME_ARQ_BIN);
   FechaArquivo(streamArv, NOME_ARQ_ARV);

   printf("\nObrigado por usar este programa. Bye.an\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

Criando arvore...

        >>> Arvore criada com sucesso

        *** Tempo gasto na operacao: 199.00 segundos ***

Checando a consistencia do programa...

>>>     Estatisticas do Banco de Dados <<<

        >>> Grau da arvore: 340
        >>> Tamanho de um no' COM preenchimento: 4076
        >>> Tamanho de um no' SEM preenchimento: 4076
        >>> Tamanho de bloco: 4096

        >>> Numero de nos da arvore: 40393
        >>> Numero de registros no arquivo: 9565483
        >>> Numero de chaves da arvore: 9565483
        >>> Profundidade da arvore: 3
        >>> Numero de chaves repetidas: 0

        >>> Registro com a menor chave <<<

Codigo do aluno: 400010594975
Codigo do curso: 1
Nome da IES: FACULDADE DE CI?NCIAS SOCIAIS APLICADAS DE BELO HORIZONTE
Nome do curso: PEDAGOGIA
Sexo do aluno: Masculino
Idade do aluno: 46
UF de nascimento: MINAS GERAIS
Ano de ingresso: 2010

        >>> Registro com a maior chave <<<

Codigo do aluno: 400012124372
Codigo do curso: 21569970
Nome da IES: INSTITUTO FEDERAL DE EDUCA??O, CI?NCIA E TECNOLOGIA DE S?O PAULO
Nome do curso: MATEM?TICA
Sexo do aluno: Masculino
Idade do aluno: 55
UF de nascimento:
Ano de ingresso: 2011

...consistencia OK

        *** Tempo gasto na operacao:  0.00 segundos ***

>>> Escolha uma das opcoes a seguir:

        [1] Apresenta chaves em ordem
        [2] Consulta dados de um aluno
        [3] Insere um aluno
        [4] Altera dados de um aluno
        [5] Remove um aluno
        [6] Encerra o programa

**********************************************************/

