/****
 *
 * Arquivo TabelaDE.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2008
 * ltima modificao: 07/06/2016
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso com encadeamento
 *
 ****/

/*********************** Includes *************************/

#include "TabelaDE.h"   /* Interface deste mdulo      */
#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include "Registros.h"  /* Processamento de registros  */
#include "Lista.h"      /* Lista encadeada */

/********************* Funes Locais *********************/

/****
 *
 * ComparaListas(): compara duas estruturas do tipo
 *                  tElementosEmLista
 *
 * Parmetros:
 *      p1 (entrada) - ponteiro para a primeira estrutura
 *      p2 (entrada) - ponteiro para a segunda estrutura
 *
 * Retorno:
 *   0, se os campos nElementos das estruturas forem iguais
 *   > 0, se o campo nElementos da primeira estrutura for maior
 *   < 0, se o campo nElementos da primeira estrutura for menor
 *
 * Observao: Esta funo  usada para ordenar as listas
 *             de acordo com seus nmeros de elementos
 *
 ****/
static int ComparaListas(const void *e1, const void *e2)
{
   tElementosEmLista ec1 = *(tElementosEmLista *)e1,
                     ec2 = *(tElementosEmLista *)e2;

   return (int) ec1.nElementos - ec2.nElementos;
}

/********************* Funes Globais ********************/

/****
 *
 * CriaTabelaDE(): Cria e inicializa uma tabela de disperso
 *                 com encadeamento
 *
 * Parmetros:
 *      nElementos (entrada) - nmero de posies da tabela
 *                             de disperso
 *
 * Retorno: A tabela de disperso criada
 *
 ****/
tTabelaDE CriaTabelaDE(int nElementos)
{
   tTabelaDE tabela;
   int       i;

      /* Aloca o array que representa a tabela de disperso */
   tabela = calloc(nElementos, sizeof(tListaSE));

      /* Certifica que houve alocao */
   ASSEGURA(tabela, "Impossivel alocar tabela de dispersao");

      /* Inicializa as listas encadeadas */
   for (i = 0; i < nElementos; ++i)
      IniciaListaSE(tabela + i);

   return tabela;
}

/****
 *
 * BuscaDE(): Executa uma busca simples numa tabela de
 *            disperso com encadeamento
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *      chave (entrada) - a chave de busca
 *      fDispersao (entrada) - ponteiro para a funo de disperso a ser usada
 *
 * Retorno: ndice do registro no arquivo de registros,
 *          se a chave for encontrada; -1, em caso contrrio
 *
 ****/
int BuscaDE(tTabelaDE tabela, int tamTabela, tCEP chave, tFDispersao fDispersao)
{
   int posColetor;

      /* Encontra a respectiva lista encadeada */
   posColetor = fDispersao(chave)%tamTabela;

      /* A lista deve ter sido encontrada */
   ASSEGURA( posColetor >= 0 && posColetor < tamTabela,
             "ERRO: Impossivel localizar lista" );

      /* A funo BuscaListaSE() completa o servio */
   return BuscaListaSE(tabela + posColetor, chave);
}

/****
 *
 * InsereDE(): Insere uma nova chave numa tabela de disperso
 *             com encadeamento
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *      conteudo (entrada) - a chave de busca e seu
 *                           respectivo ndice
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: Nada
 *
 ****/
void InsereDE( tTabelaDE tabela, int tamTabela,
               const tCEP_Ind *conteudo, tFDispersao fDispersao )
{
   int posColetor;

      /* Encontra a lista encadeada que dever conter a chave */
   posColetor = fDispersao(conteudo->chave)%tamTabela;

      /* Garante que o ndice  vlido */
   ASSEGURA( posColetor >= 0 && posColetor < tamTabela,
             "ERRO: Impossivel localizar um lista" );

      /* A funo InsereListaSE() completa o servio */
   return InsereListaSE(&tabela[posColetor], conteudo );
}

/****
 *
 * RemoveDE(): Remove um item de uma tabela disperso com
 *             encadeamento
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *      chave (entrada) - a chave de busca
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveDE(tTabelaDE tabela, int tamTabela, tCEP chave, tFDispersao fDispersao)
{
   int posColetor;

      /* Encontra a lista encadeada que contm a chave */
   posColetor = fDispersao(chave)%tamTabela;

      /* Garante que o ndice  vlido */
   ASSEGURA( posColetor > 0 && posColetor < tamTabela,
             "ERRO: Impossivel localizar lista" );

      /* A funo RemoveListaSE() completa o servio */
   return RemoveListaSE(tabela +posColetor, chave);
}

/****
 *
 * NChavesEmListasDE(): Escreve num arquivo o nmero de
 *                        elementos em cada continer da
 *                        tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *      stream (entrada) - stream associado ao arquivo onde o
 *                         resultado da operao ser escrito
 *
 * Retorno: Nada
 *
 ****/
void NChavesEmListasDE( tTabelaDE tabela, int tamTabela, FILE *stream )
{
   int                i, n, nMax, frequencia;
   tElementosEmLista *ar;

   ASSEGURA(stream, "\nStream recebido e' NULL\n");

      /* Aloca o array que ir armazenar ndices de */
      /* listas e respectivos nmeros de elementos  */
   ar = calloc(tamTabela, sizeof(tElementosEmLista));

   ASSEGURA(ar, "\nNao foi possivel alocar array\n");

      /* Preenche o contedo do array */
   for (i = 0; i < tamTabela; ++i) {
      ar[i].iLista = i;
      ar[i].nElementos = ComprimentoListaSE(tabela[i]);
   }

      /* Ordena o array de acordo com */
      /* os nmeros de elementos      */
   qsort( ar, tamTabela, sizeof(tElementosEmLista),
          ComparaListas );

   for (i = 0; i < tamTabela; ++i)
      fprintf( stream, "%d\t%d\n", ar[i].iLista,
               ar[i].nElementos );

   nMax = ar[tamTabela - 1].nElementos;

   fprintf(stream, "\n");

      /* O cdigo a seguir pode ser bastante melhorado, */
      /* uma vez que o array  ordenado pelo nmero de */
      /* elementos em cada continer                   */
   for (n = 0; n <= nMax; ++n) {
      frequencia = 0;

      for (i = 0; i < tamTabela; ++i)
         if (ar[i].nElementos == n)
            ++frequencia;

      if (frequencia)
         fprintf( stream, "Numero de listas com %d "
                  "elementos:\t%4.d\n", n, frequencia );
   }

   free(ar); /* O array no  mais necessrio */
}

/****
 *
 * NChavesDE(): Calcula o nmero de chaves na tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *
 * Retorno: O nmero de chaves na tabela de disperso
 *
 ****/
int NChavesDE(tTabelaDE tabela, int tamTabela)
{
   int i, n = 0;

   for (i = 0; i < tamTabela; ++i)
      n += ComprimentoListaSE(tabela[i]);

   return n;
}

/****
 *
 * NListasNaoVaziasDE(): Determina o nmero de listas que no
 *                     so vazias numa tabela de disperso
 *                     com encadeamento
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *
 * Retorno: O nmero de conteiners que no so vazios
 *
 ****/
int NListasNaoVaziasDE( tTabelaDE tabela, int tamTabela )
{
   int i, n = 0;

   for (i = 0; i < tamTabela; ++i)
      if (tabela[i])
         n++;

   return n;
}

/****
 *
 * DestroiTabelaDE(): Libera o espao ocupado por uma tabela
 *                  de disperso com encadeamento
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *
 * Retorno: Nada
 *
 ****/
void DestroiTabelaDE(tTabelaDE *tabela, int tamTabela)
{
   int i;

      /* Destri as listas encadeadas */
   for (i = 0; i < tamTabela; ++i) {
      DestroiListaSE(*tabela + i);
   }

      /* Libera o espao ocupado pelo array */
   free(*tabela);

   *tabela = NULL;
}
