/****
 *
 * Arquivo TabelaDEA.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2008
 * ltima modificao: 08/06/2010
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso aberto
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include "TabelaDEA.h"  /* Interface deste mdulo      */
#include "Registros.h"  /* Processamento de registros  */

/***************** Definies de Funes ******************/

/****
 *
 * RedimensionaTabDEA(): Redimensiona uma tabela de disperso com
 *                       endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso atual
 *      tamanhoTab (entrada) - tamanho da tabela de disperso atual
 *      fDispersao (entrada) - ponteiro para a funo de disperso
 *                             a ser usada
 *
 * Retorno: Endereo da nova tabela
 *
 * Observao: O redimensionamento sempre dobra o tamanho atual da
 *             tabela
 *
 ****/
static tTabelaDEA RedimensionaTabDEA( tTabelaDEA tabela, int tamanhoTab,
                                      tFDispersaoDEA fDispersao )
{
   tTabelaDEA tabelaNova;
   int        i, tamanhoNovo = 2*tamanhoTab;

   tabelaNova = calloc(tamanhoNovo, sizeof(tColetorDEA));
   ASSEGURA(tabelaNova, "Impossivel redimensionar a tabela");

      /* Inicia os coletores */
   for (i = 0; i < tamanhoNovo; ++i)
         /* Todos os coletores da nova tabela esto inicialmente desocupados */
      tabela[i].status = VAZIO;

      /* Transfere as chaves da tabela antiga para a tabela nova */
   for (i = 0; i < tamanhoTab; ++i)
      if (tabela[i].status == OCUPADO)
         ASSEGURA( InsereDEA( tabelaNova, tamanhoNovo,
                              tabela[i].chaveEIndice, fDispersao ),
                   "Erro de insercao de redimensionamento");

   free(tabela); /* Libera o espao ocupado pela tabela antiga */

   return tabelaNova;
}

/****
 *
 * CriaTabelaDEA(): Cria e inicializa uma tabela de disperso
 *                  com endereamento aberto
 *
 * Parmetros:
 *      nElementos (entrada) - nmero de posies da tabela
 *                             de disperso
 *
 * Retorno: A tabela de disperso criada
 *
 ****/
tTabelaDEA CriaTabelaDEA(int nElementos)
{
   tTabelaDEA  tabela;
   int         i;

   tabela = calloc(nElementos, sizeof(tColetorDEA));
   ASSEGURA( tabela, "Impossivel alocar a tabela de dispersao" );

   for (i = 0; i < nElementos; ++i)
         /* Todos os elementos esto inicialmente desocupados */
      tabela[i].status = VAZIO;

   return tabela;
}

/****
 *
 * BuscaDEA(): Executa uma busca simples numa tabela de disperso
 *             com endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *      chave (entrada) - a chave de busca
 *      fDispersao (entrada) - ponteiro para a funo de
 *                             disperso a ser usada
 *
 * Retorno: ndice do registro no arquivo de registros,
 *          se a chave for encontrada; -1, em caso contrrio
 *
 ****/
int BuscaDEA( tTabelaDEA tabela, int tamanhoTab,
              tCEP chave, tFDispersaoDEA fDispersao )
{
   int i, pos;

      /* Faz sucessivas sondagens at encontrar a chave, */
      /* uma posio vazia ou a ltima sondagem          */
   for (i = 0; i < tamanhoTab; ++i) {
      pos = fDispersao(chave, i, tamanhoTab);
      ASSEGURA(pos < tamanhoTab, "Dispersao invalida recebido por BuscaDEA()");

      if (tabela[pos].status == VAZIO)
         return -1; /* A chave no foi encontrada */

      if (tabela[pos].status == OCUPADO &&
          !memcmp( tabela[pos].chaveEIndice.chave, chave, sizeof(tCEP) ))
         return tabela[pos].chaveEIndice.indice; /* Chave encontrada */
   }

   return -1;  /* A chave no foi encontrada */
}

/****
 *
 * InsereDEA(): Faz insero numa tabela de disperso com
 *              endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *      chaveEIndice (entrada) - a chave de busca e seu respectivo ndice
 *      fDispersao (entrada) - ponteiro para a funo de disperso a ser usada
 *
 * Retorno: 1, se houver insero; 0, em caso contrrio
 *
 ****/
int InsereDEA( tTabelaDEA tabela, int tamanhoTab,
               tCEP_Ind chaveEIndice, tFDispersaoDEA fDispersao )
{
   int i, pos, posInsercao = -1;

      /* Faz sucessivas sondagens at encontrar a chave */
      /* ou um local onde se pode fazer a insero.     */
   for (i = 0; i < tamanhoTab; ++i) {
      pos = fDispersao(chaveEIndice.chave, i, tamanhoTab);
      ASSEGURA( pos < tamanhoTab, "Valor invalido recebido por InsereDEA()" );

      if ( tabela[pos].status == REMOVIDO && posInsercao < 0 )
            /* Se a chave no for encontrada numa nova   */
            /* sondagem, esta ser a posio de insero */
         posInsercao = pos;

      if (tabela[pos].status == VAZIO && posInsercao < 0) {
         posInsercao = pos; /* A chave no foi encontrada */
         break; /* A chave ser inserida fora do lao */
      }

      if ( tabela[pos].status == OCUPADO &&
           !memcmp( tabela[pos].chaveEIndice.chave,
                    chaveEIndice.chave, sizeof(tCEP) ) )
         return 0; /* Chave encontrada */
   }

      /* Insere a chave e respectivo ndice na  */
      /* posio determinada dentro do lao for */
   tabela[posInsercao].chaveEIndice = chaveEIndice;
   tabela[posInsercao].status = OCUPADO;

   return 1;
}

/****
 *
 * RemoveDEA(): Remove uma chave de uma tabela de disperso com
 *              endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *      chave (entrada) - a chave de busca
 *      fDispersao (entrada) - ponteiro para a funo de disperso a ser usada
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveDEA( tTabelaDEA tabela, int tamanhoTab, tCEP chave,
               tFDispersaoDEA fDispersao )
{
   int i, pos;

      /* Faz sucessivas sondagens at encontrar a chave, */
      /* uma posio vazia ou atingir a ltima sondagem. */
   for (i = 0; i < tamanhoTab; ++i) {
      pos = fDispersao(chave, i, tamanhoTab);
      ASSEGURA(pos < tamanhoTab, "Valor dispersao invalido em RemoveDEA()");

      if (tabela[pos].status == VAZIO)
         return 0; /* A chave no foi encontrada */

      if (tabela[pos].status == OCUPADO &&
          !memcmp( tabela[pos].chaveEIndice.chave, chave,sizeof(tCEP))) {
         tabela[pos].status = REMOVIDO; /* Marca como removido */

         return 1; /* Remoo foi OK */
      }
   }

   return 0;  /* A chave no foi encontrada */
}

/****
 *
 * NChavesDEA(): Calcula o nmero de chaves na tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *
 * Retorno: O nmero de chaves na tabela de disperso
 *
 ****/
int NChavesDEA(tTabelaDEA tabela, int tamanhoTab)
{
   int i, nChaves = 0;

   for (i = 0; i < tamanhoTab; ++i)
      if (tabela[i].status == OCUPADO)
         ++nChaves;

   return nChaves;
}

/****
 *
 * NColetoresVaziosDEA(): Calcula o nmero de elementos vazios
 *                     na tabela de disperso
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *
 * Retorno: O nmero de elementos vazios na tabela de disperso
 *
 ****/
int NColetoresVaziosDEA(tTabelaDEA tabela, int tamanhoTab)
{
   int i, n = 0;

   for (i = 0; i < tamanhoTab; ++i)
      if (tabela[i].status == VAZIO)
         ++n;

   return n;
}

/****
 *
 * NChavesRemovidasDEA(): Determina o nmero de chaves que foram removidas
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamanhoTab (entrada) - tamanho da tabela de disperso
 *
 * Retorno: O nmero de conteiners que no so vazios
 *
 ****/
int NChavesRemovidasDEA(tTabelaDEA tabela, int tamanhoTab)
{
   int i, n = 0;

   for (i = 0; i < tamanhoTab; ++i)
      if (tabela[i].status == REMOVIDO)
         ++n;

   return n;
}

/****
 *
 * DestroiTabelaDEA(): Libera o espao ocupado por uma tabela
 *                  de disperso com endereamento aberto
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      tamTabela (entrada) - tamanho da tabela de disperso
 *
 * Retorno: Nada
 *
 ****/
void DestroiTabelaDEA(tTabelaDEA *tabela)
{
      /* Libera o espao ocupado pelo array */
   free(*tabela);

   *tabela = NULL;
}
