/****
 *
 * Arquivo Dispersao.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 07/08/2017
 * ltima modificao: 25/05/2018
 *
 * Descrio: Implementao de funes de disperso
 *
 ****/

#include <stdio.h>     /* printf()         */
#include <stdlib.h>    /* srand() e rand() */
#include <time.h>      /* time()           */
#include <limits.h>    /* UCHAR_MAX        */
#include "Dispersao.h" /* Interface deste mdulo */

/***************** Funes de Disperso *******************/

/****
 *
 * DispersaoPolinomial(): Calcula o valor de disperso de um
 *                        string usando o mtodo polinomial
 *
 * Parmetros:
 *      chave (entrada) - o string
 *
 * Retorno: Nada
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoPolinomial(const char *chave)
{
   unsigned int dispersao = 0,
                base = 2,
                primo = 31;

   while (*chave) {
      dispersao = (dispersao*base + *chave++)%primo;
   }

   return dispersao;
}

/****
 *
 * DispersaoRotativa(): Calcula o valor de disperso de um
 *                      string usando o mtodo rotativo
 *
 * Parmetros:
 *      chave (entrada) - o string
 *
 * Retorno: Nada
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoRotativa(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
       dispersao = (dispersao << 4) ^
                   (dispersao >> 28) ^ (*chave++);
   }

   return dispersao;
}

/****
 *
 * DispersaoJenkins(): Implementao da funo de disperso JOAAT
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoJenkins(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao += *chave++;
      dispersao += (dispersao << 10);
      dispersao ^= (dispersao >> 6);
   }

   dispersao += (dispersao << 3);
   dispersao ^= (dispersao >> 11);
   dispersao += (dispersao << 15);

   return dispersao;
}

/****
 *
 * DispersaoDJB(): Implementao da funo de disperso DJB
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoDJB(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao = 33*dispersao + *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoDJB2(): Implementao da funo de disperso DJB2
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoDJB2(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao = 33*dispersao ^ *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoSAX(): Implementao da funo de disperso SAX
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoSAX(const char *chave)
{
   unsigned int dispersao = 0;

   while (*chave) {
      dispersao ^= (dispersao << 5) + (dispersao >> 2) + *chave++;
   }

   return dispersao;
}

/****
 *
 * DispersaoFNV(): Implementao da funo de disperso FNV
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoFNV(const char *chave)
{
   unsigned int dispersao = 2166136261;

   while (*chave) {
      dispersao = (dispersao*16777619) ^ *chave++;
   }

   return dispersao;
}

/****
 *
 * CriaTabela(): Cria uma tabela de valores aleatrios
 *
 * Parmetros:
 *      tab[] (sada) - array que representa a tabela
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 ****/
static void CriaTabela(int tab[], int tam)
{
   int i;

      /* Inicia o gerador de nmeros aleatrios */
   srand((unsigned)time(NULL));

      /* Cria a tabela de valores aleatrios */
   for (i = 0; i < tam; ++i) {
      tab[i] = rand();
   }
}

/****
 *
 * DispersaoJSW(): Implementao da funo de disperso JSW
 *
 * Parmetros:
 *      chave (entrada) - a chave cujo valor de disperso
 *                        ser calculado
 *
 * Retorno: O valor de disperso calculado
 *
 * Observao: Esta funo no testa se o ponteiro que
 *             representa o string  NULL (como deveria)
 *
 ****/
unsigned int DispersaoJSW(const char *chave)
{
   unsigned int dispersao = 16777551;
   static int   primeiraChamada = 1,
                tabela[UCHAR_MAX];

   if (primeiraChamada) {
      CriaTabela(tabela, UCHAR_MAX);
      primeiraChamada = 0;
   }

   while (*chave) {
      dispersao = (dispersao << 1 | dispersao >> 31) ^
                  tabela[*chave++];
   }

   return dispersao;
}

