/****
 *
 * Exemplo: Testando uma Funo de Disperso Pronta
 *          (verso alternativa)
 *
 * Seo: 7.7.1
 *
 * Descrio:
 *    -> L arquivo contendo CEPs
 *
 *    -> Armazena as chaves (CEPs) numa lista indexada
 *       sem ordenao
 *
 *    -> Aloca um array de inteiros do tamanho da lista
 *
 *    -> Arquivo de dados: CEPs.bin
 *
 * Observao: Para criar o arquivo binrio, use o programa
 *             CriaArquivoBin.c neste diretrio
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de Entrada/Sada      */
#include <string.h>    /* Processamento de strings      */
#include <stdlib.h>    /* Miscelnea                    */
#include <time.h>      /* Medio de tempo              */
#include "Lista.h"     /* Interface do tipo lista       */
#include "leitura.h"   /* Leitura via teclado           */
#include "Dispersao.h" /* Interface do mdulo Dispersao */
#include "ASSEGURA.h"  /* Macro ASSEGURA()              */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo de dados */
#define NOME_ARQUIVO_BIN "J:\\Dados\\CEPs\\CEPs.bin"

   /* Nome do arquivo contendo as chaves da tabela de busca */
#define NOME_ARQUIVO_CHAVES "ListaDeChaves.txt"

   /* Nome do arquivo que contm os resultados de um teste */
#define NOME_ARQUIVO_RES "Resultado1.txt"

#define MAX_NOME  40 /* Nmero mximo de caracteres num nome */
#define TAM_UF     2 /* Nmero de dgitos num UF */
#define TAM_CEP    8 /* Nmero de dgitos num CEP */
#define MAX_COMP  60 /* Nmero mximo de caracteres */
                     /* num complemento             */
#define MAX_TIPO_LOG 10 /* Nmero mximo de caracteres */
                        /* num tipo de logradouro      */
#define TAM_DNE   16 /* Nmero de dgitos numa chave DNE */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/****************** Definies de Tipos *******************/

   /* Tipo de ponteiro para uma funo de disperso */
typedef unsigned int (*tFDispersao)(const char *);

typedef struct {
           int  numero;
           char UF[TAM_UF];
           int  localidadeNumero;
           char nomeAbr[MAX_NOME];
           char nome[MAX_NOME];
           int  bairroInicio;
           int  bairroFim;
           char CEP[TAM_CEP];
           char complemento[MAX_COMP];
           char tipoLogradouro[MAX_TIPO_LOG];
           char statusTipoLogradouro;
           char nomeSemAcento[MAX_NOME];
           char chaveDNE[TAM_DNE];
         } tRegistro;

/************************* Aluses ************************/

extern void LeArquivo(const char *arqBin, tLista *lista);
extern void MedidaDeTempo(void);
extern void ApresentaMenu(const char *opcoes[], int nOpcoes);
extern int ComparaColisoes(const void *e1, const void *e2);
extern void AnalisaColisoes( int colisoes[], int tam,
                             FILE *stream );
extern void TestaDispersao( tLista *lista, tFDispersao funcao,
                            FILE *stream );

/***************** Definies de Funes ******************/

/****
 *
 * LeArquivo(): L arquivo de dados (binrio) e armazena
 *              a chave de cada registro numa lista indexada
 *
 * Parmetros:
 *      arqBin (entrada) - nome do arquivo binrio
 *      lista (sada) - lista que conter as chaves dos
 *                      registros lidos
 *
 * Retorno: Nada
 *
 ****/
void LeArquivo( const char *arqBin, tLista *lista )
{
   FILE      *stream;
   tElemento  umElemento;
   tRegistro  umRegistro;

      /* Tenta abrir o arquivo binrio para leitura e escrita */
   stream = fopen(arqBin, "rb");

      /* Se o arquivo no pode ser aberto, aborta o programa */
   ASSEGURA(stream, "Arquivo nao pode ser aberto");

      /* L cada registro do arquivo e acrescenta */
      /* sua chave (CEP) na lista                 */
   while (1) {
         /* L um registro no arquivo */
      fread( &umRegistro, sizeof(umRegistro), 1, stream);

         /* Verifica se ocorreu erro ou o */
         /* final do arquivo foi atingido */
      if (ferror(stream) || feof(stream)) {
         break;
      }

         /* A chave no  armazenada como string no registro. */
         /* Portanto, no se pode usar strcpy().              */
      strncpy(umElemento.chave, umRegistro.CEP, TAM_CEP);

         /* Transforma a chave num string. Isso no  */
         /* essencial, mas facilita o processamnto.   */
      umElemento.chave[TAM_CEP] = '\0';

         /* Acrescenta a chave na lista. O valor associado */
         /*  chave  sua posio (ndice) na lista.       */
      AcrescentaElemento(lista, &umElemento);
   }

      /* Se ocorreu erro de leitura no arquivo, fecha */
      /* o arquivo e aborta o programa                */
   if (ferror(stream)) {
      fclose(stream);
      ASSEGURA(0, "Erro de leitura de arquivo ");
   }

   fclose(stream);
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      printf( "\n\n\t*** Tempo gasto na operacao: "
              "%5.2f segundos ***\n", difftime(tf, t0) );
   }
}

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\nEscolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t[%d] %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * ComparaColisoes(): Compara dois elementos do tipo tColisoes
 *
 * Parmetros:
 *      e1, e2 (entrada) - os elementos
 *
 * Retorno:
 *    * Um valor positivo se e1 > e2
 *    * Um valor negativo se e1 < e2
 *    * Zero se e1 = e2
 *
 ****/
int ComparaColisoes(const void *e1, const void *e2)
{
   int c1 = *(int *)e1,
       c2 = *(int *)e2;

   return c2 - c1;
}

/****
 *
 * AnalisaColisoes(): Apresenta num stream de texto o resultado
 *                    de uma anlise da lista de colises
 *
 * Parmetros:
 *      colisoes[] (entrada/sada) - array que representa
 *                                   a lista de colises
 *      tam (entrada) - tamanho da lista de colises
 *      stream (entrada) - stream de texto associado ao meio de
 *                         sada no qual o resultado ser escrito
 *
 * Retorno: Nada
 *
 ****/
void AnalisaColisoes(int colisoes[], int tam, FILE *stream)
{
   int i,
       vazio = 0,
       nColisoes = 0;

      /* Ordena a lista de colises em ordem */
      /* decrescente de nmero de colises   */
   qsort(colisoes, tam, sizeof(colisoes[0]), ComparaColisoes);

   for (i = 0; i < tam; ++i) {
         /* Calcula o nmero de posies da */
         /* tabela que no foram usadas     */
      if (!colisoes[i]) {
         ++vazio;
      }

         /* Calcula o nmero de colies */
      if (colisoes[i] > 1) {
         nColisoes += colisoes[i];
      }
   }

      /*                                 */
      /* Escreve os resultados no stream */
      /*                                 */

   fprintf( stream, "\n>>> Tamanho da tabela: %d", tam );
   fprintf( stream, "\n>>> Posicoes nao ocupadas: %d", vazio );
   fprintf( stream, "\n>>> Numero total de colisoes: %d",
            nColisoes );
   fprintf( stream, "\n>>> Colisoes maximas numa posicao: %d",
            colisoes[0] );
   fprintf( stream, "\n>>> Desperdicio de memoria: %4.2f%%\n",
            100*((double) vazio/tam) );
}

/****
 *
 * TestaDispersao(): Testa uma funo de disperso com uma
 *                   lista de chaves
 *
 * Parmetros:
 *      lista (entrada) - a lista de chaves que ser utilizada
 *                        no teste
 *      funcao (entrada) - ponteiro para a funo de disperso que
 *                         ser testada
 *      stream (entrada) - stream de texto associado ao meio de
 *                         sada no qual o resultado ser escrito
 *
 * Retorno: Nada
 *
 ****/
void TestaDispersao( tLista *lista, tFDispersao funcao,
                     FILE *stream )
{
   int        i,
              tam; /* Nmero de chaves na lista */
   unsigned   dispersao; /* Um valor de disperso */
   tElemento  elemento;  /* Um elemento da lista  */
   int       *colisoes; /* Ponteiro para um array que conta */
                        /* o nmeros de colises            */

   tam = Comprimento(lista);

      /* Aloca o array que contar o nmero de colises */
   colisoes = calloc(tam, sizeof(int));
   ASSEGURA(colisoes, "Impossivel alocar array");

   for (i = 0; i < tam; ++i) {
         /* Obtm um elemento da lista */
      elemento = ObtemElemento(lista, i);

         /* Calcula o valor de disperso da chave */
      dispersao = funcao(elemento.chave)%tam;

         /* Conta quantas vezes esse valor */
         /*de disperso foi obtido         */
      ++colisoes[dispersao];
   }

      /* Analisa e apresenta o resultado do teste */
   AnalisaColisoes(colisoes, tam, stream);

   free(colisoes); /* Esse array no  mais necessrio */
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa as operaes sobre uma tabela de busca
 *         implementada como lista indexada sem ordenao
 *         contendo apenas as chaves dos registros
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tLista      lista; /* Lista de chaves */
   const char *opcoes[] = { "Dispersao Modular",
                            "Dispersao Aditiva",
                            "Dispersao Polinomial",
                            "Dispersao Multiplicativa",
                            "Dispersao Rotativa",
                            "Dispersao por Mistura",
                            "Testa Todas as Dispersoes",
                            "Encerra o programa"
                          };
   int         op, /* Opo escolhida pelo usurio */
               nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]);
   FILE       *stream;

   MedidaDeTempo();
   printf("\nCriando lista...");

      /* Inicia a lista que armazena a tabela de busca */
   IniciaLista(&lista);

      /* L o contedo do arquivo de */
      /* dados e o armazena as chaves na lista */
   LeArquivo(NOME_ARQUIVO_BIN, &lista);

   printf("\n...lista criada com sucesso\n");
   MedidaDeTempo();

      /* O lao a seguir encerra quando o usurio */
      /* escolher a opo de encerramento         */
   while (1) {
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("12345678");

         /* Verifica se o usurio quer encerrar o programa */
      if (op == '8') { /* Encerra o programa */
         DestroiLista(&lista);

         break; /* Sada do lao */
      }

         /* Processa as demais opes */
      switch (op) {
         case '1': /* Disperso modular */
            printf("\n\t>>> Dispersao Modular <<<\n");
            TestaDispersao(&lista, DispersaoModular, stdout);
            break;
         case '2': /* Disperso aditiva */
            printf("\n\t>>> Dispersao Aditiva <<<\n");
            TestaDispersao(&lista, DispersaoAditiva, stdout);
            break;

         case '3': /* Disperso polinomial */
            printf("\n\t>>> Dispersao Polinomial <<<\n");
            TestaDispersao(&lista, DispersaoPolinomial, stdout);
            break;

         case '4': /* Disperso multiplicativa */
            printf("\n\t>>> Dispersao Multiplicativa <<<\n");
            TestaDispersao(&lista, DispersaoMultiplicativa, stdout);
            break;

         case '5': /* Disperso rotativa */
            printf("\n\t>>> Dispersao Rotativa <<<\n");
            TestaDispersao(&lista, DispersaoRotativa, stdout);
            break;

         case '6': /* Disperso por mistura */
            printf("\n\t>>> Dispersao por Mistura <<<\n");
            TestaDispersao(&lista, DispersaoMistura, stdout);
            break;

         case '7': /* Testa todas as funes */
            stream = fopen(NOME_ARQUIVO_RES, "w");
            ASSEGURA(stream, "Arquivo nao foi aberto");

            fprintf( stream, "***** Testes de Funcoes de "
                     "Dispersao *****\n" );

            fprintf(stream, "\n\t>>> Dispersao Modular <<<\n");
            TestaDispersao(&lista, DispersaoModular, stream);

            fprintf(stream, "\n\t>>> Dispersao Aditiva <<<\n");
            TestaDispersao(&lista, DispersaoAditiva, stream);

            fprintf(stream, "\n\t>>> Dispersao Polinomial <<<\n");
            TestaDispersao(&lista, DispersaoPolinomial, stream);

            fprintf(stream, "\n\t>>> Dispersao Multiplicativa <<<\n");
            TestaDispersao(&lista, DispersaoMultiplicativa, stream);

            fprintf(stream, "\n\t>>> Dispersao Rotativa <<<\n");
            TestaDispersao(&lista, DispersaoRotativa, stream);

            fprintf(stream, "\n\t>>> Dispersao por Mistura <<<\n");
            TestaDispersao(&lista, DispersaoMistura, stream);

            fclose(stream);

            printf( "\n\t>>> Resultado escrito no arquivo "
                    "\"%s\" <<<\n", NOME_ARQUIVO_RES );

            break;

         default: /* O programa no deve chegar at aqui */
            printf("\nEste programa contem um erro\n");
            return 1;
      }
   }

         /* Despede-se do usurio */
   printf( "\n\t>>> Obrigado por usar este programa.\n");

   return 0;
}

/******************* Exemplo de Execuo ******************



**********************************************************/

