/****
 *
 * Arquivo DispersaoEstatica.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 06/09/2017
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso esttica em memria
 *            secundria
 *
 ****/

/*********************** Includes *************************/

#include <limits.h>
#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include <math.h>       /* Funes strxxx() e memxxx() */
#include "DispersaoEstatica.h" /* Interface deste mdulo */
#include "Registros.h"  /* Processamento de registros  */
#include "ASSEGURA.h"   /* Macro ASSEGURA              */

/********************* Funes Locais *********************/

/****
 *
 * FDispersao(): Obtm o valor de disperso de uma chave inteira
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *      nColetores (entrada) - o nmero de coletores primrios
 *
 * Retorno: O valor de disperso
 *
 ****/
static int FDispersao(tChave chave, int nColetores)
{
//   return DispersaoRotativa(chave) % nColetores;
   return chave % nColetores;
}

/****
 *
 * NovoColetorDEst(): Cria um novo coletor num arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que contm
 *                        os coletores
 *     tipo (entrada) - tipo do coletor que ser criado
 *
 * Retorno: O ndice do coletor criado
 *
 ****/
static int NovoColetorDEst(FILE *stream, tTipoDeColetorDEst tipo)
{
   tColetorDEst coletor;
   int          indice;

   coletor.nRegistros = 0;
   coletor.proximo = POSICAO_NULA;
   coletor.tipo = tipo;

      /* Move o apontador de posio do arquivo para seu final */
   MoveApontador(stream, 0, SEEK_END);

   indice = ftell(stream)/sizeof(coletor); /* Obtm o ndice do coletor */

   fwrite(&coletor, sizeof(coletor), 1, stream); /* Escreve o coletor no arquivo */

   return indice;
}

/****
 *
 * LeColetorDEst(): L um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo
 *                        os coletores no qual ser feita a leitura
 *     pos (entrada) - posio no arquivo na qual ser feita a
 *                     leitura
 *    *coletor (sada) - ponteiro para o coletor que conter o
 *                       resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
static void LeColetorDEst(FILE *stream, int pos, tColetorDEst *coletor)
{
      /* Tenta mover o apontador de arquivo para o local */
      /* de leitura; se no conseguir, aborta o programa */
   MoveApontador(stream, sizeof(tColetorDEst)*pos, SEEK_SET);

   fread(coletor, sizeof(tColetorDEst), 1, stream); /* Efetua a leitura */

      /* Certifica-se que no houve erro de leitura */
   ASSEGURA(!ferror(stream),  "Erro de leitura em LeColetorDEst()");
}

/****
 *
 * EscreveColetorDEst(): Escreve um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo os
 *                        coletores no qual ser feita a escrita
 *     pos (entrada) - posio no arquivo no qual ser feita a escrita
 *    *coletor (entrada) - coletor que ser escrito no arquivo
 *
 * Retorno: Nada
 *
 ****/
static void EscreveColetorDEst(FILE *stream, int pos, const tColetorDEst *coletor)
{
      /* Tenta mover o apontador de arquivo para o local */
      /* de escrita; se no conseguir, aborta o programa */
   MoveApontador(stream, sizeof(tColetorDEst)*pos, SEEK_SET);

   fwrite(coletor, sizeof(tColetorDEst), 1, stream); /* Efetua a escrita */

      /* Certifica-se que no houve erro */
   ASSEGURA(!ferror(stream), "Erro de escrita em EscreveColetorDEst()");
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaTabDEst(): Inicia uma tabela de busca com disperso esttica
 *
 * Parmetros:
 *      streamCol (entrada) - stream associado ao arquivo que
 *                            contm os coletores
 *      nRegistros (entrada) - nmero de registros iniciais da tabela
 *      maxRegs (entrada) - nmero mximo de registros num coletor
 *
 * Retorno: O nmero de coletores primrios
 *
 ****/
int IniciaTabDEst( FILE *streamCol, int nRegistros, int maxRegs )
{
   int i,
       nColetores;

   nColetores = nRegistros/maxRegs; /* Obtm o nmero de coletores */

      /* Aloca os coletores primrios no arquivo */
   for (i = 0; i < nColetores; ++i)
      (void) NovoColetorDEst(streamCol, PRIMARIO);

   return nColetores;
}

/****
 *
 * BuscaDEst(): Efetua uma busca numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *      nColetores (entrada) - nmero de coletores primrios da tabela
 *      chave (entrada) - a chave de busca
 *      reg (sada) - o registro encontrado
 *
 * Retorno: O endereo do registro, se ele for encontrado; NULL, em caso contrrio
 *
 ****/
tRegistroMEC *BuscaDEst( FILE *stream, int nColetores, tChave chave,
                         tRegistroMEC *reg )
{
   int          iColetor, i;
   tColetorDEst coletor;

      /* Obtm o ndice do coletor primrio que pode conter o registro */
   iColetor = FDispersao(chave, nColetores);

      /* Efetua uma busca no coletor primrio e em seus coletores excedentes */
   while (iColetor != POSICAO_NULA) {
         /* L um coletor que pode conter o registro */
      LeColetorDEst(stream, iColetor, &coletor);

         /* Efetua uma busca sequencial pelo registro em memria principal */
      for (i = 0; i < coletor.nRegistros; i++)
            /* Verifica se a chave foi encontrada */
         if (chave == ObtemChave(&coletor.registros[i])) {
            *reg = coletor.registros[i]; /* Chave encontrada */
            return reg; /* Servio completo */
         }

      iColetor = coletor.proximo; /* Passa para o prximo coletor excedente */
   }

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * InsereDEst(): Insere um registro numa tabela de busca com disperso esttica
 *
 * Parmetros:
 *      nColetores (entrada) - nmero de coletores primrios da
 *                             tabela na qual ser efetuada a insero
 *      registro (entrada) - o registro que ser inserido
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: Nada
 *
 ****/
void InsereDEst(int nColetores, const tRegistroMEC *registro, FILE *stream)
{
   int          iColetor;
   tChave       chave;
   tColetorDEst coletor;

   chave = ObtemChave(registro); /* Obtm a chave do registro */

   iColetor = FDispersao(chave, nColetores); /* Obtm o ndice do coletor */

      /* L o coletor primrio em arquivo */
   LeColetorDEst(stream, iColetor, &coletor);

      /* Encontra o coletor no qual ser efetuada a insero */
   while (coletor.nRegistros == M) {
         /* O coletor est repleto. Portanto passa-se para o    */
         /* prximo coletor excedente. Se o prximo coletor     */
         /* excedente ainda no existe, cria-se um novo coletor */
      if (coletor.proximo == POSICAO_NULA) {
            /* Cria um novo coletor excedente */
         coletor.proximo = NovoColetorDEst(stream, EXCEDENTE);

            /* O coletor foi alterado e precisa ser reescrito */
         EscreveColetorDEst(stream, iColetor, &coletor);
      }

         /* Guarda a posio do prximo coletor a ser lido */
      iColetor = coletor.proximo;

         /* L um coletor excedente em arquivo */
      LeColetorDEst(stream, coletor.proximo, &coletor);
   }

      /* Acrescenta o novo registro ao final do coletor */
   coletor.registros[coletor.nRegistros] = *registro;

   coletor.nRegistros++; /* O nmero de registros no coletor aumentou */

      /* O coletor foi alterado e  preciso reescrev-lo no arquivo */
   EscreveColetorDEst(stream, iColetor, &coletor);
}

/****
 *
 * RemoveDEst(): Remove um registro de uma tabela de disperso esttica
 *
 * Parmetros:
 *      nColetores (entrada/sada) - nmero de coletores primrios
 *                                   da tabela de disperso
 *      chave (entrada) - a chave do registro que ser removido
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveDEst(int nColetores, tChave chave, FILE *stream)
{
   int          i, j, iColetor, encontrada = 0;
   tColetorDEst coletor;

      /* Obtm o ndice do coletor que pode conter o registro */
   iColetor = FDispersao(chave, nColetores);

      /* Efetua uma busca no coletor primrio e em seus coletores excedentes */
   while (iColetor != POSICAO_NULA) {
         /* L o coletor que pode conter o registro */
      LeColetorDEst(stream, iColetor, &coletor);

         /* Efetua uma busca sequencial pelo registro em memria principal */
      for (i = 0; i < coletor.nRegistros && !encontrada; i++)
            /* Verifica se a chave foi encontrada */
         if (chave == ObtemChave(&coletor.registros[i]))
            encontrada = 1;

         /* Se a chave foi encontrada, encerra a busca */
      if (encontrada)
         break;

      iColetor = coletor.proximo; /* Passa para o prximo coletor excedente */
   }

      /* Verifica se a chave foi encontrada */
   if (!encontrada)
      return 0;  /* A chave no foi encontrada */

      /* Corrige o valor de i que foi incrementado e deixou */
      /* de ser o ndice do elemento a ser removido         */
   --i;

      /* O registro a ser removido encontra-se na posio i do */
      /* coletor corrente.  necessrio mover uma posio para */
      /* trs todos os registros que se encontram  sua frente */
   for (j = i; j < coletor.nRegistros - 1; ++j)
      coletor.registros[j] = coletor.registros[j + 1];

   --coletor.nRegistros; /* O nmero de registros no coletor diminuiu */

      /* O coletor foi alterado e  preciso reescrev-lo no arquivo */
   EscreveColetorDEst(stream, iColetor, &coletor);

   return 1;  /* Remoo bem-sucedida */
}

/****
 *
 * SubstituiDEst(): Substitui o contedo de um registro que
 *                      tem uma determinada chave
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *      nColetores (entrada) - nmero de coletores primrios da tabela
 *      reg (entrada) - o registro substituto
 *
 * Retorno: Nada
 *
 * Observaes:
 *      1. A chave deve ser primria
 *      2. Se a chave no for encontrada, o programa ser abortado
 *
 ****/
void SubstituiDEst( FILE *stream, int nColetores, const tRegistroMEC *reg )
{
   int          iColetor, i;
   tColetorDEst coletor;
   tChave       chave = ObtemChave(reg);

      /* Obtm o ndice do coletor primrio que pode conter o registro */
   iColetor = FDispersao(chave, nColetores);

      /* Efetua uma busca no coletor primrio e em seus coletores excedentes */
   while (iColetor != POSICAO_NULA) {
         /* L um coletor que pode conter o registro */
      LeColetorDEst(stream, iColetor, &coletor);

         /* Efetua uma busca sequencial pelo registro em memria principal */
      for (i = 0; i < coletor.nRegistros; i++)
            /* Verifica se a chave foi encontrada */
         if (chave == ObtemChave(&coletor.registros[i])) {
               /* A chave encontrada. Altera o contedo do registro que a contm */
            coletor.registros[i] = *reg;

            return; /* Servio completo */
         }

      iColetor = coletor.proximo; /* Passa para o prximo coletor excedente */
   }

      /* A chave no foi encontrada */
   ASSEGURA(0, "A chave deveria ter sido encontrada");
}

/****
 *
 * NRegistrosTabDEst(): Calcula o nmero de registros numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de registros na tabela de disperso
 *
 ****/
int NRegistrosTabDEst(FILE *stream)
{
   int          nRegistros = 0;
   tColetorDEst umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      nRegistros += umColetor.nRegistros;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NRegistrosTabDEst" );

   return nRegistros;
}

/****
 *
 * NColetoresDEst(): Calcula o nmero de coletores numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores na tabela de disperso
 *
 ****/
int NColetoresDEst(FILE *stream)
{
   int          n = 0;
   tColetorDEst umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresDEst" );

   return n;
}

/****
 *
 * NColetoresVaziosDEst(): Calcula o nmero de coletores vazios
 *                         numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores vazios
 *
 ****/
int NColetoresVaziosDEst(FILE *stream)
{
   int          n = 0;
   tColetorDEst umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (!umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresVaziosDEst" );

   return n;
}

/****
 *
 * NColetoresRepletosDEst(): Calcula o nmero de coletores repletos
 *                       numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores repletos
 *
 ****/
int NColetoresRepletosDEst(FILE *stream)
{
   int      n = 0;
   tColetorDEst umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros == M)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresRepletosDEst" );

   return n;
}

/****
 *
 * NColetoresOcupadosDEst(): Calcula o nmero de coletores ocupados
 *                           numa tabela de disperso esttica
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores ocupados
 *
 ****/
int NColetoresOcupadosDEst(FILE *stream)
{
   int          n = 0;
   tColetorDEst umColetor;

   rewind(stream); /* Assegura que a leitura comea no incio do arquivo */

   while (1) {
      fread(&umColetor, sizeof(umColetor), 1, stream); /* L um coletor */

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream), "Erro de leitura em NColetoresOcupadosDEst" );

   return n;
}

/****
 *
 * NColetoresExcedentesDEst(): Calcula o nmero de coletores excedentes
 *                         numa tabela de disperso esttica
 *
 * Parmetros:
 *      nColetores (entrada) - nmero de coletores primrios
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores excedentes
 *
 ****/
int NColetoresExcedentesDEst(int nColetores, FILE *stream)
{
   int          i, n = 0;
   tColetorDEst coletor;

      /* Para cada coletor primrio, conta quantos coletores excedentes ele possui */
   for (i = 0; i < nColetores; ++i) {
      LeColetorDEst(stream, i, &coletor);

         /* Conta os coletores excedentes do coletor primrio corrente  */
      while (coletor.proximo != POSICAO_NULA) {
          ++n;

          LeColetorDEst(stream, coletor.proximo, &coletor);
      }
   }

   return n;
}

/****
 *
 * MaiorCadeiaExcedentesDEst(): Calcula o tamanho da maior cadeia de
 *                              coletores excedentes
 *
 * Parmetros:
 *      nColetores (entrada) - nmero de coletores primrios
 *      stream (entrada) - stream associado ao arquivo que contm os coletores
 *
 * Retorno: O nmero de coletores excedentes
 *
 ****/
int MaiorCadeiaExcedentesDEst(int nColetores, FILE *stream)
{
   int          i, tamCadeia, maiorCadeia = 0;
   tColetorDEst coletor;

   for (i = 0; i < nColetores; ++i) {
      LeColetorDEst(stream, i, &coletor);

      tamCadeia = 0;

      while (coletor.proximo != POSICAO_NULA) {
          ++tamCadeia;

          LeColetorDEst(stream, coletor.proximo, &coletor);
      }

      if (tamCadeia > maiorCadeia)
         maiorCadeia = tamCadeia;
   }

   return maiorCadeia;
}

