/****
 *
 * Arquivo DispersaoExt.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 26/08/2017
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso extensvel em memria
 *            secundria
 *
 ****/

/*********************** Includes *************************/

#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include <math.h>       /* Funes strxxx() e memxxx() */
#include "DispersaoExt.h" /* Interface deste mdulo    */
#include "Registros.h"  /* Processamento de registros  */
#include "ASSEGURA.h"   /* Macro ASSEGURA              */

/********************* Funes Locais *********************/

/****
 *
 * Dispersao(): Calcula o valor de disperso de um string usando
 *              uma digital de Rabin
 *
 * Parmetros:
 *      s (entrada) - o string
 *
 * Retorno: O valor de disperso almejado
 *
 ****/
int Dispersao(const char *s)
{
   int valor = 0,
       base = 2,
       primo = 97;

   while (*s)
      valor = (valor*base + *s++)%primo;

   return valor;
}

/****
 *
 * ObtemLSBs(): Obtm os bits menos significativos de uma chave
 *
 * Parmetros:
 *      chave (entrada) - a chave da qual os bits sero obtidos
 *      nBits (entrada) - o nmero de bits menos significativos
 *                        que sero extrados da chave
 *
 * Retorno: Um valor inteiro que representa esses bits
 *
 ****/
static int ObtemLSBs(int chave, int nBits)
{
   int mascara = (1 << nBits) - 1;

   return chave & mascara;
}

/****
 *
 * ValorBit(): Retorna o valor de um bit de um valor inteiro
 *
 * Parmetros:
 *      valor (entrada) - valor que ter seu bit retornado
 *      pos (entrada) - posio do bit que ser retornado
 *
 * Retorno: O valor do bit
 *
 ****/
int ValorBit(int valor, int pos)
{
   return 1 & valor >> pos;
}

/****
 *
 * ObtemBits(): Obtm os bits menos significativos de uma chave
 *
 * Parmetros:
 *      chave (entrada) - a chave
 *      bits (entrada) - o nmero de bits que sero extrados
 *                       da chave
 *
 * Retorno: O valor inteiro que representa esses bits
 *
 ****/
static int ObtemBits(tChave chave, int bits)
{
   int dispersao = Dispersao(chave);

   return ObtemLSBs(dispersao, bits);
}

/****
 *
 * NovoColetor(): Cria um novo coletor num arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        contm os coletores
 *
 * Retorno: O ndice do coletor criado
 *
 ****/
static int NovoColetor(FILE *stream)
{
   tColetor coletor;
   int      indice;

      /* Inicialmente, o coletor est vazio */
   coletor.nRegistros = 0;
   coletor.profLocal = 0;

      /* Move o apontador de posio do arquivo para seu final */
   MoveApontador(stream, 0, SEEK_END);

      /* Obtm o ndice do coletor */
   indice = ftell(stream)/sizeof(coletor);

      /* Escreve o coletor no arquivo */
   fwrite(&coletor, sizeof(coletor), 1, stream);

   return indice;
}

/****
 *
 * PodeDividir(): Verifica se um coletor pode ser dividido
 *
 * Parmetros:
 *     coletor (entrada) - o coletor que ser examinado
 *
 * Retorno: 1, se o coletor pode ser dividido; 0, em caso contrrio
 *
 ****/
static int PodeDividir(const tColetor *coletor)
{
   tChave chave;
   int    i;

      /* Esta funo deve ser chamada apenas */
      /* quando o coletor est repleto       */
   ASSEGURA( coletor->nRegistros == M,
             "PodeDividir() nao deveria ter sido chamada" );

      /* Obtm a chave do primeiro registro */
   strcpy(chave, coletor->registros[0].chave);

   for (i = 1; i < coletor->nRegistros; ++i)
      if (strcmp(chave, coletor->registros[i].chave))
         return 1; /* Coletor pode ser dividido */

      /* As chaves deste coletor so todas iguais */
      /* e, portanto, ele no pode ser dividido   */
   return 0;
}

/****
 *
 * InsereEmDiretorio(): Insere num diretrio uma referncia para
 *                      um coletor de numa tabela de busca com
 *                      disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca
 *      pColetor (entrada) - endereo do coletor que ser inserido
 *      iColetor (entrada) - ndice do coletor no arquivo de
 *                           coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereEmDiretorio( tTabelaDExt *tab,
                               const tColetor *pColetor,
                               int iColetor )
{
   int    i, bits, pLocal;
   tChave chave;

      /* Obtm a chave do primeiro registro do coletor */
   strcpy(chave, pColetor->registros[0].chave);

      /* Obtm a profundidade local do coletor */
   pLocal = pColetor->profLocal;

      /* Duplica o tamanho do diretrio enquanto  */
      /* sua profundidade global for menor do que */
      /* a profundidade local do coletor          */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(int));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy( tab->dir + tab->tamDir, tab->dir,
              tab->tamDir*sizeof(int) );

         /* Atualiza a profundidade global */
         /* e o tamanho do diretrio       */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

   /* Se a profundidade local for menor do que a global, ser */
   /* preciso atualizar mais de uma referncia do diretrio   */

      /* Obtm os bits menos significativos */
      /* coincidentes no coletor            */
   bits = ObtemBits(chave, pLocal);

      /* Corrige as referncias para o novo coletor */
   for (i = 0; i < tab->tamDir; ++i)
      if (ObtemLSBs(i, pLocal) == bits)
         tab->dir[i] = iColetor;
}

/****
 *
 * InsereColetorVazio(): Insere num diretrio uma referncia para
 *                       um coletor vazio de numa tabela de busca
 *                       com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca
 *      pColetor (entrada) - endereo do coletor que deu origem ao
 *                           coletor vazio
 *      iColetor (entrada) - ndice do coletor que deu origem ao
 *                           coletor vazio no arquivo de coletores
 *      iVazio (entrada) - ndice do coletor vazio no arquivo de
 *                         coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereColetorVazio( tTabelaDExt *tab,
                                const tColetor *pColetor,
                                int iColetor, int iVazio )
{
   int    i, bits, pLocal;
   tChave chave;

      /* Obtm a profundidade local do coletor */
   pLocal = pColetor->profLocal;

      /* Duplica o tamanho do diretrio enquanto  */
      /* sua profundidade global for menor do que */
      /* a profundidade local do coletor          */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(int));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy( tab->dir + tab->tamDir, tab->dir,
              tab->tamDir*sizeof(int) );

         /* Atualiza a profundidade global */
         /* e o tamanho do diretrio       */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

      /* Obtm a chave do primeiro item do coletor */
      /* que deu origem ao coletor vazio           */
   strcpy(chave, pColetor->registros[0].chave);

      /* Obtm os bits menos significativos coincidentes no */
      /* coletor que deu origem ao coletor vazio            */
   bits = ObtemBits(chave, pLocal);

      /* Corrige as referncias para o coletor vazio */
   for (i = 0; i < tab->tamDir; ++i)
      if (tab->dir[i] == iColetor && ObtemLSBs(i, pLocal) != bits)
         tab->dir[i] = iVazio;
}

/****
 *
 * DivideColetor(): Divide um coletor de uma tabela de busca
 *                  com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca que contm a
 *                            referncia para o coletor que
 *                            ser dividido
 *      coletor (entrada/sada) - o coletor que ser dividido
 *      iColetor (entrada) - ndice do coletor que ser dividido
 *                           no arquivo de coletores
 *      stream (entrada) - stream associado ao arquivo contendo
 *                         os coletores
 *
 * Retorno: Nada
 *
 ****/
static void DivideColetor( tTabelaDExt *tab, tColetor *coletor,
                           int iColetor, FILE *stream )
{
   int      i, iNovo, bits;
   tChave   chave;
   tColetor novo;

      /* Cria um novo coletor e obtm seu ndice em arquivo */
   iNovo = NovoColetor(stream);

      /* L o novo coletor em arquivo */
   LeColetor(stream, iNovo, &novo);

      /* Divide registros do antigo coletor com o novo coletor */
   while (coletor->nRegistros == M) {
         /* Zera os registros do antigo e do novo coletor */
      coletor->nRegistros = 0;
      novo.nRegistros = 0;

         /* Distribui registros entre o antigo e o novo coletor */
      for (i = 0; i < M; i++) {
         /* Para cada registro do antigo coletor, verifica-se */
         /* se o bit de ordem profLocal + 1 de sua chave  0  */
         /* ou 1. Se ele for 0, o respectivo registro fica    */
         /* no antigo coletor. Caso contrrio, esse registro  */
         /* vai para o novo coletor.                          */

            /* Obtm a chave do registro corrente */
         strcpy(chave, coletor->registros[i].chave);

            /* Obtm profLocal + 1 bits menos significativos */
         bits = ObtemBits(chave, coletor->profLocal + 1);

            /* Obtm o valor do bit mais significativo */
            /* dessa sequncia de bits                 */
         if (ValorBit(bits, coletor->profLocal) == 0)
               /* Este registro fica no antigo coletor */
            coletor->registros[coletor->nRegistros++] =
               coletor->registros[i];
         else
               /* Este registro vai para o novo coletor */
            novo.registros[novo.nRegistros++] =
               coletor->registros[i];
      }

         /* O novo e o antigo coletor tero a mesma  */
         /* profundidade local, que  a profundidade */
         /* local do antigo coletor mais um          */
      novo.profLocal = ++coletor->profLocal;

         /* Se no ficou nenhum registro no coletor */
         /* antigo,  preciso repetir o processo    */
         /* usando uma profundidade local maior     */
      if (novo.nRegistros == M)
            /* Copia os registros do novo coletor */
            /* de volta para o antigo coletor     */
         *coletor = novo;
      else if (novo.nRegistros == 0) {
            /* Verifica se o antigo coletor pode ser dividi- */
            /* do. Se esse no for o caso, o novo registro   */
            /* deve ser inserido no novo coletor que dever  */
            /*  tornar-se um coletor excedente. Como esta    */
            /* implementao no lida com esse caso, aborta- */
            /* -se o programa.                               */
         ASSEGURA( PodeDividir(coletor),
                   "Erro: O coletor nao pode ser dividido\n" )

            /* Insere uma referncia para o    */
            /* novo coletor vazio no diretrio */
         InsereColetorVazio(tab, coletor, iColetor, iNovo);

            /* Escreve o coletor vazio no arquivo */
         EscreveColetor(stream, iNovo, &novo);

            /* Cria um novo coletor e o l em arquivo */
         iNovo = NovoColetor(stream);
         LeColetor(stream, iNovo, &novo);
      }
   }

      /* Escreve os dois coletores no arquivo */
   EscreveColetor(stream, iColetor, coletor);
   EscreveColetor(stream, iNovo, &novo);

      /* Insere uma referncia para o novo coletor no diretrio */
   InsereEmDiretorio(tab, &novo, iNovo);
}

/********************* Funes Globais ********************/

/****
 *
 * LeColetor(): L um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo
 *                        os coletores no qual ser feita a leitura
 *     pos (entrada) - posio no arquivo na qual ser feita a
 *                     leitura
 *    *coletor (sada) - ponteiro para o coletor que conter o
 *                       resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
void LeColetor(FILE *stream, int pos, tColetor *coletor)
{
      /* Verifica a validade do stream */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, LeColetor));

      /* Tenta mover o apontador de arquivo para o local */
      /* de leitura; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetor)*pos, SEEK_SET);

      /* Efetua a leitura */
   fread(coletor, sizeof(tColetor), 1, stream);

      /* Certifica-se que no houve erro de leitura */
   ASSEGURA(!ferror(stream),  "Erro de leitura em LeColetor()");
}

/****
 *
 * EscreveColetor(): Escreve um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo os
 *                        coletores no qual ser feita a escrita
 *     pos (entrada) - posio no arquivo no qual ser feita a
 *                     escrita
 *    *coletor (entrada) - coletor que ser escrito no arquivo
 *
 * Retorno: Nada
 *
 ****/
void EscreveColetor( FILE *stream, int pos,
                     const tColetor *coletor )
{
      /* Verifica se o stream  vlido */
   ASSEGURA(stream, "Erro em EscreveColetor(): stream e' NULL");

      /* Tenta mover o apontador de arquivo para o local */
      /* de escrita; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetor)*pos, SEEK_SET);

      /* Efetua a escrita */
   fwrite(coletor, sizeof(tColetor), 1, stream);

      /* Certifica-se que no houve erro */
   ASSEGURA( !ferror(stream),
             "Erro de escrita em EscreveColetor()" );
}

/****
 *
 * IniciaTabelaExt(): Inicia uma tabela de busca com disperso
 *                    extensvel
 *
 * Parmetros:
 *      tabela (sada) - a tabela que ser iniciada
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: Nada
 *
 ****/
void IniciaTabelaExt(tTabelaDExt *tabela, FILE *stream)
{
      /*                                                 */
      /* A tabela  iniciada com um tamanho de diretrio */
      /* igual a 1 que aponta para um coletor vazio      */
      /*                                                 */

    tabela->profGlobal = 0;
    tabela->nReg = 0;
    tabela->tamDir = 1;

    tabela->dir = malloc(tabela->tamDir*sizeof(int));

    tabela->dir[0] = NovoColetor(stream);
}

/****
 *
 * BuscaExt(): Efetua uma busca numa tabela de busca com
 *             disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *      tab (entrada) - a tabela
 *      chave (entrada) - a chave de busca
 *      reg (sada) - o registro encontrado
 *
 * Retorno: O endereo do registro, se ele for encontrado; NULL,
 *          em caso contrrio
 *
 * Observao: Os registros so ordenados em cada coletor, de
 *             modo que busca binria poderia ser utilizada
 *
 ****/
tRegistro *BuscaExt( FILE *stream, const tTabelaDExt *tab,
                     tChave chave, tRegistro *reg )
{
   int      iColetor, i;
   tColetor coletor;

      /* Obtm o ndice do coletor que contm o registro */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor que contm o registro */
   LeColetor(stream, iColetor, &coletor);

      /* Efetua uma busca sequencial pelo */
      /* registro em memria principal    */
   for (i = 0; i < coletor.nRegistros; i++)
         /* Verifica se a chave foi encontrada */
      if (!strcmp(chave, coletor.registros[i].chave)) {
         *reg = coletor.registros[i]; /* Chave encontrada */
         return reg;
      }

   return NULL; /* A chave no foi encontrada */
}

/****
 *
 * InsereExt(): Insere um registro numa tabela de busca com
 *              disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela na qual ser efetuada a
 *                            insero
 *      registro (entrada) - o registro que ser inserido
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: Nada
 *
 * Observao: Para facilitar a implementao, o nmero mximo
 *             de registros em qualquer coletor  M - 1, de modo
 *             que sempre h espao para insero
 *
 ****/
void InsereExt( tTabelaDExt *tab, const tRegistro *registro,
                FILE *stream )
{
   int      i, j, iColetor;
   tChave   chave;
   tColetor coletor;

      /* Obtm a chave do registro */
   strcpy(chave, registro->chave);

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor em arquivo */
   LeColetor(stream, iColetor, &coletor);

      /* O nmero de registros num coletor */
      /* deve ser menor do que M           */
   ASSEGURA( coletor.nRegistros < M,
             "Excesso de registros em coletor" );

      /* Como os registros so ordenados,  preciso encontrar */
      /* a posio de insero do novo registro no coletor    */
   for (i = 0; i < coletor.nRegistros; i++)
      if (strcmp(chave, coletor.registros[i].chave) < 0)
         break;

      /* Abre espao no coletor para o novo registro */
   for (j = coletor.nRegistros; j > i; j--)
      coletor.registros[j] = coletor.registros[j - 1];

      /* Insere o novo registro no coletor */
   coletor.registros[i] = *registro;

      /* O nmero de registros no coletor aumentou */
   coletor.nRegistros++;

      /* O nmero de registros na tabela tambm aumentou */
   ++tab->nReg;

      /* Verifica se o coletor ficou repleto */
   if (coletor.nRegistros == M)
         /* O coletor ficou repleto e  preciso dividi-lo */
      DivideColetor(tab, &coletor, iColetor, stream);
   else
         /* O coletor foi alterado e  preciso */
         /* reescrev-lo no arquivo            */
      EscreveColetor(stream, iColetor, &coletor);
}

/****
 *
 * RemoveExt(): Remove um registro de uma tabela de disperso
 *              extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de disperso
 *      chave (entrada) - a chave do registro que ser removido
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveExt(tTabelaDExt *tab, tChave chave, FILE *stream)
{
   int         i, j, iColetor;
   tColetor    coletor;

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor em arquivo */
   LeColetor(stream, iColetor, &coletor);

      /* Efetua uma busca sequencial pelo */
      /* registro em memria principal    */
   for (i = 0; i < coletor.nRegistros; i++)
         /* Verifica se a chave foi encontrada */
      if (!strcmp(chave, coletor.registros[i].chave))
         break;

      /* Verifica se a chave foi encontrada */
   if (i >= coletor.nRegistros)
      return 0;  /* A chave no foi encontrada */

      /* O registro a ser removido encontra-se na posio i. */
      /*  necessrio mover uma posio para trs todos os   */
      /* registros que se encontram  sua frente.            */
   for (j = i; j < coletor.nRegistros - 1; ++j)
      coletor.registros[j] = coletor.registros[j + 1];

      /* O nmero de registros no coletor diminuiu */
   --coletor.nRegistros;

      /* O nmero de registros na tabela tambm diminuiu */
   --tab->nReg;

      /* Verifica se o coletor ficou vazio */
   if (coletor.nRegistros == 0)
         /* O coletor ficou vazio. Talvez o tamanho do        */
         /* diretrio possa ser reduzido, mas  possvel que, */
         /* mesmo assim, essa operao no valha a pena???.   */
         /* Substituir a instruo vazia a seguir pela        */
         /* implementao de uma operao que reduza o        */
         /* tamanho do diretrio  deixada como exerccio.    */
      ;
   else
         /* O coletor foi alterado e  preciso */
         /* reescrev-lo no arquivo            */
      EscreveColetor(stream, iColetor, &coletor);

   return 1;  /* Remoo bem-sucedida */
}

/****
 *
 * NRegistrosTabExt(): Calcula o nmero de registros numa tabela
 *                     de disperso extensvel
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de registros na tabela de disperso
 *
 ****/
int NRegistrosTabExt(const tTabelaDExt *tabela, FILE *stream)
{
   int      nRegistros = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      nRegistros += umColetor.nRegistros;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NRegistrosTabExt" );

      /* Checa a consistncia */
   ASSEGURA( nRegistros == tabela->nReg,
             "Erro de consistencia" );

   return nRegistros;
}

/****
 *
 * NColetores(): Calcula o nmero de coletores numa tabela
 *               de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores na tabela de disperso
 *
 ****/
int NColetores(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetores" );

   return n;
}

/****
 *
 * NColetoresVazios(): Calcula o nmero de coletores vazios
 *                     numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores vazios
 *
 ****/
int NColetoresVazios(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (!umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresVazios" );

   return n;
}

/****
 *
 * NColetoresRepletos(): Calcula o nmero de coletores repletos
 *                       numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores repletos
 *
 ****/
int NColetoresRepletos(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros == M)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresRepletos" );

   return n;
}

/****
 *
 * NColetoresOcupados(): Calcula o nmero de coletores ocupados
 *                       numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores ocupados
 *
 ****/
int NColetoresOcupados(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nRegistros)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresOcupados" );

   return n;
}

/****
 *
 * ProfLocalMin(): Calcula a menor profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: A profundidade local mnima
 *
 ****/
int ProfLocalMin(FILE *stream)
{
   int      min;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

      /* L um coletor */
   fread(&umColetor, sizeof(umColetor), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMin" );

   min = umColetor.profLocal;

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal < min)
         min = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMin" );

   return min;
}

/****
 *
 * ProfLocalMax(): Calcula a maior profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: A profundidade local mxima
 *
 ****/
int ProfLocalMax(FILE *stream)
{
   int      max;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

      /* L um coletor */
   fread(&umColetor, sizeof(umColetor), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(stream), "Erro de leitura em ProfLocalMax");

      /* Assume que a maior profundidade  a do coletor corrente */
   max = umColetor.profLocal;

      /* L cada coletor e verifica qual */
      /* deles tem a maior profundidade  */
   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal > max)
            /* Encontrada uma profundidade maior */
         max = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(stream), "Erro de leitura em ProfLocalMax");

   return max;
}
