/****
 *
 * Arquivo Registros.c: Arquivo de programa do mdulo de
 *                      processamento de registros
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 20/04/2004
 * ltima modificao: 30/03/2017
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada e Sada   */
#include <string.h> /* Processamento de strings     */
#include <stdlib.h> /* strtod() e alocao dinmica */
#include "Registros.h" /* Interface deste mdulo    */
#include "leitura.h"  /* Leitura via teclado        */
#include "ASSEGURA.h" /* Macro ASSEGURA             */

/********************* Funes Globais ********************/

/****
 *
 * AbreArquivo(): Tenta abrir um arquivo e, quando isso
 *                no  possvel, aborta o programa
 *
 * Parmetros:
 *      nome (entrada) - nome do arquivo
 *      modo (entrada) - modo de abertura
 *
 * Retorno: Stream associado ao arquivo que foi aberto
 *
 ****/
FILE *AbreArquivo(const char *nome, const char *modo)
{
   FILE *stream;

   stream = fopen(nome, modo);

   if (!stream) { /* Erro de abertura */
      fprintf( stderr, "\n>>> O arquivo %s nao pode ser aberto",
               nome);
      exit(1); /* Aborta o programa */
   }

   return stream;
}

/****
 *
 * FechaArquivo(): Tenta fechar um arquivo e, quando isso
 *                 no  possvel, aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *
 * Retorno: Nada
 *
 * Observao: Se o segundo parmetro no for NULL, o nome
 *             do arquivo aparece na mensagem de erro
 *
 ****/
void FechaArquivo(FILE *stream, const char *nomeArq)
{
   /***********************************************************/
   /* Se fclose() retornar um valor diferente de zero ocorreu */
   /* algum erro na tentativa de fechamento do arquivo. Nesse */
   /* caso, apresenta mensagem de erro e aborta o programa.   */
   /***********************************************************/

   if (fclose(stream)) { /* Erro de fechamento */
      fprintf( stderr, "\a\n>>> Ocorreu erro no fechamento "
              "do arquivo %s.\n>>> O programa sera' "
              "encerrado.\n", nomeArq ? nomeArq : "");
      exit(1); /* Aborta o programa */
   }
}

/****
 *
 * MoveApontador(): Tenta mover o apontador de posio de um
 *                  arquivo e, quando isso no  possvel,
 *                  aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      bytes (entrada) - quantidade de bytes que sero saltados
 *      DeOnde (entrada) - a partir de onde ocorrer a movimentao
 *
 * Retorno: Nada
 *
 ****/
void MoveApontador(FILE *stream, long bytes, int DeOnde)
{
   int deuErrado;

   deuErrado = fseek(stream, bytes, DeOnde);

      /* Verifica se houve erro */
   if (deuErrado) {
      fprintf( stderr, "\n>>> Erro na tentativa de mover "
                       "apontador de arquivo" );
      exit(1);
   }
}

/****
 *
 * ExibeRegistro(): Exibe na tela um registro do tipo tRegistro
 *
 * Parmetros: *registro (entrada) - registro que ser exibido
 *
 * Retorno: Nada
 *
 ****/
void ExibeRegistro(const tRegistro *registro)
{
   printf("Nome: %s", registro->chave);

   printf("\nIdade: %d", registro->idade);
}

/****
 *
 * LeRegistro(): L os dados de um aluno
 *
 * Parmetros: *aluno (sada) - estrutura que receber os
 *                              dados introduzidos pelo
 *                              usurio
 * Retorno: O parmetro
 *
 ****/
tRegistro *LeRegistro(tRegistro *aluno)
{
   printf("\n\t>>> Nome: ");
   LeString(aluno->chave, MAX_NOME);

   printf("\n\t>>> Idade: ");
   aluno->idade = LeNatural();

   return aluno;
}

/****
 *
 * SubstituiRegistro(): L os dados de um registro que, se forem
 *                      vlidos, substituiro valores antigos
 *
 * Parmetros:
 *     *novo (sada) - estrutura que armazenar os novos
 *                     dados introduzidos pelo usurio
 *     *antigo (entrada) - estrutura contendo os dados antigos
 *
 * Retorno: O primeiro parmetro
 *
 * Observao: Esta funo no est rigorosamente bem
 *             implementada pois no faz parte do tema
 *             central deste exemplo
 *
 ****/
tRegistro *SubstituiRegistro( tRegistro *novo,
                              const tRegistro *antigo )
{
   return novo;
}

/****
 *
 * TamanhoDeArquivo(): Calcula o tamanho de um arquivo lendo cada
 *                     byte que o constitui e contando quantos
 *                     bytes so lidos
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *
 * Retorno: -1, se ocorrer algum erro.
 *          O tamanho do arquivo, em caso contrrio.
 *
 * Observaes:
 *     O arquivo deve estar aberto em modo binrio que permite
 *     leitura
 *
 ****/
long TamanhoDeArquivo(FILE *stream)
{
   ASSEGURA(stream, "Erro em TamanhoDeArquivo(): stream e' NULL" );

      /* Tenta mover o apontador de posio do arquivo */
      /* para o final do mesmo; se no for possvel,   */
      /* o programa ser abortado                      */
   MoveApontador(stream, 0, SEEK_END);

   return ftell(stream);
}

/****
 *
 * NumeroDeRegistros(): Calcula o nmero de registros (i.e.,
 *                      parties de tamanho especificado) de um
 *                      arquivo binrio
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *    tamRegistro (entrada) - tamanho de cada partio do arquivo
 *
 * Retorno: O nmero de parties do arquivo, se no ocorrer
 *          nenhum erro, ou -1, em caso contrrio.
 *
 * Observao: O arquivo deve estar aberto em modo binrio que
 *             permite leitura
 *
 ****/
int NumeroDeRegistros( FILE *stream, int tamRegistro )
{
   long nBytes = 0; /* Nmero de bytes do arquivo */

      /* Calcula o nmero de bytes do arquivo */
   nBytes = TamanhoDeArquivo(stream);

      /* Verifica se ocorreu erro na contagem de bytes */
   ASSEGURA(nBytes >= 0, "Erro na contagem de bytes");

      /* O nmero de registros  o nmero de bytes do arquivo */
      /* dividido pelo nmero de bytes de cada registro       */
   return nBytes/tamRegistro;
}

/****
 *
 * LeRegistroEmArquivo(): L um registro no arquivo de registros
 *
 * Parmetros:
 *     *umRegistro (sada) - registro lido
 *      stream (entrada) - stream associado ao arquivo
 *                         de registros
 *      pos (entrada) - posio do registro no arquivo
 *
 * Retorno: o primeiro parmetro, se a recuperao for
 *           bem sucedida; NULL, em caso contrrio.
 *
 ****/
tRegistro *LeRegistroEmArquivo( tRegistro *umRegistro,
                                FILE *stream, int pos )
{
   ASSEGURA(stream, "Arquivo de registros eh nulo");

   MoveApontador(stream, pos*sizeof(*umRegistro), SEEK_SET);

   return fread(umRegistro, sizeof(*umRegistro), 1, stream)
          ? umRegistro
          : NULL;
}

/****
 *
 * ArquivoExiste(): Verifica se um arquivo existe
 *
 * Parmetros:
 *     nomeArq (entrada) - nome do arquivo que se
 *                        verificar se existe
 *
 * Retorno: 1, se o arquivo existir; 0, em caso contrrio
 *
 ****/
int ArquivoExiste(const char *nomeArq)
{
   FILE *stream;
   int   existe;

   stream = fopen(nomeArq, "rb");

   existe = stream ? 1 : 0;

   if (existe) {
      FechaArquivo(stream, nomeArq);
   }

   return existe;
}
