/****
 *
 * Ttulo do Programa: A Necessidade de Coletores Excedentes
 *
 * Seo: 8.2.5
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 25/08/2017
 * ltima modificao: 28/08/2017
 *
 * Descrio Geral: Este programa demonstra o uso de coletores
 *                  excedentes em tabela de busca com disperso
 *                  extensvel armazenada em arquivo
 *
 * Entrada: Opes escolhidas atravs de menu
 *
 * Sada: Resultados das operaes escolhidas pelo usurio
 *        (v. exemplo de execuo do programa no final
 *        deste arquivo)
 *
 * Arquivos:
 *     ASSEGURA.h - macro de tratamento de exceo usadas pelo
 *                  programa
 *     DispersaoExt.h e - interface e implementao de operaes
 *     DispersaoExt.c     operaes sobre tabelas de disperso
 *                        extensvel
 *     Registros.h e - interface e implementao de operaes
 *     Registros.c     sobre registros armazenados em arquivos
 *     main.c (este arquivo) - contm main() e outras funes
 *                             que testam as operaes sobre
 *                             tabelas de disperso extensvel
 *
 * Arquivo de dados: No h
 *
 * Observao: A diferena entre este programa e o outro programa
 *             da Seo 8.7.2  que, neste programa, os registros
 *             causam aborto do programa
 *
 ****/

/*********************** Includes *************************/

#include <limits.h>       /* Limites de tipos inteiros      */
#include <stdio.h>        /* Funes de Entrada/Sada       */
#include <string.h>       /* Funes strXXX() e memXXX()    */
#include <time.h>         /* Medio de tempo               */
#include "DispersaoExt.h" /* Operaes com tabela de busca  */
#include "Registros.h"    /* Processamento de registros     */
#include "Leitura.h"      /* Funes de leitura via teclado */
#include "ASSEGURA.h"     /* Macro ASSEGURA                 */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo contendo os registros indexados */
#define NOME_ARQ_TAB "Coletores.dat"

   /* Nome do arquivo contendo o diretrio */
#define NOME_ARQ_DIR "Diretorio.dat"

/******************** Funes de Locais *******************/

/****
 *
 * ApresentaMenu(): Apresenta um menu de opes para o usurio
 *
 * Parmetros:
 *      opcoes[] (entrada) - array de strings contendo as
 *                           opes do usurio
 *      nOpcoes (entrada) - o nmero de opes
 *
 * Retorno: Nada
 *
 ****/
static void ApresentaMenu(const char *opcoes[], int nOpcoes)
{
   int i;

   printf("\n>>> Escolha uma das opcoes a seguir:\n");

   for (i = 0; i < nOpcoes; ++i)
      printf("\n\t[%d] %s", i + 1, opcoes[i]);

   printf("\n\n\t>>> ");
}

/****
 *
 * ConstroiTabela(): L arquivo de dados (binrio) contendo
 *                   os registros e constri outro arquivo
 *                   contendo os coletores de uma tabela de
 *                   disperso extensvel
 *
 * Parmetros:
 *     streamReg (entrada) - stream associado ao arquivo de
 *                           registros
 *     streamTab (entrada) - stream associado ao arquivo que
 *                           conter os coletores
 *
 * Retorno: Endereo de uma estrutura contendo o diretrio da
 *          tabela de disperso extensvel
 *
 ****/
tTabelaDExt *ConstroiTabela( tTabelaDExt *tabela, tRegistro *regs,
                             int nRegs, FILE *streamTab )
{
   int i;

   IniciaTabelaExt(tabela, streamTab);

   for (i = 0; i < nRegs; ++i)
      InsereExt(tabela, regs + i, streamTab);

   return tabela;
}

/****
 *
 * InfoTabela(): Apresenta informaes gerais sobre a tabela
 *
 * Parmetros:
 *     arvore (entrada) - ponteiro para a raiz da rvore
 *    *streamTab (entrada) - stream associado ao arquivo que
 *                           contm os coletores
 *    *streamRegs (entrada) - stream associado ao arquivo que
 *                            contm os registros
 *
 * Retorno: Nada
 *
 ****/
void InfoTabela( const tTabelaDExt* tabela, FILE *streamTab )
{
   int nRegs;

   printf("\n\n\t\t>>>  Informacoes sobre coletores  <<<\n");

   printf("\n\t>>> Tamanho do bloco lido ou escrito: %d", TB);
   printf("\n\t>>> Tamanho de um registro: %d", TR);
   printf("\n\t>>> Tamanho de um campo do tipo int: %d", TI);
   printf("\n\t>>> Numero maximo de registros num coletor: %d", M);
   printf("\n\t>>> Tamanho de uma variavel que armazena um coletor: %d",
          (int) sizeof(tColetor));
   printf( "\n\t>>> Espaco desperdicado em cada coletor: %d",
           (int) (TB - sizeof(tColetor)) );

   printf("\n\n\t\t>>>  Informacoes sobre a tabela  <<<\n");

      /* Calcula o nmero de registros na tabela */
   nRegs = NRegistrosTabExt(tabela, streamTab);

   printf("\n\t>>> Numero de registros na tabela: %d", nRegs);

   printf( "\n\n\t>>> Tamanho do diretorio: %d",
           tabela->tamDir );

   printf( "\n\t>>> Profundidade global: %d",
           tabela->profGlobal );

   printf( "\n\n\t>>> Numero total de coletores: %d",
           NColetores(streamTab) );
   printf( "\n\t>>> Numero de coletores vazios: %d",
           NColetoresVazios(streamTab) );
   printf( "\n\t>>> Numero de coletores ocupados: %d",
           NColetoresOcupados(streamTab) );
   printf( "\n\t>>> Numero de coletores repletos: %d\n",
           NColetoresRepletos(streamTab) );

   printf( "\n\t>>> Profundidade local minima: %d",
           ProfLocalMin(streamTab) );
   printf( "\n\t>>> Profundidade local maxima: %d\n",
           ProfLocalMax(streamTab) );
}

/****
 *
 * RepresentacaoBinaria(): Apresenta no meio de sada a
 *                         representao binria de um valor
 *                         do tipo int
 *
 * Parmetros:
 *      numero (entrada): valor inteiro cuja representao
 *                        binria ser apresentada
 *
 * Retorno: Nada
 *
 ****/
void RepresentacaoBinaria(int numero)
{
   int           i, bit, numeroDeBits;
   unsigned int  mascara; /* A mscara deve ser unsigned */

       /* Cada byte contm um nmero de bits igual a CHAR_BIT */
   numeroDeBits  = CHAR_BIT*sizeof(int);

       /* Coloca 1 na posio mais  esquerda da  */
       /* mscara. Todos os outros bits tero 0s. */
   mascara = 0x1 << (numeroDeBits - 1);

     /* Exibe cada bit a partir da esquerda */
   for (i = 1; i <= numeroDeBits; i++) {
      bit = (numero & mascara) ? 1 : 0;

         /* Apresenta um bit na tela */
      printf("%d", bit);

         /* Separa sequncias de bytes com espaos em branco */
         /* para melhor visualizao da representao        */
      if (i % CHAR_BIT == 0)
         printf("  ");

         /* Move bit com valor 1 da mscara */
         /* uma posio para a direita      */
      mascara >>= 1;
   }
}

/****
 *
 * ExibeTabela(): Apresenta no meio de sada a tabela de disperso
 *
 * Parmetros:
 *      regs (entrada): array que representa a tabela
 *      nRegs (entrada): nmero de registros na tabela
 *
 * Retorno: Nada
 *
 ****/
 void ExibeTabela(tRegistro *regs, int nRegs)
{
   int i;

   printf("\n\tN\tNome\tIdade\t\t\tDigital");

   for (i = 0; i < nRegs; ++i) {
      printf( "\n\t%d\t%s\t %d\t", i + 1, regs[i].chave,
              regs[i].idade );
      RepresentacaoBinaria(Dispersao(regs[i].chave));
   }

   putchar('\n');
}

/********************* Funo main() **********************/

/****
 *
 * main(): Testa operaes com tabela de disperso extensvel
 *
 * Parmetros: nenhum
 *
 * Retorno: 0
 *
 ****/
int main(void)
{
   tRegistro   umRegistro, *pRegistro;
   const char *opcoes[] = { "Tabela", /* 1 */
                            "Busca",  /* 2 */
                            "Insere", /* 3 */
                            "Remove", /* 4 */
                            "Encerra" /* 5 */
                          };
   int          op,  /* Opo do usurio  */
                nOpcoes = sizeof(opcoes)/sizeof(opcoes[0]);
   tChave       umaChave;  /* Uma chave de busca */
   FILE        *streamTab; /* Stream associado ao       */
                           /* arquivo contendo a tabela */
   tTabelaDExt  tabela;
   tRegistro    registros[] = {
                                {"Jose", 21},     /* 1 */
                                {"Joaquim", 19},  /* 2 */
                                {"Manoel", 31},   /* 3 */
                                {"Jose", 18},     /* 4 */
                                {"Maria", 22},    /* 5 */
                                {"Mario", 25},    /* 6 */
                                {"Isabel", 25},   /* 7 */
                                {"Jose", 20},     /* 8 */
                                {"Joaquim", 24}   /* 9 */
                              };
   int          nRegistros = sizeof(registros)/sizeof(registros[0]);

      /* Cria o arquivo e checa se abertura foi OK */
   streamTab = AbreArquivo(NOME_ARQ_TAB, "w+b");

      /* Constri a tabela */
   ConstroiTabela(&tabela, registros, nRegistros, streamTab);

   printf("\n\n\t>>> Tabela criada com sucesso");

   InfoTabela(&tabela, streamTab);

   while (1) { /* Lao principal do programa */
      ApresentaMenu(opcoes, nOpcoes);

      op = LeOpcao("12345");

         /* Verifica se o usurio deseja sair do programa */
      if (op == '5')
         break; /* Encerra o lao */

      if (op > '1') {
         printf("\n\t>>> Digite o nome: ");
         LeString(umaChave, MAX_NOME);
      }

      switch (op) {
         case '1': /* Exibe a tabela */
            ExibeTabela(registros, nRegistros);

            break;

         case '2': /* Procura o registro que */
                   /* possui uma dada chave  */

               /* Tenta encontrar o registro  */
               /* que contm a chave de busca */
            pRegistro = BuscaExt( streamTab, &tabela, umaChave,
                                  &umRegistro );

               /* Verifica se a chave foi encontrada */
            if (!pRegistro)
               printf("\n\t>>> A chave nao foi encontrada\n");
            else {
               printf( "\n\n\t>>> Registro com a chave %s <<<\n",
                       umaChave );
               ExibeRegistro(&umRegistro);
               putchar('\n'); /* Embelezamento */
               putchar('\n');
            }

            break;

         case '3': /* Insere um novo registro, se a */
                   /* chave dada ainda no existe   */
               /* Recebe do usurio os dados */
               /* do registro a ser inserido */
            LeRegistro(&umRegistro);

            InsereExt(&tabela, &umRegistro, streamTab);

            printf("\n\t>>> Insercao bem-sucedida\n");

            break;

         case '4': /* Remove um registro */
            if (RemoveExt(&tabela, umaChave, streamTab))
               printf( "\n\t>>> Remocao bem sucedida\n");
            else
               printf( "\n\t>>> A remocao MAL sucedida\n");

            break;
      } /* switch */
   } /* while */

      /* Fecha o arquivo que contm os coletores */
   FechaArquivo(streamTab, NOME_ARQ_TAB);

   printf("\nObrigado por usar este programa. Bye.\n\n");

   return 0;
}

/******************* Exemplo de Execuo ******************

Erro: O coletor nao pode ser dividido

**********************************************************/
