/****
 *
 * Arquivo DispersaoExt.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 04/08/2017
 * ltima modificao: 06/09/2017
 *
 * Descrio: Implementao de operaes de busca, insero e
 *            remoo usando disperso extensvel em memria
 *            secundria
 *
 ****/

/*********************** Includes *************************/

#include <limits.h>     /* Limites de tipos inteiros   */
#include <stdlib.h>     /* Miscelnea de funes       */
#include <stdio.h>      /* Funes de E/S              */
#include <string.h>     /* Funes strxxx() e memxxx() */
#include <math.h>       /* Funes strxxx() e memxxx() */
#include "DispersaoExt.h" /* Interface deste mdulo    */
#include "Registros.h"  /* Processamento de registros  */
#include "ASSEGURA.h"   /* Macro ASSEGURA              */

/********************* Funes Locais *********************/

/****
 *
 * ObtemBits(): Obtm os bits menos significativos de uma chave
 *
 * Parmetros:
 *      chave (entrada) - a chave da qual os bits sero obtidos
 *      nBits (entrada) - o nmero de bits menos significativos
 *                        que sero extrados da chave
 *
 * Retorno: Um valor inteiro que representa esses bits
 *
 ****/
static int ObtemBits(tChave chave, int nBits)
{
   int mascara = (1 << nBits) - 1;

   return chave & mascara;
}

/****
 *
 * ValorBit(): Retorna o valor de um bit de um valor inteiro
 *
 * Parmetros:
 *      valor (entrada) - valor que ter seu bit retornado
 *      pos (entrada) - posio do bit que ser retornado
 *
 * Retorno: O valor do bit
 *
 ****/
int ValorBit(int valor, int pos)
{
   return 1 & valor >> pos;
}

/****
 *
 * NovoColetor(): Cria um novo coletor num arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo que
 *                        contm os coletores
 *
 * Retorno: O ndice do coletor criado
 *
 ****/
static int NovoColetor(FILE *stream)
{
   tColetor coletor;
   int      indice;

   coletor.nElementos = 0;
   coletor.profLocal = 0;

      /* Move o apontador de posio do arquivo para seu final */
   MoveApontador(stream, 0, SEEK_END);

      /* Obtm o ndice do coletor */
   indice = ObtemApontador(stream)/sizeof(coletor);

      /* Escreve o coletor no arquivo */
   fwrite(&coletor, sizeof(coletor), 1, stream);

   return indice;
}

/****
 *
 * InsereEmDiretorio(): Insere num diretrio uma referncia para
 *                      um coletor de numa tabela de busca com
 *                      disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de disperso extensvel
 *      pColetor (entrada) - endereo do coletor que ser inserido
 *      iColetor (entrada) - ndice do coletor no arquivo de
 *                           coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereEmDiretorio( tTabelaDExt *tab,
                               const tColetor *pColetor,
                               int iColetor )
{
   int    i, bits, pLocal;
   tChave chave;

      /* Obtm a chave do primeiro item do coletor */
   chave = pColetor->elementos[0].chave;

      /* Obtm a profundidade local do coletor */
   pLocal = pColetor->profLocal;

      /* Duplica o tamanho do diretrio enquanto  */
      /* sua profundidade global for menor do que */
      /* a profundidade local do coletor          */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(int));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy( tab->dir + tab->tamDir, tab->dir,
              tab->tamDir*sizeof(int) );

         /* Atualiza a profundidade global */
         /* e o tamanho do diretrio       */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

   /***                                               ***/
   /* Se a profundidade local do novo coletor for menor */
   /* do que a profundidade global, ser preciso        */
   /* atualizar mais de uma referncia do diretrio     */
   /***                                               ***/

      /* Obtm os bits menos significativos */
      /* coincidentes no coletor            */
   bits = ObtemBits(chave, pLocal);

      /* Corrige as referncias do diretrio para o novo coletor */
   for (i = 0; i < tab->tamDir; ++i)
         if (ObtemBits(i, pLocal) == bits)
            tab->dir[i] = iColetor;
}

/****
 *
 * PodeDividir(): Verifica se um coletor pode ser dividido
 *
 * Parmetros:
 *     coletor (entrada) - o coletor que ser examinado
 *
 * Retorno: 1, se o coletor pode ser dividido; 0, em caso contrrio
 *
 ****/
static int PodeDividir(const tColetor *coletor)
{
   tChave chave;
   int    i;

      /* Esta funo deve ser chamada apenas */
      /* quando o coletor est repleto       */
   ASSEGURA( coletor->nElementos == M,
             "PodeDividir() nao deveria ter sido chamada" );

      /* Obtm a chave do primeiro elemento */
   chave = coletor->elementos[0].chave;

   for (i = 1; i < coletor->nElementos; ++i)
      if (chave != coletor->elementos[i].chave)
         return 1; /* Coletor pode ser dividido */

      /* As chaves deste coletor so todas iguais */
      /* e, portanto, ele no pode ser dividido   */
   return 0;
}

/****
 *
 * InsereColetorVazio(): Insere num diretrio uma referncia para
 *                       um coletor vazio de numa tabela de busca
 *                       com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca
 *      pColetor (entrada) - endereo do coletor que deu origem ao
 *                           coletor vazio
 *      iColetor (entrada) - ndice do coletor que deu origem ao
 *                           coletor vazio no arquivo de coletores
 *      iVazio (entrada) - ndice do coletor vazio no arquivo de
 *                         coletores
 *
 * Retorno: Nada
 *
 ****/
static void InsereColetorVazio( tTabelaDExt *tab,
                                const tColetor *pColetor,
                                int iColetor, int iVazio )
{
   int    i, bits, pLocal;
   tChave chave;

      /* Obtm a profundidade local do coletor */
   pLocal = pColetor->profLocal;

      /* Duplica o tamanho do diretrio enquanto  */
      /* sua profundidade global for menor do que */
      /* a profundidade local do coletor          */
   while (tab->profGlobal < pLocal) {
         /* Duplica o tamanho do diretrio */
      tab->dir = realloc(tab->dir, 2*tab->tamDir*sizeof(int));
      ASSEGURA(tab->dir, "Impossivel alocar novo diretorio");

         /* Copia as referncias do antigo diretrio */
         /* para a metade final do novo diretrio    */
      memcpy( tab->dir + tab->tamDir, tab->dir,
              tab->tamDir*sizeof(int) );

         /* Atualiza a profundidade global */
         /* e o tamanho do diretrio       */
      ++tab->profGlobal;
      tab->tamDir = 2*tab->tamDir;
   }

      /* Obtm a chave do primeiro item do coletor */
      /* que deu origem ao coletor vazio           */
   chave = pColetor->elementos[0].chave;

      /* Obtm os bits menos significativos coincidentes no */
      /* coletor que deu origem ao coletor vazio            */
   bits = ObtemBits(chave, pLocal);

      /* Corrige as referncias para o coletor vazio */
   for (i = 0; i < tab->tamDir; ++i)
      if (tab->dir[i] == iColetor && ObtemBits(i, pLocal) != bits)
         tab->dir[i] = iVazio;
}

/****
 *
 * DivideColetor(): Divide um coletor de uma tabela de busca
 *                  com disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de busca que contm a
 *                            referncia para o coletor que
 *                            ser dividido
 *      coletor (entrada/sada) - o coletor que ser dividido
 *      iColetor (entrada) - ndice do coletor que ser dividido
 *                           no arquivo de coletores
 *      stream (entrada) - stream associado ao arquivo contendo
 *                         os coletores
 *
 * Retorno: Nada
 *
 ****/
static void DivideColetor( tTabelaDExt *tab, tColetor *coletor,
                           int iColetor, FILE *stream )
{
   int      i, iNovo, chave, bits;
   tColetor novo;

      /* Cria um novo coletor e obtm seu ndice em arquivo */
   iNovo = NovoColetor(stream);

      /* L o novo coletor em arquivo */
   LeColetor(stream, iNovo, &novo);

      /* Divide elementos do antigo coletor com o novo coletor */
   while (coletor->nElementos == M) {
         /* Zera os elementos do antigo e do novo coletor */
      coletor->nElementos = 0;
      novo.nElementos = 0;

         /* Distribui elementos entre o antigo e o novo coletor */
      for (i = 0; i < M; i++) {
         /* Para cada item do antigo coletor, verifica-se  */
         /* se o bit de ordem profLocal + 1 de sua chave   */
         /*  0 ou 1. Se ele for 0, o respectivo item fica */
         /* no antigo coletor. Caso contrrio, esse item   */
         /* vai para o novo coletor.                       */

            /* Obtm a chave do item corrente */
         chave = coletor->elementos[i].chave;

            /* Obtm profLocal + 1 bits menos significativos */
         bits = ObtemBits(chave, coletor->profLocal + 1);

            /* Obtm o valor do bit mais significativo    */
            /* dessa sequncia de bits e testa esse valor */
         if (ValorBit(bits, coletor->profLocal) == 0)
               /* Este item fica no antigo coletor */
            coletor->elementos[coletor->nElementos++] =
               coletor->elementos[i];
         else
               /* Este item vai para o novo coletor */
            novo.elementos[novo.nElementos++] =
               coletor->elementos[i];
      }

         /* O novo e o antigo coletor tero a mesma  */
         /* profundidade local, que  a profundidade */
         /* local do antigo coletor mais um          */
      novo.profLocal = ++coletor->profLocal;

         /* Se no ficou nenhum item no coletor  */
         /* antigo,  preciso repetir o processo */
         /* usando uma profundidade local maior  */
      if (novo.nElementos == M)
            /* Copia os elementos do novo coletor */
            /* de volta para o antigo coletor     */
         *coletor = novo;
      else if (novo.nElementos == 0) {
            /* Verifica se o antigo coletor pode ser dividido.  */
            /* Se ele no pode ser dividido, o novo registro    */
            /* deve ser inserido no novo coletor que dever se  */
            /* tornar um coletor excedente. Como esta implemen- */
            /* tao no lida com esse caso, o programa ser    */
            /* abortado se o coletor no puder ser dividido.    */
         ASSEGURA( PodeDividir(coletor),
                   "Erro: O coletor nao pode ser dividido\n" );

            /* Insere uma referncia para o    */
            /* novo coletor vazio no diretrio */
         InsereColetorVazio(tab, coletor, iColetor, iNovo);

            /* Escreve o coletor vazio no arquivo */
         EscreveColetor(stream, iNovo, &novo);

            /* Cria um novo coletor e e o l em arquivo */
         iNovo = NovoColetor(stream);
         LeColetor(stream, iNovo, &novo);
      }
   }

      /* Escreve os dois coletores no arquivo */
   EscreveColetor(stream, iColetor, coletor);
   EscreveColetor(stream, iNovo, &novo);

      /* Insere uma referncia para o novo coletor no diretrio */
   InsereEmDiretorio(tab, &novo, iNovo);
}

/********************* Funes Globais ********************/

/****
 *
 * LeColetor(): L um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo
 *                        os coletores no qual ser feita a leitura
 *     pos (entrada) - posio no arquivo na qual ser feita a
 *                     leitura
 *    *coletor (sada) - ponteiro para o coletor que conter o
 *                       resultado da leitura
 *
 * Retorno: Nada
 *
 ****/
void LeColetor(FILE *stream, int pos, tColetor *coletor)
{
      /* Verifica a validade do stream */
   ASSEGURA(stream, ERRO_STREAM_NULL(stream, LeColetor));

      /* Tenta mover o apontador de arquivo para o local */
      /* de leitura; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetor)*pos, SEEK_SET);

      /* Efetua a leitura */
   fread(coletor, sizeof(tColetor), 1, stream);

      /* Certifica-se que no houve erro de leitura */
   ASSEGURA(!ferror(stream),  "Erro de leitura em LeColetor()");
}

/****
 *
 * EscreveColetor(): Escreve um coletor em arquivo
 *
 * Parmetros:
 *     stream (entrada) - stream associado ao arquivo contendo os
 *                        coletores no qual ser feita a escrita
 *     pos (entrada) - posio no arquivo no qual ser feita a
 *                     escrita
 *    *coletor (entrada) - coletor que ser escrito no arquivo
 *
 * Retorno: Nada
 *
 ****/
void EscreveColetor( FILE *stream, int pos,
                     const tColetor *coletor )
{
      /* Verifica se o stream  vlido */
   ASSEGURA(stream, "Erro em EscreveColetor(): stream e' NULL");

      /* Tenta mover o apontador de arquivo para o local */
      /* de escrita; se no conseguir aborta o programa  */
   MoveApontador(stream, sizeof(tColetor)*pos, SEEK_SET);

      /* Efetua a escrita */
   fwrite(coletor, sizeof(tColetor), 1, stream);

      /* Certifica-se que no houve erro */
   ASSEGURA( !ferror(stream),
             "Erro de escrita em EscreveColetor()" );
}

/****
 *
 * IniciaTabelaExt(): Inicia uma tabela de busca com disperso
 *                    extensvel
 *
 * Parmetros:
 *      tabela (sada) - a tabela que ser iniciada
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: Nada
 *
 ****/
void IniciaTabelaExt(tTabelaDExt *tabela, FILE *stream)
{
      /*                                                 */
      /* A tabela  iniciada com um tamanho de diretrio */
      /* igual a 1 que aponta para um coletor vazio      */
      /*                                                 */

    tabela->profGlobal = 0;
    tabela->nEl = 0;
    tabela->tamDir = 1;

    tabela->dir = malloc(tabela->tamDir*sizeof(int));

    tabela->dir[0] = NovoColetor(stream);
}

/****
 *
 * BuscaExt(): Efetua uma busca numa tabela de busca com
 *             disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *      tab (entrada) - a tabela
 *      chave (entrada) - a chave de busca
 *
 * Retorno: O ndice do registro no arquivo, se ele for encontrado;
 *          -1, em caso contrrio
 *
 * Observao: Os elementos so ordenados por chave em cada coletor,
 *             de modo que busca binria poderia ser utilizada
 *
 ****/
int BuscaExt(FILE *stream, const tTabelaDExt *tab, tChave chave)
{
   int      iColetor, i;
   tColetor coletor;

      /* Obtm ndice do coletor que contm o ndice do registro */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor que contm o ndice do registro */
   LeColetor(stream, iColetor, &coletor);

      /* Efetua uma busca sequencial pelo */
      /* registro em memria principal    */
   for (i = 0; i < coletor.nElementos; i++)
         /* Verifica se a chave foi encontrada */
      if (chave == coletor.elementos[i].chave) {
            /* Chave encontrada */
         return coletor.elementos[i].indice;
      }

   return -1; /* A chave no foi encontrada */
}

/****
 *
 * InsereExt(): Insere um item numa tabela de busca com disperso
 *              extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela na qual ser efetuada a
 *                            insero
 *      item (entrada) - o item que ser inserido
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: Nada
 *
 * Observao: Para facilitar a implementao, o nmero mximo
 *             de elementos em qualquer coletor  M - 1, de modo
 *             que sempre h espao para insero
 *
 ****/
void InsereExt( tTabelaDExt *tab, const tChaveIndice *item,
                FILE *stream )
{
   int      i, j, iColetor;
   tChave   chave;
   tColetor coletor;

      /* Obtm a chave do item */
   chave = item->chave;

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor em arquivo */
   LeColetor(stream, iColetor, &coletor);

      /* O nmero de elementos num coletor */
      /* deve ser menor do que M           */
   ASSEGURA( coletor.nElementos < M,
             "Excesso de elementos em coletor" );

      /* Como os elementos so ordenado,  precico encontrar */
      /* a posio de insero do novo item no coletor   */
   for (i = 0; i < coletor.nElementos; i++)
      if (chave < coletor.elementos[i].chave)
         break;

      /* Abre espao no coletor para o novo item */
   for (j = coletor.nElementos; j > i; j--)
      coletor.elementos[j] = coletor.elementos[j - 1];

      /* Insere o novo item no coletor */
   coletor.elementos[i] = *item;

      /* O nmero de elementos no coletor aumentou */
   coletor.nElementos++;

      /* O nmero de elementos na tabela tambm aumentou */
   ++tab->nEl;

      /* Verifica se o coletor ficou repleto */
   if (coletor.nElementos == M)
         /* O coletor ficou repleto e  preciso dividi-lo */
      DivideColetor(tab, &coletor, iColetor, stream);
   else
         /* O coletor foi alterado e  necessrio */
         /* reescrev-lo no arquivo               */
      EscreveColetor(stream, iColetor, &coletor);
}

/****
 *
 * RemoveExt(): Remove um item de uma tabela de disperso extensvel
 *
 * Parmetros:
 *      tab (entrada/sada) - a tabela de disperso
 *      chave (entrada) - a chave do item que ser removido
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: 1, se a remoo foi ok; 0, caso contrrio
 *
 ****/
int RemoveExt(tTabelaDExt *tab, tChave chave, FILE *stream)
{
   int         i, j, iColetor;
   tColetor    coletor;

      /* Obtm o ndice do coletor */
   iColetor = tab->dir[ObtemBits(chave, tab->profGlobal)];

      /* L o coletor em arquivo */
   LeColetor(stream, iColetor, &coletor);

      /* Efetua uma busca sequencial pelo */
      /* item em memria principal        */
   for (i = 0; i < coletor.nElementos; i++)
         /* Verifica se a chave foi encontrada */
      if (chave == coletor.elementos[i].chave)
         break;

      /* Verifica se a chave foi encontrada */
   if (i >= coletor.nElementos)
      return 0;  /* A chave no foi encontrada */

      /* O item a ser removido se encontra na posio i. */
      /*  necessrio mover uma posio para trs todos  */
      /* os elementos que se encontram  sua frente.     */
   for (j = i; j < coletor.nElementos - 1; ++j)
      coletor.elementos[j] = coletor.elementos[j + 1];

      /* O nmero de elementos no coletor diminuiu */
   --coletor.nElementos;

      /* O nmero de elementos na tabela tambm diminuiu */
   --tab->nEl;

      /* Verifica se o coletor ficou vazio */
   if (coletor.nElementos == 0)
         /* O coletor ficou vazio. Talvez o tamanho do        */
         /* diretrio possa ser reduzido, mas  possvel que, */
         /* mesmo assim, essa operao no valha a pena.      */
         /* Substituir a instruo vazia a seguir pela        */
         /* implementao de uma operao que reduza o        */
         /* tamanho do diretrio  deixada como exerccio.    */
      ;

      /* O coletor foi alterado e  preciso */
      /* reescrev-lo no arquivo            */
   EscreveColetor(stream, iColetor, &coletor);

   return 1;  /* Remoo bem-sucedida */
}

/****
 *
 * NRegistrosTabExt(): Calcula o nmero de registros numa tabela
 *                     de disperso extensvel
 *
 * Parmetros:
 *      tabela (entrada) - a tabela de disperso
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de registros na tabela de disperso
 *
 ****/
int NRegistrosTabExt(const tTabelaDExt *tabela, FILE *stream)
{
   int      nRegistros = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      nRegistros += umColetor.nElementos;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NRegistrosTabExt" );

      /* Checa a consistncia */
   ASSEGURA( nRegistros == tabela->nEl, "Erro de consistencia" );

   return nRegistros;
}

/****
 *
 * NColetores(): Calcula o nmero de coletores numa tabela
 *               de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores na tabela de disperso
 *
 ****/
int NColetores(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetores" );

   return n;
}

/****
 *
 * NColetoresVazios(): Calcula o nmero de coletores vazios
 *                     numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores vazios
 *
 ****/
int NColetoresVazios(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (!umColetor.nElementos)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresVazios" );

   return n;
}

/****
 *
 * NColetoresRepletos(): Calcula o nmero de coletores repletos
 *                       numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores repletos
 *
 ****/
int NColetoresRepletos(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nElementos == M)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresRepletos" );

   return n;
}

/****
 *
 * NColetoresOcupados(): Calcula o nmero de coletores ocupados
 *                       numa tabela de disperso extensvel
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: O nmero de coletores ocupados
 *
 ****/
int NColetoresOcupados(FILE *stream)
{
   int      n = 0;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.nElementos)
         ++n;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em NColetoresOcupados" );

   return n;
}

/****
 *
 * ProfLocalMin(): Calcula a menor profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: A profundidade local mnima
 *
 ****/
int ProfLocalMin(FILE *stream)
{
   int      min;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

      /* L um coletor */
   fread(&umColetor, sizeof(umColetor), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMin" );

   min = umColetor.profLocal;

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal < min)
         min = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMin" );

   return min;
}

/****
 *
 * ProfLocalMax(): Calcula a maior profundidade local
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os coletores
 *
 * Retorno: A profundidade local mxima
 *
 ****/
int ProfLocalMax(FILE *stream)
{
   int      max;
   tColetor umColetor;

      /* Assegura que a leitura comea no incio do arquivo */
   rewind(stream);

      /* L um coletor */
   fread(&umColetor, sizeof(umColetor), 1, stream);

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMax" );

   max = umColetor.profLocal;

   while (1) {
         /* L um coletor */
      fread(&umColetor, sizeof(umColetor), 1, stream);

         /* Se ocorreu erro de leitura ou o final  */
         /* do arquivo foi atingido encerra o lao */
      if (feof(stream) || ferror(stream))
         break;

      if (umColetor.profLocal > max)
         max = umColetor.profLocal;
   }

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA( !ferror(stream),
             "Erro de leitura em ProfLocalMax" );

   return max;
}
