/****
 *
 * Exemplo: Casamento de Strings Usando o Algoritmo
 *          de Knuth, Morris e Pratt (KMP)
 *
 * Seo: 9.3.4
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/************************ Macros **************************/

#define DEBUG 1

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

/***************** Definies de Funes ******************/

/****
 *
 * ExibeTabela(): Apresenta o contedo de um array de ints
 *
 * Parmetros:
 *      ar (entrada) - o array
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void ExibeTabela(const int ar[], int tam)
{
   int i;

   putchar('\n');

   for (i = 0; i < tam; ++i) {
      printf("\ttmb[%d] = %d\n", i, ar[i]);
   }
}

/****
 *
 * CriaTMB(): Atribui valores para os elementos da tabela
 *            de maiores bordas usada pelo algoritmo KMP
 *
 * Parmetros:
 *      tmb (sada) - array que armazenar a tabela
 *      p (entrada) - string que representa o padro
 *      m (entrada) - tamanho do padro
 *
 * Retorno: Endereo da tabela
 *
 ****/
int *CriaTMB(int tmb[], const char *p, int m)
{
   int i = 1,
       j = 0;

   tmb[0] = 0; /* O primeiro elemento da tabela  sempre 0 */

   while (i < m) {
      if (p[j] == p[i]) {
         tmb[i] = j + 1;
         ++j;
         ++i;
      } else {
         if (j != 0) {
            j = tmb[j - 1];
         } else {
            tmb[i] = 0;
            ++i;
         }
      }
   }

   return tmb;
}

/****
 *
 * CasamentoKMP(): Verifica se um padro ocorre num texto usando
 *                 o algoritmo de Knuth, Morris e Pratt (KMP)
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *
 * Retorno: A posio do padro no texto, se ele for
 *          encontrado; -1, em caso contrrio.
 *
 ****/
int CasamentoKMP(const char *t, const char *p)
{
   int n = strlen(t), /* Comprimento do texto  */
       m = strlen(p), /* Comprimento do padro */
       i,  /* ndice de um caractere do texto  */
       j,  /* ndice de um caractere do padro */
      *tmb; /* Ponteiro para a tabela de maiores bordas */

      /* Aloca espao para a tabela de maiores bordas */
   ASSEGURA( tmb = malloc(m*sizeof(int)),
             "Impossivel aloca tabela de bordas" );

      /* Cria a tabela de maiores bordas */
   CriaTMB(tmb, p, m);

      /* Efetua a busca pelo padro no texto */
   for (i = 0, j = 0; i < n; ) {
         /* Verifica se os caracteres ora comparados casam */
      if (t[i] == p[j]) {
            /* Caracteres comparados casaram. Se esse era */
            /* o ltimo caractere do padro, retorna-se a */
            /* posio na qual ocorreu o casamento. Caso  */
            /* contrrio, incrementam-se os ndices.      */
         if (j == m - 1) {
            free(tmb); /* A tabela tmb no  mais necessria */
            return i - j; /* Padro foi encontrado no texto */
         } else { /* Compara os dois prximos caracteres */
            ++i;
            ++j;
         }
      } else if (j > 0) {
            /* Os caracteres no casaram e j  diferente */
            /* de 0. Ajusta-se j, mas no se ajusta i.   */
         j = tmb[j - 1];
      } else {
              /* Os caracteres no casaram e j  igual a   */
              /* 0 (i.e., o primeiro caractere do padro). */
         ++i; /* Incrementa-se i, mas no se altera j.     */
      }
   }

      /* A tabela tmb no  mais necessria */
   free(tmb);

      /* Se ainda no houve retorno, o padro */
      /* no foi encontrado no texto          */
   return -1;
}

/****
 *
 * main(): Testa a funo CasamentoKMP()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   char *texto = "abacaabaccabacabaabb";
   char *padrao = "abacab";
//   char *padrao = "ANANAB#BANANA";
   int   posicao;

   printf("\n\t>>>> Usando Knuth, Morris e Pratt <<<\n");

   posicao = CasamentoKMP(texto, padrao);

   if (posicao >= 0) {
      printf( "\n>>> \"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   } else {
      printf( "\n>>> \"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );
   }

   return 0;
}

/****************** Resultado do Programa *****************

        >>>> Usando Knuth, Morris e Pratt <<<

>>> "abacab" encontrado em "abacaabaccabacabaabb" na posicao 10

**********************************************************/
