/****
 *
 * Exemplo: Casamento de Strings Usando o Algoritmo de Boyer,
 *          Moore e Horspool (BMH) (Verso Alternativa)
 *
 * Seo: 9.5.3
 *
 ****/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

   /* Alfabeto do qual so obtidos os smbolos usados nos strings */
//#define ALFABETO "abcdefghijklmnopqrstuvwxyz"
#define ALFABETO "ACGT"

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

/****
 *
 * ValorDeCaractere(): Obtm o valor de um caractere num alfabeto
 *
 * Parmetros:
 *      caractere (entrada) - o caractere
 *      alfabeto (entrada) - string que representa o alfabeto
 *
 * Retorno: O valor do caractere no alfabeto
 *
 * Observao: O valor de um caractere  considerado como sendo seu
 *             ndice no alfabeto
 *
 ****/
int ValorDeCaractere(int caractere, const char *alfabeto)
{
   char *p = strchr(alfabeto, caractere);

      /* Obtm o endereo do caractere no alfabeto */
   p = strchr(alfabeto, caractere);

      /* Se o caractere no foi encontrado, aborta o programa */
   ASSEGURA(p, "Caractere inexistente no alfabeto");

      /* Retorna o ndice do caractere no */
      /* string que representa o alfabeto */
   return p - alfabeto;
}

/****
 *
 * TabelaDeSaltos(): Cria uma tabela de saltos usada pelo algoritmo
 *                   BMH
 *
 * Parmetros:
 *      p (entrada) - string que representa o padro
 *      m (entrada) - tamanho padro
 *      alfabeto (entrada) - o alfabeto do qual os smbolos que
 *                           constituem os strings so obtidos
 *
 * Retorno: Endereo da tabela
 *
 ****/
int *TabelaDeSaltos(const char *p, int m, const char *alfabeto)
{
   int  i,
       *tab, /* Apontar para a tabela de saltos */
        indiceCaractere, /* ndice de um caractere */
        tamAlfabeto; /* Tamanho do alfabeto */

      /* Calcula o tamanho do alfabeto */
   tamAlfabeto = strlen(alfabeto);

      /* Aloca espao para a tabela e checa sua alocao */
   tab = malloc(tamAlfabeto*sizeof(int));
   ASSEGURA(tab, "Impossivel alocar tabela de saltos");

      /* Inicia o valor do salto de cada caractere */
      /* do alfabeto com o tamanho do padro       */
   for (i = 0; i < tamAlfabeto; ++i)
      tab[i] = m;

      /* Determina o valor do salto de cada caractere */
      /* do padro de acordo com sua posio nele     */
   for (i = 0; i < m - 1; ++i) {
      indiceCaractere = ValorDeCaractere(p[i], alfabeto);
      tab[indiceCaractere] = m - i - 1;
   }

   return tab;
}

/****
 *
 * CasamentoBMH(): Verifica se um padro ocorre num texto usando
 *                 o algoritmo de Boyer, Moore e Horspool (BMH)
 *
 * Parmetros:
 *      t (entrada) - string que representa o texto
 *      p (entrada) - string que representa o padro
 *
 * Retorno: A posio do padro no texto, se ele for encontrado;
 *          -1, em caso contrrio
 *
 ****/
int CasamentoBMH(const char *t, const char *p)
{
   int i = 0, /* ndice do primeiro caractere   */
              /* do texto alinhado com o padro */
       j,
      *tabela, /* Tabela de saltos */
       m = strlen(p), /* Tamanho do padro */
       n = strlen(t), /* Tamanho do texto */
       c; /* ltimo caractere do texto alinhado com o padro */

      /* Fase 1: Pr-processamento */
   tabela = TabelaDeSaltos(p, m, ALFABETO);

      /* Fase 2: Casamento */
   while (i <= n - m) {
         /* Compara, da direita para a esquerda, os      */
         /* caracteres alinhados do padro e do texto    */
         /* at todos os caracteres do padro terem sido */
         /* comparados ou at encontrar uma discordncia */
      for (j = m - 1; j >= 0 && t[i + j] == p[j]; --j)
         ; /* Vazio */

         /* Verifica se houve casamento */
      if (j < 0)
         return i; /* Padro encontrado no texto */

         /* Obtm o ltimo caractere do */
         /* texto alinhado com o padro */
      c = ValorDeCaractere(t[i + m - 1], ALFABETO);

         /* Calcula o tamanho do prximo salto */
      i += tabela[c];
   }

   return -1; /* O padro no foi encontrado no texto */
}

/****
 *
 * main(): Testa a funo CasamentoBMH()
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
//   char *texto = "abacaabaccabacabaabb";
//   char *padrao = "abacab";
   char *texto = "ACCGTTATCAGTGCCAGCCAGCAGGCCAC";
   char *padrao = "AGCCAGC";
//   char *texto = "estruturas de dados";
//   char *padrao = "dados";
   int   posicao;

   printf("\n\t>>>> Usando Boyer, Moore e Horspool <<<\n");

   posicao = CasamentoBMH(texto, padrao);

   if (posicao >= 0) {
      printf( "\n\"%s\" encontrado em \"%s\" na posicao %d\n",
              padrao, texto, posicao );
   } else {
      printf( "\n\"%s\" nao foi encontrado em \"%s\"\n",
              padrao, texto );
   }

   return 0;
}

/****************** Resultado do Programa *****************

"AGCCAGC" encontrado em "ACCGTTATCAGTGCCAGCCAGCAGGCCAC" na posicao 15

**********************************************************/
