/****
 *
 * Exemplo: rvore Trie
 *
 * Seo: 9.8.5
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/***************** Constantes Simblicas ******************/

   /* Tamanho do alfabeto */
#define TAM_ALFABETO 'z' - 'a' + 1

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

   /* Resulta no ndice de um caractere num array */
   /* que representa os caracteres do alfabeto    */
#define INDICE_CARACTERE(c) ((c) - 'a')

   /* Libera um n e torna nulo o ponteiro usado para liber-lo */
#define FREE(p) do {free(p); p = NULL;} while(0)

/****************** Definies de Tipos *******************/

   /* Tipo de n da trie */
typedef struct rotNoTrie {
           struct rotNoTrie *filhos[TAM_ALFABETO];
           int               ehFinal;
        } tNoTrie, *tTrie;

/************************* Aluses ************************/

extern tTrie IniciaTrie(void);
extern tNoTrie *NovoNoDeTrie(void);
extern void InsereEmTrie(tNoTrie *raiz, const char *chave);
extern int BuscaEmTrie(tNoTrie *raiz, const char *chave);
extern int EhNoFinalDeTrie(const tNoTrie *pNo);
extern int EhNoVazioDeTrie(const tNoTrie *pNo);
extern int RemoveEmTrieAux( tNoTrie *pNo, const char *chave,
                            int nivel, int tam );
extern void RemoveEmTrie(tTrie pTrie, const char *chave);
extern int NumeroDeChavesEmTrie(tNoTrie *raiz);

/***************** Definies de Funes ******************/

/****
 *
 * IniciaTrie(): Inicia uma trie contendo apenas uma raiz vazia
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo da raiz da trie
 *
 ****/
tTrie IniciaTrie(void)
{
   tNoTrie *pNo = NovoNoDeTrie();

   ASSEGURA(pNo, "Impossivel alocar no' de trie");

   return pNo;
}

/****
 *
 * NovoNoDeTrie(): Cria um n de uma trie
 *
 * Parmetros: Nenhum
 *
 * Retorno: Endereo do n criado, se houve alocao; NULL, em
 *          caso contrrio
 *
 ****/
tNoTrie *NovoNoDeTrie(void)
{
   int      i;
   tNoTrie *pNo = malloc(sizeof(*pNo));

   if (pNo) {
      pNo->ehFinal = 0; /* Por enquanto, o n no  final */

         /* Inicia cada filho do n com NULL */
      for (i = 0; i < TAM_ALFABETO; i++)
         pNo->filhos[i] = NULL;
   }

   return pNo;
}

/****
 *
 * InsereEmTrie(): Insere uma nova chave (string) numa trie
 *
 * Parmetros:
 *      raiz (entrada/sada) - endereo da raiz da trie
 *      chave (entrada) - chave que ser inserida
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. Se a chave j existir, ela no ser inserida
 *     2. O ltimo n visitado indica o final de um string,
 *        de modo que se a chave era apenas prefixo de alguma
 *        outra chave na trie, ele passar a ser uma chave
 *
 ****/
void InsereEmTrie(tNoTrie *raiz, const char *chave)
{
   int      indice; /* ndice de um caractere no */
                    /* array de filhos de um n  */
   tNoTrie *p; /* Ponteiro usado para descer na trie */

   for (p = raiz; *chave; ++chave) {
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(*chave);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em InsereEmTrie()" );

         /* Verifica se o caractere corrente */
         /* existe no presente n            */
      if (!p->filhos[indice])
            /* O referido caractere no existe e  preciso */
            /* criar mais um filho para o corrente n      */
         p->filhos[indice] = NovoNoDeTrie();

         /* Passa para o prximo nvel da trie */
      p = p->filhos[indice];
   }

      /* O ltimo n visitado deve indicar */
      /* o final de um string              */
   p->ehFinal = 1;
}

/****
 *
 * BuscaEmTrie(): Efetua uma busca numa trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *      chave (entrada) - a chave de busca
 *
 * Retorno: 1, se a chave for encontrada;
 *          0, em caso contrrio
 *
 ****/
int BuscaEmTrie(tNoTrie *raiz, const char *chave)
{
   int      indice; /* ndice de um caractere no array de filhos de um n  */
   tNoTrie *p; /* Ponteiro usado para descer na trie */

   for (p = raiz; *chave && p; ++chave) {
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(*chave);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em BuscaEmTrie()" );

         /* Se o caractere corrente da chave no existe   */
         /* no presente n, a chave no faz parte da trie */
      if (!p->filhos[indice])
         return 0;

      p = p->filhos[indice]; /* Passa para o prximo nvel da trie */
   }

      /* Se o lao encerrou porque p assumiu NULL, */
      /* a chave no se encontra na trie           */
   if (!p)
      return 0;

      /* A chave est presente na trie se o ltimo n visitado for um n final */
   return p->ehFinal;
}

/****
 *
 * EhNoFinalDeTrie(): Verifica se um n de uma trie indica
 *              o final de um string
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n indicar o final de um string;
 *          0, em caso contrrio
 *
 ****/
int EhNoFinalDeTrie(const tNoTrie *pNo)
{
   return pNo->ehFinal;
}

/****
 *
 * EhNoVazioDeTrie(): Verifica se um n de uma trie est vazio
 *
 * Parmetros:
 *      pNo (entrada) - endereo do n
 *
 * Retorno: 1, se o n estiver vazio;
 *          0, em caso contrrio
 *
 ****/
int EhNoVazioDeTrie(const tNoTrie *pNo)
{
   int i;

      /* Verifica se o n possui algum filho */
   for(i = 0; i < TAM_ALFABETO; i++) {
         /* Se o n tiver pelo menos um filho, ele no est vazio */
      if(pNo->filhos[i])
         return 0; /* N no est vazio */
   }

   return 1; /* N no possui nenhum filho */
}

/****
 *
 * RemoveEmTrieAux(): Funo auxiliar de remoo de chaves
 *
 * Parmetros:
 *     pNo (entrada/sada) - endereo do n que poder ser removido
 *     chave (entrada) - chave a ser removida
 *     nivel (entrada) - nvel de remoo (v. Observao)
 *     tam (entrada) - tamanho da chave a ser removida
 *
 * Retorno: 1, se o n estiver vazio;
 *          0, em caso contrrio
 *
 * Observao: O parmetro 'nivel' varia de 0, que corresponde
 *              raiz da trie at tam - 1, que corresponde ao
 *             ltimo caractere da chave
 *
 ****/
int RemoveEmTrieAux(tNoTrie *pNo, const char *chave, int nivel, int tam)
{
   int indice; /* ndice de um caractere no array de filhos de um n  */

   if (!pNo)
      return 0; /* pNo no aponta para nenhum n */

   /*                                                      */
   /* A recurso encerra quando o nvel for maior do que o */
   /* nvel do ltimo caractere da chave, dado por tam - 1 */
   /*                                                      */

      /* Verifica se o nvel corrente  igual ao tamanho da chave  */
   if (nivel == tam) { /* Base da recurso */
         /* Se o n indicar o final da chave, ele deixar de ser um n final */
      if (pNo->ehFinal) {
         pNo->ehFinal = 0; /* N deixa de ser final de string */

            /* Se o n estiver vazio, ele precisa ser removido */
         if(EhNoVazioDeTrie(pNo))
            return 1;

         return 0; /* N no deve ser removido */
      }
   } else { /* Caso recursivo */
         /* Obtm o ndice do caractere corrente da chave */
      indice = INDICE_CARACTERE(chave[nivel]);

         /* Certifica-se que o ndice obtido  vlido */
      ASSEGURA( indice >= 0 && indice < TAM_ALFABETO,
                "Indice invalido em RemoveEmTrieAux()" );

         /* Efetua a remoo no n do prximo nvel e */
         /* verifica se esse n precisa ser removido  */
      if(RemoveEmTrieAux(pNo->filhos[indice], chave, nivel + 1, tam)) {
         FREE(pNo->filhos[indice]); /* N precisa ser removido */

            /* Sobe a rvore recursivamente removendo */
            /* os ns que devem ser removidos         */
         return !EhNoFinalDeTrie(pNo) && EhNoVazioDeTrie(pNo);
      }
   }
}

/****
 *
 * RemoveEmTrie(): Remove uma chave de um trie
 *
 * Parmetros:
 *      pTrie (entrada) - raiz da trie
 *      chave (entrada) - chave a ser removida
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  apenas uma interface para
 *             a funo RemoveEmTrieAux() que realmente realiza a tarefa de remoo
 *
 ****/
void RemoveEmTrie(tTrie pTrie, const char *chave)
{
   int tam = strlen(chave); /* Tamanho da chave */

      /* Se a chave no for um string vazio, chama a funo */
      /* RemoveEmTrieAux() para remov-la. O nvel inicial  */
      /* de remoo  0, que corresponde  raiz da trie     */
   if(tam > 0)
      RemoveEmTrieAux(pTrie, chave, 0, tam);
}

/****
 *
 * NumeroDeChavesEmTrie(): Calcula o nmero de chaves de uma trie
 *
 * Parmetros:
 *      raiz (entrada) - endereo da raiz da trie
 *
 * Retorno: O nmero de chaves da trie
 *
 ****/
int NumeroDeChavesEmTrie(tNoTrie *raiz)
{
   int cont = 0, /* Armazena o nmero de chaves */
       i;

      /* Trie vazia no contm nenhuma chave */
   if (!raiz)
      return 0;

      /* Se o n for final, encontrou-se uma chave */
   if (raiz->ehFinal)
      cont++;

      /* Conta o nmero de chaves em cada subrvore */
      /* do n corrente que no  vazia             */
   for (i = 0; i < TAM_ALFABETO; ++i)
      if (raiz->filhos[i])
         cont += NumeroDeChavesEmTrie(raiz->filhos[i]);

   return cont;
}

/****
 *
 * ResultadoBusca(): Efetua uma busca numa trie e
 *                   apresenta o resultado
 *
 * Parmetros:
 *      trie (entrada) - a trie
 *      chave (entrada) - a chave a ser procurada
 *
 * Retorno: Nada
 *
 ****/
void ResultadoBusca(tTrie trie, const char *chave)
{
   int resultado;

      /* Efetua a busca */
   resultado = BuscaEmTrie(trie, chave);

      /* Apresenta o resultado da busca */
   printf( "\n>>> A chave \"%s\" %s presente na trie", chave,
           resultado ? "esta'" : "NAO esta'");
}

/****
 *
 * Indices(): Apresenta os caracteres do alfabeto usado e seus
 *            ndices
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void Indices(void)
{
   int i;

   printf("TAM_ALFABETO = %d\n", TAM_ALFABETO);

   for (i = 0; i <= 255; ++i)
      if (islower(i))
         printf("\n%c ---> %2d ---> %2d", i, i, INDICE_CARACTERE(i));
}

/****
 *
 * main(): Testa as funes de busca, insero e remoo em tries
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   const char *chaves[] = { "a", "da", "de", "do", "o", "que",
                            "rata", "rato",  "rei", "roeu",
                            "roma", "rouba", "roupa" };
   tTrie       raiz = IniciaTrie();
   int         i,
               nChaves;

      /* Inicia a trie */
   raiz = IniciaTrie();

//Indices();

   nChaves = sizeof(chaves)/sizeof(chaves[0]);

      /* Constri a trie */
   for (i = 0; i < nChaves; ++i)
      InsereEmTrie(raiz, chaves[i]);

      /* Efetua buscas na trie e apresenta os resultados */
   ResultadoBusca(raiz, "rato");
   ResultadoBusca(raiz, "ratazana");
   ResultadoBusca(raiz, "romano");
   ResultadoBusca(raiz, "rouba");

   printf( "\n\n>>> Numero de chaves na trie: %d\n",
           NumeroDeChavesEmTrie(raiz) );

   RemoveEmTrie(raiz, "rato");
   RemoveEmTrie(raiz, "ratazana");

   printf("\n>>> Removida a chave \"%s\"", "rato");
   printf("\n>>> Removida a chave \"%s\"\n", "ratazana");

      /* Efetua buscas na trie e apresenta os resultados */
   ResultadoBusca(raiz, "rato");
   ResultadoBusca(raiz, "ratazana");

   printf( "\n\n>>> Numero de chaves na trie: %d\n",
           NumeroDeChavesEmTrie(raiz) );

   return 0;
}

/****************** Resultado do Programa *****************

>>> A chave "rato" esta' presente na trie
>>> A chave "ratazana" NAO esta' presente na trie
>>> A chave "romano" NAO esta' presente na trie
>>> A chave "rouba" esta' presente na trie

>>> Numero de chaves na trie: 13

>>> Removida a chave "rato"
>>> Removida a chave "ratazana"

>>> A chave "rato" NAO esta' presente na trie
>>> A chave "ratazana" NAO esta' presente na trie

>>> Numero de chaves na trie: 12

**********************************************************/
