/****
 *
 * Ttulo: Casamento Lxico
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 30/06/2016
 * ltima modificao: 30/05/2018
 *
 * Descrio: Este programa implementa uma soluo para o casamento
 *            lxico de um conjunto de padres com um texto
 *
 * Entrada: Arquivo contendo o texto no qual se esperam encontrar
 *          os padres
 *
 * Sada: Mensagens informando o nmero de ocorrncias de cada
 *        padro no texto
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>    /* Funes de Entrada/Sada */
#include <string.h>   /* Processamento de strings */
#include <stdlib.h>   /* Alocao dinmica de memria */
#include <limits.h>   /* Alocao dinmica de memria */

/************************ Macros **************************/

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

   /* Nmero primo usado para clculo de valores de disperso */
//#define PRIMO 33554393
#define PRIMO 32887

   /* A base a ser usada para clculo de digitais de Rabin */
#define BASE 256

   /* Nome do arquivo de texto */
#define NOME_ARQUIVO "Machado.txt"
//#define NOME_ARQUIVO "Machadinho.txt" /* Depurao */

   /* Macro utilizada em tratamento */
   /* de condies de exceo       */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

/****************** Definies de Tipos *******************/

   /* Tipo de estrutura que armazena um padro do lxico */
typedef struct {
           char *padrao;      /* O padro */
           int   dispersao;   /* Seu valor de disperso */
           int   ocorrencias; /* Nmero de ocorrncias no texto */
        } tPadraoLexico;

/***************** Definies de Funes ******************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
 *
 * CasamentoLexico(): Verifica a ocorrncia de algum padro de
 *                    um conjunto de padres (lxico) num texto
 *                    usando o algoritmo KR
 *
 * Parmetros:
 *      nomeArq (entrada) - nome do arquivo que contm o texto
 *      lexico (entrada) - array que contm os padres
 *      k (entrada) - nmero de elementos do array (lxico)
 *      m (entrada) - tamanho de cada padro no lxico
 *      b (entrada) - a base numrica utilizada em clculo de
 *                    valores de disperso
 *      q (entrada) - o nmero primo utilizado em clculo de
 *                    valores de disperso
 *
 * Retorno: Nenhum
 *
 ****/
void CasamentoLexico( const char *nomeArq, tPadraoLexico lexico[], int k,
                      int m, int b, int q )
{
   int   dispersaoTexto = 0, /* Disperso de cada janela */
         maiorPotencia = 1, /* Valor da base elevado a m - 1 */
         i, j,
         nLinha; /* Tamanho de uma linha lida no arquivo */
   char *linha; /* Apontar para cada linha lida */
   FILE *stream; /* Stream associado ao arquivo no qual ocorrer a leitura  */

      /* Abre arquivo em formato de texto para leitura */
   stream = fopen(nomeArq, "r");

      /* Se o arquivo no foi aberto, nada mais pode ser feito */
   ASSEGURA(stream, "Arquivo no foi aberto");

      /* Calcula o valor da maior potncia do polinmio */
   for (i = 0; i < m - 1; ++i)
      maiorPotencia = maiorPotencia*b%q;

      /* L cada linha do arquivo de texto e conta o nmero de ocorncias */
      /* de cada padro que  compe o lxico nessa linha                  */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(&nLinha, stream);

         /* Se a linha for NULL, encerra-se o lao */
      if (!linha)
         break;

      dispersaoTexto = 0; /* Inicia o valor de disperso das janelas da linha */

         /* Calcula o valor de disperso da primeira */
         /* janela da linha lida recentemente        */
      for (i = 0; i < m; ++i)
         dispersaoTexto = (dispersaoTexto*b + linha[i])%q;

         /* Procura casamentos entre os padres */
         /* do lxico e a linha corrente        */
      for (i = 0; i <= nLinha - m; ++i ) {
            /* Checa casamentos dos padres com a */
            /* janela corrente da linha corrente  */
         for (j = 0; j < k; ++j)
            if ( lexico[j].dispersao == dispersaoTexto &&
                 !memcmp(lexico[j].padrao, linha + i, m) )
                  ++lexico[j].ocorrencias;

            /* Calcula o valor de disperso da prxima janela da linha   */
         dispersaoTexto = ( b*(dispersaoTexto - linha[i]*maiorPotencia) +
                            linha[i + m] )%q;

            /* Corrige o valor de disperso se ele for negativo */
         if(dispersaoTexto < 0)
            dispersaoTexto += q;
      }

      free(linha); /* Libera o espao ocupado pela linha */
   }

   fclose(stream); /* Fecha o arquivo */

      /* Verifica se ocorreu erro de leitura */
   ASSEGURA(!ferror(stream), "Erro de leitura");
}

/****
 *
 * DispersoesLexico(): Calcula os valores de disperso de um
 *                     array de padres (lxico) usando uma
 *                     digital de Rabin
 *
 * Parmetros:
 *      lexico (entrada/sada) - array que contm os padres
 *      n (entrada) - nmero de elementos do array
 *      m (entrada) - tamanho de cada string do array
 *      b (entrada) - base que ser usada no clculo dos valores
 *                    de disperso
 *      q (entrada) - nmero primo que ser usado no clculo dos
 *                    valores de disperso
 *
 * Retorno: Nenhum
 *
 ****/
void DispersoesLexico(tPadraoLexico lexico[], int n, int m, int b, int q)
{
   int i, j;

      /* Calcula os valores de disperso dos padres */
   for (i = 0; i < n; ++i)
      for (j = 0; j < m; ++j)
         lexico[i].dispersao = ( lexico[i].dispersao*b +
                                 lexico[i].padrao[j] )%q;
}

/****
 *
 * EhPrimo(): Verifica se um nmero  primo
 *
 * Parmetros:
 *      n (entrada) - o nmero a ser testado
 *
 * Retorno: 1, se o nmero for primo; 0, em caso contrrio.
 *
 ****/
int EhPrimo(int n)
{
   int i;

   if (n < 2) /* Por definio, no existe primo menor do que 2*/
      return 0;

   for (i = 2; i <= n/2; ++i)
         /* Se for encontrado um divisor, o nmero no  primo */
      if (!(n%i))
         return 0;

   return 1; /* O nmero  primo */
}

/****
 *
 * MaiorPrimo(): Encontra o maior nmero primo menor do que ou
 *               igual a um dado nmero natural
 *
 * Parmetros:
 *      max (entrada) - valor mximo do nmero primo a ser
 *                      encontrado
 *
 * Retorno: 1, se o nmero for primo; 0, em caso contrrio.
 *
 ****/
int MaiorPrimo(int max)
{
   int i;

      /* Encontra o maior nmero primo menor */
      /* do que ou igual ao parmetro        */
   for (i = max; i > 2; --i)
      if (EhPrimo(i))
         return i;

   return 2; /* 2  o menor nmero que existe */
}

/****
 *
 * main(): Testa as funes de casamento lxico
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   tPadraoLexico lexico[] = { {"casa", 0, 0}, {"olho", 0, 0},
                              {"vida", 0, 0}, {"hora", 0, 0},
                              {"amor", 0, 0}, {"alma", 0, 0},
                              {"nome", 0, 0}, {"modo", 0, 0},
                              {"novo", 0, 0}, {"self", 0, 0}
                            };
   int           m = strlen(lexico->padrao), /* Todos os padres tm */
                                             /* o mesmo tamanho      */
                 k = sizeof(lexico)/sizeof(lexico[0]), /* Nmero de padres */
                                                       /* no lxico         */
                 i,
                 b = BASE,
                 q = PRIMO,
                 qMax = INT_MAX/(BASE*BASE - BASE);

      /* Teste preventivo de overflow */
   if (q > qMax) {
      qMax = MaiorPrimo(qMax);
      printf("\n\t\a>>> Ocorrera' overflow!\n"
             "\n\t>>> O maior valor de q deveria ser %d\n", qMax);
      return 1;
   }

      /* Calcula os valores de disperso dos */
      /* padres que constituem o lxico     */
   DispersoesLexico(lexico, k, m, b, q);

      /* Efetua o casamento dos padres com o texto */
   CasamentoLexico(NOME_ARQUIVO, lexico, k, m, b, q);

      /* Apresenta o resultado do casamento lxico */
   for (i = 0; i < k; ++i)
      if (lexico[i].ocorrencias)
         printf("\n\t>>> \"%s\" foi encontrado %d %s no texto", lexico[i].padrao,
                lexico[i].ocorrencias, lexico[i].ocorrencias > 1 ? "vezes" : "vez");
      else
         printf("\n\t>>> \"%s\" nao foi encontrado no texto", lexico[i].padrao);

   putchar('\n'); /* Enfeite */

   return 0;
}

/****************** Resultado do Programa *****************

        >>> "casa" foi encontrado 119 vezes no texto
        >>> "olho" foi encontrado 114 vezes no texto
        >>> "vida" foi encontrado 83 vezes no texto
        >>> "hora" foi encontrado 66 vezes no texto
        >>> "amor" foi encontrado 128 vezes no texto
        >>> "alma" foi encontrado 52 vezes no texto
        >>> "nome" foi encontrado 7 vezes no texto
        >>> "modo" foi encontrado 30 vezes no texto
        >>> "novo" foi encontrado 14 vezes no texto
        >>> "self" nao foi encontrado no texto

**********************************************************/
