/****
 *
 * Ttulo: Casamento de Padro com Texto em Fluxo Contnuo
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 30/06/2016
 * ltima modificao: 30/06/2016
 *
 * Descrio: Casamento de Padro com Texto em Fluxo Contnuo
 *
 * Entrada: Caracteres digitados pelo usurio via teclado e
 *          lidos num arquivo de texto
 *
 * Sada: Uma mensagem informando quando o casamento ocorre
 *
 ****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NOME_ARQ "I:\\Dados\\DNA\\DNA.txt"

/****
 *
 * ExibeTabela(): Apresenta o contedo de um array de ints
 *
 * Parmetros:
 *      ar (entrada) - o array
 *      tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 * Observao: Esta funo  usada apenas para depurao
 *
 ****/
void ExibeTabela(const int ar[], int tam)
{
   int i;

   for (i = 0; i < tam; ++i) {
      printf("\ttab[%d] = %d\n", i, ar[i]);
   }
}

/****
 *
 * CriaTabelaKMP(): Atribui valores para os elementos da tabela
 *                  de prefixos usada pelo algoritmo KMP
 *
 * Parmetros:
 *      tabela (sada) - array que armazenar a tabela
 *      p (entrada) - string que representa o padro
 *
 * Retorno: Endereo da tabela
 *
 ****/
int *CriaTabelaKMP(int tabela[], const char *p)
{
   int i = 0,
       j = -1,
       m = strlen(p);

   tabela[0] = -1;

   while (i < m) {
      while (j >= 0 && p[i] != p[j]) {
        j = tabela[j];
      }

      ++i;
      ++j;

      tabela[i] = j;
   }

   return tabela;
}

/****
 *
 * CasaFluxoContinuo(): tenta casar um string (padro) com
 *                      um texto cujos caracteres so lidos
 *                      um a um num meio de entrada
 *
 * Parmetros:
 *      p (entrada) - string que representa o padro
 *      stream (entrada) - stream associado ao arquivo que
 *                         contm o texto
 *      num (sada) - se no for NULL, armazenar o nmero
 *                    de caracteres lidos
 *
 * Retorno: A posio da primeira ocorrncia do padro no texto,
 *          se ele for encontrado; -1, em caso contrrio.
 *
 * Observao:  importante notar que esta funo encerra
 *             apenas quando  encontrado um casamento ou
 *             quando o final do arquivo  atingido, o que
 *             pode ser um problema para o usurio desavisado
 *             que introduz um texto via teclado e no sabe
 *             simular final de arquivo. Ou seja o usurio
 *             precisa ser um tanto nerd para saber que
 *             [CTRL] + [D] e [CTRL] + [Z] so usados com
 *             esse propsito em sistemas das famlia Unix
 *             e Windows/DOS, respectivamente.
 *
 ****/
int CasaFluxoContinuo(const char *p, FILE *stream, int *num)
{
   int c,        /* Armazena cada caractere lido */
       nCar = 0, /* Nmero de caracteres lidos */
       m = strlen(p), /* Comprimento do padro */
       i = 0, /* ndice de um caractere do padro */
      *tab; /* Ponteiro para a tabela de prefixos */

      /* Cria a tabela de prefixos */
   tab = CriaTabelaKMP(malloc(m*sizeof(int)), p);

      /* L o primeiro caractere do texto */
   c = fgetc(stream);

      /* O lao a seguir encerra quando o final do arquivo */
      /* for atingido ou for encontrado um casamento       */
   while (1) {
         /* Verifica se o final de arquivo foi atingido */
         /* ou ocorreu algum erro de leitura            */
      if (c == EOF)
         break;

      ++nCar; /* Foi lido mais um caractere */

         /* Enquanto os caracteres ora sendo comparados */
         /* no casarem, associa i ao ndice do padro  */
         /* indicado pela tabela de prefixos            */
      while (i >= 0 && c != p[i])
          i = tab[i];

      ++i; /* Passa para o prximo caractere do padro */

         /* Se j ocorreu casamento, encerra o lao */
      if (i >= m)
         break;

      c = fgetc(stream); /* L o prximo caractere do texto */
   }

      /* O array que armazena a tabela de */
      /* prefixos no  mais necessrio   */
   free(tab);

      /* Se o terceiro parmetro no for NULL, armazena */
      /* nele o nmero de caracteres lidos              */
   if (num)
      *num = nCar;

      /* Se o padro foi encontrado, retorna-se sua      */
      /* posio no texto; caso contrrio, retorna-se -1 */
   return i == m ? nCar - m : -1;
}

/****
 *
 * main(): Verifica se um padro ocorre num texto
 *         digitado pelo usurio
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero, se no ocorer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   char *p1 = "amor",/* Padro usado na primeira parte do programa */
        *p2 = "ATGAGCGGCGCCTGCA"; /* Padro a ser usado na     */
                                  /* segunda parte do programa */
   int   pos, /* Possvel posio de casamento  */
         nCaracteres; /* Nmero de caracteres lidos */
   FILE *stream; /* Stream associado ao arquivo que contm um */
                 /* banco de dados de DNA em formato FASTA    */

      /*                                      */
      /* Primeira parte: casamento com texto  */
      /*                 digitado via teclado */
      /*                                      */

   printf("Digite um texto falando em \"%s\":\n> ", p1);

      /* Chama a funo CasaFluxoContinuo() para ler */
      /* o texto e verificar se ele contm o padro  */
   pos = CasaFluxoContinuo(p1, stdin, &nCaracteres);

      /* Apresenta o resultado */
   if (pos >= 0)
      printf( "\n\t>>> \"%s\" aparece na posicao %d do texto\n",
              p1, pos);
   else
      printf( "\n\t>>> \"%s\" NAO aparece no texto\n", p1);


      /* Informa quantos caracteres foram lidos */
   printf("\n\t>>> Foram lidos %d caracteres\n", nCaracteres);

      /*                                      */
      /* Segunda parte: casamento com texto   */
      /*                armazenado em arquivo */
      /*                                      */

      /* Tenta abrir arquivo de dados em   */
      /* modo de texto apenas para leitura */
   stream = fopen(NOME_ARQ, "r");

      /* Se o arquivo no foi aberto, encerra o programa */
   if (!stream) {
      printf("\nArquivo \"%s\" nao pode ser aberto\n", NOME_ARQ);
      return 1;
   }

      /* Chama a funo CasaFluxoContinuo() para ler  */
      /* o arquivo e verificar se ele contm o padro */
   pos = CasaFluxoContinuo(p2, stream, &nCaracteres);

   fclose(stream); /* Arquivo j pode ser fechado */

      /* Apresenta o resultado */
   if (pos >= 0)
      printf( "\n\t>>> \"%s\" aparece na posicao %d do \n",
              p2, pos);
   else
      printf( "\n\t>>> \"%s\" NAO aparece no \n", p2);

   printf("\t>>> arquivo \"%s\"\n", NOME_ARQ);

      /* Informa quantos caracteres foram lidos */
   printf("\n\t>>> Foram lidos %d caracteres\n", nCaracteres);

   return 0;
}

/******************* Exemplos de Execuo ******************

Digite um texto falando em "amor":
> E' o amor amor amor

        >>> "amor" aparece na posicao 5 do texto

        >>> Foram lidos 9 caracteres

        >>> "ATGAGCGGCGCCTGCA" aparece na posicao 54161148 do
        >>> arquivo "I:\Dados\DNA.txt"

        >>> Foram lidos 54161164 caracteres

*************************************************************

Digite um texto falando em "amor":
> amor

        >>> "amor" aparece na posicao 0 do texto

        >>> Foram lidos 4 caracteres

        >>> "Ulysses" NAO aparece no
        >>> arquivo "I:\Dados\DNA.txt"

        >>> Foram lidos 254235634 caracteres

**********************************************************/
