/****
 *
 * Arquivo: Heap.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 06/10/2017
 * ltima modificao: 06/10/2017
 *
 * Descrio: Implementao de heap
 *
 ****/

/*********************** Includes *************************/

#include <string.h>   /* memmove()              */
#include <stdlib.h>   /* Alocao dinmica      */
#include "Heap.h"     /* Interface deste mdulo */
#include "ASSEGURA.h" /* Macro ASSEGURA()       */

/***************** Constantes Simblicas ******************/

#define TAMANHO_INICIAL 10 /* Capacidade inicial do heap */

/****************** Definies de Macros ******************/

#define FILHO_E(x) (2*(x) + 1)
#define FILHO_D(x) (2*(x) + 2)
#define PAI(x) (((x) - 1)/2)

/********************* Funes Locais *********************/

/****
 *
 * TrocaGenerica(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void TrocaGenerica(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * RedimensionaHeapHuff(): Redimensiona um heap dobrando sua capacidade
 *
 * Parmetros:
 *      *heap (sada) - heap que ser redimensionado
 *
 * Retorno: O primeiro parmetro
 *
 ****/
static tHeapHuff *RedimensionaHeapHuff(tHeapHuff *heap)
{
   heap->capacidade *= 2; /* Tenta duplicar a capacidade do heap */

      /* Aqui no h problema em atribuir o retorno de realloc() ao ponteiro */
      /* passado como primeiro parmetro, pois se no realloc() fracassar,   */
      /* o programa ser abortado                                            */
   heap->itens = realloc( heap->itens, heap->capacidade*sizeof(tNoHeapHuff) );
   ASSEGURA(heap->itens, "Impossivel redimensionar heap");

   return heap;
}

/****
 *
 * OrdenaHeapHuff(): Restaura a propriedade de ordenao de um
 *               heap de mnimo aps uma operao de remoo
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *      indice (entrada) - ndice do elemento do heap a partir
 *                         do qual ser efetuada a restaurao
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados no heap
 *
 * Retorno: Nada
 *
 ****/
static void OrdenaHeapHuff(tHeapHuff *heap, int indice, tFComparaHuff Compara)
{
   int iEsq = FILHO_E(indice), /* ndice do filho esquerdo */
       iDir = FILHO_D(indice), /* ndice do filho direito  */
       iMenor = indice; /* Supe que o menor elemento  o pai */

      /* Compara o filho esquerdo com seu pai */
   if ( iEsq < heap->nItens &&
        Compara( heap->itens[iEsq], heap->itens[indice] ) < 0 )
      iMenor = iEsq; /* O filho esquerdo  menor do que o pai */

      /* Compara o filho direito com o menor entre seu pai e seu irmo */
   if ( iDir < heap->nItens &&
        Compara( heap->itens[iDir], heap->itens[iMenor] ) < 0 )
      iMenor = iDir; /* O filho direito  o menor de todos */

      /* Se o n no for menor do que seus filhos, */
      /* troca-o de posio com o menor deles      */
   if (iMenor != indice) {
         /* O pai e o menor filho trocam de posio */
      TrocaGenerica( &heap->itens[indice], &heap->itens[iMenor],
                     sizeof(tNoHeapHuff) );

         /* Reordena o heap a partir do filho que se tornou pai */
      OrdenaHeapHuff(heap, iMenor, Compara);
   }
}

/****
 *
 * HeapCheioHuff(): Verifica se um heap est repleto
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: 1, se o heap estiver repleto; 0, em caso contrrio
 *
 ****/
static int HeapCheioHuff(const tHeapHuff *heap)
{
   return heap->nItens == heap->capacidade;
}

/********************* Funes Globais ********************/

/****
 *
 * IniciaHeapHuff(): Inicia um heap
 *
 * Parmetros:
 *      *heap (sada) - heap que ser iniciado
 *
 * Retorno: O primeiro parmetro
 *
 ****/
tHeapHuff *IniciaHeapHuff(tHeapHuff *heap)
{
   heap->capacidade = TAMANHO_INICIAL;
   heap->nItens = 0;
   heap->itens = calloc(TAMANHO_INICIAL, sizeof(tNoHeapHuff));

   ASSEGURA(heap->itens, "Impossivel alocar heap");

   return heap;
}

/****
 *
 * DestroiHeapHuff(): Libera o espao ocupado por um heap
 *
 * Parmetros: heap (entrada/sada) - o heap
 *
 * Retorno: Nada
 *
 ****/
void DestroiHeapHuff(tHeapHuff *heap)
{
      /* Libera o espao ocupado pelos ns do heap */
   free(heap->itens);
   heap->nItens = 0;
}

/****
 *
 * InsereEmHeapHuff(): Insere um novo item num heap
 *
 * Parmetros:
 *     heap (entrada/sada) - heap no qual ser feita insero
 *     item (entrada) - o item a ser inserido
 *     Compara (entrada) - funo de comparao
 *
 * Retorno: Nada
 *
 ****/
void InsereEmHeapHuff(tHeapHuff *heap, tNoHeapHuff item, tFComparaHuff Compara)
{
   int i;

      /* Se o array que suporta o heap estiver repleto, redimensiona-o */
   if(HeapCheioHuff(heap))
      RedimensionaHeapHuff(heap);

      /* Acrescenta o novo elemento ao final do array */
      /* e incrementa o nmero de elementos no heap   */
   heap->itens[heap->nItens] = item;
   heap->nItens++;

      /* Reordena o heap se for necessrio */
   for ( i = heap->nItens - 1; i &&
         Compara( heap->itens[PAI(i)], heap->itens[i] ) > 0; ) {
      TrocaGenerica(&heap->itens[i], &heap->itens[PAI(i)], sizeof(tNoHeapHuff));
      i = PAI(i);
   }
}

/****
 *
 * RemoveMinHeapHuff(): Remove o menor elemento de um heap binrio ascendente
 *
 * Parmetros:
 *     heap (entrada/sada) - heap onde ser feita remoo
 *     Compara (entrada/sada) - funo de comparao
 *
 * Retorno: O menor elemento do heap
 *
 ****/
tNoHeapHuff RemoveMinHeapHuff(tHeapHuff *heap, tFComparaHuff Compara)
{
   tNoHeapHuff menorElemento;

   ASSEGURA( !HeapVazioHuff(heap), "Tentativa de remover elemento de heap vazio" );

      /* O menor elemento est sempre na raiz de um heap de mnimo */
   menorElemento = heap->itens[0];

      /* Coloca o ltimo elemento do heap em sua raiz e reordena o heap */
   heap->itens[0] = heap->itens[--heap->nItens];
   OrdenaHeapHuff(heap, 0, Compara);

   return menorElemento;
}

/****
 *
 * ObtemMinimoHeapHuff(): Obtm o menor elemento de um heap de mnimo
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: O menor elemento do heap recebido como parmetro
 *
 ****/
tNoHeapHuff ObtemMinimoHeapHuff(const tHeapHuff *heap)
{
      /* O menor elemento de um heap de mnimo est na raiz */
   return heap->itens[0];
}

/****
 *
 * HeapVazioHuff(): Verifica se um heap est vazio
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: 1, se o heap estiver vazio; 0, em caso contrrio
 *
 ****/
int HeapVazioHuff(const tHeapHuff *heap)
{
   return !heap->nItens;
}

/****
 *
 * TamanhoHeapHuff(): Retorna o nmero de itens de um heap
 *
 * Parmetros:
 *      heap (entrada) - ponteiro para o heap
 *
 * Retorno: O nmero de itens do heap
 *
 ****/
int TamanhoHeapHuff(const tHeapHuff *heap)
{
   return heap->nItens;
}
