/****
 *
 * Exemplo: HeapSort
 *
 * Seo: 11.3.3
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de entrada e sada     */
#include <stdlib.h> /* Funes de alocao de memria */
#include <string.h> /* memmove()                      */
#include <time.h>   /* Funes de cronometragem       */

/***************** Constantes Simblicas ******************/

   /* Nmero mnimo de elementos na tabela a ser ordenada */
#define MIN_EL   10
   /* Nmero mximo de elementos na tabela a ser ordenada */
#define MAX_EL   20

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA( x, msg ) if (!(x) ) {\
                              printf("\n%s", msg); \
                              exit(1); \
                           }

#define FILHO_E(x) (2*(x) + 1)
#define FILHO_D(x) (2*(x) + 2)
#define PAI(x) (((x) - 1)/2)

/****************** Definies de Tipos *******************/

typedef int *tHeap;

/************************* Aluses ************************/

extern void OrdenaHeap( tHeap heap, int indice, int n );
extern void HeapSort(int tabela[], int nElem);
extern int* CriaTabela(int *nElementos, int min, int max);
extern void ApresentaTabela(const int *tabela, int nElementos);
extern void LiberaTabela(int **tabela);
extern int EstaOrdenada(int tabela[], int nElementos);

/***************** Definies de Funes ******************/

/****
 *
 * TrocaGenerica(): Troca os valores de duas variveis do mesmo tamanho
 *
 * Parmetros:
 *      var1, var2 (entrada e sada) - endereos das variveis
 *                                     que sero trocadas
 *      tam (entrada) - tamanho das variveis que sero trocadas
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
static void TrocaGenerica(void *var1, void *var2, size_t tam)
{
   void *p;

   p = malloc(tam);
   ASSEGURA(p, "Impossivel trocar duas variaveis");

   memmove(p, var1, tam);
   memmove(var1, var2, tam);
   memmove(var2, p, tam);

   free(p);
}

/****
 *
 * OrdenaHeap(): Restaura a propriedade de ordenao de um
 *               heap de mnimo aps uma operao de remoo
 *
 * Parmetros:
 *      heap (entrada e sada) - ponteiro para o heap
 *      indice (entrada) - ndice do elemento do heap a partir
 *                         do qual ser efetuada a restaurao
 *      Compara (entrada) - ponteiro para uma funo que compara
 *                          dois elementos de array armazenados no heap
 *
 * Retorno: Nada
 *
 ****/
void OrdenaHeap( tHeap heap, int indice, int n )
{
   int iEsq = FILHO_E(indice), /* ndice do filho esquerdo */
       iDir = FILHO_D(indice), /* ndice do filho direito  */
       iMenor = indice; /* Supe que o maior elemento  o pai */

      /* Compara o filho esquerdo com seu pai */
   if ( iEsq < n && heap[iEsq] > heap[indice] )
      iMenor = iEsq; /* O filho esquerdo  maior do que o pai */

      /* Compara o filho direito com o maior entre seu pai e seu irmo */
   if ( iDir < n && heap[iDir] > heap[iMenor] )
      iMenor = iDir; /* O filho direito  o maior de todos */

      /* Se o n no for maior do que seus filhos, */
      /* troca-o de posio com o maior deles      */
   if (iMenor != indice) {
         /* O pai e o maior filho trocam de posio */
      TrocaGenerica(heap + indice, heap + iMenor, sizeof(int));

         /* Reordena o heap a partir do filho que se tornou pai */
      OrdenaHeap(heap, iMenor, n);
   }
}

/****
 *
 * HeapSort(): Ordena uma tabela usando ordenao por heap sort
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a ordenada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: Nada
 *
 ****/
void HeapSort(int tabela[], int nElem)
{
   int i;

      /* Constri o heap a partir do pai do ltimo elemento da tabela */
   for(i = (nElem - 2)/2; i >= 0; i--)
      OrdenaHeap(tabela, i, nElem);

      /* Neste ponto, o heap est construdo. Troca-se o primeiro elemento */
      /* do heap, diminudo de um elemento, com o ltimo elemento e        */
      /* reordena-se o heap.                                               */
   for(i = nElem-1; i >= 1; i--) {
         /* Remove o maior elemento do heap e coloca no final da tabela */
      TrocaGenerica(tabela, tabela + i, sizeof(tabela[0]));

      OrdenaHeap(tabela, 0, i);  /* Reordena o heap */
   }
}

/****
 *
 * CriaTabela(): Cria uma tabela (array de inteiros) aleatoriamente
 *
 * Parmetros:
 *      nElementos (sada) - nmero de elementos na tabela criada
 *      min (entrada) - nmero mnimo de elementos na tabela criada
 *      max (entrada) - nmero mximo de elementos na tabela criada
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int* CriaTabela(int *nElementos, int min, int max)
{
   int *tabela, nElem, i, maiorElemento;

      /* Gera nmero de ns da tabela */
   nElem = (int) rand()%(max-min+1) + min;

   ASSEGURA( nElem >= min && nElem <= max,
             "Numero de elementos fora do intervalo" );

   tabela = calloc(nElem, sizeof(int));
   ASSEGURA( tabela, "Nao foi possivel alocar a tabela" );

      /* Isto no  necessrio e  feito apenas para */
      /* facilitar a visualizao dos resultados     */
   maiorElemento = 2*max;


      /* Gera valores dos elementos entre 0 e maiorElemento */
   for (i = 0; i < nElem; ++i)
      tabela[i] = (int) rand()%maiorElemento;

   *nElementos = nElem;

   return tabela;
}

/****
 *
 * ApresentaTabela(): Apresenta na tela a tabela recebida como parmetro
 *
 * Parmetros:
 *      tabela (entrada) - a tabela cujos elementos sero
 *                         exibidos
 *      nElementos (entrada) - nmero mnimo de elementos na
 *                             tabela
 *
 * Retorno: Nada
 *
 ****/
void ApresentaTabela(const int *tabela, int nElementos)
{
   int i;

   printf("\nNumero de elementos: %d", nElementos);
   printf("\n{");

   for (i = 0; i < nElementos - 1; ++i)
      printf("%d, ", tabela[i]);

   printf("%d}", tabela[nElementos - 1]);
}

/****
 *
 * LiberaTabela(): Libera o espao ocupado por uma tabela
 *                 e torna nulo o ponteiro que a representa
 *
 * Parmetros:
 *      tabela (entrada/sada) - a tabela cujo espao ser liberado
 *
 * Retorno: Nada
 *
 * NOTA: Se o ponteiro para o incio da tabela for nulo,
 *       nada ser feito; caso contrrio, assume-se que a
 *       tabela foi alocada dinamicamente
 *
 ****/
void LiberaTabela(int **tabela)
{
   if(*tabela) {
      free(*tabela);
      *tabela = NULL;
   }
}

/****
 *
 * EstaOrdenada(): Testa se uma tabela est ordenada em ordem
 *                 crescente
 *
 * Parmetros:
 *      tabela (entrada/sada) - tabela que ser a testada
 *      nElem (entrada) - nmero de elementos na tabela
 *
 * Retorno: 1, se a tabela estiver ordenada;
 *          0, em caso contrrio
 *
 ****/
int EstaOrdenada(int tabela[], int nElementos)
{
   int i;

   for (i = 0; i < nElementos -1; ++i)
      if (tabela[i] > tabela[i+1])
         return 0;

   return 1;
}

/****
 *
 * main(): Testa ordenao usando HeapSort
 *
 * Parmetros: Nenhum
 *
 * Retorno: 0, se no ocorrer erro; 1, em caso contrrio
 *
 ****/
int main(void)
{
   int *tabela; /* Ponteiro para as tabela utilizada */
   int  nElementos;

   srand((int) time(NULL));

   tabela = CriaTabela(&nElementos, MIN_EL, MAX_EL);

   printf("\n\n>>>>>>>>>>>>>>>> HeapSort <<<<<<<<<<<<<<<<<\n");

   printf("\n*********** Antes da Ordenacao ************\n");
   ApresentaTabela(tabela, nElementos);

   HeapSort(tabela, nElementos);

   if (!EstaOrdenada(tabela, nElementos)) {
      printf("\n\n\a>>> Erro: A tabela NAO foi ordenada <<<\n");
      return 1;
   }

   printf("\n\n*********** Depois da Ordenacao ***********\n");
   ApresentaTabela(tabela, nElementos);
   putchar('\n');

   LiberaTabela(&tabela);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>>>>>>>>>>>>>>> HeapSort <<<<<<<<<<<<<<<<<

*********** Antes da Ordenacao ************

Numero de elementos: 10
{29, 3, 32, 21, 26, 29, 37, 32, 22, 4}

*********** Depois da Ordenacao ***********

Numero de elementos: 10
{3, 4, 21, 22, 26, 29, 29, 32, 32, 37}

**********************************************************/
