/****
 *
 * Exemplo: Avaliao de Mtodos de Ordenao
 *
 * Seo: 11.7
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de entrada e sada     */
#include <stdlib.h>    /* Funes de alocao de memria */
#include <time.h>      /* Funes de cronometragem       */
#include "ASSEGURA.h"  /* Macro ASSEGURA                 */
#include "Metodos.h"   /* Mtodos de ordenao           */
#include "Leitura.h"   /* Biblioteca de leitura          */
#include "Registros.h" /* Operaes sobre arquoivos      */

/***************** Constantes Simblicas ******************/

   /* Nome do arquivo original */
#define NOME_ARQUIVO_TEXTO "CEPs.txt"

   /* Nome do arquivo binrio */
#define NOME_ARQUIVO_BIN "ApenasCEPs.bin"

/****************** Definies de Tipos *******************/

   /* Tipo de ponteiro para funo de ordenao */
typedef void (*tFOrdena) (int *, int);

/***************** Definies de Funes ******************/

/****
 *
 * ComparaIntsInv(): Compara dois elementos que sero ordenados
 *                em ordem decrescente
 *
 * Parmetros:
 *      e1, e2 (entrada) - endereos dos elementos que sero comparados
 *
 * Retorno:
 *     * 0, se os elementos so iguais
 *     * > 0, se o primeiro elemento for menor do que o segundo
 *     * < 0, se o primeiro elemento for maior do que o segundo
 *
 ****/
int ComparaIntsInv(const void *e1, const void* e2)
{
   ASSEGURA(e1 && e2, "Elemento nulo recebido por ComparaInts()");

   return *(int *)e2 - *(int *)e1;
}
/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 ****/
void MedidaDeTempo(void)
{
   static int     numeroDeChamadas = 0;
   static clock_t t0; /* Instante inicial */
   clock_t        tf; /* Instante final   */
   double         intervalo;

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = clock();
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = clock();

      intervalo = ((double) (tf - t0))/CLOCKS_PER_SEC;

      printf( "\n*** Tempo gasto na operacao: "
              "%8.5f segundos ***\n", intervalo );
   }
}

/****
 *
 * CriaTabelaApenasCEPs(): L os nmeros inteiros de um arquivo e
 *               armazena-os num array alocado dinamicamente
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      nEle (entrada) - nmero de elementos da tabela
 *
 * Retorno: Endereo da tabela criada
 *
 ****/
int *CriaTabelaApenasCEPs(FILE *stream, int nEle)
{
   int *tabela; /* Ponteiro para um array que armazena a tabela */
   int  i = 0, umaChave;

      /* Tenta alocar espao para a tabela */
   tabela = calloc(nEle, sizeof(int));
   ASSEGURA( tabela, "Tabela nao foi alocada" );

   rewind(stream); /* Garante leitura a partir do incio do arquivo */

      /* L cada nmero inteiro do arquivo e armazena-o na tabela */
   while (1) {
      fread(&umaChave, sizeof(umaChave), 1, stream);

      ASSEGURA(!ferror(stream), "Erro de leitura");

      if (feof(stream))
         break;

         /* Testa se o nmero de elementos da tabela   */
         /* menor do que o nmero de valores no arquivo */
      ASSEGURA(i < nEle, "Tabela subdimensionada");

      tabela[i++] = umaChave;
   }

   return tabela;
}

/****
 *
 * ExibeResultadoDeTeste(): Apresenta na tela o resultado de um teste
 *                   de mtodo de ordenao
 *
 * Parmetros:
 *      tabela (entrada) - tabela que ser ordenada
 *      n (entrada) - nmero de elementos da tabela
 *      ordena (entrada) - ponteiro para a funo de ordenao
 *      nome (entrada) - nome do mtodo de ordenao
 *
 * Retorno: Nada
 *
 ****/
void ExibeResultadoDeTeste(int *tabela, int n, tFOrdena ordena, const char *nome)
{
   printf("\n>>>>>>>>>>>>>>>>>>>> %s <<<<<<<<<<<<<<<<<<<<", nome);

   printf("\n*** Antes da Ordenacao: ");

   if (TabelaEstaOrdenada(tabela, n))
      printf("Tabela ordenada ***\n");
   else if (TabelaEstaEmOrdemInversa(tabela, n))
      printf("Tabela em ordem inversa ***\n");
   else
      printf("Tabela desordenada ***\n");

   printf("*** Ordenando a tabela...");
   MedidaDeTempo();

   ordena(tabela, n);

   printf(" fim da ordenacao ***");
   MedidaDeTempo();

   if (!TabelaEstaOrdenada(tabela, n))
      printf("\n\n\a>>> A tabela NAO foi ordenada <<<\n");
}

/****
 *
 * TestaMetodo(): Testa um mtodo de ordenao e apresenta o resultado na tela
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         as chaves que sero ordenadas
 *      nChaves (entrada) - nmero de elementos da tabela
 *      ordena (entrada) - ponteiro para a funo de ordenao
 *      nome (entrada) - nome do mtodo de ordenao
 *
 * Retorno: Nada
 *
 ****/
void TestaMetodo( FILE *stream, int nChaves, tFOrdena ordena,
                  const char *nome )
{
   int  *tabela; /* Ponteiro para a tabela utilizada */

   tabela = CriaTabelaApenasCEPs(stream, nChaves);
   ExibeResultadoDeTeste(tabela, nChaves, ordena, nome);
   ExibeResultadoDeTeste(tabela, nChaves, ordena, nome);
   InverteTabela(tabela, nChaves, ComparaIntsInv);
   ExibeResultadoDeTeste(tabela, nChaves, ordena, nome);
   free(tabela);
}

/****
 *
 * main(): Testa mtodos de ordenao
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main(void)
{
   int   nChaves, maxRegs;
   FILE *stream;

   printf("\n>>> Numero maximo de registros (0 = todos): ");
   maxRegs = LeInteiro();

      /* Cria o arquivo binrio */
   nChaves = CriaArquivoApenasCEPs(NOME_ARQUIVO_TEXTO, NOME_ARQUIVO_BIN, maxRegs);

      /* Verifica se ocorreu erro na criao do arquivo binrio */
   ASSEGURA(!maxRegs || maxRegs == nChaves, "Ocorreu erro na criacao do arquivo");

      /* Apresenta o resultado da converso */
   printf( "\n   >>> Armazenados %d registros no arquivo "
           "\"%s\" <<<\n", nChaves, NOME_ARQUIVO_BIN );

   stream = AbreArquivo(NOME_ARQUIVO_BIN, "rb"); /* Tenta abrir o arquivo binrio */

   printf("\n      ***** Tamanho da tabela: %d *****\n", nChaves);

   TestaMetodo(stream, nChaves, BubbleSort, "BubbleSort");
   TestaMetodo(stream, nChaves, InsertionSort, "InsertionSort");
   TestaMetodo(stream, nChaves, SelectionSort, "SelectionSort");
   TestaMetodo(stream, nChaves, QuickSort1, "QuickSort Basico");
   TestaMetodo(stream, nChaves, QuickSort2, "QuickSort Mediana");
   TestaMetodo(stream, nChaves, QuickSort3,"QuickSort Aleatorio");
   TestaMetodo(stream, nChaves, MergeSort, "MergeSort");
   TestaMetodo(stream, nChaves, HeapSort, "HeapSort");

   putchar('\n');

   fclose(stream);

   return 0;
}

/******************* Exemplo de Execuo ******************

>>> Numero maximo de registros (0 = todos): 40000

   >>> Armazenados 40000 registros no arquivo "ApenasCEPs.bin" <<<

      ***** Tamanho da tabela: 40000 *****

>>>>>>>>>>>>>>>>>>>> BubbleSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  4.07600 segundos ***

>>>>>>>>>>>>>>>>>>>> BubbleSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00000 segundos ***

>>>>>>>>>>>>>>>>>>>> BubbleSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  4.43900 segundos ***

>>>>>>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.39500 segundos ***

>>>>>>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00000 segundos ***

>>>>>>>>>>>>>>>>>>>> InsertionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.78400 segundos ***

>>>>>>>>>>>>>>>>>>>> SelectionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.68500 segundos ***

>>>>>>>>>>>>>>>>>>>> SelectionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.91300 segundos ***

>>>>>>>>>>>>>>>>>>>> SelectionSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.80600 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Basico <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00500 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Basico <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.56900 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Basico <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  1.63900 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Mediana <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00600 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Mediana <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00200 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Mediana <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00400 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Aleatorio <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.01000 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Aleatorio <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00400 segundos ***

>>>>>>>>>>>>>>>>>>>> QuickSort Aleatorio <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00300 segundos ***

>>>>>>>>>>>>>>>>>>>> MergeSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00500 segundos ***

>>>>>>>>>>>>>>>>>>>> MergeSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00500 segundos ***

>>>>>>>>>>>>>>>>>>>> MergeSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00400 segundos ***

>>>>>>>>>>>>>>>>>>>> HeapSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela desordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.01000 segundos ***

>>>>>>>>>>>>>>>>>>>> HeapSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela ordenada ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.00800 segundos ***

>>>>>>>>>>>>>>>>>>>> HeapSort <<<<<<<<<<<<<<<<<<<<
*** Antes da Ordenacao: Tabela em ordem inversa ***
*** Ordenando a tabela... fim da ordenacao ***
*** Tempo gasto na operacao:  0.01000 segundos ***


**********************************************************/
