/****
 *
 * Ttulo: Ordenao externa de arquivos usando intercalao binria
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2016
 * ltima modificao: 15/05/2018
 *
 * Descrio:
 *
 * Entrada: Um arquivo contendo registros sem ordenao
 *
 * Sada: O arquivo ordenado
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>        /* Entrada e sada     */
#include <string.h>       /* strcpy() e strchr() */
#include <stdlib.h>       /* qsort()             */
#include "IntercalaBin.h" /* Interface deste mdulo */
#include "Registros.h" /* Funes de processamento de registros */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/**************** Constantes Simblicas ****************/

   /* Tamanho de um registro  */
#define TAM_REG sizeof(tRegistroMEC)

/************************ Macros ************************/

   /* Menor de dois valores */
#define MIN(a, b) ((a) <= (b)) ? (a) : (b)

/********************* Funes Locais *********************/

/****
 *
 * ComparaInts2(): Compara dois registros
 *
 * Parmetros:
 *      e1, e2 (entrada) - os dois registros
 *
 * Retorno:
 *     * 0, se os registros so iguais
 *     * > 0, se o primeiro registro for maior do que o segundo
 *     * < 0, se o primeiro registro for menor do que o segundo
 *
 ****/
static int ComparaInts2(const void *e1, const void* e2)
{
   tRegistroMEC *r1 = (tRegistroMEC *)e1,
                *r2 = (tRegistroMEC *)e2;

   return r1->chave - r2->chave;
}

/****
 *
 * CopiaRestoArquivo(): Copia contedo de um arquivo para outro byte
 *                      a byte a partir da posio em que se encontra
 *                      o apontador de posio de cada arquivo
 *
 * Parmetros:
 *      streamE (entrada) - stream associado ao arquivo que ser
 *                          copiado
 *      streamS (entrada) - stream associado ao arquivo que receber a cpia
 *      maxBytes (entrada) - nmero mximo de bytes que sero
 *                           copiados; se esse valor for zero,
 *                           no haver limite para o nmero de bytes copiados
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. O stream que ser lido deve estar aberto num modo
 *        que permita leitura e o stream que ser escrito
 *        deve estar aberto num modo que permite escrita.
 *
 *     2. O apontador de posio do arquivo lido no  movido
 *        para o incio do arquivo, de modo que o primeiro byte
 *        a ser copiado  aquele indicado pela posio corrente
 *        desse apontador quando esta funo  chamada.
 *
 ****/
static void CopiaRestoArquivo( FILE *streamE, FILE *streamS, long maxBytes )
{
   int  c; /* Armazenar cada byte lido e escrito  */
   long contaBytes = 0;

      /* O lao a seguir encerra quando:        */
      /*   * Ocorrer tentativa de leitura alm do final do arquivo de entrada */
      /*   * Ocorrer erro de leitura ou escrita em qualquer dos arquivos      */
      /*   * O nmero mximo de bytes tiver sido copiado                      */
   while (1) {
      c = fgetc(streamE); /* L um byte no arquivo de entrada */

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(streamE) || ferror(streamE))
         break; /* Processamento encerrado */

      fputc(c, streamS); /* Escreve o byte lido no arquivo de sada */

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamS))
         break;

      ++contaBytes; /* Mais um byte foi copiado */

         /* Verifica se o nmero mximo de bytes j foi copidado */
      if (maxBytes && contaBytes >= maxBytes)
         break;
   }

      /* Certifica que no houve erro de leitura ou escrita */
   ASSEGURA(!ferror(streamE), "Erro de leitura ao tentar copiar arquivo");
   ASSEGURA(!ferror(streamS), "Erro de escrita ao tentar copiar arquivo");
}

/****
 *
 * CarregaBufferMS(): Preenche o contedo de um buffer com registros
 *                  ordenados lidos de um arquivo
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que contm
 *                         os registros
 *      buffer (entrada) - o buffer que ser preenchido
 *      maxRegs (entrada) - nmero mximo de registros que sero
 *                          lidos
 *     *pos (entrada/sada) - posio da prxima leitura no arquivo
 *
 * Retorno: O nmero de registros lidos
 *
 ****/
static int CarregaBufferMS( FILE *stream, tRegistroMEC *buffer,
                            int maxRegs, long *pos )
{
   int nRegistros; /* Nmero de registros lidos */

      /* Move o apontador de posio para o local de leitura */
   MoveApontador(stream, *pos, SEEK_SET);

      /* L os registros no arquivo */
   nRegistros = fread(buffer, TAM_REG, maxRegs, stream);
   ASSEGURA(!ferror(stream), "Erro de leitura");

   *pos = ftell(stream); /* Obtm a posio da prxima leitura */

   return nRegistros;
}

/****
 *
 * IntercalaDuasSeriesMS(): Intercala duas sries armazenadas num
 *                         arquivo e armazena o resultado noutro arquivo
 *
 * Parmetros:
 *      streamE (entrada) - stream associado ao arquivo que contm as sries
 *      streamS (entrada) - stream associado ao arquivo que conter
 *                          o resultado da intercalao
 *      is1 (entrada) - ndice da primeira srie
 *      is2 (entrada) - ndice da segunda srie
 *      tamSerie (entrada) - tamanho mximo de uma srie
 *
 * Retorno: Nada
 *
 * Observao: O ndice da primeira srie deve ser menor do que o
 *             ndice da segunda srie
 *
 ****/
static void IntercalaDuasSeriesMS( FILE *streamE, FILE *streamS,
                                   long is1, long is2, int tamSerie )
{
   tRegistroMEC *buf1, *buf2, /* Buffers de entrada */
                *bufS; /* Buffer de sada */
   int           ir1 = 0, /* ndice do prximo registro a ser lido no buffer 1 */
                 ir2 = 0, /* ndice do prximo registro a ser lido no buffer 2 */
                 irs = 0, /* ndice do prximo registro a   */
                          /* ser escrito no buffer de sada */
                 tamBuffer, /* Tamanho de um buffer */
                 n1, n2, /* Nmeros de registros nos buffers associados */
                         /* respectivamente s sries 1 e 2             */
                 resto1 = tamSerie,/* Registros restantes na srie 1 */
                 resto2 = tamSerie;/* Registros restantes na srie 2 */
   long          pos1, /* Posio de leitura de um byte da srie 1 */
                 pos2; /* Posio de leitura de um byte da srie 2 */

      /* A primeira srie deve preceder a segunda srie */
   ASSEGURA(is1 < is2, "Indice da 1a serie e' maior do que indice da 2a");

      /* Divide a memria supostamente disponvel pelos trs buffers */
   tamBuffer = MAX_REGISTROS/3;

      /* Aloca espao para cada buffer */
   buf1 = calloc(tamBuffer, TAM_REG);
   buf2 = calloc(tamBuffer, TAM_REG);
   bufS = calloc(tamBuffer, TAM_REG);
   ASSEGURA( buf1 && buf2 && bufS, "Pelo menos um buffer nao foi alocado" );

      /* Calcula a posio do primeiro byte de cada srie */
   pos1 = is1*tamSerie*TAM_REG;
   pos2 = is2*tamSerie*TAM_REG;

      /* Efetua os primeiros carregamentos de buffers */
   n1 = CarregaBufferMS(streamE, buf1, MIN(tamBuffer, resto1), &pos1);
   n2 = CarregaBufferMS(streamE, buf2, MIN(tamBuffer, resto2), &pos2);

      /* Calcula o nmero de registros restantes em cada srie */
   resto1 -= n1;
   resto2 -= n2;

      /* Intercala as duas sries */
   while (1) {
         /* Se o buffer de sada estiver cheio descarrega-o para o arquivo de sada */
      if (irs >= tamBuffer) {
         fwrite(bufS, TAM_REG, tamBuffer, streamS);
         ASSEGURA(!ferror(streamS), "Erro de escrita");

         irs = 0; /* Buffer ficou novinho em folha */
      }

         /* Verifica se o buffer 1 est esgotado */
      if (ir1 >= n1) {
            /* Se ainda houver registros na srie 1, recarrega esse buffer. */
            /* Caso contrrio, o processamento dessa srie est concludo   */
         if (resto1 > 0) {
               /* Ainda h registros da srie 1 em arquivo */
            n1 = CarregaBufferMS( streamE, buf1, MIN(tamBuffer, resto1), &pos1 );
            ir1 = 0;
            resto1 -= n1;
         } else {
            /*                                                         */
            /* No h mais registros na srie 1. Copia o restante da   */
            /* segunda srie para o arquivo de sada e encerra o lao. */
            /*                                                         */

               /* Primeiro, copia os registros que se encontram no buffer de */
               /* sada e depois copia aqueles que se encontram no buffer 2  */
            fwrite(bufS, TAM_REG, irs, streamS);
            fwrite(buf2 + ir2, TAM_REG, n2 - ir2, streamS);
            ASSEGURA(!ferror(streamS), "Erro de escrita");

               /* Se houver registros da srie 2 restantes no */
               /* arquivo, copia-os para o arquivo de sada   */
            if (resto2 > 0) {
               MoveApontador(streamE, pos2, SEEK_SET);
               CopiaRestoArquivo(streamE, streamS, (long) resto2*TAM_REG);
            }
            break;
         }
      }

         /* Verifica se o buffer 2 est esgotado */
      if (ir2 >= n2) {
            /* Se ainda houver registros na srie 2, recarrega esse buffer. */
            /* Caso contrrio, o processamento da srie 2 est concludo.   */
         if (resto2 > 0) {
               /* Talvez, ainda haja registros da srie 2 em arquivo, mas o   */
               /* nmero inicial de registros dessa srie pode ter sido menor */
            n2 = CarregaBufferMS(streamE, buf2, MIN(tamBuffer, resto2), &pos2);
            ir2 = 0;
            resto2 = n2 ? resto2 - n2 : 0;
         } else {
            /*                                                       */
            /* No h mais registros na srie 2. Copia o restante da */
            /* srie 1 para o arquivo de sada e encerra o lao.     */
            /*                                                       */

               /* Primeiro, copia os registros que se encontram no buffer de */
               /* sada e depois copia aqueles que se encontram no buffer 1  */
            fwrite(bufS, TAM_REG, irs, streamS);
            fwrite(buf1 + ir1, TAM_REG, n1 - ir1, streamS);
            ASSEGURA(!ferror(streamS), "Erro de escrita");

               /* Se houver registros da srie 1 restantes no */
               /* arquivo, copia-os para o arquivo de sada   */
            if (resto1 > 0) {
               MoveApontador(streamE, pos1, SEEK_SET);
               CopiaRestoArquivo( streamE, streamS, (long) resto1*TAM_REG );
            }
            break;
         }
      }

         /* O teste a seguir  necessrio, pois a srie 2 pode */
         /* ser a ltima do arquivo e seu tamanho inicial pode */
         /* ser menor do que o tamanho das demais sries       */
      if (n2 > 0)
            /* Copia o menor dos dois registros correntes nos  */
            /* buffers de entrada 1 e 2 para o buffer de sada */
         bufS[irs++] = ComparaInts2(buf1 + ir1, buf2 + ir2) < 0 ? buf1[ir1++]
                                                                : buf2[ir2++];

   }

   free(buf1); /* Libera os buffers */
   free(buf2);
   free(bufS);
}

/********************* Funes Globais ********************/

/****
 *
 * CriaSeriesMS(): Cria sries (ordenadas) em memria secundria
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que origina as sries
 *      nomeS (entrada) - nome do arquivo que armazenar as sries
 *
 * Retorno: Nmero de sries criadas
 *
 ****/
int CriaSeriesMS(const char *nomeE, const char *nomeS)
{
   tRegistroMEC *registros; /* Ponteiro para um array que armazena */
                            /* uma srie em memria principal      */
   int           nRegistros,  /* Nmero de registros numa srie */
                 nSeries = 0; /* Nmero de sries */
   FILE         *streamS, /* Stream associado ao arquivo que armazenar as sries */
                *streamE; /* Stream associado ao arquivo original */

      /* Aloca espao para o array. Um array deste */
      /* tamanho no pode ser alocado na pilha     */
   registros = calloc(MAX_REGISTROS, TAM_REG);
   ASSEGURA( registros, "Buffer nao foi alocado" );

      /* Tenta criar o arquivo que conter as sries */
   streamS = AbreArquivo(nomeS, "wb");

      /* Tenta abrir o arquivo original */
   streamE = AbreArquivo(NOME_ARQUIVO_BIN, "rb");

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os registros da srie corrente */
      nRegistros = fread( registros, TAM_REG, MAX_REGISTROS, streamE );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(streamE), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum registro foi lido */
      if (!nRegistros)
         break;

         /* Ordena os registros lidos */
      qsort(registros, nRegistros, TAM_REG, ComparaInts2);

         /* Escreve a srie corrente no arquivo */
      fwrite(registros, TAM_REG, nRegistros, streamS);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(streamS), "Erro de escrita em arquivo de series" );

      ++nSeries; /* Atualiza o nmero de sries */
   }

   FechaArquivo(streamS, nomeS); /* Fecha os arquivos */
   FechaArquivo(streamE, nomeE);

   free(registros); /* Libera o espao ocupado pelo array */

   return nSeries;
}

/****
 *
 * IntercalaBinMS(): Intercala sries armazenadas num arquivo
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que inicialmente contm
 *                        as sries que sero intercaladas
 *      nomeS (entrada) - nome do arquivo que conter o resultado
 *                        de intercalao das sries do primeiro arquivo
 *      nSeries (entrada) - nmero inicial de sries
 *      tamSerie (entrada) - tamanho (nmero de registros) mximo
 *                           de cada srie inicial
 *
 * Retorno: O nome do arquivo que contm o resultado
 *
 * Observao: Os nomes associados aos arquivos de entrada e sada
 *             mudam durante a execuo desta funo (i.e., eles
 *             trocam de papis). Assim ao final da operao, o
 *             resultado pode estar armazenado no arquivo
 *             representado pelo primeiro ou pelo segundo parmetro.
 *             Essa  a razo pela qual o nome do arquivo que contm
 *             o resultado  retornado. Mas  fcil alterar essa
 *             funo de modo que o resultado seja sempre armazenado
 *             no segundo parmetro.
 *
 ****/
const char *IntercalaBinMS( const char *nomeE, const char *nomeS,
                            int nSeries, int tamSerie )
{
   FILE       *streamS, /* Stream associado ao arquivo que contm */
                        /* o resultado de concatenao de sries  */
              *streamE; /* Stream associado ao arquivo que contm */
                        /* sries que sero concatenadas          */
   int         ns = nSeries, /* Nmero corrente de sries */
               ts = tamSerie; /* Tamanho mximo de cada srie */
   int         i1, i2; /* ndices de duas sries que sero intercaladas */
   const char *aux;

   while (ns > 1) {
         /* Tenta abrir o arquivo contendo as sries que sero intercaladas */
      streamE = AbreArquivo(nomeE, "rb");

         /* Tenta criar o arquivo que conter as sries resultantes */
      streamS = AbreArquivo(nomeS, "wb");

#if DEPURACAO == 1
      printf("\a\n\t>>> Intercalando %d series (ts = %d)...", ns, ts);
      fflush(stdout);
#endif

      i1 = 0; /* ndice inicial da primeira srie */
      i2 = 1; /* ndice inicial da segunda srie  */

         /* As sries so indexadas de 0 a ns - 1. */
         /* Portanto o maior valor de i2  ns - 1. */
      while (i2 < ns) {
            /* Intercala as duas sries correntes */
         IntercalaDuasSeriesMS(streamE, streamS, i1, i2, ts);

         i1 += 2; /* Passa para as prximas sries */
         i2 += 2;
      }

         /* Se o nmero corrente de sries for mpar, resta */
         /* copiar a ltima srie para o arquivo resultante */
      if (i1 < ns) {
            /* Verificao de consistncia */
         ASSEGURA(ns%2, "O numero de series deveria ser impar");
         ASSEGURA(i1 == ns - 1, "Deveria ser a ultima serie");

            /* Move apontador para o primeiro byte a ser copiado */
            /* e copia a ltima srie para o arquivo resultante  */
         MoveApontador( streamE, (long)i1*ts*TAM_REG, SEEK_SET );
         CopiaRestoArquivo(streamE, streamS, 0);
      }

      FechaArquivo(streamE, nomeE); /* Fecha os arquivos */
      FechaArquivo(streamS, nomeS);

      ns = ns/2 + ns%2; /* O nmero de sries foi reduzido  metade */

         /* O nmero mximo de registros de cada srie dobra de tamanho,   */
         /* sendo que o nmero de registros da ltima srie pode ser menor */
      ts *= 2;

      aux = nomeE; /* Troca os papis dos arquivos */
      nomeE = nomeS;
      nomeS = aux;
   }

   return nomeE; /* Retorna o nome do arquivo que contm o resultado */
}
