/****
 *
 * Ttulo: Ordenao externa de arquivos usando intercalao binria
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/06/2016
 * ltima modificao: 15/05/2018
 *
 * Descrio:
 *
 * Entrada: Um arquivo contendo registros sem ordenao
 *
 * Sada: O arquivo ordenado
 *
 ****/

/************************ Includes ************************/

#include <stdio.h>        /* Entrada e sada     */
#include <string.h>       /* strcpy() e strchr() */
#include <stdlib.h>       /* qsort()             */
#include "IntercalaBin.h" /* Interface deste mdulo */
#include "Registros.h" /* Funes de processamento         */
                       /* de registros                     */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/****************** Constantes Simblicas *****************/

   /* Tamanho de um registro  */
#define TAM_REG sizeof(tRegistro)

/********************* Funes Locais *********************/

/****
 *
 * Compara(): Compara dois registros
 *
 * Parmetros:
 *      e1, e2 (entrada) - os dois registros
 *
 * Retorno:
 *     * 0, se os registros so iguais
 *     * > 0, se o primeiro registro for maior do que o segundo
 *     * < 0, se o primeiro registro for menor do que o segundo
 *
 ****/
static int Compara(const void *e1, const void* e2)
{
   tRegistro *r1 = (tRegistro *)e1,
             *r2 = (tRegistro *)e2;

   return r1->chave - r2->chave;
}

/****
 *
 * CopiaRestoArquivo(): Copia contedo de um arquivo para outro byte
 *                      a byte a partir da posio em que se encontra
 *                      o apontador de posio de cada arquivo
 *
 * Parmetros:
 *      streamE (entrada) - stream associado ao arquivo que ser
 *                          copiado
 *      streamS (entrada) - stream associado ao arquivo que receber
 *                          a cpia
 *      maxBytes (entrada) - nmero mximo de bytes que sero
 *                           copiados; se esse valor for zero,
 *                           no haver limite para o nmero de
 *                           bytes copiados
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. O stream que ser lido deve estar aberto num modo
 *        que permita leitura e o stream que ser escrito
 *        deve estar aberto num modo que permite escrita.
 *
 *     2. O apontador de posio do arquivo lido no  movido
 *        para o incio do arquivo, de modo que o primeiro byte
 *        a ser copiado  aquele indicado pela posio corrente
 *        desse apontador quando esta funo  chamada.
 *
 ****/
static void CopiaRestoArquivo( FILE *streamE, FILE *streamS,
                               long maxBytes )
{
   int  c; /* Armazenar cada byte lido e escrito  */
   long contaBytes = 0;

      /* O lao a seguir encerra quando:        */
      /*   * Ocorrer tentativa de leitura alm  */
      /*     do final do arquivo de entrada     */
      /*   * Ocorrer erro de leitura ou escrita */
      /*     em qualquer dos arquivos           */
      /*   * O nmero mximo de bytes tiver     */
      /*     sido copiado                       */
   while (1) {
         /* L um byte no arquivo de entrada */
      c = fgetc(streamE);

         /* Testa se final do arquivo de entrada foi */
         /* atingido ou ocorreu erro de leitura      */
      if (feof(streamE) || ferror(streamE))
         break; /* Processamento encerrado */

         /* Escreve o byte lido no arquivo de sada */
      fputc(c, streamS);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamS))
         break;

      ++contaBytes; /* Mais um byte foi copiado */

         /* Verifica se o nmero mximo de bytes j foi copidado */
      if (maxBytes && contaBytes >= maxBytes)
         break;
   }

      /* Certifica que no houve erro de leitura ou escrita */
   ASSEGURA(!ferror(streamE),
            "Erro de leitura ao tentar copiar arquivo");
   ASSEGURA(!ferror(streamS),
            "Erro de escrita ao tentar copiar arquivo");
}

/****
 *
 * IntercalaDuasSeries(): Intercala duas sries armazenadas num
 *                        arquivo e armazena o resultado noutro
 *                        arquivo
 *
 * Parmetros:
 *      streamE (entrada) - stream associado ao arquivo que contm
 *                          as sries
 *      streamS (entrada) - stream associado ao arquivo que conter
 *                          o resultado da intercalao
 *      is1 (entrada) - ndice da primeira srie
 *      is2 (entrada) - ndice da segunda srie
 *      tamSerie (entrada) - tamanho mximo de uma srie
 *
 * Retorno: Nada
 *
 * Observao: O ndice da primeira srie deve ser menor do que o
 *             ndice da segunda srie
 *
 ****/
static void IntercalaDuasSeries( FILE *streamE, FILE *streamS,
                                 long is1, long is2, int tamSerie )
{
   tRegistro reg1, /* Um registro da primeira srie */
             reg2; /* Um registro da segunda srie  */
   int       ir1 = 0, /* ndice do prximo registro */
                      /* a ser lido na srie 1      */
             ir2 = 0, /* ndice do prximo registro */
                      /* a ser lido na srie 2      */
             le1 = 1, /* Indica se um registro deve */
                      /* ser lido na srie 1        */
             le2 = 1; /* Indica se um registro deve */
                      /* ser lido na srie 2        */
   long      posicao, /* Posio de leitura de um registro */
             bytesRestantes; /* Nmero de bytes de uma srie */
                             /* que precisam ser copiados    */

      /* A primeira srie deve preceder a segunda srie */
   ASSEGURA(is1 < is2,
            "Indice da 1a serie e' maior do que indice da 2a");

      /* Intercala as sries */
   while (1) {
         /* Se necessrio, l um registro da primeira srie */
      if (le1) {
            /* Move o apontador de posio de  */
            /* arquivo para o local de leitura */
         posicao = (is1*tamSerie + ir1)*TAM_REG;
         MoveApontador(streamE, posicao, SEEK_SET);

            /* L um registro da primeira srie */
            /* e checa se ocorreu erro          */
         fread(&reg1, TAM_REG, 1, streamE);
         ASSEGURA( !ferror(streamE), "Erro de leitura" );

         ++ir1; /* Passa para o prximo registro desta srie */
      }

         /* Se necessrio, l um registro da segunda srie */
      if (le2) {
            /* Move o apontador de posio de  */
            /* arquivo para o local de leitura */
         posicao = (is2*tamSerie + ir2)*TAM_REG;
         MoveApontador(streamE, posicao, SEEK_SET);

            /* L um registro da segunda srie */
            /* e checa se ocorreu erro         */
         fread(&reg2, TAM_REG, 1, streamE);
         ASSEGURA( !ferror(streamE), "Erro de leitura" );

         ++ir2; /* Passa para o prximo registro desta srie */
      }

         /* Se a primeira srie foi exaurida, copia os */
         /* registros restantes da segunda srie para  */
         /* o arquivo resultante e encerra a operao  */
      if (ir1 > tamSerie) {
            /* Desfaz o incremento para levar em conta */
            /* ltimo registro lido na srie 2        */
         --ir2;

            /* Move o apontador de posio de arquivo  */
            /* para o local onde a cpia ser iniciada */
         posicao = (is2*tamSerie + ir2)*TAM_REG;
         MoveApontador(streamE, posicao, SEEK_SET);

            /* Calcula o nmero mximo de bytes que */
            /* sero copiados e efetua a cpia      */
         bytesRestantes = (tamSerie - ir2)*TAM_REG;
         CopiaRestoArquivo(streamE, streamS, bytesRestantes);

         return; /* Servio completo */
      }

         /* Se a segunda srie foi exaurida, copia os  */
         /* registros restantes da primeira srie para */
         /* o arquivo resultante e encerra a operao  */
      if (ir2 > tamSerie || feof(streamE)) {
            /* Desfaz o incremento para levar em conta */
            /* ltimo registro lido na srie 1        */
         --ir1;

            /* Move o apontador de posio de arquivo  */
            /* para o local onde a cpia ser iniciada */
         posicao = (is1*tamSerie + ir1)*TAM_REG;
         MoveApontador(streamE, posicao, SEEK_SET);

            /* Calcula o nmero mximo de bytes que */
            /* sero copiados e efetua a cpia      */
         bytesRestantes = (tamSerie - ir1)*TAM_REG;
         CopiaRestoArquivo(streamE, streamS, bytesRestantes);

         return; /* Servio completo */
      }

         /* Copia o menor dos dois registros para */
         /* o arquivo resultante da intercalao  */
      if (Compara(&reg1, &reg2) < 0) {
         fwrite(&reg1, TAM_REG, 1, streamS);
         le1 = 1; /* L o prximo registro da srie 1 */
         le2 = 0; /* No l o prximo registro da srie 2 */
      } else {
         fwrite(&reg2, TAM_REG, 1, streamS);
         le1 = 0; /* No l o prximo registro da srie 1 */
         le2 = 1; /* L o prximo registro da srie 2 */
      }
   }
}

/********************* Funes Globais ********************/

/****
 *
 * CriaSeries(): Cria sries (ordenadas)
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que origina as sries
 *      nomeS (entrada) - nome do arquivo que armazenar as sries
 *
 * Retorno: Nmero de sries criadas
 *
 ****/
int CriaSeries(const char *nomeE, const char *nomeS)
{
   tRegistro *registros; /* Ponteiro para um array que armazena */
                         /* uma srie em memria principal      */
   int        nRegistros,  /* Nmero de registros numa srie */
              nSeries = 0; /* Nmero de sries */
   FILE      *streamS, /* Stream associado ao arquivo */
                       /* que armazenar as sries    */
             *streamE; /* Stream associado ao arquivo original */

      /* Aloca espao para o array. Um array deste */
      /* tamanho no pode ser alocado na pilha     */
   registros = calloc(MAX_REGISTROS, TAM_REG);
   ASSEGURA( registros, "Buffer nao foi alocado" );

      /* Tenta criar o arquivo que conter as sries */
   streamS = AbreArquivo(nomeS, "wb");

      /* Tenta abrir o arquivo original */
   streamE = AbreArquivo(NOME_ARQUIVO_BIN, "rb");

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os registros da srie corrente */
      nRegistros = fread( registros, TAM_REG, MAX_REGISTROS,
                          streamE );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(streamE), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum registro foi lido */
      if (!nRegistros)
         break;

         /* Ordena os registros lidos */
      qsort(registros, nRegistros, TAM_REG, Compara);

         /* Escreve a srie corrente no arquivo */
      fwrite(registros, TAM_REG, nRegistros, streamS);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(streamS),
                "Erro de escrita em arquivo de series" );

      ++nSeries; /* Atualiza o nmero de sries */
   }

      /* Fecha os arquivos */
   FechaArquivo(streamS, nomeS);
   FechaArquivo(streamE, nomeE);

      /* Libera o espao ocupado pelo array */
   free(registros);

   return nSeries;
}

/****
 *
 * IntercalaBin(): Intercala sries armazenadas num arquivo
 *
 * Parmetros:
 *      nomeE (entrada) - nome do arquivo que inicialmente contm
 *                        as sries que sero intercaladas
 *      nomeS (entrada) - nome do arquivo que conter o resultado
 *                        de intercalao das sries do primeiro
 *                        arquivo
 *      nSeries (entrada) - nmero inicial de sries
 *      tamSerie (entrada) - tamanho (nmero de registros) mximo
 *                           de cada srie inicial
 *
 * Retorno: O nome do arquivo que contm o resultado
 *
 * Observao: Os nomes associados aos arquivos de entrada e sada
 *             mudam durante a execuo desta funo (i.e., eles
 *             trocam de papis). Assim ao final da operao, o
 *             resultado pode estar armazenado no arquivo
 *             representado pelo primeiro ou pelo segundo parmetro.
 *             Essa  a razo pela qual o nome do arquivo que contm
 *             o resultado  retornado. Mas  fcil alterar essa
 *             funo de modo que o resultado seja sempre armazenado
 *             no segundo parmetro.
 *
 ****/
const char *IntercalaBin( const char *nomeE, const char *nomeS,
                          int nSeries, int tamSerie )
{
   FILE       *streamS, /* Stream associado ao arquivo que contm */
                        /* o resultado de concatenao de sries  */
              *streamE; /* Stream associado ao arquivo que contm */
                        /* sries que sero concatenadas          */
   int         ns = nSeries, /* Nmero corrente de sries */
               ts = tamSerie; /* Tamanho mximo de cada srie */
   int         i1, i2; /* ndices de duas sries */
                       /* que sero intercaladas */
   const char *aux;

   while (ns > 1) {
         /* Tenta abrir o arquivo contendo   */
         /* as sries que sero intercaladas */
      streamE = AbreArquivo(nomeE, "rb");

         /* Tenta criar o arquivo que conter as sries resultantes */
      streamS = AbreArquivo(nomeS, "wb");

#if DEPURACAO == 1
      printf("\a\n\t>>> Intercalando %d series (ts = %d)...", ns, ts);
      fflush(stdout);
#endif

      i1 = 0; /* ndice inicial da primeira srie */
      i2 = 1; /* ndice inicial da segunda srie  */

         /* As sries so indexadas de 0 a ns - 1. */
         /* Portanto o maior valor de i2  ns - 1. */
      while (i2 < ns) {
            /* Intercala as duas sries correntes */
         IntercalaDuasSeries(streamE, streamS, i1, i2, ts);

            /* Passa para as prximas sries */
         i1 += 2;
         i2 += 2;
      }

         /* Se o nmero corrente de sries for mpar, resta */
         /* copiar a ltima srie para o arquivo resultante */
      if (i1 < ns) {
            /* Verificao de consistncia */
         ASSEGURA(ns%2, "O numero de series deveria ser impar");
         ASSEGURA(i1 == ns - 1, "Deveria ser a ultima serie");

            /* Move apontador para o primeiro byte a ser copiado */
            /* e copia a ltima srie para o arquivo resultante  */
         MoveApontador( streamE, (long)i1*ts*TAM_REG, SEEK_SET );
         CopiaRestoArquivo(streamE, streamS, 0);
      }

         /* Fecha os arquivos */
      FechaArquivo(streamE, nomeE);
      FechaArquivo(streamS, nomeS);

         /* O nmero de sries foi reduzido  metade */
      ns = ns/2 + ns%2;

         /* O nmero mximo de registros de cada srie */
         /* dobra de tamanho, sendo que o nmero de    */
         /* registros da ltima srie pode ser menor   */
      ts *= 2;

         /* Troca os papis dos arquivos */
      aux = nomeE;
      nomeE = nomeS;
      nomeS = aux;
   }

      /* Retorna o nome do arquivo que contm o resultado */
   return nomeE;
}
