/****
 *
 * Ttulo: Ordenao externa de arquivos por intercalao binria
 *
 * Seo: 12.2.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/04/2017
 * ltima modificao: 16/04/2017
 *
 * Descrio: Demonstra ordenao externa de arquivos por
 *            intercalao binria
 *
 * Entrada: Arquivo "CensoMEC2012.bin"
 *
 * Sada: Arquivo "CensoMEC2012_Ord.bin"
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>        /* Funes de Entrada/Sada */
#include <string.h>       /* Processamento de strings */
#include <stdlib.h>       /* Funes de alocao dinmica e */
                          /* gerao de nmeros aleatrios  */
#include <time.h>         /* Funo time() para alimentar o */
                          /* gerador de nmeros aleatrios  */
#include "IntercalaBin.h" /* Interface do mdulo de ordenao */
#include "Leitura.h"   /* Funes de leitura via teclado   */
#include "Registros.h" /* Funes de processamento de registros */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 1

/***************** Definies de Funes ******************/

/****
 *
 * CriaArquivoRegistros(): Constri o arquivo binrio de registros
 *
 * Parmetros:
 *     nomeArqTexto (entrada) - nome do arquivo texto de registros
 *     nomeArqBin (entrada) - nome do arquivo binrio de registros
 *
 * Retorno: Stream associado ao arquivo binrio de registros
 *
 ****/
void CriaArquivoRegistros( const char *nomeArqTexto,
                           const char *nomeArqBin )
{
   int nRegistros, /* Nmero de registros no arquivo */
       maxRegistros; /* Nmero mximo de registros no arquivo */

   printf("\n\t>>> Digite o numero maximo de registros (0 = TODOS): ");
   maxRegistros = LeNatural();

   printf("\nCriando arquivo...\n");
   fflush(stdout);

      /* Cria o arquivo binrio */
   nRegistros = CriaArquivoBin( nomeArqTexto, nomeArqBin,
                                maxRegistros );

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, nomeArqBin );
}

/****
 *
 * TestaOrdenacao(): Verifica se um arquivo est ordenado em ordem
 *                   crescente
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo que ser
 *                         testado
 *
 * Retorno: 1, se o arquivo estiver ordenado; 0, em caso contrrio
 *
 ****/
int TestaOrdenacao(const char *nomeArq)
{
   tRegistro registroAtual, registroAnterior;
   int       teste;
   FILE     *stream;

   stream = AbreArquivo(nomeArq, "rb");

      /* L o primeiro registro do arquivo */
   fread(&registroAnterior, sizeof(registroAnterior), 1, stream);
   ASSEGURA( !ferror(stream), "Erro de leitura" );

      /* L os demais registros e testa se eles esto ordenados */
   while (1) {
         /* L o prximo registro */
      fread(&registroAtual, sizeof(registroAtual), 1, stream);
      ASSEGURA( !ferror(stream), "Erro de leitura" );

      if (feof(stream))
         break; /* Leitura acabou */

         /* Compara a chave do registro atual */
         /* com a chave do registro anterior  */
      teste = registroAtual.chave - registroAnterior.chave;

         /* Verifica se as duas ltimas chaves */
         /* esto em ordem crescente           */
      if (teste < 0) {
         FechaArquivo(stream, nomeArq);
         return 0;
      }

         /* A chave anterior passa a ser a chave corrente */
      memcpy( &registroAnterior, &registroAtual,
              sizeof(registroAtual) );
   }

   FechaArquivo(stream, nomeArq);

      /* Se ainda no houve retorno, o arquivo est ordenado */
   return 1;
}

/****
 *
 * main(): Efetua ordenao externa de arquivos usando intercalao
 *         binria
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   int         nSeries, teste;
   const char *nomeArqOrdenado;

      /* Cria o arquivo binrio de registro se ele no for encontrado */
   if (!ArquivoExiste(NOME_ARQUIVO_BIN)) {
      MedidaDeTempo("Criando arquivo de registros...");

      CriaArquivoRegistros( NOME_ARQUIVO_REGS, NOME_ARQUIVO_BIN );

      MedidaDeTempo("...Arquivo criado");
   }

   MedidaDeTempo("Criando series ordenadas...");

      /* L o arquivo original e cria as sries */
   nSeries = CriaSeries(NOME_ARQUIVO_BIN, NOME_ARQUIVO_SERIES);

   MedidaDeTempo("...series criadas com sucesso");

   printf("\n\t>>> Numero de series: %d\n", nSeries);

   MedidaDeTempo("Intercalando series...");

   nomeArqOrdenado = IntercalaBin( NOME_ARQUIVO_SERIES,
                                   NOME_ARQUIVO_ORD, nSeries,
                                   MAX_REGISTROS );

   MedidaDeTempo("...series intercaladas com sucesso");

   printf( "\n\t>>> Nome do arquivo ordenado: \"%s\"\n",
           nomeArqOrdenado );

   MedidaDeTempo("Testando ordenacao do arquivo...");

   teste = TestaOrdenacao(nomeArqOrdenado);

   if (teste)
      MedidaDeTempo("...Ordenacao esta' OK\n");
   else
      MedidaDeTempo("\a...O arquivo nao esta ordenado\n");

   return 0;
}

/******************* Exemplos de Execuo ******************

>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 600000

Criando arquivo...

        >>> Foram armazenados 600000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 16.00 segundos ***


>>> Criando series ordenadas...


>>> ...series criadas com sucesso

        *** Tempo gasto na operacao:  3.00 segundos ***

        >>> Numero de series: 8


>>> Intercalando series...

        >>> Intercalando 8 series (ts = 80000)...
        >>> Intercalando 4 series (ts = 160000)...
        >>> Intercalando 2 series (ts = 320000)...

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao: 33.00 segundos ***

        >>> Nome do arquivo ordenado: "../CensoMEC2012_Ord.bin"


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao:  3.00 segundos ***



>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 1200000

Criando arquivo...

        >>> Foram armazenados 1200000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 32.00 segundos ***


>>> Criando series ordenadas...


>>> ...series criadas com sucesso

        *** Tempo gasto na operacao:  8.00 segundos ***

        >>> Numero de series: 15


>>> Intercalando series...

        >>> Intercalando 15 series (ts = 80000)...
        >>> Intercalando 8 series (ts = 160000)...
        >>> Intercalando 4 series (ts = 320000)...
        >>> Intercalando 2 series (ts = 640000)...

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao:  1.48 minutos ***

        >>> Nome do arquivo ordenado: "../CensoMEC2012_TMP.bin"


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao:  6.00 segundos ***


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 2400000

Criando arquivo...

        >>> Foram armazenados 2400000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao: 49.00 segundos ***


>>> Criando series ordenadas...


>>> ...series criadas com sucesso

        *** Tempo gasto na operacao: 14.00 segundos ***

        >>> Numero de series: 30


>>> Intercalando series...

        >>> Intercalando 30 series (ts = 80000)...
        >>> Intercalando 15 series (ts = 160000)...
        >>> Intercalando 8 series (ts = 320000)...
        >>> Intercalando 4 series (ts = 640000)...
        >>> Intercalando 2 series (ts = 1280000)...

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao:  3.58 minutos ***

        >>> Nome do arquivo ordenado: "../CensoMEC2012_Ord.bin"


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 13.00 segundos ***


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 4800000

Criando arquivo...

        >>> Foram armazenados 4800000 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao:  1.32 minutos ***


>>> Criando series ordenadas...


>>> ...series criadas com sucesso

        *** Tempo gasto na operacao: 28.00 segundos ***

        >>> Numero de series: 60


>>> Intercalando series...

        >>> Intercalando 60 series (ts = 80000)...
        >>> Intercalando 30 series (ts = 160000)...
        >>> Intercalando 15 series (ts = 320000)...
        >>> Intercalando 8 series (ts = 640000)...
        >>> Intercalando 4 series (ts = 1280000)...
        >>> Intercalando 2 series (ts = 2560000)...

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao:  8.90 minutos ***

        >>> Nome do arquivo ordenado: "../CensoMEC2012_TMP.bin"


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 25.00 segundos ***


>>> Criando arquivo de registros...

        >>> Digite o numero maximo de registros (0 = TODOS): 0

Criando arquivo...

        >>> Foram armazenados 9565483 registros no arquivo
        >>> "../CensoMEC2012.bin"


>>> ...Arquivo criado

        *** Tempo gasto na operacao:  2.45 minutos ***


>>> Criando series ordenadas...


>>> ...series criadas com sucesso

        *** Tempo gasto na operacao: 58.00 segundos ***

        >>> Numero de series: 120


>>> Intercalando series...

        >>> Intercalando 120 series (ts = 80000)...
        >>> Intercalando 60 series (ts = 160000)...
        >>> Intercalando 30 series (ts = 320000)...
        >>> Intercalando 15 series (ts = 640000)...
        >>> Intercalando 8 series (ts = 1280000)...
        >>> Intercalando 4 series (ts = 2560000)...
        >>> Intercalando 2 series (ts = 5120000)...

>>> ...series intercaladas com sucesso

        *** Tempo gasto na operacao: 20.28 minutos ***

        >>> Nome do arquivo ordenado: "../CensoMEC2012_Ord.bin"


>>> Testando ordenacao do arquivo...


>>> ...Ordenacao esta' OK


        *** Tempo gasto na operacao: 48.00 segundos ***

**********************************************************/
