/****
 *
 * Arquivo Registros.c: Arquivo de programa do mdulo de
 *                      processamento de registros
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 20/04/2004
 * ltima modificao: 10/05/2018
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada e Sada   */
#include <string.h> /* Processamento de strings     */
#include <stdlib.h> /* strtod() e alocao dinmica */
#include <time.h>  /* Medio de tempo              */
#include "Registros.h" /* Interface deste mdulo    */
#include "Leitura.h"  /* Leitura via teclado        */
#include "ASSEGURA.h" /* Macro ASSEGURA             */

/***************** Constantes Simblicas ******************/

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

/********************* Funes Locais *********************/

/****
 *
 * EscreveArrayDeChars(): Escreve num stream um array de elementos
 *                        do tipo char que no  string
 *
 * Parmetros:
 *      *nome (entrada) - string que ser exibido antes do array
 *       ar (entrada) - endereo do array
 *       tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 ****/
static void EscreveArrayDeChars( FILE *stream, const char *nome,
                                 const char *ar, int tam )
{
   fputc('\n', stream);

   if (*nome)
      fprintf(stream, "%s: ", nome);

   for (int i = 0; i < tam && *ar; ++i)
      fputc(ar[i], stream);
}

/****
 *
 * ExibeArrayDeChars(): Exibe na tela um array de elementos
 *                      do tipo char que no  string
 *
 * Parmetros:
 *      *nome (entrada) - string que ser exibido antes do array
 *       ar (entrada) - endereo do array
 *       tam (entrada) - tamanho do array
 *
 * Retorno: Nada
 *
 ****/
static void ExibeArrayDeChars(const char *nome, const char *ar,
                              int tam)
{
   putchar('\n');

   if (*nome) {
      printf("%s: ", nome);
   }

   for (int i = 0; i < tam && *ar; ++i) {
      putchar(ar[i]);
   }
}

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
static char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
*
* ObtemTokens(): Separa um string em partes (tokens)
*
* Parmetros:
*     str (entrada/sada) - o string que ser separado
*                           ou NULL para continuar uma
*                           separao anterior
*     sep (entrada) - string contendo os separadores de tokens
*
* Retorno: Endereo do prximo token do string, se este
*          existir, ou NULL, em caso contrrio
*
* Observao: Esta funo NO  funcionalmente equivalente 
*             funo strtok() da biblioteca padro de C
*
****/
static char *ObtemTokens(char *str, char const *sep)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, /* Apontar para o string no qual a */
                   /* busca pelo token ser efetuada   */
               *inicio; /* Guardar o incio */
                        /* do token corrente */

   if (str != NULL)
      proximoToken = str;

   if (proximoToken == NULL)
      return NULL;

   if ((s = strpbrk(proximoToken, sep)) != NULL) {
      *s  = '\0';
      inicio = proximoToken;
      proximoToken = ++s;
   } else if (*proximoToken) {
      inicio = proximoToken;
      proximoToken = NULL;
   }

   return inicio;
}

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa
 *                    estrutura do tipo tAluno.
 *
 * Parmetros: aluno (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por ObtemTokens() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por ObtemTokens().
 *
 *     3. A funo ObtemTokens()  usada em lugar de strtok()
 *        devido  presena de campos vazios no arquivo de
 *        texto, o que faz com que dois separadores apaream
 *        juntos. Nesse caso, a funo strtok() consideraria
 *        os dois separadores como sendo apenas um.
 *
 *     4. O caractere '\r' faz parte das chamadas da funo
 *        ObtemTokens() para satisfazer uma bizarrice do
 *        sistema Mac OS X, que  o nico sistema Unix que
 *        usa '\r' como quebra de linha.
 *
 ****/
static tAluno *LinhaEmRegistro(tAluno *aluno, char *linha)
{
   char *str, /* Apontar para cada token da linha */
        *final; /* Final de uma converso numrica */

      /* Obtm o cdigo do aluno, converte-o e         */
      /* acrescenta-o ao respectivo campo da estrutura */
   str = ObtemTokens(linha, "\t\n\r;");
   ++str; /* Salta espao em branco inicial */
   aluno->codigoAluno = strtoll(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm o cdigo de vnculo do aluno ao */
      /* curso, converte-o e acrescenta-o ao   */
      /* respectivo campo da estrutura         */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->chave = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm o nome da IES e acrescenta-a */
      /* ao respectivo campo da estrutura   */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->nomeIES, str, MAX_NOME_IES);

      /* Obtm o nome do curso e acrescenta-o */
      /* ao respectivo campo da estrutura     */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->nomeCurso, str, MAX_NOME_CURSO);

      /* Obtm o sexo do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura     */
   str = ObtemTokens(NULL, "\t\n\r;");
   ASSEGURA(*str == '0' || *str == '1', "Sexo indefinido");
   aluno->sexoAluno = *str == '0' ? 0 : 1;

      /* Obtm a idade do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura      */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->idadeAluno = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm a UF de nascimento e acrescenta-a */
      /* ao respectivo campo da estrutura        */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->UFNascimento, str, MAX_UF_NASCIMENTO);

      /* Obtm o ano de ingresso do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura                */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->anoIngresso = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

   return aluno;
}

/********************* Funes Globais ********************/

/****
 *
 * AbreArquivo(): Tenta abrir um arquivo e, quando isso
 *                no  possvel, aborta o programa
 *
 * Parmetros:
 *      nome (entrada) - nome do arquivo
 *      modo (entrada) - modo de abertura
 *
 * Retorno: Stream associado ao arquivo que foi aberto
 *
 ****/
FILE *AbreArquivo(const char *nome, const char *modo)
{
   FILE *stream;

   stream = fopen(nome, modo);

   if (!stream) { /* Erro de abertura */
      fprintf( stderr, "\n>>> O arquivo \"%s\" nao pode ser aberto\n",
               nome );
      exit(1); /* Aborta o programa */
   }

   return stream;
}

/****
 *
 * FechaArquivo(): Tenta fechar um arquivo e, quando isso
 *                 no  possvel, aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *
 * Retorno: Nada
 *
 * Observao: Se o segundo parmetro no for NULL, o nome
 *             do arquivo aparece na mensagem de erro
 *
 ****/
void FechaArquivo(FILE *stream, const char *nomeArq)
{
   /***********************************************************/
   /* Se fclose() retornar um valor diferente de zero ocorreu */
   /* algum erro na tentativa de fechamento do arquivo. Nesse */
   /* caso, apresenta mensagem de erro e aborta o programa.   */
   /***********************************************************/

   if (fclose(stream)) { /* Erro de fechamento */
      fprintf( stderr, "\a\n>>> Ocorreu erro no fechamento "
              "do arquivo %s.\n>>> O programa sera' "
              "encerrado.\n", nomeArq ? nomeArq : "");
      exit(1); /* Aborta o programa */
   }
}

/****
 *
 * MoveApontador(): Tenta mover o apontador de posio de um
 *                  arquivo e, quando isso no  possvel,
 *                  aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      bytes (entrada) - quantidade de bytes que sero saltados
 *      DeOnde (entrada) - a partir de onde ocorrer a movimentao
 *
 * Retorno: Nada
 *
 ****/
void MoveApontador(FILE *stream, long bytes, int DeOnde)
{
   int deuErrado;

   deuErrado = fseek(stream, bytes, DeOnde);

      /* Verifica se houve erro */
   if (deuErrado) {
      fprintf( stderr, "\n>>> Erro na tentativa de mover "
                       "apontador de arquivo" );
      exit(1);
   }
}

/****
 *
 * ExibeRegistro(): Exibe na tela um registro do tipo tRegistro
 *
 * Parmetros: *registro (entrada) - registro que ser exibido
 *
 * Retorno: Nada
 *
 ****/
void ExibeRegistro(const tRegistro *registro)
{
   printf("Codigo do aluno: %ld", registro->codigoAluno);

   printf("\nCodigo do curso: %ld", registro->chave);

   ExibeArrayDeChars( "Nome da IES", registro->nomeIES,
                      MAX_NOME_IES );

   ExibeArrayDeChars("Nome do curso", registro->nomeCurso,
                     MAX_NOME_CURSO);

   printf("\nSexo do aluno: %s",
          registro->sexoAluno ? "Masculino" : "Feminino" );

   printf("\nIdade do aluno: %d", registro->idadeAluno);

   ExibeArrayDeChars("UF de nascimento",
                     registro->UFNascimento, MAX_UF_NASCIMENTO);

   printf("\nAno de ingresso: %d", registro->anoIngresso);
}

/****
 *
 * TamanhoDeArquivo(): Calcula o tamanho de um arquivo lendo cada
 *                     byte que o constitui e contando quantos
 *                     bytes so lidos
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *
 * Retorno: -1, se ocorrer algum erro.
 *          O tamanho do arquivo, em caso contrrio.
 *
 * Observaes:
 *     O arquivo deve estar aberto em modo binrio que permite
 *     leitura
 *
 ****/
long TamanhoDeArquivo(FILE *stream)
{
   ASSEGURA(stream, "Erro em TamanhoDeArquivo(): stream e' NULL" );

      /* Tenta mover o apontador de posio do arquivo */
      /* para o final do mesmo; se no for possvel,   */
      /* o programa ser abortado                      */
   MoveApontador(stream, 0, SEEK_END);

   return ftell(stream);
}

/****
 *
 * NumeroDeRegistros(): Calcula o nmero de registros (i.e.,
 *                      parties de tamanho especificado) de um
 *                      arquivo binrio
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *    tamRegistro (entrada) - tamanho de cada partio do arquivo
 *
 * Retorno: O nmero de parties do arquivo, se no ocorrer
 *          nenhum erro, ou -1, em caso contrrio.
 *
 * Observao: O arquivo deve estar aberto em modo binrio que
 *             permite leitura
 *
 ****/
int NumeroDeRegistros( FILE *stream, int tamRegistro )
{
   long nBytes = 0; /* Nmero de bytes do arquivo */

      /* Calcula o nmero de bytes do arquivo */
   nBytes = TamanhoDeArquivo(stream);

      /* Verifica se ocorreu erro na contagem de bytes */
   ASSEGURA(nBytes >= 0, "Erro na contagem de bytes");

      /* O nmero de registros  o nmero de bytes do arquivo */
      /* dividido pelo nmero de bytes de cada registro       */
   return nBytes/tamRegistro;
}

/****
 *
 * LeRegistroEmArquivo(): L um registro no arquivo de registros
 *
 * Parmetros:
 *     *umRegistro (sada) - registro lido
 *      stream (entrada) - stream associado ao arquivo
 *                         de registros
 *      pos (entrada) - posio do registro no arquivo
 *
 * Retorno: o primeiro parmetro, se a recuperao for
 *           bem sucedida; NULL, em caso contrrio.
 *
 ****/
tRegistro *LeRegistroEmArquivo( tRegistro *umRegistro,
                                FILE *stream, int pos )
{
   ASSEGURA(stream, "Arquivo de registros eh nulo");

   MoveApontador(stream, pos*sizeof(*umRegistro), SEEK_SET);

   return fread(umRegistro, sizeof(*umRegistro), 1, stream)
          ? umRegistro
          : NULL;
}

/****
 *
 * ArquivoExiste(): Verifica se um arquivo existe
 *
 * Parmetros:
 *     nomeArq (entrada) - nome do arquivo que se
 *                        verificar se existe
 *
 * Retorno: 1, se o arquivo existir; 0, em caso contrrio
 *
 ****/
int ArquivoExiste(const char *nomeArq)
{
   FILE *stream;
   int   existe;

   stream = fopen(nomeArq, "rb");

   existe = stream ? 1 : 0;

   if (existe) {
      FechaArquivo(stream, nomeArq);
   }

   return existe;
}

/****
 *
 * CriaArquivoBin(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura
 *                   do tipo tAluno e armazena a estrutura
 *                   num arquivo binrio
 *
 * Parmetros:
 *     streamTexto (entrada) - stream associado ao arquivo que
 *                             ser lido
 *     streamBin (entrada) - stream associado ao arquivo que ser
 *                           escrito
 *     maxRegistros (entrada) - nmero mximo de registros que
 *                              sero gerados
 *
 * Retorno: Nmero de estruturas escritas no arquivo
 *          binrio ou um valor negativo, se ocorrer erro
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de
 *        texto que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoBin( const char *nomeTexto, const char *nomeBin,
                    int maxRegistros )
{
   char  *linha; /* Apontar para cada linha lida */
   tAluno umAluno; /* Dados de uma linha      */
                   /* convertida em estrutura */
   int   nRegistros = 0, /* Nmero de estruturas escritas */
                         /* no arquivo binrio            */
         ocorreuErro = 0;
   FILE *streamTexto,
        *streamBin;

      /* Tenta abrir arquivo para leitura em modo texto */
   streamTexto = AbreArquivo(nomeTexto, "r");

      /* Tenta abrir arquivo para escrita em modo binrio */
   streamBin = AbreArquivo(nomeBin, "w");

      /* L cada linha do arquivo, cria um registro do tipo     */
      /* tAluno e armazena-o no arquivo binrio. O lao encerra */
      /* quando h tentativa de leitura alm do final do        */
      /* arquivo de entrada, ou ocorre erro de leitura/escrita  */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, streamTexto);

         /* Verifica se o final do arquivo foi  */
         /* atingido ou ocorreu erro de leitura */
      if (feof(streamTexto) || ferror(streamTexto))
         break;

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umAluno, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umAluno, sizeof(umAluno), 1, streamBin);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamBin))
         break;

      ++nRegistros; /* Mais um registro lido */

      if (maxRegistros && nRegistros >= maxRegistros)
         break;
   }

   ocorreuErro = ferror(streamTexto) || ferror(streamBin);

      /* Fecha os arquivos */
   FechaArquivo(streamTexto, nomeTexto);
   FechaArquivo(streamBin, nomeBin);

   ASSEGURA(!ocorreuErro,
            "Erro de processamento de arquivo em CriaArquivoBin");

   return nRegistros; /* No ocorreu nenhum erro */
}

/****
 *
 * EscreveRegistro(): Escreve num stream um registro do tipo tRegistro
 *
 * Parmetros: *registro (entrada) - registro que ser exibido
 *
 * Retorno: Nada
 *
 ****/
void EscreveRegistro(const tRegistro *registro, FILE *stream)
{
   fprintf(stream, "Codigo do aluno: %ld", registro->codigoAluno);

   fprintf(stream, "\nCodigo do curso: %ld", registro->chave);

   EscreveArrayDeChars(stream, "Nome da IES", registro->nomeIES,
                       MAX_NOME_IES);

   EscreveArrayDeChars(stream, "Nome do curso", registro->nomeCurso,
                       MAX_NOME_CURSO);

   fprintf(stream, "\nSexo do aluno: %s",
          registro->sexoAluno ? "Masculino" : "Feminino");

   fprintf(stream, "\nIdade do aluno: %d", registro->idadeAluno);

   EscreveArrayDeChars(stream, "UF de nascimento",
                       registro->UFNascimento, MAX_UF_NASCIMENTO);

   fprintf(stream, "\nAno de ingresso: %d", registro->anoIngresso);
}

/****
 *
 * NomeTemp(): Cria nomes de arquivos temporrios
 *
 * Parmetros:
 *      nome[] (sada) - array que armazenar o nome de arquivo
 *                       criado
 *      inicio (entrada) - string que inicia cada nome de arquivo
 *      max (entrada) - nmero mximo de nomes de arquivos gerados
 *
 * Retorno: O endereo do nome (string) criado.
 *          NULL, se a operao no for possvel.
 *
 * Observao: Quando o primeiro parmetro  NULL, o contador de
 *             arquivos  zerado. O mesmo ocorre quando o nmero
 *             mximo de arquivos  atingido. Isto permite que
 *             vrias sequncias de arquivos possam ser criadas.
 *
 ****/
char *NomeTemp(char *nome, const char *inicio, int max)
{
   static int ordem = 0; /* Nmero de ordem do arquivo */

      /* Quando o primeiro parmetro  NULL, */
      /* o contador de arquivos  zerado     */
   if (!nome) {
      ordem = 0; /* Reinicia contador de arquivos */
      return NULL;
   }

   ++ordem;	/* Um novo nome de arquivo ser criado */

      /* Verifica se o nmero mximo de */
      /* arquivos gerados foi atingido  */
   if (ordem > max) {
      ordem = 0; /* Reinicia contador de arquivos */
      return NULL;
   }

      /* Copia o incio do nome do arquivo temporrio */
      /* para o parmetro que conter o resultado     */
   strcpy(nome, inicio);

      /* Insere o nmero de ordem do arquivo a */
      /* partir da posio do caractere '\0'   */
   sprintf(strchr(nome, '\0'), "%.3d", ordem);

   return nome;
}

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas chamadas
 *                  consecutivas
 *
 * Parmetros:
 *     info (entrada) - informao que ser apresentada na tela
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da
 *             contagem de tempo e na segunda chamada ela
 *             apresenta o intervalo de tempo decorrido
 *             desde a ltima chamada
 *
 ****/
void MedidaDeTempo(const char *info)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */
   double        intervalo;

   if (info) {
      printf("\n\n>>> %s\n", info);
      fflush(stdout);
   }

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      intervalo = difftime(tf, t0);

      if (intervalo < 60)
         printf( "\n\t*** Tempo gasto na operacao: "
                 "%5.2f segundos ***\n", intervalo );
      else
         printf( "\n\t*** Tempo gasto na operacao: "
                 "%5.2f minutos ***\n", intervalo/60 );
   }
}
