/****
 *
 * Arquivo: Registros.c
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/09/2017
 * ltima modificao: 13/05/2018
 *
 * Descrio: Implementao de operaes com registros
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada e Sada   */
#include <string.h> /* Processamento de strings     */
#include <stdlib.h> /* strtod() e alocao dinmica */
#include "Registros.h" /* Interface deste mdulo    */
#include "ASSEGURA.h" /* Macro ASSEGURA             */

/***************** Constantes Simblicas ******************/

   /* Tamanho do acrscimo do bloco usado para conter */
   /* uma linha a cada chamada de realloc()           */
#define TAMANHO_BLOCO 256

/********************* Funes Locais *********************/

/****
 *
 * LeLinhaIlimitada(): L uma linha de tamanho arbitrrio num
 *                     stream de texto e armazena os caracteres
 *                     lidos num array alocado dinamicamente
 *
 * Parmetros:
 *      tam (sada) - se no for NULL, apontar para uma varivel
 *                    que armazenar o tamanho do string
 *                    constitudo pelos caracteres da linha
 *      stream (entrada) - stream de texto no qual ser feita a
 *                         leitura
 *
 * Retorno: Endereo do array contendo a linha lida. NULL, se
 *          ocorrer erro ou o final do arquivo for atingido
 *          antes da leitura de qualquer caractere
 *
 * Observaes:
 *       1. O stream deve estar associado a um arquivo de texto
 *          aberto em modo de texto que permita leitura
 *       2. O caractere '\n' no  includo no string resultante
 *          da leitura
 *       3. O primeiro parmetro pode ser NULL. Nesse caso, o
 *          tamanho do string no ser armazenado
 *
 ****/
static char *LeLinhaIlimitada(int *tam, FILE *stream)
{
   char *ar = NULL,   /* Ponteiro para um array alocado */
                      /* dinamicamente que conter os   */
                      /* caracteres lidos               */
        *p; /* Usado em chamada de realloc() */
   int   tamanho = 0, /* Tamanho do array alocado */
         c,  /* Armazenar cada caractere lido */
         i; /* ndice do prximo caractere */
            /* a ser inserido no array     */

      /* L caracteres a partir da posio corrente do      */
      /* indicador de posio do arquivo e armazena-os num  */
      /* array. A leitura encerra quando '\n'  encontrado, */
      /* o final do arquivo  atingido ou ocorre erro.      */
   for (i = 0; ; ++i) {
         /* L o prximo caractere no arquivo */
      c = fgetc(stream);

         /* Se ocorreu erro de leitura, libera o  */
         /* bloco eventualmente alocado e retorna */
      if (ferror(stream)) {
         free(ar); /* Libera o bloco apontado por 'ar' */
         return NULL; /* Ocorreu erro de leitura */
      }

         /* Verifica se array est completo. O maior valor  */
         /* que i poderia assumir deveria ser tamanho - 1.  */
         /* Mas, como ao final, o caractere '\0' dever ser */
         /* inserido, limita-se o valor de i a tamanho - 2. */
      if (i > tamanho - 2) { /* Limite atingido */
            /* Tenta redimensionar o array */
         p = realloc(ar, tamanho + TAMANHO_BLOCO);

            /* Se o redimensionamento no foi     */
            /* possvel, libera o bloco e retorna */
         if (!p) {
            free(ar); /* Libera o bloco apontado por 'ar' */
            return NULL; /* Ocorreu erro de alocao */
         }

            /* Redimensionamento foi OK. Ento, faz-se */
            /* 'ar' apontar para o novo bloco.         */
         ar = p;

            /* O array aumentou de tamanho */
         tamanho = tamanho + TAMANHO_BLOCO;
      }

         /* Se final do arquivo for atingido ou o caractere */
         /* '\n' for lido,  encerra-se a leitura            */
      if (feof(stream) || c == '\n') {
         break; /* Leitura encerrada */
      }

      ar[i] = c; /* Acrescenta o ltimo caractere lido ao array */
   }

      /* Se nenhum caractere foi lido, libera */
      /* o espao alocado e retorna NULL      */
   if (feof(stream) && !i) {
      free(ar); /* Libera o bloco apontado por 'ar' */
      return NULL; /* Nenhum caractere foi armazenado no array */
   }

      /* Insere o caractere terminal de string. Neste    */
      /* instante, deve haver espao para ele porque o   */
      /* array foi sempre redimensionado deixando um     */
      /* espao a mais para o onipresente caractere '\0' */
   ar[i] = '\0';

      /* Atualiza o valor apontado pelo parmetro */
      /* 'tam', se ele no for NULL               */
   if (tam) {
         /* i  o ndice do caractere terminal do */
         /* string e corresponde ao seu tamanho   */
      *tam = i;
   }

            /*                                     */
            /* >>> NB: O tamanho do string no <<< */
            /* >>>     inclui o caractere '\0' <<< */
            /*                                     */

      /* Tenta ajustar o tamanho do array para no */
      /* haver desperdcio de memria. Como i  o  */
      /* tamanho do string, o tamanho do array que */
      /* o contm deve ser i + 1.                  */
   p = realloc(ar, i + 1);

      /*                                                 */
      /* Se a realocao foi bem sucedida, retorna-se p. */
      /* Caso contrrio, 'ar' ainda aponta para um bloco */
      /* vlido. Talvez, haja desperdcio de memria,    */
      /* mas, aqui,  melhor retornar 'ar' do que NULL.  */
      /*                                                 */
   return p ? p : ar;
}

/****
*
* ObtemTokens(): Separa um string em partes (tokens)
*
* Parmetros:
*     str (entrada/sada) - o string que ser separado
*                           ou NULL para continuar uma
*                           separao anterior
*     sep (entrada) - string contendo os separadores de tokens
*
* Retorno: Endereo do prximo token do string, se este
*          existir, ou NULL, em caso contrrio
*
* Observao: Esta funo NO  funcionalmente equivalente 
*             funo strtok() da biblioteca padro de C
*
****/
static char *ObtemTokens(char *str, char const *sep)
{
   static char *proximoToken; /* Aponta para o prximo */
                              /* token, se ele existir */
   char        *s, /* Apontar para o string no qual a */
                   /* busca pelo token ser efetuada   */
               *inicio; /* Guardar o incio */
                        /* do token corrente */

   if (str != NULL)
      proximoToken = str;

   if (proximoToken == NULL)
      return NULL;

   if ((s = strpbrk(proximoToken, sep)) != NULL) {
      *s  = '\0';
      inicio = proximoToken;
      proximoToken = ++s;
   } else if (*proximoToken) {
      inicio = proximoToken;
      proximoToken = NULL;
   }

   return inicio;
}

/****
 *
 * LinhaEmRegistro(): Converte uma linha do arquivo numa
 *                    estrutura do tipo tRegistro
 *
 * Parmetros: aluno (sada) - ponteiro para a estrutura
 *                             resultante da converso
 *             linha (entrada/sada) - linha que ser convertida
 *
 * Retorno: Endereo da estrutura que armazena o resultado
 *          da converso
 *
 * Observaes:
 *     1. Para facilitar o entendimento, esta funo assume
 *        que o parmetro 'linha' realmente  um string no
 *        formato das linhas do arquivo. Portanto, os
 *        valores retornados por ObtemTokens() no so testados
 *        como deveriam.
 *
 *     2. O parmetro 'linha'  alterado por ObtemTokens().
 *
 *     3. A funo ObtemTokens()  usada em lugar de strtok()
 *        devido  presena de campos vazios no arquivo de
 *        texto, o que faz com que dois separadores apaream
 *        juntos. Nesse caso, a funo strtok() consideraria
 *        os dois separadores como sendo apenas um.
 *
 *     4. O caractere '\r' faz parte das chamadas da funo
 *        ObtemTokens() para satisfazer uma bizarrice do
 *        sistema Mac OS X, que  o nico sistema Unix que
 *        usa '\r' como quebra de linha.
 *
 ****/
static tRegistro *LinhaEmRegistro(tRegistro *aluno, char *linha)
{
   char *str, /* Apontar para cada token da linha */
        *final; /* Final de uma converso numrica */

      /* Obtm o cdigo do aluno, converte-o e         */
      /* acrescenta-o ao respectivo campo da estrutura */
   str = ObtemTokens(linha, "\t\n\r;");
   ++str; /* Salta espao em branco inicial */
   aluno->codigoAluno = strtoll(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm o cdigo de vnculo do aluno ao */
      /* curso, converte-o e acrescenta-o ao   */
      /* respectivo campo da estrutura         */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->chave = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm o nome da IES e acrescenta-a */
      /* ao respectivo campo da estrutura   */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->nomeIES, str, MAX_NOME_IES);

      /* Obtm o nome do curso e acrescenta-o */
      /* ao respectivo campo da estrutura     */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->nomeCurso, str, MAX_NOME_CURSO);

      /* Obtm o sexo do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura     */
   str = ObtemTokens(NULL, "\t\n\r;");
   ASSEGURA(*str == '0' || *str == '1', "Sexo indefinido");
   aluno->sexoAluno = *str == '0' ? 0 : 1;

      /* Obtm a idade do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura      */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->idadeAluno = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

      /* Obtm a UF de nascimento e acrescenta-a */
      /* ao respectivo campo da estrutura        */
   str = ObtemTokens(NULL, "\t\n\r;");
   strncpy(aluno->UFNascimento, str, MAX_UF_NASCIMENTO);

      /* Obtm o ano de ingresso do aluno e acrescenta-o */
      /* ao respectivo campo da estrutura                */
   str = ObtemTokens(NULL, "\t\n\r;");
   aluno->anoIngresso = strtol(str, &final, 10);
   ASSEGURA(!*final, "Erro de conversao numerica");

   return aluno;
}

/********************* Funes Globais ********************/

/****
 *
 * NomeTemp(): Cria nomes de arquivos temporrios
 *
 * Parmetros:
 *      nome[] (sada) - array que armazenar o nome de arquivo
 *                       criado
 *      inicio (entrada) - string que inicia cada nome de arquivo
 *      max (entrada) - nmero mximo de nomes de arquivos gerados
 *
 * Retorno: O endereo do nome (string) criado.
 *          NULL, se a operao no for possvel.
 *
 * Observao: Quando o primeiro parmetro  NULL, o contador de
 *             arquivos  zerado. O mesmo ocorre quando o nmero
 *             mximo de arquivos  atingido. Isto permite que
 *             vrias sequncias de arquivos possam ser criadas.
 *
 ****/
char *NomeTemp(char *nome, const char *inicio, int max)
{
   static int ordem = 0; /* Nmero de ordem do arquivo */

      /* Quando o primeiro parmetro  NULL, */
      /* o contador de arquivos  zerado     */
   if (!nome) {
      ordem = 0; /* Reinicia contador de arquivos */
      return NULL;
   }

   ++ordem;	/* Um novo nome de arquivo ser criado */

      /* Verifica se o nmero mximo de */
      /* arquivos gerados foi atingido  */
   if (ordem > max) {
      ordem = 0; /* Reinicia contador de arquivos */
      return NULL;
   }

      /* Copia o incio do nome do arquivo temporrio */
      /* para o parmetro que conter o resultado     */
   strcpy(nome, inicio);

      /* Insere o nmero de ordem do arquivo a */
      /* partir da posio do caractere '\0'   */
   sprintf(strchr(nome, '\0'), "%.3d", ordem);

   return nome;
}

/****
 *
 * MaxArquivosAbertos(): Calcula o nmero de arquivo que podem
 *                       ser abertos simultaneamente
 *
 * Parmetros: Nenhum
 *
 * Retorno: O nmero de arquivo que podem ser abertos
 *          simultaneamente
 *
 ****/
int MaxArquivosAbertos(void)
{
   int   i,
         cont; /* Conta o nmero de arquivos abertos */
   char  nomeArq[MAX_NOME_TMP]; /* Armazena o nome de um */
                                /* arquivo temporrio    */
   FILE *fp[MAX_ARQ_TMP]; /* Array de ponteiros para FILE que */
                      /* apontaro para os streams associados */
                      /* associados aos arquivos abertos      */

      /* Verifica quantos arquivos podem ser abertos */
   for (cont = 0; cont < MAX_ARQ_TMP; ++cont) {
         /* Tenta criar um nome de arquivo temporrio */
      ASSEGURA( NomeTemp(nomeArq, INI_ARQ_TMP, MAX_ARQ_TMP),
                "Nao foi possvel criar nome de arquivo" );

      fp[cont] = fopen(nomeArq, "w"); /* Tenta criar arquivo */

      if (!fp[cont]) /* Arquivo no foi aberto */
         break; /* O lao no deve mais continuar */
   }

         /*   >>> Remove os arquivos criados <<<    */

      /* Reinicia o gerador de arquivos temporrios */
   NomeTemp(NULL, NULL, 0);

      /* Fecha e remove os arquivos abertos */
   for (i = 1; i <= cont; ++i) {
         /* Tenta criar nome de arquivo temporrio */
      ASSEGURA( NomeTemp(nomeArq, INI_ARQ_TMP, MAX_ARQ_TMP),
                "Nao foi possvel criar nome de arquivo" );

         /* Tenta fechar o arquivo correspondente */
      ASSEGURA( !fclose(fp[i - 1]),
                "Nao foi possivel fechar um arquivo" );

         /* Tenta remover o arquivo correspondente */
      ASSEGURA( !remove(nomeArq),
                "Nao foi possivel remover um arquivo" )
   }

      /* Ao final  do lao a varivel 'cont'  */
      /* conter o nmero de arquivos abertos */
   return cont;
}

/****
 *
 * AbreArquivo(): Tenta abrir um arquivo e, quando isso no 
 *                possvel, aborta o programa
 *
 * Parmetros:
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *      modo (entrada) - modo de abertura do arquivo
 *
 * Retorno: Stream associado ao arquivo aberto
 *
 ****/
FILE *AbreArquivo(const char *nomeArq, const char *modo)
{
   FILE *stream = fopen(nomeArq, modo);

      /* Se ocorreu erro de abertura, aborta o programa */
   if (!stream) {
      fprintf( stderr, "\a\n>>> Erro na abertura do arquivo %s."
              "\n>>> O programa sera' encerrado.\n", nomeArq);
      exit(1); /* Aborta o programa */
   }

   return stream;
}

/****
 *
 * FechaArquivo(): Tenta fechar um arquivo e, quando isso
 *                 no  possvel, aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      nomeArq (entrada) - nome do arquivo ou NULL
 *
 * Retorno: Nada
 *
 * Observao: Se o segundo parmetro no for NULL, o nome
 *             do arquivo aparece na mensagem de erro
 *
 ****/
void FechaArquivo(FILE *stream, const char *nomeArq)
{
   /***********************************************************/
   /* Se fclose() retornar um valor diferente de zero ocorreu */
   /* algum erro na tentativa de fechamento do arquivo. Nesse */
   /* caso, apresenta mensagem de erro e aborta o programa.   */
   /***********************************************************/

   if (fclose(stream)) { /* Erro de fechamento */
      fprintf( stderr, "\a\n>>> Ocorreu erro no fechamento "
              "do arquivo %s.\n>>> O programa sera' "
              "encerrado.\n", nomeArq ? nomeArq : "");
      exit(1); /* Aborta o programa */
   }
}

/****
 *
 * MoveApontador(): Tenta mover o apontador de posio de um
 *                  arquivo e, quando isso no  possvel,
 *                  aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      bytes (entrada) - quantidade de bytes que sero saltados
 *      DeOnde (entrada) - a partir de onde ocorrer a movimentao
 *
 * Retorno: Nada
 *
 ****/
void MoveApontador(FILE *stream, long bytes, int DeOnde)
{
   int deuErrado;

   deuErrado = fseek(stream, bytes, DeOnde);

      /* Verifica se houve erro */
   if (deuErrado) {
      fprintf( stderr, "\n>>> Erro na tentativa de mover "
                       "apontador de arquivo" );
      exit(1);
   }
}

/****
 *
 * ObtemApontador(): Obtm a posio do apontador de posio de
 *                   um arquivo e, quando isso no  possvel,
 *                   aborta o programa
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *
 * Retorno: A posio do apontador de posio do arquivo
 *
 ****/
long ObtemApontador(FILE *stream)
{
   long posicao;

   posicao = ftell(stream);

      /* Verifica se houve erro */
   if (posicao < 0L) {
      fprintf( stderr,
               "\n>>> Valor negativo retornado por ftell()" );
      exit(1);
   }

   return posicao;
}

/****
 *
 * ComparaRegistros(): Compara dois registros que sero ordenados
 *
 * Parmetros:
 *      e1, e2 (entrada) - endereos dos registros que sero
 *                         comparados
 *
 * Retorno:
 *     * 0, se os registros so iguais
 *     * > 0, se o primeiro registro for maior do que o segundo
 *     * < 0, se o primeiro registro for menor do que o segundo
 *
 * Observao: Se um ponteiro  NULL, o registro para o qual
 *             ele aponta  considerado maior do que os demais
 *
 ****/
int ComparaRegistros(const void *e1, const void* e2)
{
   tRegistro *r1 = (tRegistro *)e1,
             *r2 = (tRegistro *)e2;

   if (!r1 && !r2)
      return 0;
   else if (!r1)
      return 1;
   else if (!r2)
      return -1;

   return r1->chave - r2->chave;
}

/********************* Funes Globais ********************/

/****
 *
 * TamanhoDeArquivo(): Calcula o tamanho de um arquivo lendo cada
 *                     byte que o constitui e contando quantos
 *                     bytes so lidos
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *
 * Retorno: -1, se ocorrer algum erro.
 *          O tamanho do arquivo, em caso contrrio.
 *
 * Observaes:
 *     1. O arquivo deve estar aberto em modo binrio que permite
 *        leitura
 *     2. A posio atual do indicador de posio do arquivo 
 *        preservada
 *
 ****/
int TamanhoDeArquivo(FILE *stream)
{
   int posicaoAtual, tamanho = 0;

      /* Guarda a posio atual do indicador de posio */
      /* do arquivo para que ele possa ser restaurado   */
   posicaoAtual = ftell(stream);

      /* Verifica se ftell() foi bem sucedida */
   if (posicaoAtual < 0) {
      return -1; /* A chamada de ftell() falhou */
   }

      /* Assegura leitura a partir do incio do arquivo */
   if ( fseek(stream, 0, SEEK_SET) ) {
         /* O movimento do indicador de posio no foi   */
         /* possvel. Neste caso, no precisa restaur-lo */
      return -1;
   }

      /* O lao encerra quando o final do arquivo for */
      /* atingido ou se ocorrer erro de leitura       */
   while (1){
         /* L e descarta o prximo byte */
      fgetc(stream);

         /* Checa se o arquivo j foi inteiramente lido */
      if (feof(stream)) {
         break; /* Encerra lao */
      }

         /* Verifica se ocorreu erro de leitura */
      if (ferror(stream)) {
            /* Restaura a posio original do indicador */
            /* de posio do arquivo  antes de retornar */
         fseek(stream, posicaoAtual, SEEK_SET);
         return -1; /* Ocorreu erro de leitura */
      }

      ++tamanho; /* Mais um byte foi lido */
   }

      /* Restaura a posio original do indicador de posio */
   if ( fseek(stream, posicaoAtual, SEEK_SET) ) {
      return -1; /* O movimento no foi possvel */
   }

   return tamanho;
}

/****
 *
 * NumeroDeRegistros(): Calcula o nmero de registros (i.e.,
 *                      parties de tamanho especificado) de um
 *                      arquivo binrio
 *
 * Parmetros:
 *    stream (entrada) - stream binrio associado ao arquivo cujo
 *                       tamanho ser calculado
 *    tamRegistro (entrada) - tamanho de cada partio do arquivo
 *
 * Retorno: O nmero de parties do arquivo, se no ocorrer
 *          nenhum erro, ou -1, em caso contrrio.
 *
 * Observao: O arquivo deve estar aberto em modo binrio que
 *             permite leitura
 *
 ****/
int NumeroDeRegistros( FILE *stream, int tamRegistro )
{
   int nBytes = 0; /* Nmero de bytes do arquivo */

      /* Calcula o nmero de bytes do arquivo */
   nBytes = TamanhoDeArquivo(stream);

      /* Verifica se ocorreu erro na contagem de bytes */
   if (nBytes < 0) {
      return -1; /* Erro na contagem de bytes */
   }

      /* O nmero de registros  o nmero de bytes do arquivo */
      /* dividido pelo nmero de bytes de cada registro       */
   return nBytes/tamRegistro;
}

/****
 *
 * ArquivoExiste(): Verifica se um arquivo existe
 *
 * Parmetros:
 *     nomeArq (entrada) - nome do arquivo que se
 *                        verificar se existe
 *
 * Retorno: 1, se o arquivo existir; 0, em caso contrrio
 *
 ****/
int ArquivoExiste(const char *nomeArq)
{
   FILE *stream;
   int   existe;

   stream = fopen(nomeArq, "rb");

   existe = stream ? 1 : 0;

   if (existe)
      FechaArquivo(stream, nomeArq);

   return existe;
}

/****
 *
 * CriaArquivoBin(): L cada linha de um arquivo no formato
 *                   especificado, converte-a numa estrutura
 *                   do tipo tRegistro e armazena a estrutura
 *                   num arquivo binrio
 *
 * Parmetros:
 *     streamTexto (entrada) - stream associado ao arquivo que
 *                             ser lido
 *     streamBin (entrada) - stream associado ao arquivo que ser
 *                           escrito
 *     maxRegistros (entrada) - nmero mximo de registros que
 *                              sero gerados
 *
 * Retorno: Nmero de estruturas escritas no arquivo
 *          binrio ou um valor negativo, se ocorrer erro
 *
 * Observaes:
 *     1. O stream de entrada deve estar aberto em modo de
 *        texto que permite leitura
 *     2. O stream de sada deve estar aberto em modo "wb"
 *
 ****/
int CriaArquivoBin( const char *nomeTexto, const char *nomeBin,
                    int maxRegistros )
{
   char  *linha; /* Apontar para cada linha lida */
   tRegistro umAluno; /* Dados de uma linha      */
                      /* convertida em estrutura */
   int   nRegistros = 0, /* Nmero de estruturas escritas */
                         /* no arquivo binrio            */
         ocorreuErro = 0;
   FILE *streamTexto,
        *streamBin;

      /* Tenta abrir arquivo para leitura em modo texto */
   streamTexto = AbreArquivo(nomeTexto, "r");

      /* Tenta abrir arquivo para escrita em modo binrio */
   streamBin = AbreArquivo(nomeBin, "w");

      /* L cada linha do arquivo, cria um registro do tipo     */
      /* tRegistro e armazena-o no arquivo binrio. O lao para */
      /* quando h tentativa de leitura alm do final do        */
      /* arquivo de entrada, ou ocorre erro de leitura/escrita  */
   while (1) {
         /* L uma linha no arquivo de entrada */
      linha = LeLinhaIlimitada(NULL, streamTexto);

         /* Verifica se o final do arquivo foi  */
         /* atingido ou ocorreu erro de leitura */
      if (feof(streamTexto) || ferror(streamTexto))
         break;

         /* Converte a linha lida em estrutura */
      LinhaEmRegistro(&umAluno, linha);

         /* Escreve a estrutura resultante da */
         /* converso no arquivo binrio      */
      fwrite(&umAluno, sizeof(umAluno), 1, streamBin);

         /* Verifica se ocorreu erro de escrita */
      if (ferror(streamBin))
         break;

      ++nRegistros; /* Mais um registro lido */

      if (maxRegistros && nRegistros >= maxRegistros)
         break;
   }

   ocorreuErro = ferror(streamTexto) || ferror(streamBin);

      /* Fecha os arquivos */
   FechaArquivo(streamTexto, nomeTexto);
   FechaArquivo(streamBin, nomeBin);

   ASSEGURA(!ocorreuErro,
            "Erro de processamento de arquivo em CriaArquivoBin");

   return nRegistros; /* No ocorreu nenhum erro */
}
