/****
 *
 * Ttulo: Ordenao externa de arquivos por intercalao multplice
 *
 * Seo: 12.4.3
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/09/2017
 * ltima modificao: 15/05/2018
 *
 * Descrio: Ordena um arquivo usando ordenao por intercalao
 *            multplice
 *
 * Entrada: Um arquivo desordenado
 *
 * Sada: O arquivo ordenado
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>     /* Funes de Entrada/Sada       */
#include <string.h>    /* Processamento de strings       */
#include <stdlib.h>    /* Funes de alocao dinmica   */
#include <time.h>      /* Funo time() para medir tempo */
#include "ListaSE.h"   /* Interface do tipo tListaSE     */
#include "Registros.h" /* Processamento de registros     */
#include "Heap.h"      /* Heap usado na intercalao     */
#include "Tipos.h"     /* Tipos usadas pelo programa     */
#include "Leitura.h"   /* Funes de leitura via teclado */

/**************** Compilao Condidicional ****************/

#define DEPURACAO 0

/***************** Definies de Funes ******************/

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas
 *                  chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da contagem
 *             de tempo e na segunda chamada ela apresenta o
 *             intervalo de tempo decorrido desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */
   double        intervalo;

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      intervalo = difftime(tf, t0);

      if (intervalo < 60) {
         printf( "\n\n\t*** Tempo gasto na operacao: "
                 "%5.2f segundos ***\n", intervalo );
      } else {
         printf( "\n\n\t*** Tempo gasto na operacao: "
                 "%5.2f minutos ***\n", intervalo/60 );
      }
   }
}

/****
 *
 * CriaSeriesOrdenadas(): Cria sries de arquivos ordenados
 *
 * Parmetros:
 *      stream (entrada) - nome do arquivo que origina as sries
 *
 * Retorno: Lista contendo informaes sobre os arquivos
 *          que contm as sries
 *
 ****/
tListaSE CriaSeriesOrdenadas(const char *nomeArq)
{
   tListaSE  lista; /* Lista contendo informaes sobre */
                    /* os arquivos que contm as sries */
   tRegistro *registros; /* Ponteiro para um array que armazena */
                         /* uma srie em memria principal *    */
   tSerie    arqTmp; /* Nome e stream associados */
                     /* a um arquivo temporrio  */
   int       nRegistros; /* Nmero de registros numa srie */
   FILE     *stream;

      /* Inicia a lista */
   IniciaListaSE(&lista);

      /* Aloca espao para o array. Um array desse */
      /* tamanho no pode ser alocado na pilha     */
   registros = calloc(MAX_REGISTROS, sizeof(*registros));
   ASSEGURA( registros, "Buffer nao foi alocado" );

      /* Tenta abrir o arquivo de dados */
   stream = AbreArquivo(nomeArq, "rb");

      /* L o arquivo a ser ordenado e cria as sries */
   while (1) {
         /* L os registros da srie corrente */
      nRegistros = fread( registros, sizeof(*registros),
                          MAX_REGISTROS, stream );

         /* Verifica se ocorreu erro de leitura */
      ASSEGURA( !ferror(stream), "Erro de leitura em arquivo" );

         /* Encerra o lao se nenhum registro foi lido */
      if (!nRegistros)
         break;

         /* Ordena a srie */
      qsort( registros, nRegistros, sizeof(*registros),
             ComparaRegistros );

         /* Cria o nome do arquivo que conter a srie corrente */
      ASSEGURA(NomeTemp(arqTmp.nome, INI_ARQ_TMP, MAX_ARQ_TMP),
               "Impossivel criar nome de arquivo temporario");

#if DEPURACAO == 1
      printf("\n\t>>> Criando serie \"%s\"...", arqTmp.nome);
#endif

         /* Tenta abrir o arquivo que conter a */
         /* srie em modo de leitura e escrita  */
      arqTmp.stream = AbreArquivo(arqTmp.nome, "w+b");

         /* Escreve os registros ordenados */
         /* da srie corrente no arquivo   */
      fwrite( registros, sizeof(*registros), nRegistros,
              arqTmp.stream);

         /* Verifica se ocorreu erro de escrita */
      ASSEGURA( !ferror(arqTmp.stream),
                "Erro de escrita em arquivo temporario" );

         /* Move apontador de posio para que a primeira */
         /* leitura da srie seja no seu incio           */
      MoveApontador(arqTmp.stream, 0L, SEEK_SET);

         /* Insere informaes sobre o arquivo */
         /* que contm a srie corrente        */
      InsereNoLSE(&lista, &arqTmp);
   }

      /* Fecha o arquivo original */
   FechaArquivo(stream, nomeArq);

   return lista;
}

/****
 *
 * RenovaBuffer(): Copia registros de uma srie para o buffer
 *                associado a ela
 *
 * Parmetros:
 *      pSerie (entrada/sada) - ponteiro para uma estrutura
 *                               contendo informaes sobre a
 *                               srie associada ao buffer que
 *                               ser renovado
 *      tamBuffer (entrada) - tamanho do buffer que ser renovado
 *
 * Retorno: O nmero de registros copiados
 *
 ****/
int RenovaBuffer(tSerie *pSerie, int tamBuffer)
{
      /* Preenche o contedo do buffer de entrada corrente */
   pSerie->nRegistros = fread( pSerie->buffer, sizeof(tRegistro),
                               tamBuffer, pSerie->stream );
   ASSEGURA( !ferror(pSerie->stream),
             "Erro de leitura de uma serie" );

   return pSerie->nRegistros;
}

/****
 *
 * RemoveSerie(): Apaga o arquivo que representa uma srie e
 *                remove da lista as informaes sobre a srie
 *
 * Parmetros:
 *      pSerie (entrada) - ponteiro para uma estrutura contendo
 *                         informaes sobre a srie que ser
 *                         removida
 *      lista (entrada/sada) - lista contendo informaes sobre
 *                              as sries
 *
 * Retorno: Nada
 *
 ****/
void RemoveSerie(const tSerie *pSerie, tListaSE *lista)
{
      /* Fecha o arquivo */
   FechaArquivo(pSerie->stream, pSerie->nome);

      /* Remove o arquivo que representa a srie */
   ASSEGURA(!remove(pSerie->nome), "Impossivel remover serie");

      /* Remove da lista o n que representa a srie */
   ASSEGURA( !RemoveNoLSE(lista, pSerie),
             "Serie nao foi encontrada na lista" );
}

/****
 *
 * IntercalaMulti(): Intercala sries de arquivos ordenados cujas
 *                   informaes esto armazenadas numa lista
 *
 * Parmetros:
 *      lista (entrada/sada) - lista contendo informaes
 *                              sobre as sries
 *      arqNome (entrada) - nome do arquivo resultante da ordenao
 *
 * Retorno: Nada
 *
 ****/
void IntercalaMulti(tListaSE *lista, const char *arqNome)
{
   tSerie    *pSerie; /* Ponteiro para uma estrutura contendo */
                      /* informaes sobre uma srie          */
   int        nSeries, /* Nmero de sries */
              tamBuffer; /* Tamanho de um buffer */
                         /* de entrada ou sada  */
   FILE      *streamOrd; /* Stream associado ao arquivo que */
                         /* conter os registros ordenados  */
   tHeap      heap; /* Heap usado na ordenao dos elementos */
                    /* dos buffers de entrada                */
   int        i,
              iBufferSaida = 0, /* Indica a posio na qual o */
                                /* prximo elemento do buffer */
                                /* de sada ser armazenado   */
              iHeap = 0; /* ndice de um elemento do heap */
   tNoHeap    raiz; /* Uma raiz do heap */
   tRegistro *bufferSaida; /* Apontar para o buffer de sada */

      /* Calcula o nmero de sries de entrada */
   nSeries = ComprimentoListaSE(*lista);

      /* O nmero de sries no pode ser nulo */
   ASSEGURA(nSeries, "A lista esta' vazia");

      /* Calcula o tamanho de cada buffer */
   tamBuffer = MAX_REGISTROS/(nSeries + 1);

      /* O tamanho de um buffer no pode ser zero */
   ASSEGURA(tamBuffer, "O tamanho de cada buffer e' zero");

      /* Aloca espao para o buffer de sada */
   bufferSaida = calloc(tamBuffer, sizeof(tRegistro));
   ASSEGURA(bufferSaida, "Impossivel alocar buffer de saida");

      /* Inicia o heap */
   IniciaHeap(&heap, nSeries);

      /* Aloca espao para cada buffer de entrada,  */
      /* preenche cada um com o contedo inicial de */
      /* sua respectiva srie e associa o primeiro  */
      /* registro de cada buffer a um n do heap    */
   while ((pSerie = ProximoListaSE(*lista))) {
         /* Aloca espao para o buffer de entrada corrente */
      pSerie->buffer = calloc(tamBuffer, sizeof(tRegistro));
      ASSEGURA( pSerie->buffer,
                "Impossivel alocar um buffer de entrada" );

         /* Preenche o contedo do buffer de entrada corrente */
      pSerie->nRegistros = fread(pSerie->buffer, sizeof(tRegistro),
                                 tamBuffer, pSerie->stream);
      ASSEGURA( !ferror(pSerie->stream),
                "Erro de leitura de uma serie" );

         /* Associa dados do primeiro registro */
         /* do buffer corrente a um n do heap */
      heap.itens[iHeap].elemento = pSerie->buffer;/* Endereo do */
                                                  /* registro    */
      heap.itens[iHeap].i = 0; /* Seu ndice */
      heap.itens[iHeap].serie = pSerie;/* Endereo de sua srie */

      ++iHeap; /* Passa para o prximo elemento do heap */
   }

      /* Ordena o heap */
   for (i = (nSeries - 1)/2; i >= 0; --i)
      OrdenaHeap(&heap, i, ComparaRegistros);

      /* Abre o arquivo que conter o resultado da ordenao */
   streamOrd = AbreArquivo(arqNome, "wb");

      /* Obtm o menor elemento do heap e o substitui */
      /* pelo prximo elemento da srie que o contm  */
      /* at que todas as sries estejam esgotadas    */
   while (!EstaVaziaLSE(*lista)) {
         /* Obtm o menor valor armazenado no heap */
      raiz = ObtemMinimo(&heap);

         /* Verifica se o buffer de sada est repleto */
      if (iBufferSaida > tamBuffer - 1) {
            /* O buffer de sada est repleto e  necessrio  */
            /* descarreg-lo no stream que contm o resultado */
         fwrite( bufferSaida, sizeof(tRegistro), tamBuffer,
                 streamOrd );
         ASSEGURA(!ferror(streamOrd),
                  "Erro de escrita de resultado de intercalacao");

         iBufferSaida = 0; /* Reinicia o buffer de sada */
      }

         /* Armazena o menor valor do heap no buffer */
         /* que conter o resultado da intercalao  */
      bufferSaida[iBufferSaida++] = *raiz.elemento;

         /* O elemento que substituir a raiz pertence */
         /*  mesma srie  qual pertencia a raiz      */
      if (++raiz.i < raiz.serie->nRegistros) {
            /* Obtm o prximo elemento da   */
            /* srie  qual a raiz pertencia */
         raiz.elemento = &raiz.serie->buffer[raiz.i];
      } else { /* Esse buffer est esgotado */
            /* Tenta renovar o buffer de entrada  */
         if (RenovaBuffer(raiz.serie, tamBuffer)) {
               /* O buffer foi renovado. Obtm os */
               /* dados do seu primeiro registro. */
            raiz.elemento = raiz.serie->buffer;
            raiz.i = 0;
         } else {/* O buffer no foi renovado porque todos os   */
                 /* registros de sua srie  j foram consumidos */

               /* Remove a srie da lista e apaga */
               /* o arquivo correspondente        */
            RemoveSerie(raiz.serie, lista);

               /* A nova raiz ser o ltimo elemento do       */
               /* heap, que passar a ter um elemento a menos */
            raiz = ReduzHeap(&heap);
         }
      }

         /* Substitui a raiz do heap pelo novo */
         /* elemento e reordena o heap         */
      SubstituiMinimo(&heap, &raiz, ComparaRegistros);
   }

      /* Verifica se restam registros no buffer de sada */
   if (iBufferSaida > 0) {
         /* Escreve os registros restantes no */
         /* stream que contm o resultado     */
      fwrite( bufferSaida, sizeof(tRegistro), iBufferSaida,
              streamOrd );
      ASSEGURA(!ferror(streamOrd),
               "Erro de escrita de resultado de intercalacao");
   }

      /* Trabalho concludo. Fecha o arquivo ordenado. */
   FechaArquivo(streamOrd, arqNome);
}

/****
 *
 * TestaOrdenacao2(): Verifica se um arquivo est ordenado em
 *                    ordem crescente
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      stream (entrada) - nome do arquivo
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. Esta funo  usada para depurao
 *     2. No h tratamento de excees
 *
 ****/
void TestaOrdenacao2(FILE *stream, const char *nome)
{
   tRegistro registroAtual, registroAnterior;
   int       teste, nRegistros, ordenado = 1;

      /* Calcula o nmero de registros */
   fseek(stream, 0L, SEEK_END);
   nRegistros = ftell(stream)/sizeof(tRegistro);

   printf( "\nTestando ordenacao de %d registros do arquivo \"%s\"...",
           nRegistros, nome );
   fflush(stdout);

      /* Move apontador para o incio do arquivo */
   rewind(stream);

      /* L o primeiro registro do arquivo */
   fread(&registroAnterior, sizeof(registroAnterior), 1, stream);
   ASSEGURA( !ferror(stream), "Erro de leitura" );

      /* L os demais registros e testa se eles esto ordenados */
   while (1) {
         /* L o prximo registro */
      fread(&registroAtual, sizeof(registroAtual), 1, stream);
      ASSEGURA( !ferror(stream), "Erro de leitura" );

      if (feof(stream))
         break; /* Leitura acabou */

         /* Compara a chave do registro atual */
         /* com a chave do registro anterior  */
      teste = registroAtual.chave - registroAnterior.chave;

         /* Verifica se as duas ltimas chaves */
         /* esto em ordem crescente           */
      if (teste <= 0) {
         ordenado = 0;
         break;
      }

         /* A chave anterior passa a ser a chave corrente */
      registroAnterior = registroAtual;
   }

   rewind(stream);

   printf( "\n\t>>> O arquivo esta' %s\n\n",
           ordenado ? "ordenado" : "desordenado" );
}

/****
 *
 * TestaSeries(): Verifica se as sries esto ordenadas
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *      stream (entrada) - nome do arquivo
 *
 * Retorno: Nada
 *
 * Observaes:
 *     1. Esta funo  usada para depurao
 *     2. No h tratamento de excees
 *
 ****/
void TestaSeries(tListaSE *lista)
{
   tSerie *pSerie;

   while ((pSerie = ProximoListaSE(*lista)))
      TestaOrdenacao2(pSerie->stream, pSerie->nome);
}

/****
 *
 * CriaArquivoRegistros(): Verifica se o arquivo de registros j
 *                         existe. Se ela j existir e o usurio
 *                         desejar reconstrui-lo ou se ele no
 *                         existir, ele ser construdo.
 *
 * Parmetros:
 *     nomeArqTexto (entrada) - nome do arquivo texto de registros
 *     nomeArqBin (entrada) - nome do arquivo binrio de registros
 *
 * Retorno: Nada
 *
 ****/
void CriaArquivoRegistros( const char *nomeArqTexto,
                            const char *nomeArqBin )
{
   int   nRegistros, /* Nmero de registros no arquivo */
         maxRegistros; /* Nmero mximo de registros no arquivo */

      /* Verifica se o arquivo binrio de registros j existe */
   if (ArquivoExiste(nomeArqBin)) {
      /*                                */
      /* Arquivo j existe. Se o usurio */
      /* desejar, ele ser reconstrudo */
      /*                                */

#if DEPURACAO == 1
         /* Oferece a opo ao usurio */
      int op = LeSimNao( "Arquivo de registros ja' existe. "
                      "Deseja substitui-lo" );

         /* Verifica se o usurio deseja reconstruir o arquivo */
      if (op == 'n') /* Ele no deseja isso   */
         return;
#else
      return;
#endif
   }

   /*                                        */
   /* Neste ponto, ou o arquivo ainda no    */
   /* existe ou usurio deseja reconstrui-lo */
   /*                                        */

   printf("\n\t>>> Digite o numero maximo de registros "
             "\n\t>>> (0 = TODOS): ");
   maxRegistros = LeNatural();

   printf("\nCriando arquivo...\n");
   fflush(stdout);

      /* Cria o arquivo binrio */
   nRegistros = CriaArquivoBin( nomeArqTexto, nomeArqBin,
                                maxRegistros );

      /* Apresenta o resultado da converso */
   printf( "\n\t>>> Foram armazenados %d registros no arquivo"
           "\n\t>>> \"%s\"\n", nRegistros, nomeArqBin );
}

/****
 *
 * main(): Efetua ordenao externa de arquivos usando
 *         intercalao multplice
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   tListaSE lista;
   FILE    *stream;

      /* Cria o arquivo de registros se ele no existir */
   CriaArquivoRegistros(NOME_ARQUIVO_REGS, NOME_ARQUIVO_BIN);

   MedidaDeTempo();
   printf("\nCriando series ordenadas...");
   fflush(stdout);

      /* L o arquivo original e cria as sries */
   lista = CriaSeriesOrdenadas(NOME_ARQUIVO_BIN);

   printf("\n...series criadas com sucesso");
   MedidaDeTempo();

#if DEPURACAO == 1
   TestaSeries(&lista);
#endif

   MedidaDeTempo();
   printf("\n\nIntercalando series...");
   fflush(stdout);

   IntercalaMulti(&lista, NOME_ARQUIVO_ORD);

   printf("\n...series intercaladas com sucesso\n");
   MedidaDeTempo();

   stream = AbreArquivo(NOME_ARQUIVO_BIN, "rb");
   TestaOrdenacao2(stream, NOME_ARQUIVO_BIN);
   FechaArquivo(stream, NOME_ARQUIVO_BIN);

   stream = AbreArquivo(NOME_ARQUIVO_ORD, "rb");
   TestaOrdenacao2(stream, NOME_ARQUIVO_ORD);
   FechaArquivo(stream, NOME_ARQUIVO_ORD);

   return 0;
}

/******************* Exemplo de Execuo ******************

Criando series ordenadas...
...series criadas com sucesso

        *** Tempo gasto na operacao:  1.08 minutos ***


Intercalando series...
...series intercaladas com sucesso


        *** Tempo gasto na operacao: 41.00 segundos ***

Testando ordenacao de 9565483 registros do arquivo "../CensoMEC2012.bin"...
        >>> O arquivo esta' desordenado


Testando ordenacao de 9565483 registros do arquivo "CensoMEC2012_Ord.bin"...
        >>> O arquivo esta' ordenado

**********************************************************/
