/****
 *
 * Ttulo: Testando Ordenao em Memria Secundria
 *
 * Seo 12.7.2
 *
 * Autor: Ulysses de Oliveira
 *
 * Data de Criao: 16/09/2017
 * ltima modificao: 15/05/2018
 *
 * Descrio: Verifica se um arquivo se econtra ordenado em ordem
 *            crescente de uma chave
 *
 * Entrada: Um arquivo binrrio contendo registros dotipo tRegistro
 *          com chave de ordenao denominada 'chave' (string)
 *
 * Sada: Uma mensagem informando se o arquivo est ou no ordenado
 *
 ****/

/*********************** Includes *************************/

#include <stdio.h>  /* Funes de Entrada/Sada */
#include <string.h> /* Processamento de strings */
#include <stdlib.h> /* Funes de alocao dinmica */
#include <time.h>   /* Funo time() para medida de tempo */

/****************** Constantes Simblicas *****************/

   /* Nome do arquivo de dados original */
#define NOME_ARQUIVO_BIN "J:\\Dados\\CEPs\\CEPs.bin"

   /* Nome do arquivo de dados ordenado */
#define NOME_ARQUIVO_ORD "CEPs_Ord.bin"

#define MAX_NOME  40 /* Nmero mximo de caracteres num nome */
#define TAM_UF     2 /* Nmero de dgitos num UF */
#define TAM_CHAVE  8 /* Nmero de dgitos num CEP */
#define MAX_COMP  60 /* Nmero mximo de caracteres */
                     /* num complemento             */
#define MAX_TIPO_LOG 10 /* Nmero mximo de caracteres */
                        /* num tipo de logradouro      */
#define TAM_DNE   16 /* Nmero de dgitos numa chave DNE */

/****************** Definies de Macros ******************/

   /* Macro utilizada para verificar     */
   /* ocorrncia de condies de exceo */
#define ASSEGURA(condicao, msg) if (!(condicao)) {\
                                   fprintf(stderr, "\n%s\n", msg);\
                                   exit(1); \
                                }

/****************** Definies de Tipos *******************/

   /* Tipo da chave */
typedef char tChave[TAM_CHAVE + 1];

   /* Tipo usado para armazenamento de */
   /* registros no arquivo binrio     */
typedef struct {
           int  numero;
           char UF[TAM_UF];
           int  localidadeNumero;
           char nomeAbr[MAX_NOME];
           char nome[MAX_NOME];
           int  bairroInicio;
           int  bairroFim;
           char chave[TAM_CHAVE];
           char complemento[MAX_COMP];
           char tipoLogradouro[MAX_TIPO_LOG];
           char statusTipoLogradouro;
           char nomeSemAcento[MAX_NOME];
           char chaveDNE[TAM_DNE];
         } tRegistro;

/***************** Definies de Funes ******************/

/****
 *
 * MedidaDeTempo(): Mede intervalos de tempo entre duas chamadas consecutivas
 *
 * Parmetros: Nenhum
 *
 * Retorno: Nada
 *
 * Observao: Esta funo deve ser chamada aos pares; i.e.,
 *             na primeira chamada ela marca o incio da contagem
 *             de tempo e na segunda chamada ela apresenta o
 *             intervalo de tempo decorrido desde a ltima chamada
 *
 ****/
void MedidaDeTempo(void)
{
   static int    numeroDeChamadas = 0;
   static time_t t0; /* Instante inicial */
   time_t        tf; /* Instante final   */
   double        intervalo;

      /* Esta funo foi chamada mais uma vez */
   ++numeroDeChamadas;

      /* Se o nmero da chamada for mpar, esta */
      /* chamada  a primeira de um par         */
   if (numeroDeChamadas%2) { /* Calcula o instante inicial */
      t0 = time(NULL);
   } else { /* Calcula e apresenta o intervalo de   */
            /* tempo entre as duas ltimas chamadas */
      tf = time(NULL);

      intervalo = difftime(tf, t0);

      if (intervalo < 60) {
         printf( "\n\n\t*** Tempo gasto na operacao: "
                 "%5.2f segundos ***\n", intervalo );
      } else {
         printf( "\n\n\t*** Tempo gasto na operacao: "
                 "%5.2f minutos ***\n", intervalo/60 );
      }
   }
}

/****
 *
 * TestaOrdenacaoMS(): Verifica se um arquivo est ordenado em ordem crescente
 *
 * Parmetros:
 *      stream (entrada) - stream associado ao arquivo
 *
 * Retorno: 1, se o arquivo estiver ordenado; 0, em caso contrrio
 *
 ****/
int TestaOrdenacaoMS(FILE *stream)
{
   tRegistro registroAtual, registroAnterior;
   int       teste;

   rewind(stream);

      /* L o primeiro registro do arquivo */
   fread(&registroAnterior, sizeof(registroAnterior), 1, stream);
   ASSEGURA( !ferror(stream), "Erro de leitura" );

      /* L os demais registros e testa se eles esto ordenados */
   while (1) {
         /* L o prximo registro */
      fread(&registroAtual, sizeof(registroAtual), 1, stream);
      ASSEGURA( !ferror(stream), "Erro de leitura" );

      if (feof(stream))
         break; /* Leitura acabou */

         /* Compara a chave do registro atual */
         /* com a chave do registro anterior  */
      teste = memcmp( registroAtual.chave, registroAnterior.chave,
                      TAM_CHAVE );

         /* Verifica se as duas ltimas chaves */
         /* esto em ordem crescente           */
      if (teste < 0)
         return 0;

         /* A chave anterior passa a ser a chave corrente */
      memcpy( &registroAnterior, &registroAtual,
              sizeof(registroAtual) );
   }

      /* Se ainda no houve retorno, o arquivo est ordenado */
   return 1;
}

/****
 *
 * main(): Verifica se um arquivo est ordenado
 *
 * Parmetros: Nenhum
 *
 * Retorno: Zero
 *
 ****/
int main (void)
{
   FILE *stream;
   int   resultado;

      /* Tenta abrir o arquivo de dados ordenado */
//   stream = fopen(NOME_ARQUIVO_ORD, "rb");
   stream = fopen(NOME_ARQUIVO_BIN, "rb");
   ASSEGURA(stream, "Arquivo ordenado nao foi aberto");

   MedidaDeTempo();
   printf( "\nTestando ordenacao do arquivo %s...",
           NOME_ARQUIVO_ORD );

   resultado = TestaOrdenacaoMS(stream);

      /* Fecha o arquivo ordenado */
   fclose(stream);

   MedidaDeTempo();

   if (resultado) {
      printf("\n\t>>> Ordenacao OK\n\n");
   } else {
      printf("\a\n\t>>> O arquivo nao esta ordenado\n\n");
   }

   return 0;
}

/******************* Exemplo de Execuo ******************

Testando ordenacao do arquivo CEPs_Ord.bin...

        *** Tempo gasto na operacao:  0.00 segundos ***

        >>> O arquivo nao esta ordenado

**********************************************************/
